#!/bin/env python3
# Copyright (C) 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
# Remarks:
#   Requires python 3.3 for the subclassing of OSError.
#
# Bugs:
#  No tests!


import argparse, os, zipfile
from datetime import datetime

from sqlalchemy.exc import StatementError
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from babase.lib import connection_opts, arg_parsing, utils
from babase.lib.exceptions import Error

from babase.lib.orm import Analysis, Analysis_ND_Experiment, Analysisprop, \
                           ND_Experiment, ND_Experimentprop, \
                           ND_Experiment_Stock

from babase.lib.constants import *

class FatalError(Error):
    pass

class InternalError(FatalError):
    pass

class AnalysisDataError(FatalError):
    def __init__(self, msg, analysis_id):
        super().__init__(self, msg)
        self.analysis_id = analysis_id

class NoAnalysispropFound(AnalysisDataError):
    pass

class MultipleAnalysispropFound(AnalysisDataError):
    pass

class AnalysisFileError(AnalysisDataError):
    def __init__(self, msg, analysis_id, pathname, oserror):
        super().__init__(self, msg)
        self.pathname = pathname
        self.oserror = oserror

class ColnoDataError(AnalysisDataError):
    def __init__(self, msg, analysis_id, nd_experimentprop_id, value):
        super().__init__(self, msg, analysis_id)
        self.nd_experimentprop_id = nd_experimentprop_id
        self.value = value

class BadQueryError(FatalError):
    pass

class BadStockQuery(BadQueryError):
    pass

class NoStockFound(BadStockQuery):
    pass

class ImproperStockQuery(BadStockQuery):
    pass

class ExtraColsStockQuery(BadStockQuery):
    pass

class BadAnalysisQuery(BadQueryError):
    pass

class NoAnalysisFound(BadAnalysisQuery):
    pass

class ImproperAnalysisQuery(BadAnalysisQuery):
    pass

class ExtraColsAnalysisQuery(BadAnalysisQuery):
    pass

def remove_whitespace(string):
    return '_'.join(string.split())

def parse_args():
    '''Parse command line args, getting defaults from the environment'''

    def add_arguments(parser, req):
        '''Add supported arguments to a parser'''

        dbgroup = parser.add_argument_group(
            'Database options')
        connection_opts.add_arguments(dbgroup)

        selection_group = parser.add_argument_group(
            'Data options')

        selection_group.add_argument(
                '--analysis-query', metavar='aquery'
                , help='SQL returning ANALYSIS.Analysis_Id values'
                       'of analysis to dump')
        selection_group.add_argument(
                '--stock-query', metavar='squery'
                , help='SQL returning STOCK.Stock_Id values to include'
                       ' in dump')
        selection_group.add_argument(
                '--anonymize', action='store_true'
                , help='Replace column headings with a stock id based value')

        file_group = parser.add_argument_group(
            'File information')

        file_group.add_argument(
            '--nd-experiment-cv', metavar='cvname'
            , required=arg_parsing.req_or_default(
                parser, 'nd_experiment_cv', req)
            , help='CV.Name of the vocabularly used to classify the'
                   ' experimental results as an a column of data')
        file_group.add_argument(
            '--nd-experiment-term', metavar='termname'
            , required=arg_parsing.req_or_default(
                parser, 'nd_experiment_term', req)
            , help='CVTERM.Name which classifies the experimental results'
                   ' (the ND_EXPERIMENT.Type_Id) as representing a column'
                   ' of a file')

        file_group.add_argument(
            '--column-cv', metavar='cvname'
            , required=arg_parsing.req_or_default(
                parser, 'column_cv', req)
            , help='CV.Name of vocabularly used to classify'
                   ' ANALYSIS_ND_EXPERIMENT values as columns in an'
                   ' uploaded pathname')
        file_group.add_argument(
            '--column-term', metavar='termname'
            , required=arg_parsing.req_or_default(
                parser, 'column_term', req)
            , help='The CVTERM.Name value which designates'
                   ' ANALYSIS_ND_EXPERIMENT rows to be representations of'
                   ' the columns of the uploaded pathname'
                   ' (the value of ANALYSIS_ND_EXPERIMENT.Type_Id)')

        file_group.add_argument(
            '--column-number-cv', metavar='cvname'
            , default=BB_CFD_COLUMN_NUMBER_CV_DEFAULT
            , help='CV.Name of the vocabularly used to classify'
                   ' ND_EXPERIMENTPROP values as column numbers of the'
                   ' uploaded pathname')
        file_group.add_argument(
            '--column-number-term', metavar='termname'
            , default=BB_CFD_COLUMN_NUMBER_TERM_DEFAULT
            , help='CVTERM.Name value which classifes a ND_EXPERIMENTPROP'
                   ' value as a column number of the uploaded pathname')

        file_group.add_argument(
            '--file-cv', metavar='cvname'
            , default=BB_CFD_FILE_CV_DEFAULT
            , help='CV.Name of the vocabularly used to classify the'
                   ' pathnames of uploaded analysis files')
        file_group.add_argument(
            '--file-term', metavar='termname'
            , default=BB_CFD_FILE_TERM_DEFAULT
            , help='CVTerm.Name used to designate an ANALYSISPROP.Value'
                   ' as the pathname of an uploaded analysis file')

    parser = argparse.ArgumentParser(
               description='Produce a zip file containing files'
                           ' of functional genomic analysis results'
               , epilog=('The {0} environment variable'
                         ' supplies default arguments')
                        .format(BB_CHADO_FUNCGEN_DUMP_ENVVAR))
    arg_parsing.add_version_arg(parser)
    env_parser = argparse.ArgumentParser()
    add_arguments(env_parser, False)
    defaults = arg_parsing.env_defaults(BB_CHADO_FUNCGEN_DUMP_ENVVAR
                                        , env_parser)
    parser.set_defaults(**defaults)    # Avoid missing required arguments
    add_arguments(parser, True)
    parser.set_defaults(**defaults)    # Have parser deflts override arg deflts
    parser.add_argument(
            'filepath', type=argparse.FileType('wb')
            , help='File to write, or stdout if omitted')

    args = parser.parse_args()

    return (parser, args)

def raise_internal_error(msg):
    raise InternalError('Programming error establishing Chado default: {0}'
                        .format(msg))

def get_web_db_objs(session):
    '''Supply constants that the web interface needs but does not supply.'''
    args = argparse.Namespace()
    args.nd_experiment_cv   = BB_CFD_WEB_ND_EXPERIMENT_CV
    args.nd_experiment_term = BB_CFD_WEB_ND_EXPERIMENT_TERM
    args.column_cv          = BB_CFD_WEB_COLUMN_CV
    args.column_term        = BB_CFD_WEB_COLUMN_TERM
    args.column_number_cv   = BB_CFD_WEB_COLUMN_NUMBER_CV
    args.column_number_term = BB_CFD_WEB_COLUMN_NUMBER_TERM
    args.file_cv            = BB_CFD_WEB_FILE_CV
    args.file_term          = BB_CFD_WEB_FILE_TERM

    lookup = arg_parsing.ChadoLookup(args, session, raise_internal_error)
    return get_db_objs(lookup)

def get_db_objs(lookup):
    '''Lookup command line arguments in the db.  Return a dict.

    Parameters:

    lookup  A ChadoLookup instance.
    '''
    db_objs = {}

    db_objs['nd_experiment_term'] = lookup.cvterm_arg(
        lookup.cv_arg('nd_experiment_cv'), 'nd_experiment_term')
    db_objs['column_term'] = lookup.cvterm_arg(
        lookup.cv_arg('column_cv'), 'column_term')
    db_objs['column_number_term'] = lookup.cvterm_arg(
        lookup.cv_arg('column_number_cv'), 'column_number_term')
    db_objs['file_term'] = lookup.cvterm_arg(
        lookup.cv_arg('file_cv'), 'file_term')

    return db_objs

class Dump():
    def get_stock_ids(self, args, session):
        '''Return a set of the stock ids to include, or None if including
        all stocks.
        '''
        if args.stock_query == None:
            return None

        stock_ids = set()
        try:
            row = None
            for row in session.execute(args.stock_query).fetchall():
                stock_ids.add(row[0])
            if row == None:
                raise NoStockFound('--stock-query returned no rows')
            if len(row) > 1:
                raise ExtraColsStockQuery(
                    'Too may columns return by the --stock-query squery')
        except StatementError as err:
            raise ImproperStockQuery(
                             'Problem with supplied --stock-query:\n {0}'
                             .format(err))

        return stock_ids

    def get_analysis_ids(self, args, db_objs, session):
        '''Return a collection of unique analysis_ids.
        '''
        analysis_ids = set()
        if args.analysis_query == None:
            sql = ('select analysis.analysis_id'
                   ' from analysis'
                   ' where ('
                   '   select count(*)'
                   '     from analysisprop'
                   '     where analysisprop.analysis_id = analysis.analysis_id'
                   '           and analysisprop.type_id = {0}'
                   '   ) = 1'
                   '  and exists (' # exclude VCF files
                   '    select 1'
                   '      from analysis_nd_experiment'
                   '      where analysis_nd_experiment.analysis_id ='
                   '            analysis.analysis_id)'
                   .format(db_objs['file_term'].cvterm_id))
        else:
            sql = args.analysis_query
        try:
            row = None
            for row in session.execute(sql).fetchall():
                analysis_ids.add(row[0])
            if row == None:
                raise NoAnalysisFound('--analysis-query returned no rows')
            if len(row) > 1:
                raise ExtraColsAnalysisQuery(
                    'Too many columns returned by the --analysis-query aquery')
        except StatementError as err:
            raise ImproperAnalysisQuery(
                             'Problem with supplied --analysis-query:\n {0}'
                             .format(err))

        return analysis_ids

    def get_pathname(self, db_objs, session, analysis_id):
        try:
            return (session.query(Analysisprop.value)
                    .filter_by(analysis_id=analysis_id
                               , type_id=db_objs['file_term'].cvterm_id)
                    .scalar())
        except NoResultFound:
            raise NoAnalysispropFound(
                             'The analysis with analysis_id ({0})'
                             ' does not have a related file in the filesystem:'
                             ' There is no ANALYSISPROP row related to this'
                             ' analysis having a the Type_id'
                             ' identified by the supplied --file-cv and'
                             ' --file-type'
                             .format(str(analysis_id))
                             , analysis_id)
        except MultipleResultsFound:
            raise MultipleAnalyispropFound(
                             'The analysis with analysis_id ({0})'
                             ' has more than one related file in the'
                             ' filesystem:'
                             ' Multiple ANALYSISPROP rows found related to'
                             ' this analysis having the Type_id'
                             ' identified by the supplied --file-cv and'
                             ' --file-type'
                             .format(str(analysis_id))
                             , analysis_id)

    def get_file_info(self, db_objs, session, analysis_id, stock_ids):
        '''Return a tuple (``cols``, ``ids``), or None if the
        given ``analysis_id`` identifies an analysis that uses
        none of the given ``stock_ids``.

        ``cols``
        Dict with keys of (integer) zero-based column numbers in the data file
        that contain data on a stock in ``stock_ids``.  The corresponding
        value is the stock_id that goes wtih the column.
        
        ``ids``
        Set of those stock_ids in ``stock_ids``.
        '''
        ids = set()
        cols = {}
        for (col
             , stock_id
             , analysis_nd_experiment_id
             , nd_experiment_id
             , nd_experimentprop_id) in (
                session.query(
                     ND_Experimentprop.value
                     , ND_Experiment_Stock.stock_id
                     , Analysis_ND_Experiment.analysis_nd_experiment_id
                     , ND_Experiment.nd_experiment_id
                     , ND_Experimentprop.nd_experimentprop_id)
                .select_from(Analysis_ND_Experiment)
                .filter_by(analysis_id=analysis_id
                           , type_id=db_objs['column_term'].cvterm_id)
                .join(ND_Experiment)
                .filter_by(type_id=db_objs['nd_experiment_term'].cvterm_id)
                .join(ND_Experimentprop)
                .filter_by(type_id=db_objs['column_number_term'].cvterm_id)
                .join(ND_Experiment_Stock)
                .all()):
            if stock_ids == None or stock_id in stock_ids:
                ids.add(stock_id)
                try:
                    cols[int(col)] = stock_id
                except ValueError:
                    raise ColnoDataError(
                        'A column number value ({0}) for analysis_id ({1})'
                        ' found via analysis_nd_experiment_id ({2}),'
                        ' nd_experiment_id ({3}),'
                        ' in ND_EXPERIMENTPROP with nd_experimentprop_id ({4})'
                        ' cannot be converted to an integer'
                        .format(col, analysis_id, analysis_nd_experiment_id
                                , nd_experiment_id, nd_experimentprop_id))
        if cols:
            return (cols, ids)
        else:
            return None


    def get_file_columns(self, pathname, cols, anonymize):
        '''Return string having only the columns in ``cols`` in ``pathname``.
        '''
        colnums = list(cols.keys())
        colnums.sort(reverse=True)
        data = []
        try:
            with open(pathname, 'r') as fileobj:
                first_line = True
                for line in fileobj:
                    if line[-1] == '\n': # In case file's last char is not EOL
                        line = line[:-1]
                    target_cols = colnums.copy()
                    line_data = line.split('\t')
                    if first_line and anonymize:
                        first_line = False
                        saved_cols = line_data[:1] # include column 0
                        target_cols.sort()
                        for target_col in target_cols:
                            saved_cols.append('{0}{1}'
                                              .format(BB_CFD_ANON_STOCK_PREFIX
                                                      , cols[target_col]))
                    else:
                        target_col = 0 # Always include column 0
                        data_col = 0
                        saved_cols = []
                        try:
                            for col_data in line_data:
                                if data_col == target_col:
                                    saved_cols.append(col_data)
                                    target_col = target_cols.pop()
                                data_col += 1
                        except IndexError:
                            pass  # Running out of target columns breaks loop
                    data.append('\t'.join(saved_cols))

        except OSError as err:
            raise AnalysisFileError(
                'The pathame ({0}) for analysis_id ({1}) cannot  be opened:\n'
                .format(pathname, str(analysis_id), str(err))
                , analysis_id, pathname, err)

        data.append('')  # End in \n
        return '\n'.join(data)


    def build_file(self, db_objs, session, analysis_id, anonymize, stock_ids):
        '''Return a tuple (``data``, ``ids``)

        ``data``
        String with functional analysis results
        from the ``analysis_id`` analysis having those columns
        that are analysis of the supplied ``stock_ids``.

        ``ids``
        Set of those ``stock_ids`` that are analyzed in ``data``.

        Return None if the analysis is of none of the supplied stock.
        '''
        # Before returning, check that analysis has a pathname.
        pathname = self.get_pathname(db_objs, session, analysis_id)
        tuple = self.get_file_info(db_objs, session, analysis_id, stock_ids)
        if not tuple:
            return None
        cols, ids = tuple
        return (self.get_file_columns(pathname, cols, anonymize), ids)


    def cmdline_dump(self, args, db_objs, session, zfile):
        '''Writes files of funcgen analysis results using command line args.

        ``args``
        An argparse.ArgumentParser.parse() return value.

        ``db_objs``
        A dict containing sqlalchemy objects keyed by
        their ``args``.attribute.
        '''
        return self.dump(db_objs, session, zfile
                         , self.get_stock_ids(args, session)
                         , self.get_analysis_ids(args, db_objs, session)
                         , args.anonymize)

    def web_dump(self, session, zfile, stock_ids, analysis_ids, anonymize):
        '''Writes files of funcgen analysis results into ``zfile``.

        Called from the web interface it uses default values
        to look up the needed Chado codes.
        
        Parameters:

        ``session``
        An sqlalchemy session.

        ``zfile``
        A zipfile.ZipFile() object opened for writing.

        ``stock_ids``
        A list, set or other object which supports the "in" operator
        and contains (integer) stock ids.

        ``analysis_ids``
        An iterable of (interger) analysis ids.

        ``anonymize``
        Boolean.  When True replace returned column headings with a
        stock id based value.

        Returns:

        The tuple (``dir``, ``used_ids``).

        ``dir``
        The directory prefix in which all analysis files are placed
        inside the zipfile.

        ``used_ids``
        A set of stock_ids that were analyzed in the files placed
        in ``zfile``.
        '''
        return self.dump(get_web_db_objs(session), session, zfile
                         , stock_ids, analysis_ids, anonymize)

    def dump(self, db_objs, session, zfile, stock_ids, analysis_ids
             , anonymize):
        '''Writes files of funcgen analysis results into ``zfile``.
        
        Parameters:

        ``db_objs``
        A dict containing sqlalchemy objects keyed by
        their ``args``.attribute.

        ``session``
        An sqlalchemy session.

        ``zfile``
        A zipfile.ZipFile() object opened for writing.

        ``stock_ids``
        A list, set or other object which supports the "in" operator
        and contains (integer) stock ids.

        ``analysis_ids``
        An iterable of (interger) analysis ids.

        ``anonymize``
        Boolean.  When True replace returned column headings with a
        stock id based value.

        Returns:

        The tuple (``dir``, ``used_ids``).

        ``dir``
        The directory prefix in which all analysis files are placed
        inside the zipfile.

        ``used_ids``
        A set of stock_ids that were analyzed in the files placed
        in ``zfile``.
        '''

        dirname = ('cfd-{0}-UTC'
                   .format(datetime.utcnow().strftime('%Y-%m-%d-%H-%M-%S')))

        all_used_ids = set()

        for analysis_id in analysis_ids:
            analysis = session.query(Analysis).get(analysis_id)
            filename = str(analysis_id)
            if analysis.name != None:
                filename += '_' + remove_whitespace(analysis.name)
            tuples = self.build_file(db_objs, session, analysis_id, anonymize
                                     , stock_ids)
            if tuples:
                data, used_ids = tuples
                zfile.writestr(os.path.join(dirname, filename), data)
                all_used_ids.update(used_ids)

        return (dirname, all_used_ids)

def main():
    # Parse command line
    parser, args = parse_args()

    # Validate user input
    session = connection_opts.my_session(parser, args)

    lookup = arg_parsing.ChadoLookup(
        args=args, session=session, error_func=parser.error)
    db_objs = get_db_objs(lookup)

    try:
        try:
            with zipfile.ZipFile(args.filepath, 'w'
                                 , compression=zipfile.ZIP_DEFLATED) as zfile:
                Dump().cmdline_dump(args, db_objs, session, zfile)
        except Exception:
            # Do not leave corrupted zipfiles laying about
            try:
                os.remove(args.filepath.name)
            except OSError as err:
                if err != FileNotFoundError:
                    utils.std_error(
                        'Error deleting the (possibly corrupt) zipfile:'
                        ' {0}: {1}\n'
                        .format(args.filepath.name, str(err)))
            raise
    except Error as err:
        utils.data_error(err.msg)

if __name__ == '__main__':
    main()
