#!/bin/env python3
# Copyright (C) 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
# Remarks:
#  In the user interface column numbers are 1's based, but internally
# in this program they are 0's based.
#
# Bugs:
#  No tests!
#  TypeMaker() does not cache cvterm lookup.
#  A typeo in an argument option name causes the value to be the pathname.

import argparse \
       , itertools \
       , os \
       , sys

from sqlalchemy import func
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

from babase.lib import connection_opts, arg_parsing, utils
from babase.lib.exceptions import Error

from babase.lib.orm import Stock, Stockprop, Stock_CVTerm, \
    CV, CVTerm, DB, DBXref, Stock_DBXref, stock_stock_id_seq

from babase.lib.constants import *

FIRST_DATA_LINE = 3 # 1 relative
FIRST_DATA_COL = 3  # zero relative

# The "fixed" column indexes
DB_COL        = 0
ACCESSION_COL = 1
ID_COL        = 2

class FatalError(Error):
    pass

class NonFatalError(Error):
    pass

def parse_args():
    '''Parse command line args, getting defaults from the environment'''

    def add_arguments(parser, req):
        '''Add supported arguments to a parser'''

        dbgroup = parser.add_argument_group(
            'Database options')
        connection_opts.add_arguments(dbgroup)

        dbgroup.add_argument(
                '--dry-run', action='store_true'
                , help='Rollback transaction upon completion')

        data_group = parser.add_argument_group(
            'Data options')

        data_group.add_argument(
                '--organism-id', metavar='oid', type=int
                , required=arg_parsing.req_or_default(
                    parser, 'organism_id', req)
                , help='STOCK.Organism_Id value')

        data_group.add_argument(
                '--pub-id', metavar='pubid', type=int
                , required=arg_parsing.req_or_default(
                    parser, 'pub_id', req)
                , help='STOCK_CVTERM.Pub_Id value')

        data_group.add_argument(
                '--primary-db', metavar='name'
                , required=arg_parsing.req_or_default(
                    parser, 'primary_db', req)
                , help='DB.Name of db containing the ids of'
                       ' the 3rd column')

        processing_group = parser.add_argument_group(
            'Processing options')

        processing_group.add_argument(
                '--allow-duplicates', action='store_true'
                , help=('Create new STOCK row when columns 1 and 2 match'
                        ' but the remaining columns do not'))

    parser = argparse.ArgumentParser(
               description='Load stock information into Babase/Chado'
               , epilog=('The {0} environment variable'
                         ' supplies default arguments')
                        .format(BB_CHADO_STOCK_LOAD_ENVVAR))
    arg_parsing.add_version_arg(parser)
    env_parser = argparse.ArgumentParser()
    add_arguments(env_parser, False)
    defaults = arg_parsing.env_defaults(BB_CHADO_STOCK_LOAD_ENVVAR, env_parser)
    parser.set_defaults(**defaults)    # Avoid missing required arguments
    add_arguments(parser, True)
    parser.set_defaults(**defaults)    # Have parser deflts override arg deflts
    parser.add_argument(
            'pathname', nargs='?', type=argparse.FileType('r')
            , default=sys.stdin
            , help='File containing stock data, or stdin if omitted')

    args = parser.parse_args()

    return (parser, args)

class TypeMaker():
    '''Make a STOCK_CVTERM row in the db.'''
    def __init__(self, args, session, user_colno, col, parts):
        if len(parts) != 2:
            raise FatalError('Unrecoginized column heading ({0}) in column {1}'
                             .format(col, user_colno))
        self.pub_id = args.pub_id
        self.session = session
        self.user_colno = user_colno
        self.cv = self.lookup_cv(self.user_colno, col, parts[1])

    def lookup_cv(self, user_colno, col, cvname):
        try:
            return self.session.query(CV).filter_by(name=cvname).one()
        except NoResultFound:
            raise FatalError(('Invalid colum heading ({0}) in column {1}:'
                              ' No such CV ({2})')
                             .format(col, self.user_colno, cvname))

    def lookup_cvterm(self, val):
        try:
            return (self.cv.cvterms
                    .filter_by(name=val, cv_id=self.cv.cv_id)
                    .one())
        except NoResultFound:
            raise FatalError('Column {0}:'
                             ' CVTERM.Name ({1}) not found in CV where'
                             ' CV.Name = ({2})'
                             . format(self.user_colno, val, self.cv.name))

    def make_row(self, stock_id, name):
        '''Make a row in STOCK_CVTERM.

        Parameters:

        ``stock_id``
        The STOCK.Stock_Id of the related STOCK row.

        ``name``
        The CVTERM.Name of the CVTERM to put in the STOCK_CVTERM.CVTerm_Id.
        '''
        stock_cvterm = Stock_CVTerm()
        stock_cvterm.stock_id = stock_id
        stock_cvterm.cvterm_id = self.lookup_cvterm(name).cvterm_id
        stock_cvterm.pub_id = self.pub_id
        stock_cvterm.is_not = False
        stock_cvterm.rank = BB_CHADO_RANK_SINGLETON

        self.session.add(stock_cvterm)
        
class PropMaker(TypeMaker):
    '''Make a STOCKPROP row in the db.'''
    def __init__(self, session, user_colno, col, parts):
        if len(parts) != 3:
            raise FatalError('Unrecoginized column heading ({0}) in column {1}'
                             .format(col, user_colno))
        self.session = session
        self.user_colno = user_colno
        self.cv = self.lookup_cv(self.user_colno, col, parts[1])
        self.cvterm = self.lookup_cvterm(parts[2])

    def make_row(self, stock_id, val):
        '''Make a row in STOCKPROP.

        Parameters:

        ``stock_id``
        The STOCK.Stock_Id of the related STOCK row.

        ``val``
        The STOCKPROP.Value.
        '''
        val = val.strip()
        if val == BB_CSL_NO_PROP_VALUE:
            val = None

        stockprop = Stockprop()
        stockprop.stock_id = stock_id
        stockprop.type_id = self.cvterm.cvterm_id
        stockprop.value = val
        stockprop.rank = BB_CHADO_RANK_SINGLETON

        self.session.add(stockprop)

class Cache():
    '''Abstract cache class.'''
    def __init__(self, session):
        self.session = session
        self.cache = {}

class DBCache(Cache):
    '''Cache db.db_id, by name.'''
    def lookup(self, dbname):
        dbname = dbname.strip()
        if dbname not in self.cache:
            self.cache[dbname] = self._lookup(dbname)
        return self.cache[dbname]
    def _lookup(self, dbname):
        try:
            return self.session.query(DB).filter_by(name=dbname).one().db_id
        except NoResultFound:
            raise FatalError('No DB with the given Name ({0})'.format(dbname))

class DBXrefCache(Cache):
    '''Cache dbxref_id by db_id, accession.'''
    def lookup(self, dbname, db_id, target_accession):
        target_accession = target_accession.strip()
        if db_id not in self.cache:
            self.cache[db_id] = {}
        if target_accession not in self.cache[db_id]:
            self.cache[db_id][target_accession] = (
                self._lookup(dbname, db_id, target_accession))
        return self.cache[db_id][target_accession]
    def _lookup(self, dbname, db_id, target_accession):
        try:
            return (self.session.query(DBXref)
                    .filter_by(db_id=db_id, accession=target_accession).one()
                    .dbxref_id)
        except NoResultFound:
            raise FatalError(('No DBXREF found in the ({0}) DB with an'
                        ' Accession of ({1})')
                       .format(dbname, target_accession))

class DescribeCVTerm(Cache):
    '''Cache cvterm.name and cv.name by cvterm.cvterm_id'''
    def __init__(self, session):
        self.cvcache = {}
        super().__init__(session)

    def lookup(self, cvterm_id):
        if cvterm_id not in self.cache:
            self.cache[cvterm_id] = self._lookup(cvterm_id)
        return self.cache[cvterm_id]
    def _lookup(self, cvtid):
        try:
            cvterm = (self.session.query(CVTerm)
                      .filter_by(cvterm_id=cvtid).one())
        except NoResultFound:
            raise FatalError('System error: CVTERM.CVTerm_Id = ({0}):'
                             ' No such CVTerm_Id'
                             .format(cvtid))
        if cvterm.cv_id not in self.cvcache:
            # Trust the db and don't bother to catch the exception.
            self.cvcache[cvterm.cv_id] = (self.session.query(CV)
                                          .filter_by(cv_id=cvterm.cv_id).one()
                                          .name)
        return (self.cvcache[cvterm.cv_id], cvterm.name)

class DescribeDBXref(DescribeCVTerm):
    '''Cache dbxref.accession and db.name by dbxref.dbxref_id'''
    def _lookup(self, dbxref_id):
        try:
            dbxref = (self.session.query(DBXref)
                      .filter_by(dbxref_id=dbxref_id).one())
        except NoResultFound:
            raise FatalError('System error: DBXREF.DBXref_Id = ({0}):'
                             ' No such DBXref_Id'
                             .format(dbxref_id))
        if dbxref.dbxref_id not in self.cvcache:
            self.cvcache[dbxref.dbxref_id] = (self.session.query(DB)
                                              .filter_by(db_id=dbxref.db_id)
                                              .one()
                                              .name)
        return (self.cvcache[dbxref.dbxref_id], dbxref.accession)

class MatchType():
    def __init__(self, cv, load_stock):
        self.cv = cv
        self.load_stock = load_stock

    def match(self, session, colno, stock, val):
        '''Does the given ``stock`` have a type of ``val``?'''
        try:
            (session.query(Stock_CVTerm)
             .filter_by(stock_id=stock.stock_id)
             .join(CVTerm)
             .filter(CVTerm.cv_id == self.cv.cv_id, CVTerm.name == val)
             .one())
        except NoResultFound:
            return ('Column number {colno}:'
                    ' STOCK has no matching STOCK_CVTERM.Type_Id:'
                    ' STOCK.Stock_Id = ({stock_id}):'
                    ' CV.Name = ({cv_name}):'
                    ' CVTERM.Name = ({val}):'
                    .format(stock_id=stock.stock_id
                            , val=val
                            , cv_name=self.cv.name
                            , colno=colno + 1))
        else:
            return ''

class MatchProp():
    def __init__(self, cv, cvterm, load_stock):
        self.cv = cv
        self.cvterm = cvterm
        self.load_stock = load_stock

    def match(self, session, colno, stock, val):
        '''Does the given ``stock`` have a type of ``cvterm``.cvterm_id and a
        value of ``val``?'''
        try:
            (session.query(Stockprop)
             .filter_by(stock_id=stock.stock_id
                        , type_id=self.cvterm.cvterm_id
                        , value=val.strip()
                        , rank=BB_CHADO_RANK_SINGLETON)
             .one())
        except NoResultFound:
            return ('Column number {colno}:'
                    ' STOCK has no matching STOCKPROP:'
                    ' STOCK.Stock_Id = ({stock_id}):'
                    ' CV.Name = ({cv_name}):'
                    ' CVTERM.Name = ({cvterm_name}):'
                    ' STOCKPROP.Value = ({value}):'
                    ' STOCKPROP.Rank = ({rank})'
                    .format(stock_id=stock.stock_id
                            , cv_name=self.cv.name
                            , cvterm_name=self.cvterm.name
                            , value=val.strip()
                            , colno=colno + 1
                            , rank=BB_CHADO_RANK_SINGLETON))
        else:
            return ''

class LoadStock():
    def __init__(self):
        self.file_error = None
        self.serror = False
        self.lineno = None

        # Set in process_headers()
        self.col_processors = []
        self.primary_type_colno = None
        self.column_number_colno = None
        self.primary_type_finder = None
        self.match_objs = []

        self.column_numbers = []

        self.describe_cvterm = None


    def process_headers(self, args, session):
        '''Examine header lines in the ``args``.pathname and return
        a 2 tuple.  The first element is a processor for the primary type
        and the second a list of processors for columns 4 and up.

        Sets:
        self.col_processors
        self.column_number_colno
        self.primary_type_colno
        self.primary_type_finder
        self.match_objs
        '''
        try:
            line = next(args.pathname)
        except StopIteration:
            raise FatalError('No lines in file')

        if line[-1] == '\n':
            line = line[:-1]

        cols = line.split('\t')
        if len(cols) < FIRST_DATA_COL:
            raise FatalError('The file must contain at least {0} columns.'
                             .format(FIRST_DATA_COL))

        user_colno = 4
        for col in cols[FIRST_DATA_COL:]:
            parts = col.split(BB_CSL_HDG_DELIM)
            if parts[0] == BB_CSL_TYPE_PREFIX_HDG:
                tp_obj = TypeMaker(args, session, user_colno, col, parts)
                self.col_processors.append(tp_obj)
                matcher = MatchType(tp_obj.cv, self)
                self.match_objs.append(matcher)
            elif parts[0] == BB_CSL_PROP_PREFIX_HDG:
                pm_obj = PropMaker(session, user_colno, col, parts)
                self.col_processors.append(pm_obj)
                matcher = MatchProp(pm_obj.cv, pm_obj.cvterm, self)
                self.match_objs.append(matcher)
            elif parts[0] == BB_CSL_COLUMN_NUMBER_HDG:
                self.column_number_colno = user_colno - 1
            elif parts[0] == BB_CSL_PRIMARYTYPE_PREFIX_HDG:
                if self.primary_type_finder:
                    raise FatalError(('Column {0} ({1}) declares a second'
                                      ' primary type, only one is allowed')
                                     .format(user_colno, col))
                self.primary_type_colno = user_colno - 1
                lookup_obj = TypeMaker(args, session, user_colno, col, parts)
                self.primary_type_finder = lookup_obj
                # Use a slot to maintain a 1-to-1 mapping between
                # data columns and functions.
                self.col_processors.append(None)
                self.match_objs.append(None)
            else:
                raise FatalError(('Column {0} ({1}) has an invalid heading:'
                                  ' ({2}) unrecognized')
                                 .format(user_colno, col, parts[0]))
            user_colno += 1

        if not self.primary_type_colno:
            raise FatalError('Line 1 does not contain a heading prefixed with'
                             ' "{0}{1}"'
                             .format(BB_CLS_PRIMARYTYPE_PREFIX_HDG
                                     , BB_CSL_HDG_DELIM))

        if not self.column_number_colno:
            raise FatalError('Line 1 does not contain a "{0}"'
                             ' heading'
                             .format(BB_CSL_COLUMN_NUMBER_HDG))

        try:
            line = next(args.pathname)
        except StopIteration:
            raise FatalError('Only one line in file')

    def stock_matches(self, session, old_stock, new_stock, cols):
        '''Does stock match what's in file?'''
        errors = []
        # Does old stock match what's on the line?
        # Match organisim here instead of in query so as to report errors.
        if old_stock.organism_id != new_stock.organism_id:
            errors.append('Mismatch in STOCK.Organism_Id:'
                          ' Old STOCK.Stock_Id = ({stock_id}):'
                          ' Old STOCK.Organism_Id = ({old_val}):'
                          ' New STOCK.Organism_Id = ({new_val})'
                          .format(stock_id=old_stock.stock_id
                                  , old_val=old_stock.organism_id
                                  , new_val=new_stock.organism_id))
        if old_stock.type_id != new_stock.type_id:
            old_cv_name, old_cvterm_name = self.describe_cvterm(
                                                    old_stock.cvterm_id)
            new_cv_name, new_cvterm_name = self.describe_cvterm(
                                                    new_stock.cvterm_id)
            errors.append('Mismatch in STOCK.Type_Id:'
                          ' Old STOCK.Stock_Id = ({stock_id}):'
                          ' Old STOCK.Type_Id = ({old_type_id}):'
                          ' Old CV.Name = ({old_cv_name}):'
                          ' Old CVTerm.Name = ({old_cvterm_name}):'
                          ' New STOCK.Type_Id = ({new_type_id})'
                          ' New CV.Name = ({new_cv_name}):'
                          ' New CVTerm.Name = ({new_cvterm_name}):'
                          .format(stock_id=old_stock.stock_id
                                  , old_type_id=old_stock.type_id
                                  , old_cv_name=old_cv_name
                                  , old_cvterm_name=old_cvterm_name
                                  , new_type_id=new_stock.type_id
                                  , new_cv_name=new_cv_name
                                  , new_cvterm_name=new_cvterm_name))
        if old_stock.dbxref_id != new_stock.dbxref_id:
            old_db_name, old_dbxref_accession = self.describe_dbxref(
                                                      old_stock.dbxref_id)
            new_db_name, new_dbxref_accession = self.describe_dbxref(
                                                      new_stock.dbxref_id)
            errors.append('Mismatch in STOCK.Type_Id:'
                          ' Old STOCK.Stock_Id = ({stock_id}):'
                          ' Old STOCK.DBXref_Id = ({old_dbxref_id}):'
                          ' Old DB.Name = ({old_db_name}):'
                          ' Old DBXREF.Accession = ({old_dbxref_accession}):'
                          ' New STOCK.DBXref_Id = ({new_dbxref_id})'
                          ' New DB.Name = ({new_db_name}):'
                          ' New DBXREF.Accession = ({new_dbxref_accession}):'
                          .format(stock_id=old_stock.stock_id
                                  , old_dbxref_id=old_stock.dbxref_id
                                  , old_db_name=old_db_name
                                  , old_dbxref_accession=old_dbxref_accession
                                  , new_dbxref_id=new_stock.dbxref_id
                                  , new_db_name=new_db_name
                                  , new_dbxref_accession=new_dbxref_accession))
        if old_stock.is_obsolete:
            errors.append('Existing stock is obsolete:'
                          ' Old STOCK.Stock_Id = ({stock_id})'
                          .format(old_stock_id=old_stock.stock_id))
        
        # Do the related tables match?
        colno = FIRST_DATA_COL
        for val, matcher in itertools.zip_longest(cols[FIRST_DATA_COL:]
                                                  , self.match_objs):
            if (colno != self.column_number_colno
                and colno != self.primary_type_colno):
                if matcher == None:
                    FatalError('Too many columns of data')
                if val == None:
                    val = ''
                error = matcher.match(session, colno, old_stock, val)
                if error:
                    errors.append(error)
            colno += 1

        return errors

    def soft_msg(self, args, msgs):
        '''Report soft errors to the user.'''
        for msg in msgs:
            utils.std_error('{0}: Line {1}: Matching stock row found: {2}\n'
                            .format(args.pathname.name, self.lineno, msg))

    def soft_error(self, args, msgs):
        '''Report and record soft errors to the user.'''
        self.soft_msg(args, msgs)
        self.serror = True

    def save_column_number(self, val):
        try:
            column_number = int(val)
        except ValueError:
            FatalError('The "{1}" value ({0})'
                       ' cannot be converted to an integer'
                       .format(val, BB_CSL_COLUMN_NUMBER_HDG))
        self.column_numbers.append(column_number)

    def make_new(self, session, stock, dbxref_id, cols):
        '''Save line as a STOCK row and related rows.'''
        # The stock row
        stock.stock_id = session.execute(func.next_value(
                stock_stock_id_seq)).scalar()
        stock.uniquename = stock.stock_id

        session.add(stock)

        # The stock_dbxref row that records the freezer sample
        stock_dbxref = Stock_DBXref()
        stock_dbxref.stock_id = stock.stock_id
        stock_dbxref.dbxref_id = dbxref_id
        stock_dbxref.is_current = True

        session.add(stock_dbxref)

        # Add related rows
        colno = FIRST_DATA_COL
        for val, processor in itertools.zip_longest(cols[FIRST_DATA_COL:]
                                               , self.col_processors):
            if colno != self.primary_type_colno:
                if colno == self.column_number_colno:
                    self.save_column_number(val)
                else:
                    if processor == None:
                        raise FatalError('Too many columns of data')
                    if val == None:
                        val = ''
                    processor.make_row(stock.stock_id, val)
            colno += 1

        return stock

    def by_column_number(self, stock_ids):
        '''Check self.column_numbers to see that they start with 2
        and increase by 1 without gaps and return the stock_ids
        sorted by the self.column_numbers values.
        '''
        c_nums = self.column_numbers[:]
        c_nums.sort()
        if c_nums[0] != BB_CSL_MIN_COL_NUM:
            self.file_error('The minimum "column_number" is ({0}), it'
                            ' must be {1}'.format(c_nums[0]
                                                  , BB_CSL_MIN_COL_NUM))
        val = BB_CSL_MIN_COL_NUM
        for column_number in c_nums:
            if column_number != val:
                self.file_error(
                    'line {0}: The "{2}" value ({1})'
                    ' is out of sequence or duplicated'
                    .format(self.column_numbers.index(column_number)
                            + FIRST_DATA_LINE
                            , column_number
                            , BB_CSL_COLUMN_NUMBER_HDG))
            val += 1

        return [stock_ids[column_number - BB_CSL_MIN_COL_NUM]
                for column_number in self.column_numbers]


    def load_stock(self, parser_error, args, session):
        '''Loads stock table with new stock or finds existing stock.

        Returns a list of loaded or found stock ids.
        '''

        self.file_error = utils.make_file_error(args.pathname.name)

        self.describe_cvterm = DescribeCVTerm(session).lookup
        self.describe_dbxref = DescribeDBXref(session).lookup
        db_cache = DBCache(session)
        dbxref_cache = DBXrefCache(session)

        primary_db_name = args.primary_db
        try:
            primary_db = db_cache.lookup(primary_db_name)
        except FatalError as e:
            parser_error('Bad --primary-db name: {0}'.format(e))

        # Get the first line, which tells us how to store the data.
        # (And throw out the 2nd header line, which is for the user.)
        try:
            self.process_headers(args, session)
        except FatalError as e:
            self.file_error(e)

        stock_ids = []
        self.lineno = FIRST_DATA_LINE - 1
        try:
            for line in args.pathname:
                self.lineno += 1
                # Be forgiving of files that do not end in EOL.
                if line[-1] == '\n':
                    line = line[:-1]

                cols = line.split('\t')

                if cols[DB_COL] == BB_CSL_SPECIAL_COL1_DB:
                    # Validate the dbxref_id
                    did = cols[ACCESSION_COL]
                    db_sname = session.execute(
                        'select sname from genetic_inventory.dna'
                        " where did = '{0}';"
                        .format(did)).scalar()
                    if db_sname == None:
                        raise FatalError('The did ({0}) is not found on'
                                         ' genetic_inventory.dna'
                                         .format(did))
                    file_sname = cols[ID_COL]
                    if (db_sname != '' and db_sname != None
                        and file_sname != ''
                        and file_sname != BB_CSL_MISSING_UNIQUE_ID
                        and db_sname != file_sname):
                        raise FatalError(
                            'The sname ({0}) found in genetic_inventory.dna'
                            ' where did is ({1}) does not match the'
                            ' sname ({2}) found in column 3'
                            .format(db_sname, did, file_sname))

                # Build the stock row we're looking for
                new_stock = Stock()
                new_stock.dbxref_id = dbxref_cache.lookup(
                    primary_db_name, primary_db, cols[ID_COL])
                new_stock.organism_id = args.organism_id
                new_stock.type_id = self.primary_type_finder.lookup_cvterm(
                    cols[self.primary_type_colno]).cvterm_id
                new_stock.is_obsolete = False

                accession_dbxref_id = dbxref_cache.lookup(
                    cols[DB_COL]
                    , db_cache.lookup(cols[DB_COL])
                    , cols[ACCESSION_COL])

                # Order results so matching on multiple rows is predictable.
                # (Use lowest stock_id value when there are multiple matches.)
                query = (session.query(Stock).order_by(Stock.stock_id.desc())
                         .join(Stock_DBXref)
                         .filter_by(dbxref_id=accession_dbxref_id))
                try:
                    old_stock = query.one()
                    # Query finds one row, make sure it matches
                    mismatch_msgs = self.stock_matches(
                        session, old_stock, new_stock, cols)
                    if mismatch_msgs:
                        if args.allow_duplicates:
                            stock_ids.append(
                                self.make_new(session, new_stock
                                              , accession_dbxref_id, cols)
                                .stock_id)
                        else:
                            self.soft_error(args, mismatch_msgs)
                    else:
                        stock_ids.append(old_stock.stock_id)
                        self.save_column_number(
                            cols[self.column_number_colno])
                except NoResultFound:
                    # Query finds no rows, make a row
                    stock_ids.append(self.make_new(session, new_stock
                                                   , accession_dbxref_id, cols)
                                     .stock_id)
                except MultipleResultsFound:
                    # Query finds more than one row, see if any match.
                    match = False
                    for old_stock in query.all():
                        mismatch_msgs = self.stock_matches(
                            session, old_stock, new_stock, cols)
                        if mismatch_msgs:
                            if not args.allow_duplicates:
                                self.soft_error(args, mismatch_msgs)
                        else:
                            match = old_stock
                            # Keep going, to report on everything
                    if match:
                        if args.allow_duplicates:
                            stock_ids.append(match.stock_id)
                            self.save_column_number(
                                cols[self.column_number_colno])
                        else:
                            self.soft_msg(args
                                          , [('Line matches stock_id = ({0})'
                                             .format(match.stock_id))])
                    else:
                        if args.allow_duplicates:
                            stock_ids.append(
                                self.make_new(session, new_stock
                                              , accession_dbxref_id, cols)
                                .stock_id)
        except FatalError as e:
            self.file_error('line {0}: {1}'.format(self.lineno, e))

        if self.lineno == 2:
            self.file_error('File contains no data lines.')

        if self.serror:
            session.rollback()
            self.file_error('Aborted: Matching stock information found')

        return self.by_column_number(stock_ids)

def print_stockids(stock_ids):
    '''Print a list of stock ids in a form suitable for chado-funcgen-load.'''
    print(':'.join([str(id) for id in stock_ids]))

def main():
    try:
        # Parse command line
        parser, args = parse_args()

        # Validate user input
        session = connection_opts.my_session(parser, args)
        arg_parsing.validate_stock_args(parser, args, session)

        ids = LoadStock().load_stock(parser.error, args, session)

        # Commit or rollback
        connection_opts.finish(args, session)

        print_stockids(ids)
    except KeyboardInterrupt:
        utils.report_keyboard_interrupt()

if __name__ == '__main__':
    main()
