# Copyright (C) 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
# Remarks:
#
# Bugs:

import argparse \
       , shlex \
       , os

from sqlalchemy.orm.exc import NoResultFound

from babase._version import __version__

from babase.lib.orm import CV, DB, CVTerm, Organism, Pub

from babase.lib.constants import *

def env_defaults(evar, env_parser):
    '''Return a dict of defaults by getting command line arguments
       from the environment var, evar
       
       Input:
         evar  An environment variable name
         env_parser An argparse.ArgumentParser() object that is
                    initalized so that env_parser.parse_args()
                    will parse whatever command line args are in
                    the environment variable.
       Returns:
         Dict keyed with argument names and having default values.'''

    if evar in os.environ:
        envargs = env_parser.parse_args(shlex.split(os.environ[evar]))
    else:
        envargs = argparse.Namespace()
    return vars(envargs)

def bool_arg(arg):
    '''Convert a user supplied argument to a boolean value.
    '''
    val = arg.lower()
    return not(val == BB_FALSE_FALSE or val == BB_FALSE_F
               or val == BB_FALSE_N or val ==BB_FALSE_NO
               or val == BB_FALSE_0 or val == BB_FALSE_EMPTYSTRING)

def req_or_default(parser, dest, req):
    '''Return req if parser has no default for dest,
       or False otherwise.'''
    if parser.get_default(dest) == None:
        return req
    else:
        return False

def add_analysis_args(anal_group, parser, req):
    '''Add standard args supporting creation of an ANALYSIS row.
       anal_group  An argparse group
       parser      An argparse parser
       req         Boolean as to whether there are required arguments.
       '''
    anal_group.add_argument(
        '--analysis-description', metavar='description'
        , help='ANALYSIS.Description value (default: NULL)')
    anal_group.add_argument(
        '--analysis-program', metavar='program'
        , required=req_or_default(parser, 'analysis_program', req)
        , help='ANALYSIS.Program value')
    anal_group.add_argument(
        '--analysis-programversion', metavar='version'
        , required=req_or_default(parser
                                  , 'analysis_programversion'
                                  , req)
        , help='ANALYSIS.Programversion value')
    anal_group.add_argument(
        '--analysis-algorithm', metavar='algorithm'
        , help='ANLYSIS.Algorithm value (default: NULL)')
    anal_group.add_argument(
        '--analysis-sourcename', metavar='sourcename'
        , help='ANALYSIS.Sourcename (default: NULL)')
    anal_group.add_argument(
        '--analysis-sourceversion', metavar='sourceversion'
        , help='ANALYSIS.Sourceversion (default: NULL)')
    anal_group.add_argument(
        '--analysis-sourceuri', metavar='sourceuri'
        , help='ANALYSIS.Sourceuri (default: NULL)')
    anal_group.add_argument(
        '--analysis-timeexecuted', metavar='timestamp'
        , required=req_or_default(parser, 'analysis_timeexecuted', req)
        , help='ANALYSIS.Timeexecuted value')

    anal_group.add_argument(
        '--analysis-cv', metavar='cvname'
        , help='CV.Name of vocabularly used to classify the analysis')
    anal_group.add_argument(
        '--analysis-term', metavar='termname'
        , help='CVTERM.Name which classifies the analysis')

def validate_analysis_args(parser, args):
    '''Extra validation of those arguments involving creation of ANALYSIS
       rows.'''
    if ((args.analysis_cv == None
         or args.analysis_term == None)
        and (args.analysis_cv != None
             or args.analysis_term != None)):
        parser.error('--analysis-cv and --analysis-term must both be used'
                     ' or neither may be specified'
                     '\nHint: Either or both may be in an'
                     ' environment variable')

def validate_stock_args(parser, args, session):
    if args.organism_id != None:
        if session.query(Organism).get(args.organism_id) == None:
            parser.error('Invalid --organism-id: ({0})'
                         .format(args.organism_id))
    if args.pub_id != None:
        if session.query(Pub).get(args.pub_id) == None:
            parser.error('Invalid --pub-id: ({0})'.format(args.pub_id))

def add_version_arg(arg_group):
    '''Add a --version argument'''
    arg_group.add_argument(
        '-v', '--version', action='version'
        , version='%(prog)s: Babase-Chado {0}'.format(__version__))

class ChadoLookup():
    '''Encapsulate state and provide methods to lookup ids in Chado.'''

    def __init__(self, args=None, session=None, error_func=None):
        '''Parameters:

        ``args``
        An Argparse argument object.

        ``session``
        An SQLAlchemy session.

        ``parse_error``
        A function of one argument, ``msg``, to call when there's an
        error looking up a value.  ``msg`` is the error message.
        '''
        self.args = args
        self.session = session
        self.parse_error = error_func

    def deconvert_dest(self, dest):
        '''Given an argparse ``dest```, returns the argument option'''
        return '--{0}'.format(dest.replace('_','-'))

    def by_value(self, table, value, arg):
        '''Return the ``table`` db object with the table.name of the
        supplied ``value``.

        Parameters:

        ``table``
        An SQLAlchemy table class.

        ``value``
        The value to match against ``table``.name.

        ``arg``
        An argparse arg attribute name.
        '''
        try:
            obj = (self.session.query(table)
                   .filter_by(name = value)
                   .one())
        except NoResultFound:
            self.parse_error('The supplied {0} ({1}) does not exist'
                             .format(self.deconvert_dest(arg), value))
        return obj

    def by_argname(self, table, arg):
        '''Return the ``table`` db object with the table.name of the
        supplied ``arg``.

        Parameters:

        ``table``
        An SQLAlchemy table class.

        ``arg``
        An argparse arg attribute name.
        '''
        value = getattr(self.args, arg)
        if value == None:
            obj = None
        else:
            obj = self.by_value(table, value, arg)
        return obj

    def cv_arg(self, arg):
        '''Return the CV object with the cv.name of the supplied arg

        Parameters:

        ``arg``
        An argparse arg attribute name.
        '''
        return self.by_argname(CV, arg)

    def db_arg(self, arg):
        '''Return the DB object with the db.name of the supplied arg

        Parameters:

        ``arg``
        An argparse arg attribute name.
        '''
        return self.by_argname(DB, arg)

    def cvterm_value(self, cv, value, arg, is_rel=False):
        '''Return a cvterm in the given ``cv``, by term name.

        Parameters:

        ``cv``
        An SQLAlchemy instance of a CV.

        ``value``
        The name to lookup.  This is the value of the ``arg`` argument.

        ``arg``
        An argparse arg attribute name.

        ``is_rel``
        Boolean.  True when the searched for term is a Chado relationship
        term (when
        CVTERM.Is_Relationshiptype <> BB_CVTERM_NOT_RELATIONSHIPTYPE).
        '''
        try:
            query = cv.cvterms.filter_by(name = value)
            if is_rel:
                query = query.filter(CVTerm.is_relationshiptype
                                     != BB_CVTERM_NOT_RELATIONSHIPTYPE)
            return query.one()
        except NoResultFound:
            if is_rel:
                self.parse_error(('The supplied {0} ({1}) does not exist'
                                  ' in the {2} CV as a is_relationshiptype'
                                  ' term')
                                 .format(self.deconvert_dest(arg)
                                         , value, cv.name))
            else:
                self.parse_error(('The supplied {0} ({1}) does not exist'
                                  ' in the {2} CV')
                                 .format(self.deconvert_dest(arg)
                                         , value, cv.name))

    def cvterm_arg(self, cv, arg):
        '''Return a cvterm in the given ``cv``, by name, from an argument

        Parameters:

        ``cv``
        An SQLAlchemy instance of a CV.

        ``arg``
        An argparse arg attribute name.
        '''
        value = getattr(self.args, arg)
        if value == None:
            return None
        else:
            return self.cvterm_value(cv, value, arg)

    def relationship_cvterm_arg(self, cv, arg):
        '''Return a cvterm in the given cv that is a relationship,
        from an argument.

        Parameters:

        ``cv``
        An SQLAlchemy instance of a CV.

        ``arg``
        An argparse arg attribute name.
        '''
        value = getattr(self.args, arg)
        if value == None:
            return None
        else:
            return self.cvterm_value(cv, value, arg, True)

    def dbxref_arg(self, db, arg):
        '''Return a dbxref in ``db``, by accession.

        Parameters:

        ``db``
        An SQLAlchemy DB instance.

        ``arg``
        An argparse arg attribute name.
        '''
        value = getattr(self.args, arg)
        if value == None:
            obj = None
        else:
            try:
                obj = db.dbxrefs.filter_by(accession = value).one()
            except NoResultFound:
                self.parse_error(('The supplied {0} ({1}) does not exist'
                                  ' in the {2} DB')
                                 .format(self.deconvert_dest(arg)
                                         , value, db.name))
        return obj

class CollectAction(argparse.Action):
    '''An argparse Action class for collecting repeated occurences
    of an argument into a list.
    '''
    def __init__(self, option_strings, dest, nargs=None, const=None
                 , default=None, type=None, choices=None
                 , required=False, help=None, metavar=None):
        if default == None:
            # Setting default here makes new list every time.
            default =[]
        super().__init__(option_strings, dest, nargs, const
                       , default, type, choices
                       , required, help, metavar)
    def __call__(self, parser, namespace, values, option_strings=None):
        getattr(namespace, self.dest).append(values)

