# Copyright (C) 2014 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
import argparse, psycopg2
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

class PgConnectionAction(argparse.Action):
    '''An argparse Action class for creating a psycopg2
       connection dict'''
    def __init__(self, option_strings, dest, nargs=None, const=None
                 , default={}, type=None, choices=None
                 , required=False, help=None, metavar=None):
        super().__init__(option_strings, dest, nargs, const
                       , default, type, choices
                       , required, help, metavar)
    def __call__(self, parser, namespace, values, option_strings=None):
        if (hasattr(namespace, self.dest)
            and getattr(namespace, self.dest) != None):
            data = getattr(namespace, self.dest)
        else:
            data = {}

        if option_strings == '--username':
            data['user'] = values
        else:
            data[option_strings[2:]] = values

        setattr(namespace, self.dest, data)

def add_arguments(parser):
    '''Add the args babase needs to connect to pg to the parser'''
    parser.add_argument(
            '--dbname', action=PgConnectionAction, dest='pg_conn_parms'
            , metavar='dbname'
            , help='Database name or service connection string')
    parser.add_argument(
            '--host', action=PgConnectionAction, dest='pg_conn_parms'
            , metavar='hostname'
            , help='Database server host or socket directory')
    parser.add_argument(
            '--port', action=PgConnectionAction, dest='pg_conn_parms'
            , metavar='port'
            , help='Database server port')
    parser.add_argument(
            '--username', action=PgConnectionAction, dest='pg_conn_parms'
            , metavar='user'
            , help='Database user name')


def connect(parser, args):
    '''Connect to pg with psycopg2, return connection'''
    conn_args = args.pg_conn_parms
    try:
        if 'dbname' in conn_args:
            if ('=' in conn_args['dbname']
                or conn_args['dbname'].lstrip()[:13].lower()
                   == 'postgresql://'):
                if len(conn_args) > 1:
                    k = list(conn_args.keys())
                    k.remove('dbname')
                    parser.error(('--dbname cannot contain a connection'
                                  ' string when other (i.e. {0})'
                                  ' pg connection arguments are supplied'
                                 ).format(deconvert_dest(k[0])))
                return psycopg2.connect(conn_args['dbname'])
            else:
                conn_args['database'] = conn_args['dbname']
                del conn_args['dbname']
                return psycopg2.connect(**conn_args)
        else:
            if conn_args:
                return psycopg2.connect(**conn_args)
            else:
                return psycopg2.connect('')
    except psycopg2.OperationalError as e:
        parser.error(e)

def my_session(parser, args):
    '''Return an SQLAlchemy session'''

    def my_connector():
        '''Connect to pg with psycopg2'''
        return connect(parser, args)

    #engine = create_engine('postgresql://', creator=my_connector, echo=True)
    engine = create_engine('postgresql://', creator=my_connector)
    session_factory = sessionmaker(bind=engine)
    return session_factory()

def finish(args, session):
    '''Finish up the session as appropriate.'''
    if args.dry_run:
        session.rollback()
    else:
        session.commit()
