# Copyright (C) 2014, 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
# Remarks:
#   SQLAlchemy ORM mappings of selected Chado tables and other db objects.
#
# Bugs:

# Whole bunch 'o mapping stuff
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Sequence
from sqlalchemy.dialects.postgresql import *

Base = declarative_base()

class DB(Base):
    __tablename__ = 'db'

    db_id = Column(INTEGER, primary_key=True)
    name = Column(VARCHAR(255), unique=True)
    description = Column(VARCHAR(255))
    urlprefix = Column(VARCHAR(255))
    url = Column(VARCHAR(255))

    dbxrefs = relationship('DBXref', lazy='dynamic', passive_deletes=True)

class DBXref(Base):
    __tablename__ = 'dbxref'

    dbxref_id = Column(INTEGER, primary_key=True)
    db_id = Column(INTEGER, ForeignKey('db.db_id'))
    accession = Column(VARCHAR(255))
    version = Column(VARCHAR(255))
    description = Column(TEXT)

    db = relationship('DB')
    cvterms = relationship('CVTerm', lazy='dynamic', passive_deletes=True)
    features = relationship('Feature', lazy='dynamic', passive_deletes=True)
    feature_dbxrefs = relationship('Feature_DBXref', lazy='dynamic'
                                   , passive_deletes=True)
    stocks = relationship('Stock', lazy='dynamic', passive_deletes=True)
    stock_dbxrefs = relationship(
        'Stock_DBXref', lazy='dynamic', passive_deletes=True)

class CV(Base):
    __tablename__ = 'cv'

    cv_id = Column(INTEGER, primary_key=True)
    name = Column(VARCHAR(255), unique=True)
    definition = Column(TEXT)

    cvterms = relationship('CVTerm', lazy='dynamic', passive_deletes=True)

class CVTerm(Base):
    __tablename__ = 'cvterm'

    cvterm_id = Column(INTEGER, primary_key=True)
    cv_id = Column(INTEGER, ForeignKey('cv.cv_id'))
    name = Column(VARCHAR(1024))
    definition = Column(TEXT)
    dbxref_id = Column(INTEGER, ForeignKey('dbxref.dbxref_id'), unique=True)
    is_obsolete = Column(INTEGER)
    is_relationshiptype = Column(INTEGER)

    cv = relationship('CV')
    dbxref = relationship('DBXref')
    analysisprops = relationship('Analysisprop'
                                 , lazy='dynamic', passive_deletes=True)
    analysis_cvterms = relationship(
        'Analysis_CVTerm', lazy='dynamic', passive_deletes=True)
    features = relationship('Feature', lazy='dynamic', passive_deletes=True)
    analysisfeatureprops = relationship(
        'Analysisfeatureprop', lazy='dynamic', passive_deletes=True)
    feature_relationships = relationship(
        'Feature_Relationship', lazy='dynamic', passive_deletes=True)
    stocks = relationship(
        'Stock', lazy='dynamic', passive_deletes=True)
    stock_cvterms = relationship(
        'Stock_CVTerm', lazy='dynamic', passive_deletes=True)
    stockprops = relationship(
        'Stockprop', lazy='dynamic', passive_deletes=True)
    stock_dbxrefprops = relationship(
        'Stock_DBXrefprop', lazy='dynamic', passive_deletes=True)
    stock_relationships = relationship(
        'Stock_Relationship', lazy='dynamic', passive_deletes=True)
    nd_experiments = relationship(
        'ND_Experiment', lazy='dynamic', passive_deletes=True)
    nd_experiment_stocks = relationship(
        'ND_Experiment_Stock', lazy='dynamic', passive_deletes=True)
    analysis_nd_experiments = relationship(
        'Analysis_ND_Experiment', lazy='dynamic', passive_deletes=True)
    nd_experimentprops = relationship(
        'ND_Experimentprop', lazy='dynamic', passive_deletes=True)
    pubs = relationship(
        'Pub', lazy='dynamic', passive_deletes=True)

    types = relationship(
        'CVTerm_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='CVTerm_Relationship.type_id==CVTerm.cvterm_id')
    subjects = relationship(
        'CVTerm_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='CVTerm_Relationship.subject_id==CVTerm.cvterm_id')
    objects = relationship(
        'CVTerm_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='CVTerm_Relationship.object_id==CVTerm.cvterm_id')

class CVTerm_Relationship(Base):
    __tablename__ = 'cvterm_relationship'

    cvterm_relationship_id = Column(INTEGER, primary_key=True)
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    subject_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    object_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))

    type = relationship(
        'CVTerm'
        , primaryjoin='CVTerm.cvterm_id==CVTerm_Relationship.type_id')
    subject = relationship(
        'CVTerm'
        , primaryjoin='CVTerm.cvterm_id==CVTerm_Relationship.subject_id')
    object = relationship(
        'CVTerm'
        , primaryjoin='CVTerm.cvterm_id==CVTerm_Relationship.object_id')

class Organism(Base):
    __tablename__ = 'organism'

    organism_id = Column(INTEGER, primary_key=True)
    abbreviation = Column(VARCHAR(255))
    genus = Column(VARCHAR(255))
    species = Column(VARCHAR(255))
    common_name = Column(VARCHAR(255))
    comment = Column(TEXT)

    features = relationship('Feature', lazy='dynamic', passive_deletes=True)
    stocks = relationship('Stock', lazy='dynamic', passive_deletes=True)

analysis_analysis_id_seq = Sequence('analysis_analysis_id_seq', optional=True)

class Analysis(Base):
    __tablename__ = 'analysis'

    analysis_id = Column(INTEGER, primary_key=True)
    name = Column(VARCHAR(255))
    description = Column(TEXT)
    program = Column(VARCHAR(255))
    programversion = Column(VARCHAR(255))
    algorithm = Column(VARCHAR(255))
    sourcename = Column(VARCHAR(255))
    sourceversion = Column(VARCHAR(255))
    sourceuri = Column(TEXT)
    timeexecuted = Column(TIMESTAMP)
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))

    analysisprops = relationship('Analysisprop', passive_deletes=True)
    analysis_cvterms = relationship('Analysis_CVTerm', passive_deletes=True)
    analysisfeatures = relationship('Analysisfeature', passive_deletes=True)
    type = relationship('CVTerm')
    analysis_nd_experiments = relationship(
        'Analysis_ND_Experiment', lazy='dynamic', passive_deletes=True)

class Analysisprop(Base):
    __tablename__ = 'analysisprop'

    analysisprop_id = Column(INTEGER, primary_key=True)
    analysis_id = Column(INTEGER, ForeignKey('analysis.analysis_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    analysis = relationship('Analysis')
    cvterm = relationship('CVTerm')

class Analysis_CVTerm(Base):
    __tablename__ = 'analysis_cvterm'

    analysis_cvterm_id = Column(INTEGER, primary_key=True)
    analysis_id = Column(INTEGER, ForeignKey('analysis.analysis_id'))
    cvterm_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    is_not = Column(BOOLEAN)

    analysis = relationship('Analysis')
    cvterm = relationship('CVTerm')

feature_feature_id_seq = Sequence('feature_feature_id_seq', optional=True)

class Feature(Base):
    __tablename__ = 'feature'

    feature_id = Column(INTEGER, feature_feature_id_seq, primary_key=True)
    dbxref_id = Column(INTEGER, ForeignKey('dbxref.dbxref_id'))
    organism_id = Column(INTEGER, ForeignKey('organism.organism_id'))
    name = Column(VARCHAR(255))
    uniquename = Column(TEXT)
    residues = Column(TEXT)
    seqlen = Column(INTEGER)
    md5checksum = Column(CHAR(32))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    is_analysis = Column(BOOLEAN)
    is_obsolete = Column(BOOLEAN)
    timeaccessioned = Column(TIMESTAMP)
    timelastmodified =Column(TIMESTAMP)

    dbxref = relationship('DBXref')
    organism = relationship('Organism')
    cvterm = relationship('CVTerm')
    featurelocs = relationship(
        'Featureloc', passive_deletes=True
        , primaryjoin='Featureloc.feature_id == Feature.feature_id')
    srcfeaturelocs = relationship(
        'Featureloc', passive_deletes=True, lazy="dynamic"
        , primaryjoin='Featureloc.srcfeature_id == Feature.feature_id')
    feature_dbxrefs = relationship('Feature_DBXref', passive_deletes=True)
    analysisfeatures = relationship('Analysisfeature', passive_deletes=True)
    subjects = relationship(
        'Feature_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='Feature_Relationship.subject_id==Feature.feature_id')
    objects = relationship(
        'Feature_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='Feature_Relationship.object_id==Feature.feature_id')

class Featureloc(Base):
    __tablename__ = 'featureloc'

    featureloc_id = Column(INTEGER, primary_key=True)
    feature_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    srcfeature_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    fmin = Column(INTEGER)
    is_fmin_partial = Column(BOOLEAN)
    fmax = Column(INTEGER)
    is_fmax_partial = Column(BOOLEAN)
    strand = Column(SMALLINT)
    phase = Column(INTEGER)
    residue_info = Column(TEXT)
    locgroup = Column(INTEGER)
    rank = Column(INTEGER)

    feature = relationship(
        'Feature'
        , primaryjoin='Feature.feature_id == Featureloc.feature_id')
    srcfeature = relationship(
        'Feature'
        , primaryjoin='Feature.feature_id == Featureloc.srcfeature_id')

class Feature_Relationship(Base):
    __tablename__ = 'feature_relationship'

    feature_relationship_id = Column(INTEGER, primary_key=True)
    subject_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    object_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    subject = relationship(
        'Feature'
        , primaryjoin='Feature.feature_id==Feature_Relationship.subject_id')
    object = relationship(
        'Feature'
        , primaryjoin='Feature.feature_id==Feature_Relationship.object_id')
    cvterm = relationship('CVTerm')

class Feature_DBXref(Base):
    __tablename__ = 'feature_dbxref'

    feature_dbxref_id = Column(INTEGER, primary_key=True)
    feature_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    dbxref_id = Column(INTEGER, ForeignKey('dbxref.dbxref_id'))
    is_current = Column(BOOLEAN)

    # It'd be nice to have this be "eager" but sqlalchemy 0.7.9 is
    # not smart enough to know that one end is already loaded.
    #feature = relationship('Feature', lazy='joined', innerjoin=True)
    #dbxref = relationship('DBXref', lazy='joined', innerjoin=True)
    feature = relationship('Feature')
    dbxref = relationship('DBXref')

class Analysisfeature(Base):
    __tablename__ = 'analysisfeature'

    analysisfeature_id = Column(INTEGER, primary_key=True)
    feature_id = Column(INTEGER, ForeignKey('feature.feature_id'))
    analysis_id = Column(INTEGER, ForeignKey('analysis.analysis_id'))
    rawscore = Column(DOUBLE_PRECISION)
    normscore = Column(DOUBLE_PRECISION)
    significance = Column(DOUBLE_PRECISION)
    identity = Column(DOUBLE_PRECISION)

    feature = relationship('Feature')
    analysis = relationship('Analysis')
    analysisfeatureprops = relationship('Analysisfeatureprop'
                                        , passive_deletes=True)

class Analysisfeatureprop(Base):
    __tablename__ = 'analysisfeatureprop'

    analysisfeatureprop_id = Column(INTEGER, primary_key=True)
    analysisfeature_id = Column(
        INTEGER
        , ForeignKey('analysisfeature.analysisfeature_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    analysisfeature = relationship('Analysisfeature')
    cvterm = relationship('CVTerm')

class Pub(Base):
    __tablename__ = 'pub'

    pub_id = Column(INTEGER, primary_key=True)

    title = Column(TEXT)
    volumetitle = Column(TEXT)
    volume = Column(VARCHAR(255))
    series_name = Column(VARCHAR(255))
    issue = Column(VARCHAR(255))
    pyear = Column(VARCHAR(255))
    pages = Column(VARCHAR(255))
    miniref = Column(VARCHAR(255))
    uniquename = Column(TEXT)
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    is_obsolete = Column(BOOLEAN)
    publisher = Column(VARCHAR(255))
    pubplace = Column(VARCHAR(255))

    cvterm = relationship('CVTerm')
    stock_cvterms = relationship(
        'Stock_CVTerm', lazy="dynamic", passive_deletes=True)

class Stock(Base):
    __tablename__ = 'stock'

    stock_id = Column(INTEGER, primary_key=True)
    dbxref_id = Column(INTEGER, ForeignKey('dbxref.dbxref_id'))
    organism_id = Column(INTEGER, ForeignKey('organism.organism_id'))
    name = Column(VARCHAR(255))
    uniquename = Column(TEXT)
    description = Column(TEXT)
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    is_obsolete = Column(BOOLEAN)

    dbxref = relationship('DBXref')
    organism = relationship('Organism')
    cvterm = relationship('CVTerm')
    nd_experiment_stocks = relationship(
        'ND_Experiment_Stock', lazy='dynamic', passive_deletes=True)
    stock_cvterms = relationship('Stock_CVTerm', passive_deletes=True)
    stockprops = relationship('Stockprop', passive_deletes=True)
    stock_dbxrefs = relationship('Stock_DBXref', passive_deletes=True)
    subjects = relationship(
        'Stock_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='Stock_Relationship.subject_id==Stock.stock_id')
    objects = relationship(
        'Stock_Relationship', passive_deletes=True, lazy="dynamic"
        , primaryjoin='Stock_Relationship.object_id==Stock.stock_id')

stock_stock_id_seq = Sequence('stock_stock_id_seq', optional=True)

class Stock_CVTerm(Base):
    __tablename__ = 'stock_cvterm'

    stock_cvterm_id = Column(INTEGER, primary_key=True)
    stock_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    cvterm_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    pub_id = Column(INTEGER, ForeignKey('pub.pub_id'))
    is_not = Column(INTEGER)

    stock = relationship('Stock')
    cvterm = relationship('CVTerm')
    pub = relationship('Pub')

class Stockprop(Base):
    __tablename__ = 'stockprop'

    stockprop_id = Column(INTEGER, primary_key=True)
    stock_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    stock = relationship('Stock')
    cvterm = relationship('CVTerm')

class Stock_DBXref(Base):
    __tablename__ = 'stock_dbxref'

    stock_dbxref_id = Column(INTEGER, primary_key=True)
    stock_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    dbxref_id = Column(INTEGER, ForeignKey('dbxref.dbxref_id'))
    is_current = Column(BOOLEAN)

    stock = relationship('Stock')
    dbxref = relationship('DBXref')
    stock_dbxrefprops = relationship('Stock_DBXrefprop', passive_deletes=True)

class Stock_DBXrefprop(Base):
    __tablename__ = 'stock_dbxrefprop'

    stock_dbxrefprop_id = Column(INTEGER, primary_key=True)
    stock_dbxref_id = Column(
        INTEGER
        , ForeignKey('stock_dbxref.stock_dbxref_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    stock_dbxref = relationship('Stock_DBXref')
    cvterm = relationship('CVTerm')

class Stock_Relationship(Base):
    __tablename__ = 'stock_relationship'

    stock_relationship_id = Column(INTEGER, primary_key=True)
    subject_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    object_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    subject = relationship(
        'Stock'
        , primaryjoin='Stock.stock_id==Stock_Relationship.subject_id')
    object = relationship(
        'Stock'
        , primaryjoin='Stock.stock_id==Stock_Relationship.object_id')
    cvterm = relationship('CVTerm')

class ND_Geolocation(Base):
    __tablename__ = 'nd_geolocation'
    
    nd_geolocation_id = Column(INTEGER, primary_key=True)
    description = Column(VARCHAR(255))
    latitude = Column(REAL)
    longitude = Column(REAL)
    geodetic_datum = Column(VARCHAR(32))
    altitude = Column(REAL)

    nd_experiments = relationship(
        'ND_Experiment', lazy='dynamic', passive_deletes=True)

nd_experiment_nd_experiment_id_seq = Sequence(
    'nd_experiment_nd_experiment_id_seq', optional=True)

class ND_Experiment(Base):
    __tablename__ = 'nd_experiment'

    nd_experiment_id = Column(INTEGER, primary_key=True)
    nd_geolocation_id = Column(
        INTEGER
        , ForeignKey('nd_geolocation.nd_geolocation_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))

    nd_geolocation = relationship('ND_Geolocation')
    cvterm = relationship('CVTerm')
    nd_experiment_stocks = relationship(
        'ND_Experiment_Stock', lazy='dynamic', passive_deletes=True)
    analysis_nd_experiments = relationship(
        'Analysis_ND_Experiment', lazy='dynamic', passive_deletes=True)
    nd_experimentprops = relationship(
        'ND_Experimentprop', lazy='dynamic', passive_deletes=True)

class ND_Experiment_Stock(Base):
    __tablename__ = 'nd_experiment_stock'

    nd_experiment_stock_id = Column(INTEGER, primary_key=True)
    nd_experiment_id = Column(
        INTEGER
        , ForeignKey('nd_experiment.nd_experiment_id'))
    stock_id = Column(INTEGER, ForeignKey('stock.stock_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))

    nd_experiment = relationship('ND_Experiment')
    stock = relationship('Stock')
    cvterm = relationship('CVTerm')

class Analysis_ND_Experiment(Base):
    __tablename__ = 'analysis_nd_experiment'

    analysis_nd_experiment_id = Column(INTEGER, primary_key=True)
    analysis_id = Column(INTEGER, ForeignKey('analysis.analysis_id'))
    nd_experiment_id = Column(
        INTEGER
        , ForeignKey('nd_experiment.nd_experiment_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))

    analysis = relationship('Analysis')
    nd_experiment = relationship('ND_Experiment')
    cvterm = relationship('CVTerm')

class ND_Experimentprop(Base):
    __tablename__ = 'nd_experimentprop'

    nd_experimentprop_id = Column(INTEGER, primary_key=True)
    nd_experiment_id = Column(INTEGER
                              , ForeignKey('nd_experiment.nd_experiment_id'))
    type_id = Column(INTEGER, ForeignKey('cvterm.cvterm_id'))
    value = Column(TEXT)
    rank = Column(INTEGER)

    nd_experiment = relationship('ND_Experiment')
    cvterm = relationship('CVTerm')
