dnl Copyright (C) 2012, 2013 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Copy data from Catherine's swerb table to Babase's swerb tables.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

CREATE OR REPLACE FUNCTION swerb_copy(tbl TEXT, fromdate DATE, todate DATE)
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Copies swerb data from Catherine's table into babase.
  --
  -- GPL_notice(`  --', `2012, 2013', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: swerb_copy(table, fromdate, todate)
  --
  -- Arguments:
  --  tbl     Table to copy from.  Must have same datastructure as
  --          amarkham.swerb20120808
  --  fromdate  Date to start copy (inclusive)
  --  todate    Date to stop copy (inclusive)
  --
  -- Remarks:
  --   One-off program.  Once it works we won't run it again.
  --
  --   Copies a day a a time, using a fictious departure row.
  --
  --   Stops as soon as a transaction fails.
  --
  --   Catherine's docs say that only begin/end/watherhole events
  --  can have locations, but this is not true.  In these cases
  --  we associate the ('H') rows with locations anyway.
  --
  --   Catherine's docs say that begin/end/waterhole events
  --  must have locations, but this is not true.
  --  We let the Babase integrity rules worry about this.
  --
  --   Although Catherine's docs say otherwise you can have
  --  W events (drinking events) and no related physical landscape
  --  feature.  (swerb_loc_data row)  (This is fine with Babase,
  --  it does not require the existance of the unknown waterhole.)
  --
  -- Returns:
  --   Number of days processed before error.
  --
  -- Bugs:
  --   Trying to copy a day where 2 teams observe the same group
  -- will likely (but may not) break because of the requirements
  -- for a single sleeping grove (unless there's subgroups).
  --
  -- The solution is to copy the relevant rows from the source
  -- table into a temporary table and copy from there, once for
  -- each observation team.
  --

  DECLARE
    this_date DATE;
    days      INT := 0;
    c         INT;

    this_row RECORD;  -- Why tie us down to a particular table?

    half_be INT;

  BEGIN
    IF (todate < fromdate) THEN
      RAISE EXCEPTION 'Fromdate %: Todate %: Todate is before Fromdate'
                    , fromdate
                    , todate
                      USING ERRCODE='SWERB';
      RETURN days;
    END IF;

    this_date := fromdate;
    BEGIN
      WHILE this_date <= todate LOOP

        EXECUTE 'SELECT COUNT(*)
                   FROM ' || tbl::regclass || ' AS tbl
                   WHERE tbl.date = $1
                   LIMIT 1'
          INTO c
          USING this_date;
        IF c > 0 THEN
          -- There's something to be done for this date.

          half_be := NULL;

          DECLARE
            last_grp swerb_bes.focal_grp%TYPE;
            departure_created BOOLEAN := FALSE;

            this_beid swerb_bes.beid%TYPE;
            this_timesource swerb_bes.bsource%TYPE; -- and esource too.

            this_event swerb_data.event%TYPE;
            this_subgroup swerb_data.subgroup%TYPE;
            this_xyloc swerb_data.xyloc%TYPE;
            this_accuracy swerb_data.accuracy%TYPE;
            this_pdop swerb_data.pdop%TYPE;
            this_gps_datetime swerb_data.gps_datetime%TYPE;

            this_adcode swerb_loc_data.adcode%TYPE;
            this_conf swerb_loc_data.conf%TYPE;

            some_timeest BOOLEAN;
            make_beginend BOOLEAN;
            focal_grp swerb_bes.focal_grp%TYPE;

            ignore_swid INT;
            slgps_row RECORD;
            make_swerb_loc_gps BOOLEAN;

            saved_grp swerb_bes.focal_grp%TYPE;
          BEGIN

            -- Create all the rows.
            -- Create the other group rows last.           

            FOR this_row
              IN EXECUTE 'SELECT *
                            FROM ' || tbl::regclass || ' AS tbl
                            WHERE tbl.date = $1
                                  AND tbl.event != ''D''
                            ORDER BY CASE
                                       WHEN tbl.event = ''O''
                                            AND tbl.sort_be IS NULL
                                         THEN 2
                                       ELSE
                                         1
                                      END
                                    , tbl.sort_be
                                    , tbl.grp
                                    , tbl.time
                                    , tbl.event
                                    , tbl.swid'
                  USING this_date
              LOOP

              IF this_row.swid = ignore_swid THEN
                -- We've already processed this row as part of a begin
                -- or end row.
                CONTINUE;
              END IF;

              IF this_row.grp IS NULL THEN
                RAISE EXCEPTION 'Null group'
                      USING ERRCODE='SWERB';

              END IF;

              -- Keep track of the last "second begin/end" row we see that
              -- we don't know what to do with and hope that it matches
              -- up with a future begin/end.
              IF saved_grp IS NULL
                 OR saved_grp != this_row.grp THEN
                -- Group changed.
                saved_grp := this_row.grp;

                IF half_be IS NOT NULL THEN
                  -- We saw a row we thought was a begin/end "other half"
                  -- but have not processed it.
                  RAISE EXCEPTION 'Swid %: Found a row that looked like a "second" begin or end row but did not find a begin or end to go with it'
                                , half_be
                      USING ERRCODE='SWERB';
                END IF;

              END IF;

              -- Create a departure row if we need one.
              IF (departure_created AND last_grp != this_row.grp)
                 OR NOT(departure_created) THEN
                -- New group means new departure row.
                INSERT INTO swerb_departs_data (date)
                 VALUES (this_date);

                last_grp := this_row.grp;
                departure_created := TRUE;
              END IF;

              --
              -- Convert from Catherine to Babase.
              --

              -- timeest
              this_timesource := trim(trailing from this_row.sourcetime);
              IF this_timesource IS NULL AND this_row.timeest = 'N' THEN
                -- Babase sez: "Gotta have a source."
                this_timesource := 'S'; -- Catherine's "SWERB" code.
              END IF;

              -- subgroup
              this_subgroup := this_row.subgrp IS NOT NULL;

              -- X Y location.
              IF this_row.coordconvert = 'Y' THEN
                -- The utm coordinates are from a quad conversion.
                -- Babase users get the coordinates from the quads
                -- table (or have it done for them with the swerb_xy view).
                this_xyloc = NULL;
              ELSE
                -- The utm measurements are from gps units.
                -- Keep them.
                this_xyloc = bb_makepoint(this_row.x, this_row.y);
              END IF;

              -- Accuracy and pdop.
              this_accuracy := NULL;
              this_pdop := NULL;
              CASE this_row.errorcode
                WHEN 'P' THEN
                  this_pdop = this_row.error;
                WHEN 'A' THEN
                  this_accuracy = this_row.error;
                ELSE NULL;
              END CASE;
                
              -- Conf
              this_conf := COALESCE(this_row.locdconf, 'bb_swerb_certain');

              -- gps_datetime
              IF this_date < '2004-01-01'::DATE THEN
                this_gps_datetime := NULL;
              ELSE
                IF this_row.garmincode = '' OR this_row.garmincode IS NULL THEN
                  this_gps_datetime := NULL;
                ELSE
                  this_gps_datetime := this_date + this_row.time;
                END IF;
              END IF;

              --
              -- What bout are we using?
              --
              IF this_row.event != 'B' AND this_row.event != 'O' THEN
                this_beid := currval('swerb_bes_beid_seq');
              END IF;

              --
              -- Create rows based on the event code.
              --
              make_beginend := FALSE;
              make_swerb_loc_gps := FALSE;

              CASE this_row.event
                WHEN 'B' THEN
                  -- Begin row.  Create swerb_bes row as well
                  -- as a swerb_data row.

                  -- Create swerb_bes row.
                  INSERT INTO swerb_bes (did
                                      , focal_grp
                                      , start
                                      , btimeest
                                      , bsource
                                      , stop
                                      , etimeest
                                      , esource
                                      , is_effort
                                      , notes)
                    VALUES (currval('swerb_departs_data_did_seq')
                          , this_row.grp
                          , this_row.time
                          , this_row.timeest::boolean
                          , this_timesource
                          , NULL
                          , NULL
                          , NULL
                          , TRUE
                          , NULL);

                  -- Create the swerb_data row. 
                  this_event := 'bb_swerb_begin';
                  this_beid := currval('swerb_bes_beid_seq'::regclass);

                  -- ADcode
                  IF this_row.ascdesccode = 'UNK'
                     OR this_row.ascdesccode = 'TIME' THEN
                    this_adcode = 'bb_swerb_descent_code';
                  ELSE
                    -- Not a value that had to be translated.
                    this_adcode := TRIM(TRAILING FROM this_row.ascdesccode);
                  END IF;

                  IF this_date >= '2004-01-01'::DATE
                     AND this_row.locd IS NOT NULL THEN
                    BEGIN
                      -- Find the second begin row that records ascent/descent.
                      EXECUTE 'SELECT *
                                 FROM ' || tbl::regclass || ' AS tbl
                                 WHERE tbl.date = $1
                                   AND tbl.grp = $2
                                   AND CASE
                                         WHEN $5 IS NULL
                                              AND $4 IS NOT NULL
                                           THEN TRUE
                                         ELSE
                                           SUBSTRING(tbl.garmincode FROM 1
                                                                    FOR 1)
                                           = $3
                                       END
                                   AND (tbl.sort_be = $4 + 1
                                        OR (tbl.sort_be IS NULL
                                            AND $4 IS NULL))
                                   AND tbl.event = ''H''
                                   AND SUBSTRING(tbl.garmincode FROM 2 FOR 3)
                                       = ''MDT'''
                         INTO STRICT slgps_row
                         USING this_date
                             , this_row.grp
                             , SUBSTRING(this_row.garmincode FROM 1 FOR 1)
                             , this_row.sort_be
                             , this_row.garmincode;

                      ignore_swid := slgps_row.swid;
                      make_swerb_loc_gps := TRUE;

                      IF half_be IS NOT NULL
                         AND ignore_swid != half_be THEN
                        RAISE EXCEPTION 'Swid %: Time %: Grp %: Expected to use the saved "second begin/end row" (swid: %) but found a different "second begin/end row" instead'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                                      , half_be
                              USING ERRCODE='SWERB';
                      END IF;
                      half_be := NULL;

                    EXCEPTION
                      WHEN NO_DATA_FOUND THEN
                        -- No second begin row.
                        IF half_be IS NOT NULL THEN
                          RAISE EXCEPTION 'Swid %: Time %: Grp %: Expected to use the saved "second begin/end row" (swid: %) but found no "second begin/end row" to use'
                                        , this_row.swid
                                        , this_row.time
                                        , this_row.grp
                                        , half_be
                                USING ERRCODE='SWERB';
                        END IF;

                        ignore_swid := NULL;
                      WHEN TOO_MANY_ROWS THEN
                        RAISE EXCEPTION 'Swid %: Time %: Grp %: More than one hourly/half-hourly waypoint that''s really a second begin row'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                           USING ERRCODE='SWERB';
                    END;
                 END IF;

                WHEN 'E' THEN
                  -- End row.  Update swerb_bes row, create a swerb_data row.
                  UPDATE swerb_bes
                    SET stop = this_row.time
                      , etimeest = this_row.timeest::boolean
                      , esource = this_timesource
                    WHERE beid = this_beid;

                  -- Create the swerb_data row. 
                  this_event := 'bb_swerb_end';

                  -- ADcode
                  IF this_row.ascdesccode = 'UNK'
                     OR this_row.ascdesccode = 'TIME' THEN
                    this_adcode = 'bb_swerb_ascent_code';
                  ELSE
                    -- Not a value that had to be translated.
                    this_adcode := TRIM(TRAILING FROM this_row.ascdesccode);
                  END IF;

                  IF this_date >= '2004-01-01'::DATE
                     AND this_row.locd IS NOT NULL THEN
                    BEGIN
                      -- Find the second end row that records ascent/descent.
                      EXECUTE 'SELECT *
                                 FROM ' || tbl::regclass || ' AS tbl
                                 WHERE tbl.date = $1
                                   AND tbl.grp = $2
                                   AND CASE
                                         WHEN $5 IS NULL
                                              AND $4 IS NOT NULL
                                           THEN TRUE
                                         ELSE
                                           SUBSTRING(tbl.garmincode FROM 1
                                                                    FOR 1)
                                           = $3
                                       END
                                   AND (tbl.sort_be = $4 + 1
                                        OR (tbl.sort_be IS NULL
                                            AND $4 IS NULL))
                                   AND tbl.event = ''H''
                                   AND SUBSTRING(tbl.garmincode FROM 2 FOR 3)
                                       = ''MAT'''
                         INTO STRICT slgps_row
                         USING this_date
                             , this_row.grp
                             , SUBSTRING(this_row.garmincode FROM 1 FOR 1)
                             , this_row.sort_be
                             , this_row.garmincode;

                      ignore_swid := slgps_row.swid;
                      make_swerb_loc_gps := TRUE;

                      IF half_be IS NOT NULL
                         AND ignore_swid != half_be THEN
                        RAISE EXCEPTION 'Swid %: Time %: Grp %: Expected to use the saved "second begin/end row" (swid: %) but found a different "second begin/end row" instead'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                                      , half_be
                              USING ERRCODE='SWERB';
                      END IF;
                      half_be := NULL;

                    EXCEPTION
                      WHEN NO_DATA_FOUND THEN
                        -- No second end row.
                        IF half_be IS NOT NULL THEN
                          RAISE EXCEPTION 'Swid %: Time %: Grp %: Expected to use the saved "second begin/end row" (swid: %) but found no "second begin/end row" to use'
                                        , this_row.swid
                                        , this_row.time
                                        , this_row.grp
                                        , half_be
                              USING ERRCODE='SWERB';
                        END IF;

                        ignore_swid := NULL;
                      WHEN TOO_MANY_ROWS THEN
                        RAISE EXCEPTION 'Swid %: Time %: Grp %: More than one hourly/half-hourly waypoint that''s really a second end row'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                              USING ERRCODE='SWERB';
                    END;
                  END IF;

                WHEN 'H' THEN
                  -- Regular swerb waypoint.

                  IF this_row.garmincode ~'^[[:upper:]]M[AD]T' THEN
                    -- The garmincode says it's a begin or end.
                    IF half_be IS NOT NULL THEN
                      RAISE EXCEPTION 'Swid %: Time %, Grp %: Hourly/halfhourly waypoint that looks like MAT or MDT begin/end data but there is already a pending row (%) of this sort'
                                    , this_row.swid
                                    , this_row.time
                                    , this_row.grp
                                    , half_be
                              USING ERRCODE='SWERB';
                     END IF;

                     -- Save the row and hope to match it with a 
                     -- begin/end later.
                     half_be := this_row.swid;
                     CONTINUE;

                  END IF;

                  -- Create swerb data row.
                  this_event := 'bb_swerb_obs';

                  this_adcode = 'bb_swerb_notad';

                WHEN 'O' THEN
                  -- Other group observation.  Create a swerb_data row.
                  this_event = 'bb_swerb_ogrp';

                  this_adcode = 'bb_swerb_notad';

                  IF this_row.sort_be IS NULL THEN
                    -- To what bout of observation does it belong?
                    -- The swerb_bes has already been created with a begin/end.
                    DECLARE
                      c INT;

                    BEGIN
                      SELECT swerb_bes.beid
                        INTO this_beid
                        FROM swerb_departs_data
                             JOIN swerb_bes
                                  ON (swerb_bes.did = swerb_departs_data.did)
                        WHERE swerb_departs_data.date = this_date
                              AND (swerb_bes.start IS NULL
                                   OR swerb_bes.start <= this_row.time)
                              AND (swerb_bes.stop IS NULL
                                   OR this_row.time <= swerb_bes.stop);
                   
                      GET DIAGNOSTICS c = ROW_COUNT;

                      IF c = 0 THEN
                        -- Make a begin/end row pair for the other 
                        -- group event.
                        make_beginend := TRUE;
                        focal_grp := 'bb_unk_grp';
                      ELSIF c > 1 THEN
                        RAISE EXCEPTION 'Swid %: Time %, Grp %: Unable to determine which bout of observation the observation of the non-focal group should be asscociated with'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                              USING ERRCODE='SWERB';
                      END IF;
                    END;
                  END IF;
                  
                WHEN 'W' THEN
                  -- Drinking event.

                  -- See if we're in an ongoing bout of observation.
                  PERFORM 1
                    FROM swerb_bes
                         JOIN swerb_departs_data
                              ON (swerb_departs_data.did = swerb_bes.did)
                         LEFT OUTER JOIN swerb_data
                                         ON (swerb_data.beid = swerb_bes.beid)
                    WHERE swerb_bes.beid = currval('swerb_bes_beid_seq')
                          AND swerb_departs_data.date = this_date
                          AND ((swerb_data.event = 'bb_swerb_end'
                                AND swerb_data.time >= this_row.time)
                               OR NOT EXISTS (SELECT 1
                                                FROM swerb_data
                                                WHERE swerb_data.beid
                                                      = swerb_bes.beid
                                                      AND swerb_data.event
                                                          = 'bb_swerb_end'))
                    LIMIT 1;

                  IF NOT FOUND THEN
                    -- No ongoing bout of observation.
                    -- Create a zero-length, does not count toward
                    -- observer effort, bout of observation for the
                    -- drinking event.
                    make_beginend := TRUE;
                    focal_grp := this_row.grp;
                  END IF;

                  --  Create swerb_data row
                  -- and swerb_loc_data row to place at a waterhole.
                  this_event := 'bb_swerb_water';
                
                  this_adcode = 'bb_swerb_notad';

              ELSE
                RAISE EXCEPTION 'Swid %: Time %: Grp %: Event % : Unrecognized event'
                              , this_row.swid
                              , this_row.time
                              , this_row.grp
                              , this_row.event
                              USING ERRCODE='SWERB';

              END CASE;

              IF make_beginend THEN
                IF this_row.time IS NOT NULL
                   AND this_row.timeest IS NULL THEN
                  some_timeest := FALSE;
                ELSE
                  some_timeest := this_row.timeest::boolean;
                END IF;

                -- timesource
                this_timesource := trim(trailing from this_row.sourcetime);
                IF this_timesource IS NULL AND NOT(some_timeest) THEN
                  -- Babase sez: "Gotta have a source."
                  this_timesource := 'S'; -- Catherine's "SWERB" code.
                END IF;

                -- Create the swerb_bes row.
                INSERT INTO swerb_bes (did
                                    , focal_grp
                                    , start
                                    , btimeest
                                    , bsource
                                    , stop
                                    , etimeest
                                    , esource
                                    , is_effort
                                    , notes)
                  VALUES (currval('swerb_departs_data_did_seq')
                        , focal_grp
                        , this_row.time
                        , some_timeest
                        , this_timesource
                        , this_row.time
                        , some_timeest
                        , this_timesource
                        , FALSE
                        , NULL);

                this_beid := currval('swerb_bes_beid_seq'::regclass);

                -- Create a swerb_data row for the begin.
                INSERT INTO swerb_data(beid
                                    , seen_grp
                                    , event
                                    , time
                                    , quad
                                    , xyloc
                                    , altitude
                                    , pdop
                                    , accuracy
                                    , subgroup
                                    , ogdistance
                                    , gps_datetime
                                    , garmincode)
                  VALUES (this_beid
                        , focal_grp
                        , 'bb_swerb_begin'
                        , this_row.time
                        , this_row.quad
                        , this_xyloc
                        , this_row.altitude
                        , this_pdop
                        , this_accuracy
                        , this_subgroup
                        , NULL
                        , this_gps_datetime
                        , this_row.garmincode);                          

                -- Create a swerb end row for the end.
                INSERT INTO swerb_data(beid
                                    , seen_grp
                                    , event
                                    , time
                                    , quad
                                    , xyloc
                                    , altitude
                                    , pdop
                                    , accuracy
                                    , subgroup
                                    , ogdistance
                                    , gps_datetime
                                    , garmincode)
                  VALUES (this_beid
                        , focal_grp
                        , 'bb_swerb_end'
                        , this_row.time
                        , this_row.quad
                        , this_xyloc
                        , this_row.altitude
                        , this_pdop
                        , this_accuracy
                        , this_subgroup
                        , NULL
                        , this_gps_datetime
                        , this_row.garmincode);                          
              END IF;

              -- Create the swerb_data row. 
              INSERT INTO swerb_data(beid
                                  , seen_grp
                                  , event
                                  , time
                                  , quad
                                  , xyloc
                                  , altitude
                                  , pdop
                                  , accuracy
                                  , subgroup
                                  , ogdistance
                                  , gps_datetime
                                  , garmincode)
                VALUES (this_beid
                      , this_row.grp
                      , this_event
                      , this_row.time
                      , this_row.quad
                      , this_xyloc
                      , this_row.altitude
                      , this_pdop
                      , this_accuracy
                      , this_subgroup
                      , this_row.ogdistance
                      , this_gps_datetime
                      , this_row.garmincode);                          

              IF this_row.locd IS NOT NULL THEN
                -- There's a physical landscape feature associated with
                -- this row.  Make a swerb_loc_data row for it.
                INSERT INTO swerb_loc_data (swid, loc, adcode, conf, adtime)
                  VALUES(currval('swerb_data_swid_seq')
                       , this_row.locd
                       , this_adcode
                       , this_conf
                       , this_row.ascdesctime);

                IF make_swerb_loc_gps THEN
                  -- There is a second begin/end row.

                  -- Test to see that the 2nd row's been "cleaned out".
                  IF slgps_row.becode IS NOT NULL
                     OR slgps_row.timeest IS NOT NULL
                     OR slgps_row.sourcetime IS NOT NULL
                     OR slgps_row.locd IS NOT NULL
                     OR slgps_row.locdconf IS NOT NULL
                     OR slgps_row.ascdesccode IS NOT NULL
                     OR slgps_row.ascdesctime IS NOT NULL
                     OR slgps_row.quad IS NOT NULL
                     OR slgps_row.coordconvert IS NOT NULL
                     OR slgps_row.errorcode IS NOT NULL
                     OR slgps_row.error IS NOT NULL
                     OR slgps_row.ogdistance IS NOT NULL THEN
                    RAISE EXCEPTION 'Swid %: Time %, Grp %: There is a second begin/end row (and x/y coordinates for same?) but there is data in this row that will be lost (one of becode, timeest, sourcetime, locd, locdconf, ascdesccode, ascdesctime, quad, coordconvert, errorcode, error, ogdistance is not NULL)'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                              USING ERRCODE='SWERB';
                  END IF;

                  -- Make a swerb_loc_gps row.
                  INSERT INTO swerb_loc_gps (swid
                                           , xyloc
                                           , altitude
                                           , pdop
                                           , accuracy
                                           , gps_datetime
                                           , garmincode)
                         VALUES(currval('swerb_data_swid_seq')
                              , bb_makepoint(slgps_row.x, slgps_row.y)
                              , slgps_row.altitude
                              , CASE slgps_row.errorcode
                                  WHEN 'P' THEN this_row.error
                                  ELSE NULL
                                END
                              , CASE slgps_row.errorcode
                                  WHEN 'A' THEN this_row.error
                                  ELSE NULL
                                END
                              , this_date + slgps_row.time
                              , slgps_row.garmincode);
                END IF;
              ELSE
                IF make_swerb_loc_gps THEN
                  RAISE EXCEPTION 'Swid %: Time %, Grp %: There is a second begin/end row (and x/y coordinates for same?) but there is no ascent/descent grove (locd is NULL)'
                                      , this_row.swid
                                      , this_row.time
                                      , this_row.grp
                              USING ERRCODE='SWERB';
                END IF;
              END IF;

            END LOOP;
          END;

          IF half_be IS NOT NULL THEN
            -- We saw a row we thought was a begin/end "other half"
            -- but have not processed it.
            RAISE EXCEPTION 'Swid %: Found a row that looked like a "second" begin or end row but did not find a begin or end to go with it'
                          , half_be
                              USING ERRCODE='SWERB';
          END IF;

          days := days + 1;

        END IF;

        -- Prep for next iteration.
        this_date := this_date + 1;
      END LOOP;

    EXCEPTION
--      WHEN SQLSTATE 'SWERB' THEN
      WHEN OTHERS THEN
        -- Add the date being processed to any error message that shows up.
        RAISE EXCEPTION 'Date: %: %', this_date, SQLERRM;
    END;

  RETURN days;
  END;
$$;
grant_func_priv(`swerb_copy(TEXT, DATE, DATE)')
