dnl Copyright (C) 2013 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Copy data from Catherine's swerb table to Babase's swerb tables.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

CREATE OR REPLACE FUNCTION swerb_copybe(tbl TEXT, this_swid INT)
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Copies a single begin or end from Catherine's table into babase.
  --
  -- GPL_notice(`  --', `2013', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: swerb_copybe(table, swid)
  --
  -- Arguments:
  --  tbl     Table to copy from.  Must have same datastructure as
  --          amarkham.swerb20120808
  --  swid    Id (swid) of begin or end row to copy
  --
  -- Remarks:
  --   One-off program.  Once it works we won't run it again.
  --
  --   Stops as soon as a transaction fails.
  --
  -- Returns:
  --   Number of days processed before error.
  --
  -- Bugs:
  --

  DECLARE
    this_row RECORD;  -- Why tie us down to a particular table?

    this_did swerb_departs_data.did%TYPE;

    this_timesource swerb_bes.bsource%TYPE; -- and esource too.
    this_subgroup swerb_data.subgroup%TYPE;
    this_xyloc swerb_data.xyloc%TYPE;
    this_accuracy swerb_data.accuracy%TYPE;
    this_pdop swerb_data.pdop%TYPE;
    this_conf swerb_loc_data.conf%TYPE;
    this_gps_datetime swerb_data.gps_datetime%TYPE;

    this_adcode swerb_loc_data.adcode%TYPE;

    this_date DATE;

  BEGIN

    -- Get the row.
    BEGIN
      EXECUTE 'SELECT *
                 FROM ' || tbl::regclass || ' AS tbl
                 WHERE tbl.SWID = $1'
        INTO STRICT this_row
        USING this_swid;
    EXCEPTION
      WHEN TOO_MANY_ROWS THEN
        RAISE EXCEPTION 'SWID %: Duplicate swids'
                      , this_swid;
      WHEN NO_DATA_FOUND THEN
        RAISE EXCEPTION 'SWID %: Missing swid'
                      , this_swid;
    END;

    this_date := this_row.date;

    -- Sanity checks on the row.

    IF this_row.event != 'B' AND this_row.event != 'E' THEN
      RAISE EXCEPTION '(Old/Catherine) SWID %: Date %: Time %: Event %: Not a begin or end row'
                    , this_swid
                    , this_row.date
                    , this_row.time
                    , this_row.event;
      RETURN 1;
    END IF;

    IF this_row.subgrp != 'S' THEN
      RAISE EXCEPTION 'SWID %: Not a subgroup', this_swid;
      RETURN 1;
    END IF;

    -- Find the departure row to go with this group/date/time.
    BEGIN
      SELECT swerb_departs_data.did
        INTO STRICT this_did
        FROM swerb_departs_data, swerb_bes
        WHERE swerb_departs_data.date = this_row.date
              AND swerb_bes.did = swerb_departs_data.did
              AND swerb_bes.focal_grp = this_row.grp
              AND (swerb_bes.start IS NULL
                   OR swerb_bes.start <= this_row.time)
              AND (swerb_bes.stop IS NULL
                   OR swerb_bes.stop >= this_row.time)
        GROUP BY swerb_departs_data.did;
    EXCEPTION
      WHEN TOO_MANY_ROWS THEN
        RAISE EXCEPTION 'Date %: Group %: Time %: More than one departure row for the group/day/time'
                      , this_row.date
                      , this_row.grp
                      , this_row.time;
      WHEN NO_DATA_FOUND THEN
        RAISE EXCEPTION 'Date %: Group %: Time % No departure row for the group/day/time'
                      , this_row.date
                      , this_row.grp
                      , this_row.time;
    END;

    --
    -- Convert from Catherine to Babase.
    --

    -- timeest
    this_timesource := trim(trailing from this_row.sourcetime);
    IF this_timesource IS NULL AND this_row.timeest = 'N' THEN
      -- Babase sez: "Gotta have a source."
      this_timesource := 'S'; -- Catherine's "SWERB" code.
    END IF;

    -- subgroup
    this_subgroup := this_row.subgrp IS NOT NULL;

    -- X Y location.
    IF this_row.coordconvert = 'Y' THEN
      -- The utm coordinates are from a quad conversion.
      -- Babase users get the coordinates from the quads
      -- table (or have it done for them with the swerb_xy view).
      this_xyloc = NULL;
    ELSE
      -- The utm measurements are from gps units.
      -- Keep them.
      this_xyloc = bb_makepoint(this_row.x, this_row.y);
    END IF;

    -- Accuracy and pdop.
    this_accuracy := NULL;
    this_pdop := NULL;
    CASE this_row.errorcode
      WHEN 'P' THEN
        this_pdop = this_row.error;
      WHEN 'A' THEN
        this_accuracy = this_row.error;
      ELSE NULL;
    END CASE;
                
    -- Conf
    this_conf := COALESCE(this_row.locdconf, 'bb_swerb_certain');

    -- gps_datetime
    IF this_row.date < '2004-01-01'::DATE THEN
      this_gps_datetime := NULL;
    ELSE
      IF this_row.garmincode = '' OR this_row.garmincode IS NULL THEN
        this_gps_datetime := NULL;
      ELSE
        this_gps_datetime := this_row.date + this_row.time;
      END IF;
    END IF;

    -- Create the begin row.  Create swerb_bes row as well
    -- as a swerb_data row.

    -- Create swerb_bes row.
    INSERT INTO swerb_bes (did
                        , focal_grp
                        , start
                        , btimeest
                        , bsource
                        , stop
                        , etimeest
                        , esource
                        , is_effort
                        , notes)
      VALUES (this_did
            , this_row.grp
            , this_row.time
            , this_row.timeest::boolean
            , this_timesource
            , NULL
            , NULL
            , NULL
            , TRUE
            , NULL);

    -- Create the swerb_data begin row. 
    INSERT INTO swerb_data(beid
                        , seen_grp
                        , event
                        , time
                        , quad
                        , xyloc
                        , altitude
                        , pdop
                        , accuracy
                        , subgroup
                        , ogdistance
                        , gps_datetime
                        , garmincode)
      VALUES (currval('swerb_bes_beid_seq')
            , this_row.grp
            , 'bb_swerb_begin'
            , this_row.time
            , this_row.quad
            , this_xyloc
            , this_row.altitude
            , this_pdop
            , this_accuracy
            , this_subgroup
            , this_row.ogdistance
            , this_gps_datetime
            , this_row.garmincode);                          

    IF this_row.event = 'B' AND this_row.locd IS NOT NULL THEN

      -- ADcode

      IF this_row.ascdesccode = 'UNK'
         OR this_row.ascdesccode = 'TIME' THEN
        this_adcode = 'bb_swerb_descent_code';
      ELSE
        -- Not a value that had to be translated.
        this_adcode := TRIM(TRAILING FROM this_row.ascdesccode);
      END IF;

      -- There's a physical landscape feature associated with
      -- this row.  Make a swerb_loc_data row for it.
      INSERT INTO swerb_loc_data (swid, loc, adcode, conf, adtime)
        VALUES(currval('swerb_data_swid_seq')
            , this_row.locd
            , this_adcode
            , this_conf
            , this_row.ascdesctime);
    END IF;

    --
    -- Now do the end row.
    --

    -- End row.  Update swerb_bes row, create a swerb_data row.
    UPDATE swerb_bes
      SET stop = this_row.time
        , etimeest = this_row.timeest::boolean
        , esource = this_timesource
      WHERE beid = currval('swerb_bes_beid_seq');

    -- Create the swerb_data row. 
    INSERT INTO swerb_data(beid
                         , seen_grp
                         , event
                         , time
                         , quad
                         , xyloc
                         , altitude
                         , pdop
                         , accuracy
                         , subgroup
                         , ogdistance
                         , gps_datetime
                         , garmincode)
      VALUES (currval('swerb_bes_beid_seq')
            , this_row.grp
            , 'bb_swerb_end'
            , this_row.time
            , this_row.quad
            , this_xyloc
            , this_row.altitude
            , this_pdop
            , this_accuracy
            , this_subgroup
            , this_row.ogdistance
            , this_gps_datetime
            , this_row.garmincode);                          

    IF this_row.event = 'E' AND this_row.locd IS NOT NULL THEN

      -- ADcode
      IF this_row.ascdesccode = 'UNK'
         OR this_row.ascdesccode = 'TIME' THEN
        this_adcode = 'bb_swerb_ascent_code';
      ELSE
        -- Not a value that had to be translated.
        this_adcode := TRIM(TRAILING FROM this_row.ascdesccode);
      END IF;

      -- There's a physical landscape feature associated with
      -- this row.  Make a swerb_loc_data row for it.
      INSERT INTO swerb_loc_data (swid, loc, adcode, conf, adtime)
        VALUES(currval('swerb_data_swid_seq')
             , this_row.locd
             , this_adcode
             , this_conf
             , this_row.ascdesctime);
    END IF;
 
--    -- Remove the row that we converted.
--    EXECUTE 'DELETE
--               FROM ' || tbl::regclass || ' AS tbl
--               WHERE tbl.SWID = $1'
--            USING this_swid;

    RETURN 0;

  EXCEPTION
    WHEN check_violation THEN
      -- Add the date being processed to any error message that shows up.
      RAISE EXCEPTION 'Date: %: %', this_date, SQLERRM;

  END;
$$;

grant_func_priv(`swerb_copybe(TEXT, INT)')
