dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2005, 2006, 2008 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Create the rows on members that interpolate over an interval.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')

CREATE OR REPLACE FUNCTION _interp(  this_sname      census.sname%TYPE
                                   , alive           BOOLEAN
                                   , early_bound     census.date%TYPE
                                   , late_bound      census.date%TYPE
                                   , early_group     census.grp%TYPE
                                   , late_group      census.grp%TYPE
                                   , birth_interp    BOOLEAN
                                   , statdate_interp BOOLEAN
                                   , interp_left     BOOLEAN
                                   , interp_right    BOOLEAN)
  RETURNS VOID
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Create the rows on members that interpolate over an interval.
  --
  -- GPL_notice(`  --', `2005, 2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax: _interp(  this_sname
  --                 , alive
  --                 , early_bound
  --                 , late_bound
  --                 , early_group
  --                 , late_group
  --                 , birth_interp
  --                 , statdate_interp
  --                 , interp_left
  --                 , interp_right)
  --
  -- Arguments:
  --   this_sname      Sname of individual
  --   alive           Whether or not the individual is alive.
  --   early_bound     Census date where we start interpolating.
  --   late_bound      Census date where we end interpolating.
  --   early_group     Group of the census date where we start interpolating.
  --   late_group      Group of the census date where we end interpolating.
  --                   Not used when interp_right is false and so may be
  --                   NULL in that case.
  --   birth_interp    Does the interpolation start at birth?
  --   statdate_interp Does the interpolation end with the statdate?
  --   interp_left     Does the census row on the early end of the interval
  --                   interpolate?
  --   interp_right    Does the census row on the late end of the interval
  --                   interpolate?
  --
  -- Remarks:
  --  The "early" and "late" bounds are dates of locating and interpolating
  -- censuses, unless we're at a corner case in which case they are
  -- birth and statdates.
  --
  --  The comments use mathematical interval notation.  [ or ]
  -- mean a closed interval bound (lower and upper, respectively)
  -- where the endpoint is included in the interval.  ( or )
  -- mean an open interval bound (lower and upper, respectively)
  -- where the endpoint is not included in the interval.
  --
  DECLARE
    -- Late bound we actually use.
    this_late_bound census.date%TYPE NOT NULL := late_bound;

  BEGIN
    -- Debug start
    -- RAISE NOTICE '_interp(%, %, %, %, %, %, %, %, %)'
    --       , this_sname
    --       , early_bound
    --       , late_bound
    --       , early_group
    --       , late_group
    --       , birth_interp
    --       , statdate_interp
    --       , interp_left
    --       , interp_right;
    -- Debug end

    --
    -- Interpolate over the interval.
    -- The bounds themselves are excluded (open interval).
    -- (General case, accounts for changing birth
    -- and statdate.)
    --

    --
    -- Remove all the pre-existing interplated rows.
    --

    -- Debug
    -- RAISE INFO 'Deleting date > % and date < %', early_bound, this_late_bound;
    -- End debug
    DELETE FROM MEMBERS
      WHERE sname = this_sname
            AND origin = 'bb_interpolated'
            AND date > early_bound
            AND date < this_late_bound;

    IF birth_interp THEN
      -- The birth date could be an interpolated date,
      -- or there could be earlier interpolated rows
      -- as the birthdate may have changed.
      -- (General case.)
      -- Debug
      -- RAISE INFO 'Deleting date > % and date < %', early_bound, this_late_bound;
      -- End debug
      DELETE FROM MEMBERS
        WHERE sname = this_sname
              AND origin = 'bb_interpolated'
              AND date <= early_bound;
    END IF;

    IF statdate_interp THEN
      -- The statdate could be (in the general case)
      -- an interpolated date (it _is_ in the delete case)
      -- or the statdate could have changed so there might
      -- be subsequent interpolated rows.
      DELETE FROM MEMBERS
        WHERE sname = this_sname
              AND origin = 'bb_interpolated'
              AND date >= this_late_bound;
    END IF;

    --
    -- Deal with creating rows for the corner cases.
    -- (We must create after all deletes so the deletes don't
    -- accidently destroy what we insert.)
    --

    IF birth_interp THEN
      -- Make sure there's a row in members for the birth.
      PERFORM *
        FROM census
        WHERE sname = this_sname
              AND date = early_bound
              AND status <> 'bb_absent'
        LIMIT 1;
      IF NOT FOUND THEN
        -- Insert a Interp=0 endpoint at birth.
        INSERT INTO members (sname, date, grp, origin, interp)
          VALUES (this_sname
                  , early_bound
                  , early_group
                  , 'bb_interpolated'
                  , 0);
      END IF;
    END IF;    

    IF statdate_interp 
      AND NOT (birth_interp AND early_bound = this_late_bound) THEN
      -- Make sure there's a row in members for the statdate.
      -- (except when birth = statdate)
      PERFORM *
        FROM census
        WHERE sname = this_sname
              AND date = this_late_bound
              AND status <> 'bb_absent'
       LIMIT 1;
      IF NOT FOUND THEN
        -- Be sure to insert a Interp<>0 endpoint at statdate.
        -- interp_right is false so increasing the right side
        -- of the interval won't mess with midpoint calculations.
        -- To get the Grp and Interp value correct we just let the
        -- counting loop do its job.
        this_late_bound := this_late_bound + 1;
      END IF;
    END IF;    

    --
    -- Degenerate case early_bound = late_bound. Do nothing
    -- but create the corner case (birth/statdate) members.
    --

    IF early_bound = late_bound THEN
      -- Have not yet created the row on the statdate.
      -- But the only time that early_bound = late_bound
      -- is when there's a locating census on that date
      -- or it's the birthdate anyhow.
      NULL; -- So we can have these comments.
      -- But, we do want to interpolate past the statdate.
    ELSE


      --
      -- Now finally interpolate.
      --

      DECLARE
        -- End of first half of the interval.
        from_end DATE;
        -- Beginning of the last half of the interval
        through_start DATE;

      BEGIN

        IF interp_left AND interp_right THEN
          -- Perform a midpoint calculation to get an upper and
          -- lower interval so we can interpolate the group both
          -- from the upper and from the lower bound.
          through_start := _interp_midpoint(early_bound, this_late_bound);
          from_end := through_start - 1;
        ELSEIF interp_left THEN
          -- Interpolate the group on the left half of the interval
          -- toward the right half.
          through_start := this_late_bound;
          from_end := this_late_bound - 1;
        ELSEIF interp_right THEN
          -- Interpolate the group on the right half of the interval
          -- toward the left half.
          from_end := early_bound;
          through_start := early_bound + 1;
        ELSE
          -- No interpolation.
          -- This is easy.  All the in-between rows place the individual
          -- in the unknown group with a Interp value of NULL.
          -- Further, because we looked for the endpoints
          -- we know that there are no (non-absent) census
          -- rows in the interval so we don't have to worry
          -- about existing rows.  *WRONG*

          -- Just fill in the unknown group rows and we're done.

          -- Interpolate from (early_bound, this_late_bound)
          FOR i IN 1 .. this_late_bound - early_bound - 1 LOOP
            PERFORM 1
              FROM members
              WHERE sname = this_sname
                    AND date = early_bound + i;
            IF NOT FOUND THEN
              -- There's no non-interpolating row on that day.
              INSERT INTO members (sname, date, grp, origin, interp)
                VALUES (this_sname
                        , early_bound + i
                        , 'bb_unk_grp'
                        , 'bb_interpolated'
                        , NULL);
            END IF;
          END LOOP;

          RETURN;

        END IF;

        DECLARE
          -- The date of the first non-absent census after the
          -- beginning of the interval.
          after_start DATE;
          -- The date of the prior non-absent census before
          -- tne end of the interval.
          before_end DATE;

          -- When to stop incrementing the Interp value
          -- on the lower half of the interpolation interval.
          -- (Counting up from early_bound.)
          early_istop DATE := from_end;
          -- When to stop incrementing the Interp value
          -- on the upper half of the interpolation interval.
          -- (Counting down from this_late_bound.)
          late_istop DATE := through_start;

          -- When to stop placeing the individual in the early_group
          -- in the lower half of the interpolation interval.
          -- (Counting up from early_bound.)
          early_gstop DATE := from_end;
          -- When to stop placeing the individual in the late_group
          -- in the upper half of the interpolation interval.
          -- (Counting down from this_late_bound.)
          late_gstop DATE := through_start;

        BEGIN

          DECLARE
            -- The date the individual was marked absent from
            -- his group.
            absent_date DATE;
            -- The midpoint betwen the interpolation's endpoint
            -- and its absence.
            absent_midpoint DATE;

          BEGIN

            -- Figure out where to stop putting the individual into
            -- the interval endpoint group when we're interpolating,
            -- and figure out where to stop incrementing the Interp value.

            IF interp_left THEN
              -- Have "old" census rows stop the interpolation.

              -- Lower half of the interval.
              SELECT INTO after_start date
                FROM census
                WHERE sname = this_sname
                      AND date > early_bound
                      AND status <> 'bb_absent'
                ORDER BY date
                LIMIT 1;
              IF after_start IS NOT NULL THEN
                IF after_start <= from_end THEN
                  -- Another locating census was found in the
                  -- lower half of the interval.
                  early_istop := after_start - 1;
                  early_gstop := early_istop;
                END IF;
              END IF;

              -- Absences in the same group as the endpoint
              -- cut the interpolation short "halfway" to the absence.

              -- Do the left, lower, side of the interval.
              SELECT INTO absent_date date
                FROM census
                WHERE sname = this_sname
                      AND status = 'bb_absent'
                      AND date > early_bound
                      AND grp = early_group
                ORDER BY date
                LIMIT 1;
              IF absent_date IS NOT NULL THEN
                -- Subtract 1 as the routine returns low end of upper bound.
                absent_midpoint
                     := _interp_midpoint(early_bound, absent_date) - 1;
                IF absent_midpoint < early_gstop THEN
                  early_gstop := absent_midpoint;
                END IF;
              END IF;
            END IF;

            IF interp_right THEN
              -- Upper half of the interval.

              -- Have "old" census rows stop the interpolation.
              SELECT INTO before_end date
                FROM census
                WHERE sname = this_sname
                      AND date < late_bound
                      AND status <> 'bb_absent'
                ORDER BY date DESC
                LIMIT 1;
              IF before_end IS NOT NULL THEN
                IF before_end >= through_start THEN
                  -- Another locating census was found in the
                  -- upper half of the interval.
                  late_istop := before_end + 1;
                  late_gstop := late_istop;
                END IF;
              END IF;

              -- Absences in the same group as the endpoint
              -- cut the interpolation short "halfway" to the absence.

              -- Do the right, upper, side of the interval.
              SELECT INTO absent_date date
                FROM census
                WHERE sname = this_sname
                      AND status = 'bb_absent'
                      AND date <= late_bound
                      AND grp = late_group
                ORDER BY date DESC
                LIMIT 1;
              IF absent_date IS NOT NULL THEN
                absent_midpoint := _interp_midpoint(absent_date, late_bound);
                IF late_gstop < absent_midpoint THEN
                   late_gstop := absent_midpoint;
                END IF;
              END IF;
            END IF;
          END;

          -- Debug start
          -- RAISE NOTICE 'from_end %: through_start %: this_late_bound %: after_start %: before_end %: early_istop %: late_istop %: early_gstop %: late_gstop %'
          --              , from_end
          --              , through_start
          --              , this_late_bound
          --              , after_start
          --              , before_end
          --              , early_istop
          --              , late_istop
          --              , early_gstop
          --              , late_gstop;
          -- Debug end

          --
          -- Finally, we actually insert rows into the members table.
          -- (Degenerate case of no interpolation handled above.)
          --

          DECLARE

            -- The Interp value to use
            interp_value members.interp%TYPE;
            -- The group to assign
            interped_group members.grp%TYPE;
            -- The date in members
            this_date DATE;
            -- The group to assign in members
            this_group members.grp%TYPE;
            -- Whether or not a row already exists in members
            do_insert BOOLEAN;

          BEGIN

            -- Interpolate (early_bound, from_end]
            -- (When from_end = early_bound do nothing.)
            FOR i IN 1 .. from_end - early_bound LOOP
              this_date := early_bound + i;

              IF this_date <= early_istop THEN
                -- In (early_bound, early_istop]
                -- Increment Interp
                interp_value := i;
                do_insert := TRUE;
              ELSE
                interp_value := NULL;
                -- On (early_istop, from_end]
                -- look out for non absent census rows
                -- (don't overwrite them.)
                PERFORM *
                  FROM census
                  WHERE sname = this_sname
                        AND date = this_date
                        AND status <> 'bb_absent'
                  LIMIT 1;
                do_insert := NOT FOUND;
              END IF;

              -- Absences take individuals out of their group.
              IF this_date <= early_gstop THEN
                this_group := early_group;
              ELSE
                this_group := 'bb_unk_grp';
              END IF;

              IF i > bb_interpolation_max THEN
                -- Don't exceed the interpolation limit.
                this_group := 'bb_unk_grp';
              END IF;

              IF do_insert THEN
                INSERT INTO members (sname, date, grp, origin, interp)
                  VALUES (this_sname
                          , this_date
                          , this_group
                          , 'bb_interpolated'
                          , interp_value);
              END IF;
            END LOOP;

            -- Interpolate [through_start, this_late_bound)
            -- (When through_start = this_late_bound do nothing.)
            FOR i IN 1 .. this_late_bound - through_start LOOP
              this_date := this_late_bound - i;

              IF this_date >= late_istop THEN
                -- In [late_istop, this_late_bound) interval
                -- Increment Interp
                interp_value := i;
                do_insert := TRUE;
              ELSE
                interp_value := NULL;
                -- On [through_start, late_istop]
                -- look out for non absent census rows
                -- (don't overwrite them.)
                PERFORM *
                  FROM census
                  WHERE sname = this_sname
                        AND date = this_date
                        AND status <> 'bb_absent'
                  LIMIT 1;
                do_insert := NOT FOUND;
              END IF;

              -- Absences take individuals out of their group.
              IF this_date >= late_gstop THEN
                this_group := late_group;
              ELSE
                this_group := 'bb_unk_grp';
              END IF;

              IF i > bb_interpolation_max THEN
                -- Don't exceed the interpolation limit.
                this_group := 'bb_unk_grp';
              END IF;

              IF do_insert THEN
                INSERT INTO members (sname, date, grp, origin, interp)
                  VALUES (this_sname
                          , this_date
                          , this_group
                          , 'bb_interpolated'
                          , interp_value);
              END IF;
            END LOOP;


          END;
        END;
      END;
    END IF;

    --
    -- Deal with interpolation past the statdate.
    --
    IF alive AND statdate_interp AND interp_right THEN
      -- That last census is an interpolating one and
      -- the individual's alive so see if there's an
      -- absence past the last census.
      DECLARE

        next_absence census.date%TYPE;

      BEGIN

        SELECT INTO next_absence
                    date
          FROM census
          WHERE sname = this_sname
                AND date > late_bound
                AND grp = late_group
                AND status = 'bb_absent'
          ORDER BY sname, date
          LIMIT 1;

        IF next_absence IS NOT NULL THEN
          -- There is addtional census data, the statdate
          -- does not just happen to be when data entry
          -- stopped.
          -- Interpolate past the statdate.
          -- (Subtract 1 because _interp_midpoint
          -- returns lower endpoint of upper interval.)

          FOR i IN 1 .. _interp_midpoint(late_bound, next_absence)
                        - 1
                        - late_bound LOOP
            INSERT INTO members (sname, date, grp, origin, interp)
              VALUES (  this_sname
                      , late_bound + i
                      , late_group
                      , 'bb_interpolated'
                      , i);
          END LOOP;
        END IF;
      END;
    END IF;

    RETURN;
  END;
$$;

SELECT '_interp(  this_sname      census.sname%TYPE
                , alive           BOOLEAN
                , early_bound     census.date%TYPE
                , late_bound      census.date%TYPE
                , early_group     census.grp%TYPE
                , late_group      census.grp%TYPE
                , birth_interp    BOOLEAN
                , statdate_interp BOOLEAN
                , interp_left     BOOLEAN
                , interp_right    BOOLEAN)' AS done_with;

