dnl Copyright (C) 2016 Jake Gordon <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2005, 2008 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Interpolate on the interval which contains a date and adjust statdate.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')

CREATE OR REPLACE FUNCTION
       _interp_about_date(  this_sname  members.sname%TYPE
                          , this_date   date
                          , this_group  census.grp%TYPE
                          , this_status census.status%TYPE)
  RETURNS VOID
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Interpolate on the interval which contains a date and adjust statdate.
  --
  -- GPL_notice(`  --', `2005, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax: _interp_about_date(  this_sname
  --                            , this_date
  --                            , this_group
  --                            , this_status)
  --
  -- Arguments:
  --   this_sname  Sname of individual
  --   this_date   Date of census row
  --   this_group  Group in which the individual is located on this_date
  --               When this is NULL, there is no locating census
  --               on this_date and there is one interval over
  --               which to interpolate.  Otherwise, there are two
  --               intervals, one on either side of the locating census.
  --               (Each interval has a "count up" and a "count down"
  --               half.)
  --   this_status Status of the individual's (possibly former)
  --               census on this_date.  When this_group is
  --               supplied must be the status that belongs to
  --               the group.
  --
  -- Remarks:
  --
  --   The assumption when calling this routine is that the
  -- date is contained in an interval that must be interpolated
  -- over.  Find the bounds of the interval and re-compute the
  -- interpolation.
  --
  --   When there is a locating census on the given date, then
  -- that group should be supplied, along with the status.
  -- That is, this_group should only be NULL when there is
  -- no locating census on the given date.
  --
  -- Adjust the statdate of the individual when necessary.
  --
  -- Bugs:
  --  The recursive calls will do a little extra work.  Tough.
  --
  DECLARE
    -- Census date where we start interpolating.
    early_bound census.date%TYPE;
    -- Census date where we end interpolating.
    late_bound census.date%TYPE;

    -- Group of the census date where we start interpolating.
    early_group census.grp%TYPE;
    -- Group of the census date where we end interpolating.
    late_group census.grp%TYPE;

    -- Status of the census where we start interpolating.
    early_status census.status%TYPE;

    -- Does the interpolation start at birth?
    birth_interp BOOLEAN NOT NULL := FALSE;
    -- Does the interpolation end with the statdate?
    statdate_interp BOOLEAN NOT NULL := FALSE;

    -- Does the census row on the early end of the interval
    -- interpolate?
    interp_left BOOLEAN;
    -- Does the census row on the late end of the interval
    -- interpolate?
    interp_right BOOLEAN;

    -- Whether the individual is alive.
    alive BOOLEAN;

  BEGIN
    -- Debug start
    -- RAISE NOTICE '_interp_about_date(%, %, %, %)'
    --       , this_sname
    --       , this_date
    --       , this_group
    --       , this_status;
    -- Debug end

    -- Find the lower and upper bounds of the interval affected.

    -- As the census interpolates, it affects the nearest
    -- rows that also interpolate (and are not absent).
    interp_left := TRUE;

    --
    -- Find the early bound.
    --
    SELECT INTO early_bound, early_group, early_status
                date,        grp,         status
      FROM census
      WHERE sname = this_sname
            AND date < this_date
            AND (status = 'bb_census'
                 OR status = 'bb_demog'
                 OR status = 'bb_manual')
      ORDER BY date DESC
      LIMIT 1;
    -- IF early_bound IS NOT NULL
    -- Start bound at a interpolating census.
    -- What is the group at the start?
    IF early_bound IS NULL THEN
      -- Start bound at birthdate.
      birth_interp := TRUE;
      SELECT INTO early_bound, early_group
                  birth,       matgrp
        FROM biograph
        WHERE sname = this_sname;
      -- Does the birthdate interpolate?
      DECLARE
        -- Scratch variable.  Status at birth.
        birth_status census.status%TYPE;
      BEGIN
       SELECT INTO birth_status
                   status
              FROM census
              WHERE sname = this_sname
                    AND date = early_bound           -- the birthdate
                    AND status <> 'bb_absent'
              LIMIT 1;
       -- Does the birthdate interpolate?
       interp_left := birth_status IS NULL
                      OR birth_status = 'bb_census'
                      OR birth_status = 'bb_demog'
                      OR birth_status = 'bb_manual';
      END;
    END IF;

    --
    -- Upper bound corner case at end of census series.
    --
    DECLARE
      -- The entrydate of the individual.
      this_entrydate biograph.entrydate%TYPE;

      -- The statdate of the individual.
      this_statdate biograph.statdate%TYPE;

      -- Date of last locating census.
      last_censused census.date%TYPE;
      -- Whether there are any locating censuses.
      censused BOOLEAN;

    BEGIN
      -- Figure out if the individual is alive.
      SELECT INTO this_statdate, this_entrydate
                  statdate,      entrydate
        FROM biograph
        WHERE sname = this_sname
              AND status = 'bb_alive'
        LIMIT 1;
      alive := FOUND;

      -- The corner case at the end interval.

      SELECT INTO last_censused date
        FROM census
        WHERE sname = this_sname
              AND status <> 'bb_absent'
        ORDER BY date DESC
        LIMIT 1;
      censused = last_censused IS NOT NULL;

      IF NOT censused OR last_censused <= this_date THEN
        -- this_group is, or is after, the last locating census
        -- of the individual.
        IF this_status = 'bb_absent' OR this_group IS NULL THEN
          -- Census is 'missing' on this_date, that is:
          -- Inserting (or updateing?) an absence after
          -- the last locating census.  Or, deleting
          -- the last locating census.

          IF NOT censused THEN
            -- There are no locating censuses.
            -- Make sure statdate is entrydate.
            -- (Don't use early_bound--which at this point
            -- must be birth--because if entrydate <> birth
            -- then statdate will be updated to before the
            -- entrydate, which is not allowed.)
            IF NOT this_statdate = this_entrydate THEN
              -- This automatically re-interpolates
              -- (because we changed the statdate.)
              -- (The trigger constraints should ensure that
              -- we only get to this point if the individual
              -- is alive because otherwise there'd have
              -- been a census "present" after a death date.)
              UPDATE biograph
                SET statdate = entrydate
                WHERE sname = this_sname;
            END IF;
          ELSE
            IF last_censused = this_date THEN
              -- There's (another) locating census on
              -- the same date as "our" "missing" census.
              -- Interpolate using this locating census.
              DECLARE
                -- The group of the other census.
                center_group census.grp%TYPE;
                -- The status of the other census.
                center_status census.status%TYPE;

              BEGIN
                SELECT INTO center_group, center_status
                            grp,          status
                  FROM census
                  WHERE sname = this_sname
                        AND date = this_date
                        AND status <> 'bb_absent'
                  LIMIT 1;
                PERFORM _interp_about_date(  this_sname
                                           , this_date
                                           , center_group
                                           , center_status);
              END;
            ELSE
              IF alive AND this_statdate <> last_censused THEN
                -- The last locating census was an interpolating
                -- census but now it's a non-interpolating census.
                -- Readjust the statdate and interpolate at the
                -- statdate.
                UPDATE biograph
                  SET statdate = last_censused
                  WHERE sname = this_sname; 
              END IF;

              -- Re-interpolate the interval before the statdate.
              PERFORM _interp_about_date(  this_sname
                                         , early_bound
                                         , early_group
                                         , early_status);
            END IF;
          END IF;

          RETURN;

        ELSE
          -- "our" census is a locating census.
          -- (last_censused = this_date, "our" census is the last.)
          IF alive THEN
            -- Individual is alive
            IF this_statdate = this_date THEN
              -- There's only the lower interval.
              interp_right := this_status = 'bb_census'
                              OR this_status = 'bb_demog'
                              OR this_status = 'bb_manual';
              PERFORM _interp(  this_sname
                              , alive
                              , early_bound
                              , this_date
                              , early_group
                              , this_group
                              , birth_interp
                              , TRUE
                              , interp_left
                              , interp_right);
            ELSE
              -- The statdate must be set.
              -- This will interpolate, so we needn't.
              -- (We know this_group IS NOT NULL because
              -- we tested above.)
              -- (Insert into census (or update?))
              UPDATE biograph
                SET statdate = this_date
                WHERE sname = this_sname;
            END IF;

            RETURN;

          END IF; -- Alive or dead
        END IF; -- Absent or locating
      END IF; -- Otherwise, there's a later census.
    END;

    --
    -- Find the late bound.
    --
    SELECT INTO late_bound, late_group
                date,       grp
      FROM census
      WHERE SNAME = this_sname
            AND date > this_date
            AND (status = 'bb_census'
                 OR status = 'bb_demog'
                 OR status = 'bb_manual')
      ORDER BY date
      LIMIT 1;
    IF late_bound IS NOT NULL THEN
      -- Late bound at a census.
      -- Locating census at late bound
      interp_right := TRUE;
    ELSE
      -- No interpolating and locating censuses
      -- after the deleted row.  Do we need to
      -- adjust the individual's statdate because
      -- he's still alive?

      -- No, because we already dealt with this corner case above.

      -- End bound at statdate.
      statdate_interp := TRUE;
      -- The statdate does not interpolate.
      interp_right := FALSE;
      SELECT INTO late_bound statdate
        FROM biograph
        WHERE sname = this_sname;
    END IF;

    --
    -- Interpolate
    --

    IF this_group IS NULL
       OR (this_status <> 'bb_census'
           AND this_status <> 'bb_demog'
           AND this_status <> 'bb_manual') THEN
      -- There's no locating and interpolating census on the
      -- interpolation date.
      PERFORM _interp(  this_sname
                      , alive
                      , early_bound
                      , late_bound
                      , early_group
                      , late_group
                      , birth_interp
                      , statdate_interp
                      , interp_left
                      , interp_right);
    ELSE
      -- There is a locating and interpolating census on the
      -- interpolation date.

      -- Interpolate before the locating census.
      PERFORM _interp(  this_sname
                      , alive
                      , early_bound
                      , this_date
                      , early_group
                      , this_group
                      , birth_interp
                      , FALSE
                      , interp_left
                      , TRUE);

      -- Interpolate after the locating census.
      PERFORM _interp(  this_sname
                      , alive
                      , this_date
                      , late_bound
                      , this_group
                      , late_group
                      , FALSE
                      , statdate_interp
                      , TRUE
                      , interp_right);
    END IF;
    RETURN;
  END;
$$;

SELECT '_interp_about_date(  this_sname  members.sname%TYPE
                           , this_date   date
                           , this_group  census.grp%TYPE
                           , this_status census.status%TYPE)' AS done_with;

