dnl Copyright (C) 2016 Jake Gordon <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2005, 2006, 2008 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Interpolate about a date with a census that either has or
dnl had a non-interplating status.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')

CREATE OR REPLACE FUNCTION
      _interpish_about_date(  this_sname  members.sname%TYPE
                            , this_date   date
                            , this_group  census.grp%TYPE
                            , this_status census.status%TYPE)
  RETURNS VOID
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Interpolate about a date with a census that either has or
  -- had a non-interplating, non-absent status.
  --
  -- GPL_notice(`  --', `2005, 2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax: _interpish_about_date(this_sname
  --                             , this_date
  --                             , this_group
  --                             , this_status);
  --
  -- Arguments:
  --   this_sname  Sname of individual
  --   this_date   Date of census row
  --   this_group  Group in which the indvidual is located on this_date
  --               When this is NULL, there is no locating census
  --               on this_date and there is one interval over
  --               which to interpolate.  Otherwise, there are two
  --               intervals, one on either side of the locating census.
  --               (Each interval has a "count up" and a "count down"
  --               half.)
  --   this_status Status of the individual's (possibly former)
  --               census on this_date.  When this_group is
  --               supplied must be the status that belongs to
  --               the group.
  --
  -- Remarks:
  --
  -- This routine is an optmization.  In theory we could always call
  -- _interp_about_date() instead.  But this would be really slow
  -- in those cases where the individual has no interpolated censuses.
  -- In order for rebuild_members() to work in a timely fashion, and
  -- so we can load the data initially in a timely fashion, we need
  -- this optimization.
  --
  --   Interpolate over as narrow a range as possible.  
  --
  --   The assumption when calling this routine is that the
  -- date is contained in an interval that must be interpolated
  -- over.  Find the bounds of the interval and re-compute the
  -- interpolation.
  --
  -- Compare with _interp_about_date().
  --

  DECLARE
    -- Census date where we start interpolating.
    early_bound census.date%TYPE;
    -- Census date where we end interpolating.
    late_bound census.date%TYPE;

    -- Group of the census date where we start interpolating.
    early_group census.grp%TYPE;
    -- Group of the census date where we end interpolating.
    late_group census.grp%TYPE;

    -- Does the interpolation start at birth?
    birth_interp BOOLEAN NOT NULL := FALSE;
    -- Does the interpolation end with the statdate?
    statdate_interp BOOLEAN NOT NULL := FALSE;

    -- Does the census row on the early end of the interval
    -- interpolate?
    interp_left BOOLEAN;
    -- Does the census row on the late end of the interval
    -- interpolate?
    interp_right BOOLEAN;

    -- The individual's status
    biograph_status biograph.status%TYPE;
    -- Individual's birthdate
    this_birth biograph.birth%TYPE;
    -- Individual's matgrp
    this_matgrp biograph.matgrp%TYPE;

    -- Scratch variable. Status of census at endpoint.
    bound_status census.status%TYPE;

  BEGIN
    -- RAISE INFO '_interpish_about_date(%, %, %, %)'
    --          , this_sname
    --          , this_date
    --          , this_group
    --          , this_status;

    -- The interval might not interpolate, its affects might
    -- reach only to the nearest non-absent census row.

    -- Get birthdate and matgrp for endpoint calculations.
    -- Get the status so we can tell if the
    -- individual is alive.
    SELECT INTO biograph_status, this_birth, this_matgrp
                status,      birth,      matgrp
      FROM biograph
      WHERE sname = this_sname;

    IF this_date = this_birth THEN
      -- Because the birthdate can interpolate, interpolating
      -- on the birthdate is like interpolating on a regular
      -- census.  Call the regular routine.
      PERFORM _interp_about_date(  this_sname
                                 , this_date
                                 , this_group
                                 , this_status);
      RETURN;
    END IF;

    --
    -- Find the early bound, wheather it interpolates and its group.
    --
    SELECT INTO early_bound, bound_status, early_group
                date,        status,       grp
      FROM census
      WHERE sname = this_sname
            AND date < this_date
            AND status <> 'bb_absent'
      ORDER BY date DESC
      LIMIT 1;
    IF early_bound IS NOT NULL THEN
      -- Does the early bound census interpolate?
      IF bound_status <> 'bb_census'
         AND bound_status <> 'bb_demog'
         AND bound_status <> 'bb_manual' THEN
        -- The lower bound does not interpolate.
        interp_left := FALSE;
      ELSE
        -- The lower bound does interpolate.
        interp_left := TRUE;
      END IF;
    ELSE
      -- Start early bound at birthdate.
      birth_interp := TRUE;
      early_bound := this_birth;
      early_group := this_matgrp;
      -- Does the birthdate interpolate?
      SELECT INTO bound_status
                  status
             FROM census
             WHERE sname = this_sname
                   AND date = early_bound           -- the birthdate
                   AND status <> 'bb_absent'
             LIMIT 1;
      interp_left := bound_status IS NULL
                     OR bound_status = 'bb_census'
                     OR bound_status = 'bb_demog'
                     OR bound_status = 'bb_manual';
    END IF;

    --
    -- Find the late bound.
    --
    SELECT INTO late_bound, bound_status, late_group
                date,       status,       grp
      FROM census
      WHERE SNAME = this_sname
            AND date > this_date
            AND status <> 'bb_absent'
      ORDER BY date
      LIMIT 1;
    IF late_bound IS NOT NULL THEN
      -- Does the late bound census interpolate?
      IF bound_status = 'bb_census'
         OR bound_status = 'bb_demog'
         OR bound_status = 'bb_manual' THEN
        --
        -- The late bound does interpolate. (census found)
        --
        interp_right := TRUE;

        IF NOT(interp_left) THEN
          -- Because the right side interpolates find
          -- the left side that interpolates too,
          -- the left side's interpolation could affect the right.
          DECLARE
            -- The early bound we might use.
            temp_early_bound DATE;
            temp_early_grp census.grp%TYPE;

          BEGIN
            SELECT INTO temp_early_bound, temp_early_grp
                        date,             grp
              FROM census
              WHERE sname = this_sname
                    AND date < this_date
                    AND (status = 'bb_census'
                         OR status = 'bb_demog'
                         OR status = 'bb_manual')
              ORDER BY date DESC
              LIMIT 1;
            IF FOUND THEN
              -- There is a earlier interpolating census.
              early_bound := temp_early_bound;
              early_group := temp_early_grp;
              -- Note that we're now interpolating from early_bound.
              interp_left := TRUE;
            ELSE
              -- All the earlier censuses are non-interpolating.
              -- Interpolate back to the birthdate.
              birth_interp := TRUE;
              early_bound := this_birth;
              early_group := this_matgrp;
              -- Does the birthdate interpolate?
              PERFORM 1
                      FROM census
                      WHERE census.sname = this_sname
                            AND census.date = this_birth
                            AND census.status <> 'bb_absent'
                      LIMIT 1;
              IF NOT FOUND THEN
                -- No non-absent census on birthdate, birth interpolates.
                interp_left := TRUE;
              END IF;
            END IF;
          END;
        END IF;
      ELSE
        --
        -- Late bound is not an interpolating census.
        --
        interp_right := FALSE;
        IF interp_left THEN
          -- Find the right side that interpolates, it could
          -- effect the left side's interpolation.

          DECLARE
            -- The late bound we might use.
            temp_late_bound DATE;
            temp_late_grp census.grp%TYPE;

          BEGIN
              SELECT INTO temp_late_bound, temp_late_grp
                          date,            grp
              FROM census
              WHERE sname = this_sname
                    AND date > this_date
                    AND (status = 'bb_census'
                         OR status = 'bb_demog'
                         OR status = 'bb_manual')
              ORDER BY date
              LIMIT 1;
            IF FOUND THEN
              -- There is a later interpolating census.
              late_bound := temp_late_bound;
              late_group := temp_late_grp;
              -- Note that we're now interpolating from late_bound.
              interp_right := TRUE;
            END IF;
            -- If all the later censuses are non-interpolating then
            -- we can leave the right side non-interpolating as it
            -- won't affect the left side's interpolation.
          END;
        END IF;
      END IF;
    ELSE
      --
      -- No locating censuses after this_date.
      --

      -- Do we need to adjust the individual's statdate because
      -- he's still alive?

      IF biograph_status = 'bb_alive' THEN
        -- The individual is alive.

        -- The deleted census was the last census that
        -- located the individual.
        -- We need to update the individual's
        -- statdate.
        -- Changing a statdate will trigger an intepolation so
        -- we are done after that.

        DECLARE
          new_statdate census.date%TYPE;
        BEGIN

          -- Find the new statdate for the individual.
          SELECT INTO new_statdate date
            FROM census
            WHERE sname = this_sname
                  AND status <> 'bb_absent'
            ORDER BY date DESC
            LIMIT 1;
          IF new_statdate IS NULL THEN
            -- There are no locating censuses for
            -- the (alive) individual.  So his statdate
            -- should be his entrydate.
            SELECT INTO new_statdate entrydate
              FROM biograph
              WHERE sname = this_sname;
          END IF;

          -- Update the statdate.
          UPDATE biograph
            SET statdate = new_statdate
            WHERE sname = this_sname
                  AND statdate <> new_statdate;

          IF FOUND THEN
            -- Changed the statdate.
            -- Updating the statdate interpolates the
            -- statdate corner case for us.
            RETURN;
          END IF;
        END;

      END IF;

      -- End bound at statdate.
      statdate_interp := TRUE;
      SELECT INTO late_bound statdate
        FROM biograph
        WHERE sname = this_sname;
      -- Statdate does not interpolate when there's no locating
      -- census to go with it.
      interp_right := FALSE;
    END IF;

    --
    -- Interpolate over the interval.
    -- The bounds themselves are excluded (open interval).
    -- (General case, accounts for changing birth
    -- and statdate.)
    --

    PERFORM _interp(  this_sname
                    , biograph_status = 'bb_alive'
                    , early_bound
                    , late_bound
                    , early_group
                    , late_group
                    , birth_interp
                    , statdate_interp
                    , interp_left
                    , interp_right);
    RETURN;
  END;
$$;

SELECT '_interpish_about_date(  this_sname  members.sname%TYPE
                              , this_date   date
                              , this_group  census.grp%TYPE)' AS done_with;

