dnl Copyright (C) 2014, 2015 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Chado_analysis_delete function, delete a vcf analysis undoing chado-vcf-load.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

CREATE OR REPLACE FUNCTION chado_analysis_delete(this_analysis_id INT)
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  SET client_min_messages TO WARNING
  bb_function_set_search_path
  AS $$
  -- Undo chado-vcf-load or chado-funcgen-load.
  --
  -- GPL_notice(`  --', `2014, 2015', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: chado_analysis_delete(INT this_analysis_id)
  --
  -- Arguments:
  --   this_analysis_id  Id of analysis row to delete.
  --
  -- Remarks:
  --
  --   Deletes related features rows if they are associated with
  --  the analysis and are associated with no other analysis
  --  and have a feature.dbxref_id that matches the analysisprop.value
  --  of the analysis where the cv.name = 'babase_chado_terms' and
  --  the cvterm.name = 'vcf_column_source' (the anaylsisprop.value
  --  is text as "db.name:dbxref.accession") and feature.is_analysis is true
  --  and have a residue if related to the dbxref
  -- 'unique_baboons:reference_baboon' or have no residue if
  --  related to one of the vcf_column_source dbxrefs.
  --
  --  Stock are never deleted.
  --
  -- Bugs:
  --   Chado has built-in cascading deletes.  This program does not
  -- rely on them.
  --
  --   The choice of exception raised (foreign_key_violation) may not
  -- be correct.

  DECLARE
    this_accession TEXT;
    id INT;

    genotype_cvterm_id cvterm.cvterm_id%TYPE;
    reference_baboon_dbxref_id dbxref.dbxref_id%TYPE;

  BEGIN
    -- Lookup the cvterm.cvterm_id that says which analysisprop rows
    -- contain genotype dbxref accession values.
    SELECT cvterm.cvterm_id
      INTO STRICT genotype_cvterm_id
      FROM cvterm
           JOIN cv ON (cv.cv_id = cvterm.cv_id)
      WHERE cv.name = 'bb_cad_sample_id_trans_cv'
            AND cvterm.name = 'bb_cad_sample_id_trans_term';

    SELECT dbxref.dbxref_id
      INTO STRICT reference_baboon_dbxref_id
      FROM dbxref
           JOIN db ON (db.db_id = dbxref.db_id)
      WHERE db.name = 'bb_cad_feature_db_name'
            AND dbxref.accession
                = 'bb_cad_reference_base_feature_dbxref_accession';


    -- Find the dbxref.db_id of the feature.dbxref_ids associated with
    -- the analysis.
    CREATE TEMP TABLE IF NOT EXISTS genotype_dbxrefs (
      id INT UNIQUE
    ) ON COMMIT DROP;

    FOR this_accession 
      IN SELECT analysisprop.value
           FROM analysisprop
           WHERE analysisprop.analysis_id = this_analysis_id
                 AND analysisprop.type_id = genotype_cvterm_id
      LOOP

      INSERT INTO genotype_dbxrefs (id)
        SELECT dbxref.dbxref_id
          FROM dbxref
               JOIN db ON (db.db_id = dbxref.db_id)
          WHERE db.name = split_part(this_accession, ':', 1)
                AND dbxref.accession = split_part(this_accession, ':', 2);
      IF NOT FOUND THEN
        RAISE foreign_key_violation USING
          MESSAGE = 'Cannot delete ACCESSION'
          , DETAIL = 'Key (Accession_Id) = (' || this_analysis_id
                     || '): Value (ANALYSISPROP.Value) = (' || this_accession
                     || '): Value (ANALYSISPROP.Type_Id) = ('
                     || genotype_cvterm_id
                     || '): Value (CVTERM.Name) = (bb_cad_sample_id_trans_term'
                     || '): Value (CV.Name) = (bb_cad_sample_id_trans_cv'
                     || '): The ANALYSISPROP.Value (of the form'
                     || ' db.name:dbxref.accession) does not resolve to an'
                     || ' existing DBXREF row'
          , HINT = 'Without a DBXREF row it is impossible to tell whether'
                   ' a FEATURE row is that of an analyzed genotype';
      END IF;
    END LOOP;

    -- analysisfeature

    CREATE TEMP TABLE IF NOT EXISTS analysisfeature_ids (
      id INT, feature_id INT)
      ON COMMIT DROP;

    INSERT INTO analysisfeature_ids (id, feature_id)
      SELECT analysisfeature.analysisfeature_id
           , analysisfeature.feature_id
        FROM analysisfeature
        WHERE analysisfeature.analysis_id = this_analysis_id;

    -- analysisfeatureprop
    DELETE FROM analysisfeatureprop
      USING analysisfeature_ids
      WHERE analysisfeatureprop.analysisfeature_id = analysisfeature_ids.id;

    -- Finish analysisfeature
    DELETE FROM analysisfeature
      USING analysisfeature_ids
      WHERE analysisfeature.analysisfeature_id = analysisfeature_ids.id;


    -- feature

    CREATE TEMP TABLE IF NOT EXISTS feature_ids (id INT)
      ON COMMIT DROP;

    INSERT INTO feature_ids (id)
      SELECT feature.feature_id
        FROM feature
        JOIN analysisfeature_ids
             ON (feature.feature_id = analysisfeature_ids.feature_id)
        WHERE
          NOT EXISTS (
            SELECT 1
              FROM analysisfeature
              WHERE analysisfeature.feature_id = feature.feature_id)
          AND feature.is_analysis
          AND ((EXISTS (SELECT 1
                          FROM genotype_dbxrefs
                           WHERE genotype_dbxrefs.id = feature.dbxref_id)
                AND feature.residues IS NULL)
               OR (feature.dbxref_id = reference_baboon_dbxref_id
                   AND feature.residues IS NOT NULL));

    TRUNCATE genotype_dbxrefs;
    TRUNCATE analysisfeature_ids;

    -- featureloc
    DELETE FROM featureloc
      USING feature_ids
      WHERE featureloc.feature_id = feature_ids.id;

    -- featureprop
    DELETE FROM featureprop
      USING feature_ids
      WHERE featureprop.feature_id = feature_ids.id;

    -- feature_cvterm
    DELETE FROM feature_cvterm
      USING feature_ids
      WHERE feature_cvterm.feature_id = feature_ids.id;

    -- feature_dbxref
    DELETE FROM feature_dbxref
      USING feature_ids
      WHERE feature_dbxref.feature_id = feature_ids.id;

    -- feature_synonym
    DELETE FROM feature_synonym
      USING feature_ids
      WHERE feature_synonym.feature_id = feature_ids.id;

    -- feature_relationship
    DELETE FROM feature_relationship
      USING feature_ids
      WHERE feature_relationship.subject_id = feature_ids.id
            OR feature_relationship.object_id = feature_ids.id;


    -- Finish feature
    DELETE FROM feature
      USING feature_ids
      WHERE feature.feature_id = feature_ids.id;

    TRUNCATE feature_ids;


    -- analysis_nd_experiment

    CREATE TEMP TABLE IF NOT EXISTS analysis_nd_experiment_ids (
      id INT, nd_experiment_id INT)
      ON COMMIT DROP;

    INSERT INTO analysis_nd_experiment_ids (id, nd_experiment_id)
      SELECT analysis_nd_experiment.analysis_nd_experiment_id
           , analysis_nd_experiment.nd_experiment_id
        FROM analysis_nd_experiment
        WHERE analysis_nd_experiment.analysis_id = this_analysis_id;

    -- nd_experiment

    CREATE TEMP TABLE IF NOT EXISTS nd_experiment_ids (id INT)
      ON COMMIT DROP;

    INSERT INTO nd_experiment_ids (id)
      SELECT analysis_nd_experiment_ids.nd_experiment_id
        FROM analysis_nd_experiment_ids
        GROUP BY analysis_nd_experiment_ids.nd_experiment_id;

    -- Finish analysis_nd_experiment
    DELETE FROM analysis_nd_experiment
      USING analysis_nd_experiment_ids
      WHERE analysis_nd_experiment.analysis_nd_experiment_id
            = analysis_nd_experiment_ids.id;

    TRUNCATE analysis_nd_experiment_ids;


    -- nd_experimentprop
    DELETE FROM nd_experimentprop
      USING nd_experiment_ids
      WHERE nd_experimentprop.nd_experiment_id = nd_experiment_ids.id;


    -- nd_experiment_stock

    CREATE TEMP TABLE IF NOT EXISTS nd_experiment_stock_ids (id INT)
      ON COMMIT DROP;

    INSERT INTO nd_experiment_stock_ids (id)
      SELECT nd_experiment_stock.nd_experiment_stock_id
        FROM nd_experiment_stock
             JOIN nd_experiment_ids
                  ON (nd_experiment_stock.nd_experiment_id
                      = nd_experiment_ids.id);

    -- nd_experiment_stockprop
    DELETE FROM nd_experiment_stockprop
      USING nd_experiment_stock_ids
      WHERE nd_experiment_stockprop.nd_experiment_stock_id
            = nd_experiment_stock_ids.id;

    -- Finish nd_experiment_stock
    DELETE FROM nd_experiment_stock
      USING nd_experiment_stock_ids
      WHERE nd_experiment_stock.nd_experiment_stock_id
            = nd_experiment_stock_ids.id;

    TRUNCATE nd_experiment_stock_ids;

    -- Finish nd_experiment
    DELETE FROM nd_experiment
      USING nd_experiment_ids
      WHERE nd_experiment.nd_experiment_id = nd_experiment_ids.id;

    TRUNCATE nd_experiment_ids;


    -- Tables directly related to analysis.

    -- analysisprop
    DELETE FROM analysisprop
      WHERE analysisprop.analysis_id = this_analysis_id;

    -- analysis_cvterm
    DELETE FROM analysis_cvterm
      WHERE analysis_cvterm.analysis_id = this_analysis_id;

    -- There is no analysis_dbxref (yet)
    -- -- analysis_dbxref
    --DELETE FROM analysis_dbxref
    --  WHERE analysis_dbxref.analysis_id = this_analysis_id;

    -- There is no analysis_relationship (yet)
    -- -- analysis_relationship
    --DELETE FROM analysis_relationship
    --  WHERE analysis_relationship.subject_id = this_analysis_id
    --        OR analysis_relationship.object_id = this_analysis_id;

    -- Delete the analysis
    DELETE FROM analysis WHERE analysis.analysis_id = this_analysis_id;
    IF FOUND THEN
      RETURN this_analysis_id;
    ELSE
      RETURN NULL;
    END IF;
  END;
$$;

CREATE OR REPLACE FUNCTION chado_analysis_delete(this_name VARCHAR(255))
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Undo chado-vcf-load or chado-funcgen-load.
  --
  -- GPL_notice(`  --', `2015', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: chado_analysis_delete(VARCHAR(255) this_name)
  --
  -- Arguments:
  --   this_name   Analysis.name of analysis to delete
  --
  -- Remarks:
  --   Glue that calls chado_analysis_delete(INT this_analysis_id)
  --
  -- Bugs:

  DECLARE
    result INT;

  BEGIN

    SELECT chado_analysis_delete(analysis.analysis_id)
      INTO STRICT result
      FROM analysis
      WHERE analysis.name = this_name;

  RETURN result;
  END;
$$;

grant_func_priv(`chado_analysis_delete(INT)')
grant_func_priv(`chado_analysis_delete(VARCHAR(255))')
