dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004, 2005, 2008 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Reinterpolate an individual's portion of the MEMBERS table.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl
dnl

CREATE OR REPLACE FUNCTION rebuild_members(this_sname members.sname%TYPE)
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Reinterpolate an individual's portion of the MEMBERS table.
  --
  -- GPL_notice(`  --', `2004, 2005, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax: rebuild_members(sname)
  --
  -- Arguments:
  --  sname  Sname of individual to rebuild.
  --
  -- Returns: integer  0 on success.
  --
  -- Remarks:
  --  Does not clear out members first, and assumes that the
  -- census triggers have copied the census rows into members.
  -- In other words, this only reinterpolates the members table
  -- by removing and rebuilding all interpolated rows.
  --
  -- Bugs:
  --  Sometimes re-interpolates the birth and statdate twice.

  DECLARE 
    -- Info about the last interpolating, locating census.
    last_group census.grp%TYPE;
    last_date census.date%TYPE;
    this_statdate biograph.statdate%TYPE;
    this_status biograph.status%TYPE;

  BEGIN

  -- Some basic error checking, as this might be used by a person.

  -- Start the first interpolation with the birthdate.
  SELECT INTO last_date, last_group, this_statdate, this_status
              birth,     matgrp,     statdate,      status
         FROM biograph WHERE sname = this_sname;
  IF NOT FOUND THEN
    RAISE EXCEPTION 'rebuild_members: %: no such Sname on BIOGRAPH'
                    , this_sname;
    RETURN 1;
  END IF;

  -- Do the deed.

  PERFORM *
     FROM census
     WHERE sname = this_sname
           AND (status = 'bb_census'
                OR status = 'bb_demog'
                OR status = 'bb_manual')
     LIMIT 1;
  IF NOT FOUND THEN
    -- No interpolating, locating census rows.  Special case.
    -- Interpolate from birth to statdate.
    PERFORM _interp(  this_sname
                    , this_status = 'bb_alive'
                    , last_date
                    , this_statdate
                    , last_group
                    , NULL
                    , TRUE
                    , TRUE
                    , TRUE
                    , FALSE);
    RETURN 0;
  END IF;

  -- There are locating, interpolating census rows.

  DECLARE
    cenrow census%ROWTYPE;
    birth_interp BOOLEAN := TRUE;
    interp_left BOOLEAN;

  BEGIN

    -- See if the birthdate interpolates.
    PERFORM *
      FROM census
      WHERE sname = this_sname
            AND date = last_date    -- the birthdate
            AND status <> 'bb_absent'        
            AND status <> 'bb_census'
            AND status <> 'bb_demog'
            AND status <> 'bb_manual'
      LIMIT 1;
    interp_left := NOT FOUND;

    FOR cenrow 
      IN SELECT *
                FROM census
                WHERE sname = this_sname
                      AND (status = 'bb_census'
                           OR status = 'bb_demog'
                           OR status = 'bb_manual')
                ORDER BY date LOOP
      PERFORM _interp(this_sname
                      , this_status = 'bb_alive'
                      , last_date
                      , cenrow.date
                      , last_group
                      , cenrow.grp
                      , birth_interp
                      , FALSE
                      , interp_left
                      , TRUE);
      -- Prepare for next iteration.
      last_date := cenrow.date;
      last_group := cenrow.grp;
      birth_interp := FALSE;
      interp_left := TRUE;
    END LOOP;

    -- Does the census on the statdate interpolate?
    DECLARE
      interp_right BOOLEAN;
    BEGIN
      PERFORM *
        FROM census
        WHERE sname = this_sname
              AND date = this_statdate
              AND (status = 'bb_census'
                   OR status = 'bb_demog'
                   OR status = 'bb_manual')
        LIMIT 1;
      interp_right = FOUND;

      -- Now interpolate.
      PERFORM _interp(this_sname
                      , this_status = 'bb_alive'
                      , last_date
                      , this_statdate
                      , last_group
                      , last_group
                      , FALSE
                      , TRUE
                      , TRUE
                      , interp_right);
    END;
  END;
  RETURN 0;
  END;
$$;
grant_func_priv(`rebuild_members(members.sname%TYPE)')
