dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2006, 2008, 2010 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Rebuild one individual's REPSTATS rows in batch.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl Bugs:
dnl  This will only work if the triggers are turned off on CYCPOINTS!
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

CREATE OR REPLACE FUNCTION rebuild_repstats(this_sname cycles.sname%TYPE)
  RETURNS INT
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Rebuild one individual's REPSTATS rows.
  --
  -- GPL_notice(`  --', `2006, 2008, 2010', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax: rebuild_repstats(this_sname)
  --
  -- Arguments:
  --  this_sname   Individual to rebuild
  --
  -- Remarks:
  --  Alternate method from that of triggers so the two can be run
  -- independently and compared.
  --
  -- Bugs:

  DECLARE
    this_date maturedates.matured%TYPE;
    last_date maturedates.matured%TYPE;
    statdate biograph.statdate%TYPE;
    this_sex biograph.sex%TYPE;

    this_state repstats.state%TYPE;
    last_state repstats.state%TYPE;
    this_dins repstats.dins%TYPE;
    this_pid repstats.pid%TYPE;
    a_pid pregs.pid%TYPE;

    obsstart cycgaps.state%TYPE;

  BEGIN

  -- Delete all the individual's repstats.
  DELETE
    FROM repstats
    WHERE sname = this_sname;

  --
  -- Iterate over the days from the matured date to the statdate
  -- (inclusive).
  --

  -- Start with matured date.
  SELECT INTO this_date matured FROM maturedates WHERE sname = this_sname;
  IF NOT FOUND THEN
    -- Individual is not mature, all done.
    RETURN 0;
  END IF;

  -- Go through the statdate.
  SELECT INTO statdate,          this_sex
              biograph.statdate, biograph.sex
    FROM biograph
    WHERE sname = this_sname;
  IF this_sex <> 'bb_female' THEN
    -- Individual is not female, all done.
    RETURN 0;
  END IF;

  --
  -- Start cycling at maturity.
  --
  last_date := this_date;
  this_state := 'bb_cycling';
  last_state := 'bb_cycling';
  this_dins := 1;
  this_pid := NULL;

  WHILE this_date <= statdate LOOP

    -- We only want a row if we're not in a gap.
    PERFORM 1
      FROM cycgapdays
      WHERE sname = this_sname
            AND date = this_date
      LIMIT 1;
    IF NOT FOUND THEN
      -- We're not in a gap.  There will be other data.

      --
      -- See if the state changes.
      --

      --
      -- Is this a conception date?
      --
      SELECT INTO a_pid pregs.pid
        FROM cycles, cycpoints, pregs
        WHERE cycles.sname = this_sname
              AND cycpoints.cid = cycles.cid
              AND cycpoints.date = this_date
              AND pregs.conceive = cycpoints.cpid
        LIMIT 1;
      IF FOUND THEN
        -- Yes, this is a conception date.
        this_state := 'bb_pregnant';
        this_pid := a_pid;
        this_dins := 1;
      ELSE

        --
        -- Is this a resumption date?
        -- 
        PERFORM 1
          FROM cycles, cycpoints, pregs
          WHERE cycles.sname = this_sname
                AND cycpoints.cid = cycles.cid
                AND cycpoints.date = this_date
                AND pregs.resume = cycpoints.cpid
          LIMIT 1;
        IF FOUND THEN
          -- Yes, this is a resumption of cycling date.
          this_state := 'bb_cycling';
          this_pid := NULL;
          this_dins := 1;
        ELSE

          --
          -- Is this a birth date?
          --
          PERFORM 1
            FROM biograph, pregs, cycpoints, cycles
            WHERE biograph.birth = this_date
                  AND pregs.pid = biograph.pid
                  AND cycpoints.cpid = pregs.conceive
                  AND cycles.cid = cycpoints.cid
                  AND cycles.sname = this_sname
            LIMIT 1;
          IF FOUND THEN
            -- Yes, this is a birth date for some progeny of the individual.
            this_state := 'bb_lactating';
            -- (this_pid stays the same)
            this_dins := 1;
          ELSE

            -- Are we resuming observation and have to change
            -- our state on that basis?
            SELECT INTO obsstart state
              FROM cycgaps
              WHERE sname = this_sname
                    AND date = this_date
                    AND (code = 'bb_cg_start' OR code = 'bb_cg_point');
            IF FOUND THEN
              IF obsstart = 'bb_cg_pregnant' THEN
                this_state := 'bb_pregnant';
              ELSEIF obsstart = 'bb_cg_lactating' THEN
                this_state := 'bb_lactating';
              ELSE
                this_state := 'bb_cycling';
              END IF;

              IF this_state = 'bb_pregnant' OR this_state = 'bb_lactating' THEN
                -- We've got to find the pid that goes with the
                -- pregnancy.
                -- Assume it's the most recent conception prior
                -- to the this_date (which we know is not a conception.)
                SELECT INTO this_pid pregs.pid
                  FROM pregs, cycpoints, cycles
                  WHERE pregs.conceive = cycpoints.cpid
                        AND cycpoints.date < this_date
                        AND cycles.cid = cycpoints.cid
                        AND cycles.sname = this_sname
                  ORDER BY cycpoints.date DESC
                  LIMIT 1;

              END IF;

              -- We've been in a gap and don't know when the state started.
              this_dins := NULL;

            END IF;
          END IF;
        END IF;
      END IF;

      -- Convenient error reporting to isolate problems.
      IF this_state IS NULL THEN
        -- Let the db raise the error and reject the row, need be.
        RAISE NOTICE '% has a NULL state for % due to a gap'
                     , this_sname
                     , this_date;
      END IF;

      -- Insert the repstats row.
      INSERT INTO repstats (sname, date, state, dins, pid)
             VALUES (this_sname, this_date, this_state, this_dins, this_pid);

      this_dins := this_dins + 1;
    ELSE
      -- In a gap.
      -- Forget what we know.
      this_state := NULL;
      this_dins := NULL;
      this_pid := NULL;

    END IF;     

    IF this_state <> last_state
       OR (this_state IS NULL AND last_state IS NOT NULL)
       OR (this_state IS NOT NULL AND last_state IS NULL) THEN
      -- The state has changed, ...

      IF last_state IS NOT NULL THEN
        -- ... but not because we just came out of a gap.
        -- Go back and update the prior repstats
        -- with the days remaining.
        IF this_dins IS NOT NULL THEN
          -- We didn't end with a gap, so we know when the
          -- state ended.
          -- Go back supply days remaining.
          UPDATE repstats
            SET dr = (this_date - '1 day'::interval)::date - date
            WHERE sname = this_sname
                  AND date < this_date
                  AND date >= last_date;
        END IF;

        -- Update the pid
        UPDATE repstats
          SET pid = this_pid
            WHERE sname = this_sname
                  AND date < this_date
                  AND date >= last_date
                  AND pid IS NULL;

        END IF;

      -- Whether we came out of a gap or not we need
      -- to track the state change.
      last_date := this_date;
      last_state := this_state;

    END IF;

    this_date := this_date + '1 day'::interval;

  END LOOP;

  RETURN 0;
  END;
$$;
grant_func_priv(`rebuild_repstats(cycles.sname%TYPE)')
