dnl Copyright (C) 2014, 2016-2017 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2011-2014, 2016 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004 - 2006, 2008, 2010, 2011 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl
include(`func_macros.m4')dnl

dnl Don't output anything while defining macros.
divert(-1)

changequote({,})
dnl
dnl Update mother's cycle's for birth
dnl
dnl syntax: update_mothers_cycles(which)
dnl
dnl Input:
dnl  which   NEW to use the new pregnancy pid
dnl          OLD to use the old pregnancy pid
dnl
define({update_mothers_cycles}, {dnl
    -- Update mother's cycle's for birth.
    DECLARE
      mother_sname biograph.sname%TYPE;

    BEGIN
      SELECT INTO mother_sname cycles.sname
        FROM pregs, cycles
        WHERE pregs.pid = $1.pid
              AND pregs.conceive = cycles.cid
        LIMIT 1;
      IF FOUND THEN
-- Do something clever.
--        PERFORM rebuild_cyclestats(mother_sname);
      END IF;
    END;
})
changequote(`,')


changequote({,})
dnl
dnl Error when a pregnancy with a resume has no progney.
dnl
dnl syntax: pid_check()
dnl
define({pid_check}, {dnl
      -- Got rid of the link to the mother
      -- Mother's pregnancy must have an offspring to have a resume.
      PERFORM 1
        FROM pregs
        WHERE pregs.pid = OLD.pid
              AND pregs.resume IS NOT NULL
              AND NOT EXISTS (SELECT 1 FROM biograph
                                       WHERE biograph.pid = pregs.pid)
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION '%: Cannot change BIOGRAPH.Pid, %: PREGS.Resume must be NULL when there is no progeny'
                      , OLD.sname
                      , OLD.pid;
        RETURN NULL;
      END IF;
})
changequote(`,')

dnl Plpgsql fragment for checking changes to BIOGRAPH.Entrydate
dnl and ensuring that a column does not have a value less than
dnl (i.e. before) the biograph entrydate.
dnl
dnl Syntax: rev_check_entrydate(table, column)
dnl
dnl Restrictions: Must be used in a FOR EACH ROW trigger function.
changequote({,})
define({rev_check_entrydate},{
  PERFORM $1.$2 FROM $1
    WHERE $1.sname = NEW.sname
          AND NEW.entrydate > CAST($1.$2 AS DATE)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Value (Entrydate) = (' || NEW.Entrydate
                     || '): BIOGRAPH.Entrydate cannot be '
                     || 'greater than $1.$2';
    RETURN NULL;
  END IF;
})
changequote(`,')


dnl plpgsql fragment for checking changes to biograph.statdate
dnl and ensuring that a column does not
dnl have a value bigger than the biograph statdate,
dnl when the individual has a status showing them dead.
dnl Syntax: rev_check_statdate(table, column)
dnl Restrictions: Must be used in a FOR EACH ROW trigger function
dnl on biograph.
changequote({,})
define({rev_check_statdate},{
  PERFORM $1.$2 FROM $1
    WHERE $1.sname = NEW.sname
          AND NEW.Statdate < CAST($1.$2 AS DATE)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: BIOGRAPH.Statdate cannot be less than $1.$2'
                    , NEW.sname;
    RETURN NULL;
  END IF;
})
changequote(`,')



dnl plpgsql fragment for checking that an individual is not
dnl a dad on DAD_DATA.
dnl
dnl Syntax: _not_a_dad(column)
dnl
dnl column  A dad column on DAD_DATA.
dnl
dnl Restrictions: Must be used in a FOR EACH ROW trigger function
dnl on biograph.
changequote({,})
define({_not_a_dad},{
      -- Must be a male to be a ($1) dad.
      PERFORM 1
        FROM dad_data
        WHERE dad_data.$1 = NEW.sname
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION '%: Sex %: Individual must be male because he appears as a DAD_DATA.$1 value'
                      , NEW.sname
                      , NEW.sex;
        RETURN NULL;
      END IF;
})
changequote(`,')



dnl plpgsql fragment for checking that an individual is 
dnl dad that's alive at the time of conception.
dnl
dnl Syntax: _is_living_dad(column)
dnl
dnl column  A dad column on DAD_DATA.
dnl
dnl Restrictions: Must be used in a FOR EACH ROW trigger function
dnl on biograph.
changequote({,})
define({_is_living_dad},{

      -- The ($1) dad must be alive during the mother''s fertile period.
      PERFORM 1
        FROM dad_data
             JOIN biograph ON (biograph.sname = dad_data.kid)
             JOIN pregs ON (pregs.pid = biograph.pid)
             JOIN cycpoints ON (cycpoints.cpid = pregs.conceive)
        WHERE dad_data.$1 = NEW.sname
              AND NEW.statdate
                    < cycpoints.date
                      - bb_ovulation_len
                      - CASE WHEN NEW.status = bb_alive
                               THEN bb_interpolation_max
                             ELSE 0
                        END
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 'Bioid %: Sname %: Status %: Statdate %: Dad % (DAD_DATA.$1) is not alive during the mom''s fertile period -- on or after the Zdate - bb_ovulation_len (- another bb_interpolation_max if the dad is alive)'
                      , NEW.bioid
                      , NEW.sname
                      , NEW.Status
                      , NEW.Statdate
                      , NEW.Sname;
        RETURN NULL;
      END IF;
})
changequote(`,')




dnl plpgsql fragment for checking that an individual has a
dnl dad that's alive at the time of conception.
dnl
dnl Syntax: _has_living_dad(column)
dnl
dnl column  A dad column on DAD_DATA.
dnl
dnl Restrictions: Must be used in a FOR EACH ROW trigger function
dnl on biograph.
changequote({,})
define({_has_living_dad},{

    -- Is the $1 dad of this individual alive?
    PERFORM 1
      FROM dad_data
           JOIN biograph ON (biograph.sname = dad_data.$1)
         , pregs
           JOIN cycpoints ON (cycpoints.cpid = pregs.conceive)          
      WHERE dad_data.kid = NEW.sname
            AND pregs.pid = NEW.pid
            AND biograph.statdate
                < cycpoints.date
                  - bb_ovulation_len
                  - CASE WHEN biograph.status = bb_alive
                         THEN bb_interpolation_max
                         ELSE 0
                    END
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Bioid %: Sname %: Has a related DAD_DATA.$1 dad who is not alive during the mom''s fertile period -- on or after the Zdate - bb_ovulation_len (- another bb_interpolation_max if the dad is alive)'
                      , NEW.bioid
                      , NEW.sname;
      RETURN NULL;
    END IF;

    -- Is the $1 dad of this individual mature?
    PERFORM 1
      FROM dad_data
           JOIN biograph ON (biograph.sname = dad_data.$1)
           LEFT OUTER JOIN maturedates ON (maturedates.sname = dad_data.$1)
         , pregs
           JOIN cycpoints ON (cycpoints.cpid = pregs.conceive)          
      WHERE dad_data.kid = NEW.sname
            AND pregs.pid = NEW.pid
            AND (maturedates.matured IS NULL
                 OR cycpoints.date < maturedates.matured)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Bioid %: Sname %: Has a related DAD_DATA.$1 dad who is not mature on the Zdate'
                      , NEW.bioid
                      , NEW.sname;
      RETURN NULL;
    END IF;
})
changequote(`,')

dnl Re-enable output
divert`'dnl

--
-- biograph
-- $Id: biograph.m4,v 1.18 2004/11/08 02:23:09 kop Exp $
--
SELECT 'biograph' AS table;


SELECT 'biograph_before_insert_func' AS biograph_before_insert_func;
CREATE FUNCTION biograph_before_insert_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  -- Function for biograph before insert trigger
  --
  -- Copyright (C) 2011 The Meme Factory, Inc.
  -- GPL_notice(`  --', `2004-2006, 2008, 2010, 2011', `Karl O. Pinc <kop@meme.com>')

  BEGIN

    -- Assign the Alt_Snames value.
    PERFORM 1
      FROM alternate_snames
      WHERE alternate_snames.sname = NEW.sname
            OR alternate_snames.sname_alternate = NEW.sname
      LIMIT 1;

    NEW.alt_snames = FOUND;

  RETURN NEW;
  END;
$$;


SELECT 'biograph_func' AS biograph_func;
CREATE FUNCTION biograph_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  -- Function for biograph insert or update trigger
  --
  -- Copyright (C) 2011-2013, 2016 The Meme Factory, Inc.  http://www.meme.com/
  -- GPL_notice(`  --', `2004-2006, 2008, 2010, 2011', `Karl O. Pinc <kop@meme.com>')

  BEGIN

  -- sname can"t change

  IF TG_OP = 'UPDATE' THEN
    cannot_change(`BIOGRAPH', `Bioid')
    cannot_change(`BIOGRAPH', `Sname') -- Allows change from/to NULL.

    -- Don't allow a sname to change to NULL, that will leave
    -- members rows for a stillborn.  (We could delete the
    -- members rows in such a case, but that'd take a separate
    -- procedure as the ordinary user's don't have the permissions.
    -- Besides, the whole thing has not been thought through.
    -- The less magic the better.)
    IF OLD.sname IS NOT NULL AND NEW.sname IS NULL THEN
      RAISE EXCEPTION '%: Cannot change BIOGRAPH.Sname from not NULL to NULL'
                    , OLD.sname;
      RETURN NULL;
    END IF;
  END IF;

  -- sname can't be a unksname
  IF TG_OP = 'INSERT' THEN
    PERFORM 1
      FROM unksnames
      WHERE unksnames.unksname = NEW.sname
    LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Cannot have a BIOGRAPH.Sname that is also a UNKSNAMES.Unksname'
                    , NEW.sname;
      RETURN NULL;
    END IF;
  END IF;

  -- sname can't be a partunks

  IF TG_OP = 'INSERT' THEN
    PERFORM 1
      FROM partunks
      WHERE partunks.unksname = NEW.sname
    LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Cannot have a BIOGRAPH.Sname that is also a PARTUNKS.Unksname'
                    , NEW.sname;
      RETURN NULL;
    END IF;
  END IF;

  -- sname can't be a alternate_sname

  IF TG_OP = 'INSERT' THEN
    PERFORM 1
      FROM alternate_snames
      WHERE alternate_snames.sname_alternate = NEW.sname
    LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Cannot have a BIOGRAPH.Sname that is also a ALTERNATE_SNAMES.Sname_Alternate'
                    , NEW.sname;
      RETURN NULL;
    END IF;
  END IF;

  -- Some quick updating of other tables, so as not to confuse
  -- the remaining tests.
  IF TG_OP = 'UPDATE' THEN
    IF (OLD.status = 'bb_alive' AND NEW.status <> 'bb_alive')
       OR (NEW.status <> 'bb_alive' AND NEW.statdate < OLD.statdate) THEN
      -- Automatic Mdates after the statdate should automatically
      -- go away.
      DELETE FROM cycpoints
             USING cycles
             WHERE cycles.sname = NEW.sname
                   AND cycpoints.cid = cycles.cid
                   AND cycpoints.source = 'bb_auto'
                   AND cycpoints.code = 'bb_menses'
                   AND cycpoints.date > NEW.statdate;
    END IF;
  END IF;

  -- Back to validating data.

  -- sex
  IF TG_OP = 'UPDATE' THEN
    IF OLD.sex <> NEW.sex THEN
      -- Trying to change the sex.

      -- Can't change the sex of a female with related cycpoints
      PERFORM * FROM cycles WHERE sname = OLD.sname LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION
              '%: Cannot change sex from female, related CYCLES exist'
              , NEW.sname;
        RETURN NULL;
      END IF;

      -- Can't change the sex of a female with related cycgaps
      PERFORM * FROM cycgaps WHERE sname = OLD.sname LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION
              '%: Cannot change sex from female, related CYCGAPS exist'
              , NEW.sname;
        RETURN NULL;
      END IF;

      -- Can't change the sex of a male with a dispersal date.
      PERFORM * FROM dispersedates WHERE sname = OLD.sname LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION
              '%: Cannot change the sex from male, related DISPERSEDATES exist'
              , NEW.sname;
        RETURN NULL;
      END IF;

      -- Can't change the sex of a female with female point samples.
      PERFORM 1
        FROM samples
        WHERE samples.sname = NEW.sname
              AND samples.stype = 'bb_female_point'
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION '%: Cannot change the sex from female, related female SAMPLES exist'
                        , NEW.sname;
        RETURN NULL;
      END IF;

      -- Can't change the sex of a male in the MPI consorts table.
      PERFORM 1
        FROM consorts
        WHERE consorts.had = NEW.sname
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION '%: Cannot change the sex from male, related CONSORTS.Had exist'
                        , NEW.sname;
        RETURN NULL;
      END IF;
      PERFORM 1
        FROM consorts
        WHERE consorts.got = NEW.sname
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION '%: Cannot change the sex from male, related CONSORTS.Got exist'
                        , NEW.sname;
        RETURN NULL;
      END IF;

      -- Confirm sex-specific darting samples match the individual's sex.
      PERFORM 1
        FROM dartings
             JOIN dart_samples ON (dartings.dartid = dart_samples.dartid)
             JOIN dart_sample_types ON (dart_samples.ds_type = dart_sample_types.ds_type)
        WHERE dartings.sname = NEW.sname
             AND dart_sample_types.sex IS NOT NULL
             AND dart_sample_types.sex <>  NEW.sex
        LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION integrity_constraint_violation USING
                MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
              , DETAIL = 'Key (Sname) = (' || NEW.sname
                           || '): Value (Sex) = (' || NEW.sex
                           || '): Cannot have sex-specific tissues in DART_SAMPLES that'
                           || ' do not match the individual''s sex';
          RETURN NULL;
        END IF;

      -- Cannot have related TESTES rows unless you're male.
      IF NEW.sex <> 'bb_male' THEN
        PERFORM 1
          FROM dartings
               JOIN testes_arc ON (testes_arc.dartid = dartings.dartid)
          WHERE dartings.sname = NEW.sname
          LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION 'Bioid %: Sname %: Sex %: Cannot have related TESTES_ARC rows unless the individual is male'
                        , NEW.bioid
                        , NEW.sname
                        , NEW.sex;
          RETURN NULL;
        END IF;

        PERFORM 1
          FROM dartings
               JOIN testes_diam ON (testes_diam.dartid = dartings.dartid)
          WHERE dartings.sname = NEW.sname
          LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION 'Bioid %: Sname %: Sex %: Cannot have related TESTES_DIAM rows unless the individual is male'
                        , NEW.bioid
                        , NEW.sname
                        , NEW.sex;
          RETURN NULL;
        END IF;
      END IF;
    END IF;
  END IF;

  --
  -- birthdate
  --

  -- MATURES.Maturedate

  -- Maturedate must be at least 2.7ish years of age.
  rev_check_age_ge_days(`MATUREDATES', `Matured', bb_min_maturity_age)

  -- When we observed maturity then it must be reached before 7.5ish
  -- years of age.
  PERFORM 0
    FROM maturedates
    WHERE sname = NEW.sname
          AND maturedates.mstatus = 'bb_on'
    LIMIT 1;
  IF FOUND THEN
    rev_check_age_lt_days(`MATUREDATES', `Matured', bb_max_maturity_age)
  END IF;

  -- Maturedate must be on or before statdate
  rev_check_statdate(`MATUREDATES', `Matured')

  -- RANKDATES.Ranked

  -- Ranked must be at least 5 years of age.
  rev_check_age_ge_years(`RANKDATES', `Ranked', `bb_min_rankdate_age')

  -- When we observed ranking then it must be reached before 12
  -- years of age.
  PERFORM 0
    FROM rankdates
    WHERE sname = NEW.sname
          AND rankdates.rstatus = 'bb_on'
    LIMIT 1;
  IF FOUND THEN
    rev_check_age_lt_years(`RANKDATES', `Ranked', `bb_max_rankdate_age')
  END IF;

  -- Ranked must be on or before statdate
  rev_check_le(`RANKDATES', `Ranked', `BIOGRAPH', `Statdate')

  -- CONSORTDATES.Consorted

  -- Consorted must be at least 5 years of age.
  rev_check_age_ge_years(`CONSORTDATES', `Consorted', `bb_min_rankdate_age')

  -- Consorted must be on or after entrydate
  rev_check_entrydate(`CONSORTDATES', `Consorted')

  -- Consorted must be on or before statdate
  rev_check_statdate(`CONSORTDATES', `Consorted')

  --
  -- DISPERSEDATES.Dispersed
  --

  IF TG_OP = 'UPDATE' THEN
    -- Otherwise we can't have dispersedate at all.

    -- Dispersed must be on or after Entrydate
    rev_check_entrydate(`DISPERSEDATES', `Dispersed')

    IF NEW.statdate < OLD.statdate   -- Optimization
       OR NEW.status <> OLD.status THEN
      DECLARE
        this_dispersed dispersedates.dispersed%TYPE;

        this_group      census.grp%TYPE;
        last_absence    census.date%TYPE;

      BEGIN
        -- Find the individual's dispersal date.
        SELECT INTO this_dispersed dispersed
          FROM dispersedates
          WHERE dispersedates.sname = NEW.sname;

        IF FOUND THEN
          IF NEW.statdate < this_dispersed THEN
            IF NEW.status = 'bb_alive' THEN
              -- Find the group in which the individual was last censused.
              SELECT INTO this_group census.grp
                FROM census
                WHERE census.sname = NEW.sname
                      AND census.status <> 'bb_absent'
                ORDER BY census.date DESC
                LIMIT 1;
              IF NOT FOUND THEN
                -- Error, can't be after statdate when no census.
                RAISE EXCEPTION '%: DISPERSEDATES.Dispersed cannot be after BIOGRAPH.Statdate (%) when the individual has never been censused present in a group'
                              , NEW.sname
                              , NEW.statdate;
                RETURN NULL;
              ELSE
                -- Find the last absence in the censused group.
                SELECT INTO last_absence census.date
                  FROM census
                  WHERE census.sname = NEW.sname
                        AND census.grp = this_group
                        AND census.date > NEW.statdate
                        AND census.status = 'bb_absent'
                  ORDER BY census.date
                  LIMIT 1;
                IF NOT FOUND OR last_absence < this_dispersed THEN
                  RAISE EXCEPTION '%: DISPERSEDATES.Dispersed that are after the Statdate (%) must be on or before the next recorded absence from the group in which the individual was last censused'
                          , NEW.sname
                          , NEW.statdate;
                  RETURN NULL;
                END IF;
              END IF;
            ELSE
              -- Individual is not alive.
              -- Dispersed must be on or before statdate
              RAISE EXCEPTION '%: DISPERSEDATES.Dispersed cannot be after the Statdate (%) unless the individual is alive'
                            , NEW.sname
                            , NEW.statdate;
              RETURN NULL;
            END IF;
          END IF;
        END IF;
      END;
    END IF;
  END IF;

  IF TG_OP = 'UPDATE' THEN
    -- Since biograph rows must exist first there's no point in checking
    -- on INSERT.

    -- SWERB observation of a lone animal can't be before the entrydate.
    IF NEW.entrydate > OLD.entrydate THEN
      PERFORM 1
        FROM swerb_data
             JOIN swerb_bes ON (swerb_bes.beid = swerb_data.beid)
             JOIN swerb_departs_data ON (swerb_departs_data.did = swerb_bes.did)
        WHERE swerb_data.lone_animal = NEW.sname
              AND NEW.entrydate > swerb_departs_data.date
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
            , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.Sname
                     || '): Value (Entrydate) = (' || NEW.Entrydate
                     || '): Cannot have a SWERB lone animal sighting before '
                     || 'an individual enters the study population';
        RETURN NULL;
      END IF;
    END IF;

    -- SWERB observation of a lone male can't be after the statdate.
    IF NEW.statdate < OLD.statdate THEN
      PERFORM 1
        FROM swerb_data
             JOIN swerb_bes ON (swerb_bes.beid = swerb_data.beid)
            JOIN swerb_departs_data ON (swerb_departs_data.did = swerb_bes.did)
        WHERE swerb_data.lone_animal = NEW.sname
              AND NEW.statdate < swerb_departs_data.date
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 'Sname %: Statdate %: Must still be alive when observed as a SWERB lone animal'
                      , NEW.sname
                      , NEW.statdate;
        RETURN NULL;
      END IF;
    END IF;
  END IF;

  --
  -- Birthdate comparison to start and end of pregnancy.
  -- 

  -- Conception cycpoints date not later than birthdate.
  PERFORM NEW.sname FROM pregs, cycpoints
    WHERE pregs.pid = NEW.pid
          AND cycpoints.cpid = pregs.conceive
          AND (cycpoints.date > NEW.birth
                OR cycpoints.ldate > NEW.birth)
    LIMIT 1;
--          AND ( ( (NEW.birth < cycpoints.date) IS TRUE )
--                OR ( (NEW.birth < cycpoints.ldate) IS TRUE ) );
  IF FOUND THEN
    RAISE EXCEPTION '%: BIOGRAPH.Birth is before one of the PREGS.Conceive CYCPOINTS dates'
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- Birthdate cannot follow resumption cycpoints date.
  PERFORM NEW.sname FROM pregs, cycpoints
    WHERE pregs.pid = NEW.pid
          AND cycpoints.cpid = pregs.resume
          AND (cycpoints.date < NEW.birth
                OR cycpoints.edate < NEW.birth)
    LIMIT 1;
--          AND ( ( (NEW.birth > cycpoints.date) IS TRUE )
--                OR ( (NEW.birth > cycpoints.edate) IS TRUE ) );
  IF FOUND THEN
    RAISE EXCEPTION '%: BIOGRAPH.Birth is after one of the PREGS.Resume CYCPOINTS dates'
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- Pid -- Can't have a resume with no progeny.
  IF TG_OP = 'UPDATE' THEN
    IF OLD.pid IS NOT NULL
       AND OLD.pid <> NEW.pid THEN
      pid_check()
    END IF;
  END IF;

  --
  -- DAD_DATA
  --

  IF TG_OP = 'UPDATE' THEN

    IF NEW.pid IS NOT NULL THEN
      -- Check to see that listed dads are alive at the time
      -- of conception.
      _has_living_dad(`dad_excl')    
      _has_living_dad(`dad_1perr')    
      _has_living_dad(`dad_5perr')    
      _has_living_dad(`dad_allmales')    
      _has_living_dad(`dad_consensus')    
    END IF;
    --
    -- Check to see if this individual can have a dad.
    --

    -- Can't have DAD_DATA row for the offspring unless the kid exists
    -- at the time the genetic were analyzed.
    IF NEW.birth IS NULL THEN
      PERFORM 0
        FROM dad_data
        WHERE dad_data.kid = NEW.sname
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 'Sname %: Bioid %: Cannot have a DAD_DATA row for the offspring when there is no birth date'
                      , NEW.sname
                      , NEW.bioid;
        RETURN NULL;
      END IF;
    ELSE
      -- There is a birth date.
      IF NEW.birth <> OLD.birth THEN
        -- DAD_DATA.Date
        PERFORM 0
          FROM dad_data
          WHERE dad_data.kid = NEW.sname
                AND dad_data.date < NEW.birth
          LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION 'Sname %: Bioid %: Birth %: The birthdate of the offspring cannot be after DAD_DATA.Date'
                        , NEW.sname
                        , NEW.bioid
                        , NEW.birth;
          RETURN NULL;
        END IF;
      END IF;
    END IF;

    --
    -- Be sure this individual can be selected as a dad.
    --
    IF NEW.sex = 'bb_male' THEN
      -- A dad must be alive during the mother's fertile period.
      _is_living_dad(`Dad_excl')
      _is_living_dad(`Dad_1perr')
      _is_living_dad(`Dad_5perr')
      _is_living_dad(`Dad_allmales')
      _is_living_dad(`Dad_consensus')
    ELSE
      -- Gotta be male to be a dad.
      _not_a_dad(`Dad_excl')
      _not_a_dad(`Dad_1perr')
      _not_a_dad(`Dad_5perr')
      _not_a_dad(`Dad_allmales')
      _not_a_dad(`Dad_consensus')
    END IF;

  END IF;

  -- CYCPOINTS
  IF NEW.sex = 'bb_female' THEN
    -- Birthdate must be strictly before all cycpoints.
    -- (Really the check could be more restrictve as most
    -- cycpoints must be after maturity but this is easiest and
    -- probably most efficient.)
    -- Check Edates too (which is not sufficent for
    -- a Date check as some are NULL.)
    PERFORM 1
      FROM cycles, cycpoints
      WHERE cycles.sname = NEW.sname
            AND cycpoints.cid = cycles.cid
            AND (cycpoints.date <= NEW.birth
                         OR cycpoints.edate <= NEW.birth)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: %: BIOGRAPH.Birth must be before CYCPOINTS.Date and CYCPOINTS.Edate'
                    , NEW.sname
                    , NEW.birth;
      RETURN NULL;
    END IF;

  -- All Edates and Dates must be on or before the statdate, excepting
  -- automatic Mdates when the individual is alive.
    PERFORM 1
      FROM cycles, cycpoints
      WHERE cycles.sname = NEW.sname
            AND cycpoints.cid = cycles.cid
            AND (NEW.statdate < cycpoints.date
                 OR NEW.statdate < cycpoints.edate)
            AND ((cycpoints.code <> 'bb_menses'
                  AND cycpoints.source <> 'bb_auto')
                 OR NEW.status <> 'bb_alive')
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: %: BIOGRAPH.Statdate may not be before CYCPOINTS.Date or CYCPOINTS.Edate unless the individual is alive and the CYCPOINTS row is an automatic Mdate'
                    , NEW.sname
                    , NEW.statdate;
      RETURN NULL;
    END IF;
  END IF;

  -- CYCLES

  -- SEXSKINS
  -- Must be after the birthdate.
  PERFORM 1
    FROM sexskins, cycles
    WHERE cycles.sname = NEW.sname
          AND sexskins.cid = cycles.cid
          AND NEW.birth >= sexskins.date;
  IF FOUND THEN
    RAISE EXCEPTION ' %: BIOGRAPH.Birth must be before SEXSKINS.Date'
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- Must be before the statdate.
  PERFORM 1
    FROM sexskins, cycles
    WHERE cycles.sname = NEW.sname
          AND sexskins.cid = cycles.cid
          AND NEW.statdate < sexskins.date;
  IF FOUND THEN
    RAISE EXCEPTION ' %: BIOGRAPH.statdate cannot be before SEXSKINS.Date'
                    , NEW.sname;
    RETURN NULL;
  END IF;


  --
  -- CENSUS
  --

  -- Birth must be on or before first non-absent census.
  -- PERFORM date
  --   FROM census
  --   WHERE sname = NEW.sname
  --         AND date < NEW.birth
  --         AND status <> 'bb_absent';

  -- This query performs very badly as of 8.0.3.  It would
  -- require an index on name and date, in addition to the
  -- existing index on date and name, to perform.  Unfortunately
  -- the planner does not use such an index if I create it.
  -- So, do this query as infrequently as possible.
  -- (All this only really matters when inserting lots
  -- of census rows which change the statdate of live individuals.
  -- By lots I mean thousands, not just your regular census update.
  -- There's apparently a planner in the works that will fix this.
  -- See:
  -- http://archives.postgresql.org/pgsql-performance/2005-06/msg00546.php
  DECLARE
    do_test BOOLEAN := TRUE;

  BEGIN
    IF TG_OP = 'UPDATE' THEN
      -- Only test when birth changes.
      do_test := OLD.birth <> NEW.birth;
    END IF;

    IF do_test THEN
      PERFORM date
        FROM census
        WHERE sname = NEW.sname
              AND date < NEW.birth
              AND status <> 'bb_absent'
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 
              '%: %: Birth cannot be after first non-absent census'
              , NEW.sname
              , NEW.birth;
        RETURN NULL;
      END IF;
    END IF;
  END;

  -- Now that we know birth is ok vis census, we can check the statdate.
  DECLARE
    last_census census.date%TYPE;

  BEGIN
    SELECT INTO last_census date
      FROM census
      WHERE sname = NEW.sname
            AND status <> 'bb_absent'
      ORDER BY date DESC
      LIMIT 1;

    IF last_census IS NOT NULL THEN
      IF NEW.status <> 'bb_alive' THEN
        -- Statdate for dead individuals must be on or after last census.
        IF NEW.statdate < last_census THEN
          RAISE EXCEPTION '%: %: Statdate of dead individuals must be >= %, the last census date'
                          , NEW.sname
                          , NEW.statdate
                          , last_census;
          RETURN NULL;
        END IF;
      ELSE
        IF NEW.statdate <> last_census THEN
          RAISE EXCEPTION '%: %: Statdate of live individuals must be = %, the last census date'
                          , NEW.sname
                          , NEW.statdate
                          , last_census;
          RETURN NULL;
        END IF;
      END IF;
    ELSE
      -- No locating censuses, they're none or all absences.
      IF NEW.status = 'bb_alive' AND NEW.statdate <> NEW.entrydate THEN
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
            , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                         || '): Value (Sname) = (' || NEW.sname
                         || '): Value (Entrydate) = (' || NEW.entrydate
                         || '): Value (Statdate) = (' || NEW.statdate
                         || '): Statdate of live individuals without '
                         || 'census rows (absences excepted), must be '
                         || 'the Entrydate';
        RETURN NULL;
      END IF;
    END IF;
  END;

  --
  -- INTERACT_DATA and PARTS
  --

  -- Entrydate cannot be later than interaction dates.
  PERFORM 1
    FROM parts, interact_data, acts
    WHERE NEW.sname = parts.sname
          AND interact_data.iid = parts.iid
          AND acts.act = interact_data.act
          AND (((acts.class = 'bb_grooming'
                  AND interact_data.date < 'bb_grooming_transition')
                AND rnkdate(NEW.entrydate) > interact_data.date)
               OR (NOT (acts.class = 'bb_grooming'
                         AND interact_data.date < 'bb_grooming_transition')
                   AND NEW.entrydate > interact_data.date))
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Value (Entrydate) = (' || NEW.entrydate
                     || '): BIOGRAPH.Entrydate cannot be bigger than '
                     || 'INTERACT_DATA.Date, or in the case of bb_grooming before '
                     || 'bb_grooming_transition the month and year of '
                     || 'BIOGRAPH.Entrydate cannot be bigger than '
                     || 'INTERACT_DATA.Date';
    RETURN NULL;
  END IF;

  -- Statdate cannot be earlier than interaction dates.
  -- (We don't care about bb_grooming_transition date because this
  -- just makes the interaction date earlier.)
  PERFORM 1
    FROM parts, interact_data --, acts
    WHERE NEW.sname = parts.sname
          AND interact_data.iid = parts.iid
          -- AND acts.act = interact_data.act
          -- AND (((acts.class = 'bb_grooming'
          --         AND interact_data.date < 'bb_grooming_transition')
          --       AND date_trunc('month', NEW.statdate) < interact_data.date)
          --      OR (NOT ((acts.class = 'bb_grooming'
          --                 AND interact_data.date < 'bb_grooming_transition')
                   AND NEW.statdate < interact_data.date
          --                                            ))
    LIMIT 1;
  IF FOUND THEN
    -- RAISE EXCEPTION '%: %: BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date, or in the case of bb_grooming before bb_grooming_transition the month and year of BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date'
    RAISE EXCEPTION '%: %: BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date'
                    , NEW.sname
                    , NEW.statdate;
    RETURN NULL;
  END IF;

  -- Males can't have mounts, consortships, or ejaculations before 4
  -- years of age.
  PERFORM 1
    FROM parts, interact_data, acts
    WHERE NEW.sex = 'bb_male'
          AND parts.sname = NEW.sname
          AND interact_data.iid = parts.iid
          AND acts.act = interact_data.act
          AND (acts.class = 'bb_mounts'
               OR acts.class = 'bb_consortship'
               OR acts.class = 'bb_ejaculation')
          AND interact_data.date - CAST('bb_min_male_mce_age' AS interval)
              < NEW.birth
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: BIOGRAPH.Birth cannot be within bb_min_male_mce_age of INTERACT_DATA.Date for those interactions that are mounts, consortships, or ejaculations'
                    , NEW.sname
                    , NEW.birth;
     RETURN NULL;
  END IF;

  --
  -- MPIS, MPI_DATA and MPI_PARTS
  --

  -- Entrydate cannot be later than MPI interaction dates.
  PERFORM 1
    FROM mpi_parts, mpi_data, mpis
    WHERE NEW.sname = mpi_parts.sname
          AND mpi_data.mpidid = mpi_parts.mpidid
          AND mpis.mpiid = mpi_data.mpiid
          AND mpis.date < NEW.entrydate
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.Sname
                     || '): Value (Entrydate) = (' || NEW.Entrydate
                     || '): BIOGRAPH.Entrydate cannot be bigger than '
                     || 'MPIS.Date';
    RETURN NULL;
  END IF;

  -- Statdate cannot be earlier than mpi interaction dates.
  PERFORM 1
    FROM mpi_parts, mpi_data, mpis
    WHERE NEW.sname = mpi_parts.sname
          AND mpi_data.mpidid = mpi_parts.mpidid
          AND mpis.mpiid = mpi_data.mpiid
          AND NEW.statdate < mpis.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: BIOGRAPH.Statdate cannot be less than MPIS.Date.'
                    , NEW.sname
                    , NEW.statdate;
    RETURN NULL;
  END IF;

  -- Point samples SAMPLES, NEIGHBORS.

  -- SAMPLES

  -- Point sample data must be taken on a date when the individual
  -- is alive.
  rev_check_statdate(`SAMPLES', `Date')

  -- Point sample data must be taken on or after the focal
  -- individual's entrydate. (also their birth)
  rev_check_entrydate(`SAMPLES', `Date')

  -- NEIGHBORS

  -- Neighbor must have entered the study population.
  PERFORM 1 FROM neighbors, point_data, samples
    WHERE neighbors.sname = NEW.sname
          AND point_data.pntid = neighbors.pntid
          AND samples.sid = point_data.sid
          AND NEW.entrydate > samples.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Value (Entrydate) = (' || NEW.entrydate
                     || '): Neighbor in NEIGHBORS cannot have a BIOGRAPH.'
                     || 'Entrydate after the SAMPLES.Date';
    RETURN NULL;
  END IF;

  -- Neighbor must still be alive and uncensored on SAMPLES.date.
  PERFORM 1 FROM neighbors, point_data, samples
    WHERE neighbors.sname = NEW.sname
          AND point_data.pntid = neighbors.pntid
          AND samples.sid = point_data.sid
          AND NEW.Statdate < samples.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Value (Statdate) = (' || NEW.statdate
                     || '): BIOGRAPH.Statdate for a neighbor cannot be less '
                     || 'than SAMPLES.Date';
    RETURN NULL;
  END IF;

  --
  -- DARTINGS
  --

  -- Cannot be darted after the statdate.
  rev_check_statdate(`DARTINGS', `Date')

  -- Cannot be darted before entrydate.
  rev_check_entrydate(`DARTINGS', `Date')

  --
  -- Hybrid Scores
  --

  -- We don't care about statdate. Hybrid Scores can happen after death.

  -- Cannot have a hybrid score on or before entrydate
  PERFORM 1
    FROM hybridgene_analyses, hybridgene_scores
    WHERE hybridgene_analyses.hgaid = hybridgene_scores.hgaid
          AND hybridgene_scores.sname = NEW.sname
          AND hybridgene_analyses.date < NEW.entrydate
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of BIOGRAPH'
        , DETAIL = 'Key (Bioid) = (' || NEW.bioid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Value (Entrydate) = (' || NEW.entrydate
                     || '): Individual''s Entrydate cannot be after the '
                     || 'HYBRIDGENE_ANALYSES.Date of an analysis that '
                     || 'scores him or her in HYBRIDGENE_SCORES'
        , HINT = 'Can''t have a hybrid score before entering population';
    RETURN NULL;
  END IF;

  IF TG_OP = 'UPDATE' then
    --
    -- Alt_snames
    --
    -- This need only be checked on update, since the before insert trigger
    -- does the check on insert.
    PERFORM 1
      FROM alternate_snames
      WHERE alternate_snames.sname = NEW.sname
            OR alternate_snames.sname_alternate = NEW.sname
      LIMIT 1;
    IF (FOUND AND NEW.alt_snames = FALSE)
       OR (NOT FOUND AND NEW.alt_snames = TRUE) THEN
      RAISE EXCEPTION data_exception USING
            MESSAGE = 'Cannot update BIOGRAPH'
          , DETAIL = 'Key (Sname) = ('
                     || NEW.sname
                     || '): Value (Alt_snames) = ('
                     || NEW.alt_snames
                     || '): Alt_Snames must be true if and only if the individual''s sname matches either a Sname value or an Sname_Alternate value on ALTERNATE_SNAMES';
      RETURN NULL;
    END IF;
  END IF;

  --
  -- Done all the checking, now update the rest of the
  -- database when necessary.
  -- (Separate checks for INSERT and UPDATE as otherwise
  -- the compiler will complain about references to OLD
  -- in an INSERT trigger.
  --

  --
  -- Update members.
  --
  -- This comes first as it may affect other updates.
  --
  IF NEW.sname IS NOT NULL THEN
    -- Abortions don't get rows in members.
    IF TG_OP = 'UPDATE' THEN
      IF NEW.birth <> OLD.birth
         OR NEW.matgrp <> OLD.matgrp THEN
        --
        -- Changes to birthdate result in changes to members.
        --

        -- Find the first non-absent census.
        DECLARE
          -- The date of that census.
          first_date census.date%TYPE;
          -- The group of that census.
          first_group census.grp%TYPE;
          -- The status of that census.
          first_status census.status%TYPE;
          -- Whether or on the first census interpolates.
          interpolates BOOLEAN;
          -- Whether or not the birthdate interpolates.
          interp_birth BOOLEAN;

        BEGIN
          -- SELECT INTO first_date MIN(date)
          --   FROM census
          --   WHERE sname = NEW.sname
          --         AND status <> 'bb_absent';
          SELECT INTO first_date, first_group, first_status
                      date,       grp,         status
            FROM census
            WHERE sname = NEW.sname
                  AND status <> 'bb_absent'
            ORDER BY date
            LIMIT 1;
          IF first_date IS NOT NULL THEN
            interpolates := first_status = 'bb_census'
                             OR first_status = 'bb_demog'
                             OR first_status = 'bb_manual';
            -- It may _seem_ silly to interpolate here when
            -- the birthdate is the first census, but it takes
            -- care of clearing out all the old interpolation
            -- stuff that may be on MEMBERS.
            -- interp_birth := first_date <> NEW.birth
            --                 OR (first_date = NEW.birth
            --                     AND interpolates);
            -- (same as below)
            interp_birth := NOT(first_date = NEW.birth
                                AND NOT(interpolates));
            PERFORM _interp(  NEW.sname
                            , NEW.status = 'bb_alive'
                            , NEW.birth
                            , first_date
                            , NEW.matgrp
                            , first_group
                            , TRUE
                            , FALSE
                            , interp_birth
                            , NEW.status = 'bb_alive');
          ELSE
            -- No locating censuses.
            -- Interpolate from birthdate to statdate.
            PERFORM _interp(  NEW.sname
                            , NEW.status = 'bb_alive'
                            , NEW.birth
                            , NEW.statdate
                            , NEW.matgrp
                            , NULL
                            , TRUE
                            , TRUE
                            , TRUE
                            , FALSE);
          END IF;
        END;
      END IF;

      IF NEW.statdate <> OLD.statdate THEN
        --
        -- Changes to statdate result in changes to members.
        --

        -- Are there non-absent censuses?
        DECLARE
          last_date census.date%TYPE;

        BEGIN

          SELECT INTO last_date date
            FROM census
            WHERE sname = NEW.sname
                  AND status <> 'bb_absent'
            ORDER BY date DESC
            LIMIT 1;
          IF last_date IS NOT NULL THEN

            DECLARE
              -- The group of the last census.
              last_group census.grp%TYPE;
              -- The status of that census.
              last_status census.status%TYPE;
              -- Whether or on the last census interpolates.
              last_interpolates BOOLEAN;

              -- The date of the locating, interpolating census before
              -- the last census.
              prior_date census.date%TYPE;
              -- The group of that census.
              prior_group census.grp%TYPE;
              -- The status of that census.
              prior_status census.status%TYPE;
              -- Whether or on the last census interpolates.
              prior_interpolates BOOLEAN;

              -- Whether or not to optimize by not interpolating
              -- back to the birthdate.
              optimize BOOLEAN;

              -- Whether or not the birthdate interpolates
              birth_interp BOOLEAN;

            BEGIN
              -- Collect the rest of the info about the last census.
              SELECT INTO last_group, last_status
                          grp,        status
                     FROM census
                     WHERE sname = NEW.sname
                           AND date = last_date
                           AND status <> 'bb_absent'
                     LIMIT 1;
              last_interpolates := last_status = 'bb_census'
                                   OR last_status = 'bb_demog'
                                   OR last_status = 'bb_manual';
              IF NEW.status = 'bb_alive' THEN
                -- Find the next to last locating, interpolating census.
                -- (The last is on the statdate (last_date = NEW.statdate)
                -- and we've already got all the info.)
                -- The order by sname does nothing but make the
                -- pg 8.0.3 optimizer behave.
                SELECT INTO prior_date, prior_group, prior_status
                            date,       grp,         status
                  FROM census
                  WHERE sname = NEW.sname
                        AND date < NEW.statdate
                        AND (status = 'bb_census'
                             OR status = 'bb_demog'
                             OR status = 'bb_manual')
                  ORDER BY sname, date DESC
                  LIMIT 1;
                IF prior_date IS NOT NULL THEN
                  -- A locating, interpolating census before the last census.

                  -- Interpolate between the prior locating,
                  -- interpolating census and the last
                  -- census.  As the individual is alive we
                  -- may have changed the statdate
                  -- in response to a change in the last CENSUS row
                  -- and here are responsible for interpolating about that
                  -- census.  

                  prior_interpolates := prior_status = 'bb_census'
                                        OR prior_status = 'bb_demog'
                                        OR prior_status = 'bb_manual';
                  -- As we's alive, the last census is the statdate
                  -- so no interpolation is necessary past the census date.
                  PERFORM _interp(  NEW.sname
                                  , NEW.status = 'bb_alive'
                                  , prior_date
                                  , last_date
                                  , prior_group
                                  , last_group
                                  , FALSE
                                  , TRUE
                                  , prior_interpolates
                                  , last_interpolates);
                ELSE
                  -- No locating, interpolating census before the last census.

                  -- Optimize for data loading.  During the data
                  -- load we make the individuals alive, and set their
                  -- statdate to their birthdate.  Then we load the census
                  -- rows in date order.  When done we go back
                  -- and put in the correct statdate, status, and dcause.

                  -- We could just...
                  -- Interpolate from birth to the last census (the statdate).
                  -- This will take a lot of resources if the birth to
                  -- statdate interval is long.  But all those individuals
                  -- are long dead so we won't do this very often.
                  -- It does happen all the time during the initial data
                  -- load, so we want this to be fast.

                  optimize := NOT(last_interpolates);

                  IF optimize THEN
                    -- Need interpolate only back to the last
                    -- locating census, which is non-interpolating.

                    SELECT INTO prior_date, prior_group, prior_status
                                date,       grp,         status
                      FROM census
                      WHERE sname = NEW.sname
                            AND date < NEW.statdate
                            AND status <> 'bb_absent'
                      ORDER BY date DESC
                      LIMIT 1;
                    IF prior_date IS NULL THEN
                      -- No 2nd locating census.  Give up.
                      -- and interpolate back to the birth after all.
                      optimize := FALSE;
                      birth_interp := TRUE;
                    ELSE  
                      PERFORM _interp(  NEW.sname
                                      , NEW.status = 'bb_alive'
                                      , prior_date
                                      , last_date
                                      , prior_group
                                      , last_group
                                      , FALSE
                                      , TRUE
                                      , FALSE
                                      , last_interpolates);
                    END IF;
                  ELSE
                    -- Optimize flag has been twiddled, can't use ELSE
                    -- to interpolate, but we do need to know if the
                    -- birthdate interpolates or not.

                    PERFORM 1
                      FROM census
                      WHERE sname = NEW.sname
                            AND date = NEW.birth
                            AND status <> 'bb_absent'
                      LIMIT 1;
                    -- IF FOUND THEN
                    -- We already eliminated bb_manual, bb_census, and bb_demog,
                    -- what's left is a non-interpolating census row on the
                    -- birthdate.
                    birth_interp := NOT(FOUND);
                  END IF;

                  IF NOT optimize THEN
                    -- Have to interpolate back to the birth.
                    PERFORM _interp(  NEW.sname
                                    , NEW.status = 'bb_alive'
                                    , NEW.birth
                                    , last_date
                                    , NEW.matgrp
                                    , last_group
                                    , TRUE
                                    , TRUE
                                    , birth_interp
                                    , last_interpolates);
                  END IF;
                END IF;
              ELSE
                -- The individual is dead, the interpolating
                -- has already been done between the last and next to
                -- last census, but the statdate is not necessarly
                -- the last census date.  So, interpolate between
                -- the census date and the statdate.
                -- (The group at the statdate is not used and
                -- so NULL may be supplied.)
                PERFORM _interp(  NEW.sname
                                , NEW.status = 'bb_alive'
                                , last_date
                                , NEW.statdate
                                , last_group
                                , NULL
                                , FALSE
                                , TRUE
                                , last_interpolates
                                , FALSE);
              END IF;
            END;
          ELSE
            -- No locating censuses.
            -- Interpolate from birthdate to statdate.
            PERFORM _interp(  NEW.sname
                            , NEW.status = 'bb_alive'
                            , NEW.birth
                            , NEW.statdate
                            , NEW.matgrp
                            , NULL
                            , TRUE
                            , TRUE
                            , TRUE
                            , FALSE);
          END IF;
        END;
      ELSE  -- Statdate did not change.
        IF OLD.status <> NEW.status THEN
          -- Changing the status means we re-interpolate the
          -- last locating census to statdate.
          -- Fortunately, we don't need to optimize this
          -- and can use our standard routines.
          DECLARE

            -- Date of last locating census
            last_date census.date%TYPE;
            -- Group of last locating census
            last_group census.grp%TYPE;
            -- Status of last locating census
            last_status census.status%TYPE;

          BEGIN

            SELECT INTO last_date, last_group, last_status
                        date,      grp,        status
              FROM census
              WHERE sname = NEW.sname
                    AND status <> 'bb_absent'
              ORDER BY sname, date DESC
              LIMIT 1;

            IF last_date IS NOT NULL THEN
              -- There is a locating census.
              -- (If there isn't then we know that
              -- changing the Statdate or the birth
              -- is all that affects MEMBERS.)

              PERFORM _interp_about_date(NEW.sname
                                         , last_date
                                         , last_group
                                         , last_status);
            END IF;
          END;
        END IF;
      END IF;
    ELSE
      -- Inserting.

      -- If alive, birth = statdate so be sure to get a row in members.
      -- If dead interpolate from birth to death.  Regardless, there
      -- are no census rows yet.
      PERFORM _interp(  NEW.sname
                      , NEW.status = 'bb_alive'
                      , NEW.birth
                      , NEW.statdate
                      , NEW.matgrp
                      , NULL
                      , TRUE
                      , TRUE
                      , TRUE
                      , FALSE);
    END IF;
  END IF;

  --
  -- Check for automatic Mdate generation.
  --
  IF TG_OP = 'UPDATE' AND NEW.sex = 'F' THEN
    IF (OLD.status <> 'bb_alive' AND NEW.status = 'bb_alive')
       OR (NEW.status <> 'bb_alive' AND OLD.statdate < NEW.statdate) THEN
      -- There's new opportunity for an automatic Mdate,
      -- see about making one.

      -- Find the last Ddate
      DECLARE
        ddate_row cycpoints;

      BEGIN
        SELECT INTO ddate_row
                    cycpoints.*
          FROM cycles, cycpoints
          WHERE cycles.sname = NEW.sname
                AND cycpoints.cid = cycles.cid
                AND cycpoints.code = 'bb_deturgesence'
          ORDER BY cycpoints.date DESC
          LIMIT 1;
        IF FOUND THEN
          -- Make a new automatic Mdate.
          PERFORM _add_autom(NEW.sname, ddate_row);
        END IF;
      END;
    END IF;
  END IF;

  --
  -- Check for updating REPSTATS and CYCSTATS.
  --
  IF TG_OP = 'INSERT' THEN
    -- New birth (birthdate or link to mom through pregs) updates
    -- mom's cycles.
    update_mothers_cycles(`NEW')

    IF NEW.sex = 'bb_female' AND NEW.sname IS NOT NULL THEN
      -- Update this individual's cycles for statdate.
      -- Do something clever here.
      -- PERFORM rebuild_cyclestats(NEW.sname);
    END IF;
  END IF;


  IF TG_OP = 'UPDATE' THEN
    -- Changing birthdate changes mother's cycle records.
    IF OLD.birth <> NEW.birth THEN
       update_mothers_cycles(`NEW')
    END IF;

    -- Changing pregnancy changes mother's cycle records.
    IF OLD.pid <> NEW.pid THEN
       update_mothers_cycles(`NEW')

       PERFORM * FROM pregs AS oldpregs
                      , pregs AS newpregs
                      , cycles AS oldcycles
                      , cycles AS newcycles
               WHERE OLD.pid = oldpregs.pid
                     AND NEW.pid = newpregs.pid
                     AND oldpregs.conceive = oldcycles.cid
                     AND newpregs.conceive = newcycles.cid
                     AND oldcycles.sname <> newcycles.sname
               LIMIT 1;
       IF FOUND THEN
         -- Mother has changed!
         -- Update the old mother' cycles.
         update_mothers_cycles(`OLD')
       END IF;
    END IF;

    --
    -- Changing statdate updates this individual's sexual cycle records.
    --
    IF NEW.sex = 'bb_female' 
       AND OLD.statdate <> NEW.statdate
       AND NEW.sname IS NOT NULL THEN

      PERFORM _biograph_update_cycgapdays(NEW, OLD);

      -- Statdate changed.  Update this individual's cycles.
      -- Do something clever here.
      -- PERFORM rebuild_cyclestats(NEW.sname);
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;

CREATE FUNCTION biograph_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  -- Function for biograph delete trigger
  --
  -- GPL_notice(`  --', `2005, 2008', `Karl O. Pinc <kop@meme.com>')

  DECLARE

  BEGIN
    -- Check that a pregnancy has progeny if there's a resume.
    pid_check()    

    RETURN NULL;
  END;
$$;


CREATE FUNCTION biograph_before_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for biograph BEFORE delete trigger
  --
  -- GPL_notice(`  --', `2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --

  DECLARE

  BEGIN
    -- Delete all the members rows that go with the individual.
    -- (Unless you're stillborn there's always members rows
    -- even when there's no census rows.)
    DELETE FROM members WHERE members.sname = OLD.sname;

    RETURN OLD;
  END;
$$;


CREATE FUNCTION _biograph_update_cycgapdays(newrow biograph, oldrow biograph)
  RETURNS VOID
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Update the cycgapdays table when update happens in the biograph trigger.
  --
  -- GPL_notice(`  --', `2006, 2008, 2010', `Karl O. Pinc <kop@meme.com>')
  --
  -- Syntax:  _biograph_update_cycgapdays(newrow biograph, oldrow biograph)
  --
  -- Input:
  --  newrow   The triggers NEW row.
  --  oldrow   The trigger's OLD row.
  --
  -- Side Effects:
  --   Updates the cycagpdays table based on the changes between the
  -- new and old rows.
  --
  -- Remarks:
  --  Exists as a function because the regular users can't update the
  -- cycgapdays table.
  --
  -- The biograph trigger needs to update the cycgapdays table when
  -- the statdate changes.  (Someday we might get fancy with rows
  -- after the statdate when dead and the status will then matter too.)
  --

  DECLARE

  BEGIN
    --
    -- Update the individual's cycgapdays rows.
    --
    IF newrow.statdate < oldrow.statdate THEN
      -- Gotta delete cycgapdays rows.
      -- No need to be picky, they may or may not even exist.
      DELETE
        FROM cycgapdays
        WHERE sname = newrow.sname
              AND date > newrow.statdate + bb_autom_interval;

    ELSE --newrow.statdate > oldrow.statdate
      IF EXISTS (SELECT 1 FROM cycgaps WHERE cycgaps.sname = newrow.sname) THEN
        -- Might have to add new cycgapdays.

        -- Iterate over those intervals between the old and new statdate+
        -- where there's a gap and create cycgapdays.
        <<add_cycgapdays>>
        DECLARE
          this_date DATE;
          this_cg_code cycgaps.code%TYPE;
          -- old end-of-cycgapdays date
          -- So far past the statdate+ that there are no further 'gap' effects
          old_eocgd_date cycgapdays.date%TYPE
                         := oldrow.statdate + bb_autom_interval;
          -- new end-of-cycgapdays date
          new_eocgd_date cycgapdays.date%TYPE
                         := newrow.statdate + bb_autom_interval;

        BEGIN
          -- Are we in a gap on the old statdate+?
          PERFORM 1
            FROM cycgapdays
            WHERE cycgapdays.sname = newrow.sname
                  AND cycgapdays.date = old_eocgd_date
            LIMIT 1;

          IF FOUND THEN
            -- We are in a gap on the old statdate+.
            this_date := old_eocgd_date + 1;
          ELSE
            -- Maybe we're not in a gap because we're in a point
            -- observation?
            PERFORM 1
              FROM cycgaps
              WHERE cycgaps.sname = newrow.sname
                    AND cycgaps.date = old_eocgd_date
                    AND cycgaps.code = 'bb_cg_point'
              LIMIT 1;

            IF FOUND THEN
              -- We really are in a gap anyway and the statdate+ just
              -- happened to be on a day in which there was a point
              -- observation.
              this_date := old_eocgd_date + 1;

            ELSE
              -- Find the date at which we are actually in a gap.
              SELECT INTO this_date cycgaps.date + 1
                FROM cycgaps
                WHERE cycgaps.sname = newrow.sname
                      AND cycgaps.date >= old_eocgd_date
                      AND cycgaps.code = 'bb_cg_end'
                ORDER BY cycgaps.date
                LIMIT 1;
              IF NOT FOUND THEN
                -- We never do get into a gap in observation.
                EXIT add_cycgapdays;
              END IF;
            END IF;
          END IF;

          WHILE this_date <= new_eocgd_date LOOP

            -- See if we change from observed to unobserved, or vice versa.
            SELECT INTO this_cg_code cycgaps.code
              FROM cycgaps
              WHERE cycgaps.sname = newrow.sname
                    AND cycgaps.date = this_date;

            IF this_cg_code <> 'bb_cg_point'
               OR this_cg_code IS NULL THEN
              -- When it is a point observation, don't write to cycgapdays
              -- and just go to the next date.
              IF this_cg_code = 'bb_cg_start' THEN
                -- Find the next time when we're not under observation.
                SELECT INTO this_date cycgaps.date + 1
                  FROM cycgaps
                  WHERE cycgaps.sname = newrow.sname
                        AND cycgaps.date > this_date
                        AND cycgaps.code = 'bb_cg_end'
                        AND cycgaps.date < new_eocgd_date
                  ORDER BY cycgaps.date
                  LIMIT 1;
                IF NOT FOUND THEN
                  -- We remain under observation until the statdate+.
                  EXIT add_cycgapdays;
                END IF;
              END IF;

              -- Note that we're not under observation.
              INSERT INTO cycgapdays (sname, date)
                          values (newrow.sname, this_date);

            END IF;
            -- Prepare for next iteration.
            this_date := this_date + 1;
          END LOOP;

        END;
      END IF;
    END IF;

    RETURN;
  END;
$$;
grant_func_priv(`_biograph_update_cycgapdays(biograph, biograph)')

SELECT 'biograph_before_insert_trigger' AS trigger;
CREATE TRIGGER biograph_before_insert_trigger
  BEFORE INSERT
  ON biograph FOR EACH ROW
  EXECUTE PROCEDURE biograph_before_insert_func();

SELECT 'biograph_trigger' AS trigger;
CREATE TRIGGER biograph_trigger
  AFTER INSERT OR UPDATE
  ON biograph FOR EACH ROW
  EXECUTE PROCEDURE biograph_func();

SELECT 'biograph_delete_trigger' AS trigger;
CREATE TRIGGER biograph_delete_trigger
  AFTER DELETE
  ON biograph FOR EACH ROW
  EXECUTE PROCEDURE biograph_delete_func();

SELECT 'biograph_before_delete_trigger' AS trigger;
CREATE TRIGGER biograph_before_delete_trigger
  BEFORE DELETE
  ON biograph FOR EACH ROW
  EXECUTE PROCEDURE biograph_before_delete_func();
