dnl Copyright (C) 2012, 2017 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004-2006, 2008 Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl Remarks:

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

dnl
dnl
dnl plpgsql fragment for deleting members rows when not interpolating
dnl
dnl Syntax: delete_non_absent_member()
dnl
dnl Variables required:
dnl
changequote({,})
define({delete_non_absent_member},{
    IF OLD.status <> 'bb_absent' THEN
      -- The old census row is not an absence row
      -- so was just inserted into the members table.
      -- Remove it so the new copy can be inserted (or
      -- equalivently interpolated).
      DELETE FROM members
        WHERE sname = OLD.sname
              AND date = OLD.date
              AND grp = OLD.grp
              AND origin = OLD.status;
    END IF;
})dnl
changequote(`,')dnl


dnl
dnl
dnl plpgsql fragment for checking that the dispersedate does not come
dnl         after the statdate unless there's a subsequent absence on
dnl         census.  (We don't bother checking here to ensure that the
dnl         individual is also alive.)
dnl
dnl Syntax: check_dispersedate(rec)
dnl
dnl Arguments:
dnl   rec  Either NEW or OLD, depending on which should be used.
dnl
dnl Variables required:
dnl
changequote({,})
define({check_dispersedate},{

      DECLARE
        this_group     census.grp%TYPE;
        this_statdate  biograph.statdate%TYPE;
        this_dispersed dispersedates.dispersed%TYPE;

        -- First absence in this_group after statdate
        some_absence   census.date%TYPE;

      BEGIN
        SELECT INTO this_statdate,     this_dispersed
                    biograph.statdate, dispersedates.dispersed
          FROM biograph, dispersedates
          WHERE biograph.sname = $1.sname
                AND dispersedates.sname = $1.sname
                AND biograph.statdate < dispersedates.dispersed
          LIMIT 1;
        IF FOUND THEN
          -- The individual has a dispersedate after the statdate.
          -- Find the group in which the individual was last censused.
          SELECT INTO this_group census.grp
            FROM census
            WHERE census.sname = $1.sname
                  AND census.status <> 'bb_absent'
            ORDER BY census.date DESC
            LIMIT 1;
          IF NOT FOUND THEN
            RAISE EXCEPTION 'Cenid %: %: %: Status %: DISPERSEDATES.Dispersed (%)cannot be after BIOGRAPH.Statdate (%) unless the individual is alive and has a census that locates the individual in a group'
                          , $1.cenid
                          , $1.sname
                          , $1.date
                          , $1.status
                          , this_dispersed
                          , this_statdate;
            RETURN NULL;
          END IF;

          -- Find the first absence in this_group after the statdate.
          SELECT INTO some_absence census.date
            FROM census
            WHERE census.sname = $1.sname
                  AND census.date > this_statdate
                  AND census.grp = this_group
                  AND census.status = 'bb_absent'
            ORDER BY census.date
            LIMIT 1;
          IF NOT FOUND OR some_absence < this_dispersed THEN
            RAISE EXCEPTION integrity_constraint_violation USING
                  MESSAGE = 'Error on ' || TG_OP || ' of CENSUS'
                , DETAIL = 'Key (CenId) = (' || $1.cenid
                             || '): Value (Sname) = (' || $1.sname
                             || '): Value (Date) = (' || $1.date
                             || '): Value (Status) = (' || $1.status
                             || '): Value (DISPERSEDATES.Dispersed) = ('
                             || this_dispersed
                             || '): Value (BIOGRAPH.Statdate) = ('
                             || this_statdate
                             || '): DISPERSEDATES.Dispersed cannot be after '
                             || 'BIOGRAPH.Statdate unless the individual is '
                             || 'alive and there is an absence in the last '
                             || 'recorded group (Group '
                             || this_group
                             || ') on or after the dispersedate';
            RETURN NULL;
          END IF;
        END IF;        
      END;
})dnl
changequote(`,')dnl


--  
-- census
-- $Id: census.m4,v 1.10 2004/10/12 02:02:16 kop Exp $
--
SELECT 'census' AS table;

SELECT 'census_func' AS function;
CREATE FUNCTION census_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for census insert and update triggers
  --
  -- Copyright (C) 2017 The Meme Factory, Inc.,  http://www.meme.com/
  -- GPL_notice(`  --', `2004, 2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  --   Have the interpolation routines update the statdate.
  -- That way they can figure out when statdate changes
  -- mean interpolation must happen.  Otherwise
  -- we'd call the interpolation routines even though
  -- updating the statdate in biograph would have already
  -- triggered another interpolation.

  --
  -- First the integrity checking.
  --

  --
  -- Things to check only on update.
  --
  IF TG_OP = 'UPDATE' THEN
    cannot_change(`CENSUS', `Cenid')

    -- Make life easier for updating members, Sname cannot change.
    cannot_change(`CENSUS', `Sname')

  END IF;

  --
  -- Checks for insert or update.
  --

  -- Can't have rows in the group if the GROUPS.Start
  -- column is NULL.
  PERFORM 1
    FROM groups
    WHERE groups.gid = NEW.grp
          AND groups.start IS NULL
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Cenid %: Grp %: Cannot have a CENSUS row for a group with a NULL GROUPS.Start column'
                    , NEW.cenid
                    , NEW.grp;
    RETURN NULL;
  END IF;

  -- Can't have rows the group before the GROUPS.Start date.
  PERFORM 1
    FROM groups
    WHERE groups.gid = NEW.grp
          AND groups.start > NEW.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Cenid %: Grp %: %: Cannot have a CENSUS row for a group before the group''s GROUPS.Start date'
                    , NEW.cenid
                    , NEW.grp
                    , NEW.date;
    RETURN NULL;
  END IF;

  -- Can't have rows in the group after the GROUPS.Cease_To_Exist date.
  PERFORM 1
    FROM groups
    WHERE groups.gid = NEW.grp
          AND groups.cease_to_exist < NEW.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Cenid %: Grp %: %: Cannot have a CENSUS row for a group after the group''s GROUPS.Cease_To_Exist date'
                    , NEW.Cenid
                    , NEW.grp
                    , NEW.date;
    RETURN NULL;
  END IF;

  -- Cannot be marked absent in the unknown group.
  -- (Interpolation needs this enforced.)
  IF NEW.grp = 'bb_unk_grp'
     AND NEW.status = 'bb_absent' THEN
    RAISE EXCEPTION '%: %: CENSUS.Grp cannot be bb_unk_grp when CENSUS.Status is bb_absent.'
                    , NEW.sname
                    , NEW.date;
    RETURN NULL;
  END IF;

  -- Cannot have a non-absent census row before the individual's birthdate.

  PERFORM NEW.sname FROM biograph
    WHERE NEW.sname = biograph.sname
          AND NEW.status <> 'bb_absent'
          AND NEW.date < biograph.birth
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: CENSUS.Date is before BIOGRAPH.Birth, and Status is not bb_absent'
                    , NEW.sname
                    , NEW.date;
    RETURN NULL;
  END IF;

  -- Cannot be "present" in two (different or same) groups on the same day.
  PERFORM NEW.sname FROM census
    WHERE census.sname = NEW.sname
          AND census.date = NEW.date
          AND census.cenid <> NEW.cenid
          AND census.status <> 'bb_absent'
          AND NEW.status <> 'bb_absent'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Cenid %: Cannot be CENSUSed in two groups on the same day'
    , NEW.sname
    , NEW.date
    , NEW.cenid;
    RETURN NULL;
  END IF;

  -- Cannot be censused more than once in the same group on the same day.
  PERFORM NEW.sname FROM census
    WHERE census.sname = NEW.sname
          AND census.date = NEW.date
          AND census.grp = NEW.grp
          AND census.cenid <> NEW.cenid
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: %: Cenid %: Cannot be CENSUSed in the same group twice on the same day'
      , NEW.sname
      , NEW.date
      , NEW.grp
      , NEW.cenid;
    RETURN NULL;
  END IF;


  -- 'D' census rows should have a demography note, but then there'd
  -- be no way to enter them (census rows go in first).  So, check
  -- for this in the warning scripts.

  -- Census rows with D status and no demog row must not be from
  -- an actual census, they must be a row that's waiting to have
  -- a demog row added.
  PERFORM NEW.status
    WHERE NEW.status = 'bb_demog'
          AND NEW.cen
          AND NEW.cenid NOT IN (SELECT cenid
                                       FROM demog 
                                       WHERE cenid=NEW.cenid)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Cenid %: CENSUS rows with a Status of bb_demog and a Cen of TRUE must have an associated DEMOG row -- just create a DEMOG row when Cen is TRUE'
          , NEW.sname
          , NEW.date
          , NEW.cenid;
    RETURN NULL;
  END IF;

  --
  -- When the individual is dead (the death test here must
  -- match that in biograph trigger) then changing the census table must
  -- not cause a change in the statdate, below.  If it does, then
  -- interpolation would happen twice as both this trigger
  -- and the biograph trigger would interpolate -- and I just
  -- don't want to have to think of the implications.
  -- (Maybe it does not matter, but I'm being safe.)
  --

  PERFORM 1
    FROM biograph
    WHERE sname = NEW.sname
          AND statdate < NEW.date
          AND status <> 'bb_alive'
          AND NEW.status <> 'bb_absent'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: %: Individual CENSUSed after death date'
                    , NEW.sname
                    , NEW.date
                    , NEW.status;
    RETURN NULL;
  END IF;

  --
  -- Then updating of other tables.
  --

  --
  -- Update members.
  --

  IF NEW.status <> 'bb_absent' THEN
    -- All interpolated rows that existed for the individual
    -- on the date of census are removed by the BEFORE trigger.

    -- All 'present' census rows get inserted straight into members.
    IF NEW.status = 'bb_census' 
       OR NEW.status = 'bb_demog'
       OR NEW.status = 'bb_manual' THEN
      -- Interpolating census rows get 0 for Interp.
      INSERT INTO members (sname, date, grp, origin, interp)
             VALUES (NEW.sname, NEW.date, NEW.grp, NEW.status, 0);
    ELSE
      -- Non-interpolating census rows get NULL for Interp.
      INSERT INTO members (sname, date, grp, origin, interp)
             VALUES (NEW.sname, NEW.date, NEW.grp, NEW.status, NULL);
    END IF;
  END IF;


  --  Interpolate and all that.

  IF TG_OP = 'UPDATE' THEN
    IF OLD.date <> NEW.DATE
       OR OLD.grp <> NEW.grp
       OR OLD.status <> NEW.status THEN
      PERFORM _interp_delete(OLD.sname, OLD.date, OLD.status);
      PERFORM _interp_insert(NEW.sname, NEW.date, NEW.grp, NEW.status);
    END IF;
  ELSE
    PERFORM _interp_insert(NEW.sname, NEW.date, NEW.grp, NEW.status);
  END IF;

  --
  -- DISPERSEDATES.Dispersed
  -- Check after interpolation because interpolation updates the statdate.
  --
  -- Can't be dispersed after the statdate if you're
  -- alive unless the dispersal date is on or before the absence after
  -- the last census.
  IF TG_OP = 'UPDATE' THEN
    IF (NEW.date < OLD.date AND NEW.status = 'bb_absent')
       OR (NEW.status = 'bb_absent' AND NEW.status <> OLD.status) THEN
      check_dispersedate(`NEW')
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'census_before_func' AS function;
CREATE FUNCTION census_before_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for census BEFORE update or insert trigger
  --
  -- GPL_notice(`  --', `2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Need to remove all the old stuff from the members table
  -- so that we can do things to multiple census rows in one statement.
  -- (Especially when it comes to the "old" census rows.)

  DECLARE

  BEGIN
  --
  -- Update the members table.
  --

  IF TG_OP = 'UPDATE' THEN
    -- Delete corresponding OLD non-interpolated member rows.
    delete_non_absent_member()
  END IF;

  IF NEW.status <> 'bb_absent' THEN
    -- Remove any interpolated row that exists for the individual
    -- on the date of census.
    DELETE FROM members
           WHERE sname = NEW.sname
                 AND date = NEW.date
                 AND origin = 'bb_interpolated';
  END IF;

  RETURN NEW;
  END;
$$;



SELECT 'census_before_delete_func' AS function;
CREATE FUNCTION census_before_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for census BEFORE delete trigger
  --
  -- GPL_notice(`  --', `2006, 2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Need to remove all the old stuff from the members table
  -- so that we can do things to multiple census rows in one statement.
  -- (Especially when it comes to the "old" census rows.)

  DECLARE

  BEGIN
  --
  -- Update the members table.
  --

  -- Delete corresponding non-interpolated member rows.
  delete_non_absent_member()

  RETURN OLD;
  END;
$$;


SELECT 'census_delete_func' AS function;
CREATE FUNCTION census_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for census delete trigger
  --
  --
  -- Copyright (C) 2017 The Meme Factory, Inc.,  http://www.meme.com/
  -- GPL_notice(`  --', `2004, 2006, 2008', `Karl O. Pinc <kop@meme.com>')

  DECLARE

  BEGIN

  --
  -- Check that dispersal does not happen after statdate
  -- unless the individual is alive and there's a subsequent
  -- absence in the group where the individual was last censused.
  --
  -- We can check this before interoplation even though interpolation
  -- updates the statdate because when the statdate is updated
  -- the biograph trigger does the check too.  That takes care of
  -- the condition where we delete a locating census.  And this
  -- check takes care of deleting absences so that covers all the
  -- cases.
  --  

  IF OLD.status = 'bb_absent' THEN
    check_dispersedate(`OLD')    
  END IF;

  --
  -- Done checking things.  Now update the rest of the database
  -- where appropriate.
  --

  --
  -- Update the members table.
  --

  -- Interpolate.
  PERFORM _interp_delete(OLD.sname, OLD.date, OLD.status);

  RETURN NULL;
  END;
$$;


SELECT 'census_trigger' AS trigger;
CREATE TRIGGER census_trigger
  AFTER INSERT OR UPDATE
  ON census FOR EACH ROW
  EXECUTE PROCEDURE census_func();


SELECT 'census_before_trigger' AS trigger;
CREATE TRIGGER census_before_trigger
  BEFORE INSERT OR UPDATE
  ON census FOR EACH ROW
  EXECUTE PROCEDURE census_before_func();

SELECT 'census_before_delete_trigger' AS trigger;
CREATE TRIGGER census_before_delete_trigger
  BEFORE DELETE
  ON census FOR EACH ROW
  EXECUTE PROCEDURE census_before_delete_func();

SELECT 'census_delete_trigger' AS trigger;
CREATE TRIGGER census_delete_trigger
  AFTER DELETE
  ON census FOR EACH ROW
  EXECUTE PROCEDURE census_delete_func();
