dnl Copyright (C) 2014, 2017 Jake Gordon <jacob.gordon@duke.edu>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl    Jake Gordon <jacob.gordon@duke.edu>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

--  
-- dart_samples
--
SELECT 'dart_samples' AS table;

SELECT 'dart_samples_func' AS function;
CREATE FUNCTION dart_samples_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for dart_samples insert and update triggers
  --
  -- GPL_notice(`  --', `2014, 2017', `Jake Gordon, <jacob.gordon@duke.edu>')

  IF TG_OP = 'UPDATE' THEN
    cannot_change(`DART_SAMPLES', `DS_Id')
    cannot_change(`DART_SAMPLES', `Dartid')
    cannot_change(`DART_SAMPLES', `DS_Type')

    -- Check for WBC_COUNTS rows with Slide_number greater than the
    -- number of blood smears collected for this Dartid.
    IF NEW.ds_type = bb_dst_bloodsmear
       AND NEW.num <> OLD.num THEN

      PERFORM 1
        FROM wbc_counts
        WHERE wbc_counts.dartid = NEW.dartid
          AND NEW.ds_type = bb_dst_bloodsmear
          AND wbc_counts.slide_number > NEW.num
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of DART_SAMPLES'
            , DETAIL = 'Key (DS_Id) = (' || NEW.ds_id
                         || '): Value (Dartid) = (' || NEW.dartid
                         || '): Value (DS_Type) = (' || NEW.ds_type
                         || '): Value (Num) = (' || NEW.num
                         || '): Cannot reduce Num of blood smears to '
                         || 'a value less than the Slide_number of '
                         || 'any related WBC_COUNTS rows';
        RETURN NULL;
      END IF;
    END IF;
  END IF;

  -- Sex-specific samples should match the individual's
  -- sex.
  PERFORM 1
    FROM dartings
         JOIN biograph ON (biograph.sname = dartings.sname)
         JOIN dart_sample_types ON (dart_sample_types.ds_type = NEW.ds_type)
    WHERE dartings.dartid = NEW.dartid
          AND dart_sample_types.sex IS NOT NULL
          AND biograph.sex <> dart_sample_types.sex
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of DART_SAMPLES'
        , DETAIL = 'Key (DS_Id) = (' || NEW.ds_id
                     || '): Value (Dartid) = (' || NEW.dartid
                     || '): Value (DS_Type) = (' || NEW.ds_type
                     || '): Sex-specific samples must match the'
                     || ' individual''s sex';
    RETURN NULL;
  END IF;

  -- Enforce mins and maxes
  PERFORM 1
    FROM dart_sample_types
    WHERE dart_sample_types.ds_type = NEW.ds_type
          AND (dart_sample_types.minimum > NEW.num
              OR dart_sample_types.maximum < NEW.num)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of DART_SAMPLES'
        , DETAIL = 'Key (DS_Id) = (' || NEW.ds_id
                     || '): Value (Dartid) = (' || NEW.dartid
                     || '): Value (DS_Type) = (' || NEW.ds_type
                     || '): Value (Num) = (' || NEW.num
                     || '): Num must be between this DS_Type''s'
                     || ' Minimum and Maximum values in'
                     || ' DART_SAMPLE_TYPES, inclusive';
    RETURN NULL;
  END IF;

  RETURN NULL;
  END;
 $$;

SELECT 'dart_samples_delete_func' AS function;
CREATE FUNCTION dart_samples_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for DART_SAMPLES delete trigger
  --
  -- GPL_notice(`  --', `2017', `Jake Gordon, <jacob.gordon@duke.edu>')

  DECLARE

  BEGIN

  -- Check for WBC_COUNTS rows that are associated with the deleted
  -- row(s)
  IF OLD.ds_type = bb_dst_bloodsmear THEN
    PERFORM 1
      FROM wbc_counts
      WHERE wbc_counts.dartid = OLD.dartid
        AND OLD.ds_type = bb_dst_bloodsmear
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of DART_SAMPLES'
          , DETAIL = 'Key (DS_Id) = (' || OLD.ds_id
                       || '): Value (Dartid) = (' || OLD.dartid
                       || '): Value (DS_Type) = (' || OLD.ds_type
                       || '): Value (Num) = (' || OLD.num
                       || '): Cannot delete a blood smear row '
                       || 'when there are any related rows in '
                       || 'WBC_COUNTS';
      RETURN NULL;
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;


 SELECT 'dart_samples_trigger' AS trigger;
 CREATE TRIGGER dart_samples_trigger
  AFTER INSERT OR UPDATE
  ON dart_samples FOR EACH ROW
  EXECUTE PROCEDURE dart_samples_func();

SELECT 'dart_samples_delete_trigger' AS trigger;
CREATE TRIGGER dart_samples_delete_trigger
  AFTER DELETE
  ON dart_samples FOR EACH ROW
  EXECUTE PROCEDURE dart_samples_delete_func();
