dnl Copyright (C) 2017 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2010, 2011, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl


dnl plpgsql fragment for ensuring that a time value on a related table
dnl occurs while the animal is sedated.
dnl
dnl Syntax: _check_sedation(table, column, descr)
dnl
dnl table  The related table
dnl column The column containing a time value
dnl descr  Description of the action represented by the related row
dnl        for inclusion in the error message.
dnl        (First word should be capitalized)
dnl
changequote({,})
define({_check_sedation},{dnl
    PERFORM 1
      FROM $1
      WHERE $1.dartid = NEW.dartid
            AND $1.$2
                   > NEW.darttime + 'bb_max_knockout_interval'::interval
            AND NOT EXISTS (SELECT 1
                              FROM anesths AS other
                              WHERE other.dartid = NEW.dartid
                                    AND other.antime < $1.$2
                                    AND other.antime
                                          >= $1.$2
                                                - 'bb_max_knockout_interval'::interval
                              LIMIT 1)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'DARTINGS.Dartid %: DARTINGS.Darttime %:: $3 must occur within bb_max_knockout_interval of the previous anesthetic administration (or the darting)',
                      NEW.dartid,
                      NEW.darttime;
      RETURN NULL;
    END IF;
})dnl
changequote(`,')dnl

--  
-- dartings
--
SELECT 'dartings' AS table;

SELECT 'dartings_func' AS function;
CREATE FUNCTION dartings_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for dartings insert and update triggers
  --
  -- GPL_notice(`  --', `2010, 2011', `Karl O. Pinc <kop@meme.com>')

  IF TG_OP = 'UPDATE' THEN
    -- Cannot change the Dartid.
    cannot_change(`DARTINGS', `Dartid')
  END IF;

  -- The individual must be alive when darted.

  -- Can't be darted after the statdate.
  check_statdate(`DARTINGS', `Date')

  -- Must be born and in the population when darted.
  check_entrydate(`DARTINGS', `Date')

  --
  -- ANESTHS
  --

  -- ANESTHS.Antime cannot be before DARTINGS.Darttime.
  rev_check_keyed_ge(`ANESTHS'
                     , `Anesthid'
                     , `Antime'
                     , `DARTINGS'
                     , `Dartid'
                     , `Darttime')

  --
  -- BODYTEMPS
  --

  -- BODYTEMPS.Bttime cannot be before DARTINGS.Pickuptime.
  rev_check_keyed_ge(`BODYTEMPS'
                     , `Btid'
                     , `Bttime'
                     , `DARTINGS'
                     , `Dartid'
                     , `Pickuptime')

  -- BODYTEMPS.Bttime cannot be before DARTINGS.Darttime.
  rev_check_keyed_ge(`BODYTEMPS'
                     , `Btid'
                     , `Bttime'
                     , `DARTINGS'
                     , `Dartid'
                     , `Darttime')

  IF TG_OP = 'UPDATE' THEN
    IF OLD.darttime <> NEW.darttime THEN
      -- ANESTHS.Antime cannot be more than bb_max_knockout_interval
      -- after the later of DARTINGS.Darttime and the prior
      -- ANESTHS.Antime for the darting.
      _check_sedation(`ANESTHS', `Antime', `Anesthetic administration')

      -- BTEMPS.Bttime cannot be more than bb_max_knockout_interval
      -- after the later of DARTINGS.Darttime and the prior
      -- ANESTHS.Antime for the darting.
      _check_sedation(`BODYTEMPS', `Bttime', `Body temperature measurement')
    END IF;

    IF OLD.sname <> NEW.sname THEN
      -- Individuals must be male to have rows on TESTES.
      PERFORM 1
        FROM biograph, testes_arc
        WHERE biograph.sname = NEW.sname
              AND biograph.sex <> 'bb_male'
              AND testes_arc.dartid = NEW.dartid
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 'DARTINGS.Sname %: DARTINGS.Date %: DARTINGS.Dartid %: Individuals must be male to have related rows on TESTES_ARC'
                      , NEW.sname
                      , NEW.date
                      , NEW.dartid;
        RETURN NULL;
      END IF;

      PERFORM 1
        FROM biograph, testes_diam
        WHERE biograph.sname = NEW.sname
              AND biograph.sex <> 'bb_male'
              AND testes_diam.dartid = NEW.dartid
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION 'DARTINGS.Sname %: DARTINGS.Date %: DARTINGS.Dartid %: Individuals must be male to have related rows on TESTES_DIAM'
                      , NEW.sname
                      , NEW.date
                      , NEW.dartid;
        RETURN NULL;
      END IF;

      -- Individuals must be the correct sex to have sex-specific
      -- samples.
      PERFORM 1
        FROM dart_samples, biograph, dart_sample_types
        WHERE dart_samples.dartid = NEW.dartid
              AND dart_samples.ds_type = dart_sample_types.ds_type
              AND biograph.sname = NEW.sname
              AND dart_sample_types.sex IS NOT NULL
              AND biograph.sex <> dart_sample_types.sex
        LIMIT 1;
      IF FOUND THEN
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of DARTINGS'
            , DETAIL = 'Key (Dartid) = (' || NEW.dartid
                         || '): Value (Sname) = (' || NEW.sname
                         || '): Cannot change sname to one with a'
                         || ' different sex when there are sex-'
                         || 'specific samples recorded in'
                         || ' DART_SAMPLES';
        RETURN NULL;
      END IF;

    END IF;

    -- WBC_COUNTS
    --
    -- Make sure Date is still before or equal to Count_Date
    PERFORM 1
      FROM wbc_counts
      WHERE wbc_counts.dartid = NEW.dartid
        AND wbc_counts.count_date < NEW.date
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of DARTINGS'
          , DETAIL = 'Key (Dartid) = (' || NEW.dartid
                       || '): Value (Sname) = (' || NEW.sname
                       || '): Value (Date) = (' || NEW.date
                       || '): Date must be on or before the '
                       || 'Count_Date of any related rows in '
                       || 'WBC_COUNTS';
      RETURN NULL;
    END IF;

    IF OLD.date > NEW.date OR OLD.sname <> NEW.sname THEN
      -- The darting occurred when the individual was older than thought.

       PERFORM 1
         FROM biograph
         WHERE biograph.sname = NEW.sname
               AND AGE(NEW.date, biograph.birth)
                   >= 'bb_adult_testes_age'::INTERVAL
         LIMIT 1;
       IF FOUND THEN
         -- Check to see that minimum adult testes sizes are met.
         -- TESTES_ARC
         PERFORM 1
           FROM testes_arc
           WHERE testes_arc.dartid = NEW.dartid
                 AND (testes_arc.testlength < bb_min_adult_arc_testlength
                      OR testes_arc.testwidth < bb_min_adult_arc_testwidth)
           LIMIT 1;
         IF FOUND THEN
           RAISE EXCEPTION 'DARTINGS.Sname %: DARTINGS.Dartid %: Individuals at least bb_adult_testes_age old when darted must have a TESTES_ARC.Testwidth of at least bb_min_adult_arc_testwidth and a TESTES_ARC.Testlength of at least bb_min_adult_arc_testlength'
                         , NEW.sname
                         , NEW.dartid;
           RETURN NULL;
         END IF;

         -- TESTES_DIAM
         PERFORM 1
           FROM testes_diam
           WHERE testes_diam.dartid = NEW.dartid
                 AND (testes_diam.testlength < bb_min_adult_diam_testlength
                      OR testes_diam.testwidth < bb_min_adult_diam_testwidth)
           LIMIT 1;
         IF FOUND THEN
           RAISE EXCEPTION 'DARTINGS.Sname %: DARTINGS.Dartid %: Individuals at least bb_adult_testes_age old when darted must have a TESTES_DIAM.Testwidth of at least bb_min_adult_diam_testwidth and a TESTES_DIAM.Testlength of at least bb_min_adult_diam_testlength'
                         , NEW.sname
                         , NEW.dartid;
           RETURN NULL;
         END IF;
       END IF;
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'dartings_trigger' AS trigger;
CREATE TRIGGER dartings_trigger
  AFTER INSERT OR UPDATE
  ON dartings FOR EACH ROW
  EXECUTE PROCEDURE dartings_func();
