dnl Copyright (C) 2016 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004, 2006-2008, 2010, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

--  
-- interact_data
-- $Id: interact.m4,v 1.3 2004/11/08 02:23:10 kop Exp $
--
SELECT 'interact_data' AS table;


SELECT 'interact_data_before_func' AS function;
CREATE FUNCTION interact_data_before_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  DECLARE

  BEGIN
  -- Function for interact_data before insert and update triggers
  --
  -- GPL_notice(`  --', `2010', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  --  Serves to validate and automatically assign INTERACT_DATA.Observer.

  IF TG_OP = 'INSERT' THEN
    IF NEW.sid IS NOT NULL THEN
      IF NEW.observer IS NOT NULL THEN
        -- An Sid and an Observer exist so the observer must be that
        -- of the related SAMPLES row.
        PERFORM 1
         FROM samples
          WHERE samples.sid = NEW.sid
                AND samples.observer = NEW.observer
          LIMIT 1;
        IF NOT FOUND THEN
          RAISE EXCEPTION 'Iid %: Sid %: Observer %: The observer is not that recorded in the Observer column of the related SAMPLES row'
                        , NEW.Iid
                        , NEW.Sid
                        , NEW.Observer;
          RETURN NULL;
        END IF;

        RETURN NEW;

      END IF;

      -- Automatically assign the Observer.
      SELECT INTO NEW.observer samples.observer
        FROM samples
        WHERE samples.sid = NEW.sid;
    END IF;
  ELSE -- TG_OP = 'UPDATE'
    -- Can't relate to a SAMPLES row if there's an existing Observer,
    -- without setting the Observer to NULL and allowing the system
    -- to automatically assign an Observer.

    IF OLD.sid IS NULL
       AND NEW.sid IS NOT NULL
       AND OLD.observer IS NOT NULL
       AND NEW.observer IS NOT NULL THEN
      RAISE EXCEPTION 'Iid %: Sid %: Observer %: Can not assign an Sid where there previously was none (NULL) when Observer is non-NULL'
                    , NEW.iid
                    , NEW.sid
                    , OLD.observer;
      RETURN NULL;
    END IF;

    IF OLD.sid IS NULL
       AND NEW.sid IS NOT NULL
       AND OLD.observer IS NULL
       AND NEW.observer IS NOT NULL THEN
      RAISE EXCEPTION 'Iid %: Sid %: Observer %: Can not manually assign a non-NULL Observer at the same time as assigning an Sid when there previously was no Sid (when Sid was NULL)'
                    , NEW.iid
                    , NEW.sid
                    , OLD.observer;
      RETURN NULL;
    END IF;

    IF NEW.sid IS NOT NULL
       AND (NEW.sid <> OLD.sid
            OR OLD.sid IS NULL) THEN
      -- Automatically assign the Observer.
      SELECT INTO NEW.observer samples.observer
        FROM samples
        WHERE samples.sid = NEW.sid;
    END IF;
  END IF;

  RETURN NEW;
  END;
$$;


SELECT 'interact_data_func' AS function;
CREATE FUNCTION interact_data_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  DECLARE
    this_class acts.class%TYPE;

  BEGIN
  -- Function for interact_data insert and update triggers
  --
  -- GPL_notice(`  --', `2004, 2006-2008, 2010', `Karl O. Pinc <kop@meme.com>')

  IF TG_OP = 'INSERT' THEN
    -- Can't insert if the ACTS is retired.
    PERFORM 1
      FROM acts
      WHERE acts.act = NEW.act
            AND acts.retired
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Iid %: Cannot use an ACTS.Act value (%) that is retired in a new INTERACT_DATA row'
                    , NEW.iid
                    , NEW.act;
      RETURN NULL;
    END IF;
  ELSE -- TG_OP = 'UPDATE'
    -- Can't change Observer unless you can modify triggers.
    IF NEW.sid IS NOT NULL
       AND NEW.observer <> OLD.observer
       AND NOT(has_table_privilege('interact_data', 'trigger')) THEN
      RAISE EXCEPTION 'Iid %: Sid%: Cannot change INTERACT_DATA.Observer when Sid is non-NULL, when there is a related SAMPLES row'
                    , NEW.iid
                    , NEW.sid;
      RETURN NULL;
    END IF;
  END IF;

  -- We have different rules for different classes of interactions
  -- so the structure here divides by interaction class.
  SELECT INTO this_class acts.class
    FROM acts
    WHERE act = NEW.act;

  <<interaction_class>>
  BEGIN

    --
    -- Interaction of agonism or grooming.
    --
    IF this_class = 'bb_agonism' OR this_class = 'bb_grooming' THEN
      IF this_class = 'bb_grooming'
         AND NEW.date < 'bb_grooming_transition'
         AND NEW.date <> date_trunc('month', NEW.date)
         AND NEW.sid IS NULL THEN
        -- For grooming interactions the date must be the first of the
        -- month.
        -- This is an arbitrary restriction based on the way the data
        -- has been collected.
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of INTERACT_DATA'
            , DETAIL = 'Key (IId) = (' || NEW.iid
                         || '): Value (Act) = (' || NEW.act
                         || '): Value (Date) = (' || NEW.date
                         || '): When Sid is NULL, grooming dates '
                         || 'before bb_grooming_transition must be '
                         || 'the first of the month';
        RETURN NULL;
      END IF;

      IF NEW.sid IS NULL THEN
        IF (NEW.start IS NOT NULL
            OR NEW.stop IS NOT NULL) THEN
          -- For agonism and grooming interactions the times must be NULL,
          -- unless we get the data from the handhelds (all-occurences data.)
          -- This is an arbitrary restriction based on the way the data
          -- has been collected.
          RAISE EXCEPTION 'INTERACT_DATA %: %: %: %: %: Agonism and grooming start and stop times must be NULL when there''s no Sid'
                          , NEW.iid
                          , NEW.act
                          , NEW.date
                          , NEW.start
                          , NEW.stop;
          RETURN NULL;
        END IF;
      ELSE
        -- Sid is not NULL.
        IF NEW.start IS NULL OR NEW.stop IS NULL OR NEW.stop <> NEW.start THEN
          -- Start times must equal Stop times and niether be null.
          RAISE EXCEPTION 'INTERACT_DATA %: %: %: %: %: Start times must equal stop times, and neither can be NULL'
                          , NEW.iid
                          , NEW.act
                          , NEW.date
                          , NEW.start
                          , NEW.stop;
          RETURN NULL;
        END IF;
      END IF;

      -- Done with agonism and grooming.
      EXIT interaction_class;
    END IF;

    --
    -- Interaction is mount or ejaculation.
    --

    IF this_class = 'bb_mounts' OR this_class = 'bb_ejaculation' THEN
      IF NEW.start <> NEW.stop THEN
        -- For mounts and ejaculation interactions the start and stop times
        -- must be the same.
        RAISE EXCEPTION 'INTERACT_DATA %: %: %: %: %: Mounts and ejaculation start and stop times must be equal'
                        , NEW.iid
                        , NEW.act
                        , NEW.date
                        , NEW.start
                        , NEW.stop;
        RETURN NULL;
      END IF;

      -- Done with mounts and ejaculations.
      EXIT interaction_class;
    END IF;

    --
    -- Some other kind of interaction.
    --
    IF NEW.stop < NEW.start THEN

      IF this_class <> 'bb_consortship' THEN
        RAISE EXCEPTION 'INTERACT_DATA %: %: %: %: %: Start must be before Stop unless Act is bb_consortship and Stop is NULL'
                        , NEW.iid
                        , NEW.act
                        , NEW.date
                        , NEW.start
                        , NEW.stop;
        RETURN NULL;
      END IF;
    END IF;

    IF NEW.sid IS NULL
       AND (NEW.act = 'bb_approach'
            OR NEW.act = 'bb_request_to_groom')
       AND (NEW.start IS NULL
            OR NEW.stop IS NULL
            OR NEW.start <> NEW.stop) THEN
      RAISE EXCEPTION 'INTERACT_DATA %: %: %: %: %: Start must equal Stop and niether can be NULL when Sid is not NULL and Act is bb_approach or bb_request_to_groom'
                      , NEW.iid
                      , NEW.act
                      , NEW.date
                      , NEW.start
                      , NEW.stop;
      RETURN NULL;
    END IF;
    -- End of interaction specific stuff.
  END;

  --
  -- Check the interaction date against the entrydate and statdate.
  -- Really, this is mostly checked in the PARTS triggers but there's
  -- no special rule saying date or iid can't change, etc.

  -- Entrydate
  -- Groomings compare only against the first of the month where the
  -- other interaction classes compare directly against the entrydate.
  PERFORM biograph.sname FROM biograph, parts, acts
    WHERE parts.iid = NEW.iid
          AND biograph.sname = parts.sname
          AND ((this_class = 'bb_grooming'
                  AND NEW.date < 'bb_grooming_transition'
                  AND rnkdate(biograph.entrydate) > NEW.date)
               OR (NOT (this_class = 'bb_grooming'
                          AND NEW.date < 'bb_grooming_transition')
                   AND biograph.entrydate > NEW.date))
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of INTERACT_DATA'
        , DETAIL = 'Key (IId) = (' || NEW.iid
                     || '): Value (Date) = (' || NEW.date
                     || '): Value (Act) = (' || NEW.act
                     || '): BIOGRAPH.Entrydate cannot be bigger than '
                     || 'INTERACT_DATA.Date, or in the case of bb_grooming '
                     || 'before bb_grooming_transition the month and year '
                     || 'of BIOGRAPH.Entrydate cannot be bigger than '
                     || 'INTERACT_DATA.Date';
    RETURN NULL;
  END IF;

  -- Statdate
  -- (No need to worry about bb_grooming transition date because it
  -- decreases the interaction date.)
  PERFORM biograph.sname FROM biograph, parts --, acts
    WHERE parts.iid = NEW.iid
          -- AND acts.act = NEW.act
          AND biograph.sname = parts.sname
          -- AND ((acts.class = 'bb_grooming'
          --       AND NEW.date < 'bb_grooming_transition'
          --       AND date_trunc('month', biograph.statdate) < NEW.date)
          --      OR (NOT(acts.class = 'bb_grooming'
          --              AND NEW.date < 'bb_grooming_transition')
                   AND biograph.statdate < NEW.date
          --                                       ))
    LIMIT 1;
  IF FOUND THEN
    -- RAISE EXCEPTION 'INTERACT_DATA %: %: %: BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date, or in the case of bb_grooming before bb_grooming_transition the month and year of BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date'
    RAISE EXCEPTION 'INTERACT_DATA %: %: %: BIOGRAPH.Statdate cannot be less than INTERACT_DATA.Date'
                    , NEW.iid
                    , NEW.date
                    , NEW.act;
    RETURN NULL;
  END IF;

  --
  -- Checks done for mounts, consortships, and ejaculations.
  --
  IF NEW.act = 'bb_mounts'
     OR NEW.act = 'bb_consortship'
     OR NEW.act = 'bb_ejaculation' THEN
    -- Males cannot have mounts, consortships, or ejaculations
    -- before 4 years of age.
    PERFORM 1
      FROM biograph, parts
      WHERE parts.iid = NEW.iid
            AND biograph.sname = parts.sname
            AND biograph.sex = 'bb_male'
            AND NEW.date - CAST('bb_min_male_mce_age' AS interval)
                < biograph.birth
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'INTERACT_DATA %: %: %: Males cannot have mounts, consortships, or ejaculations before bb_min_male_mce_age of age.'
                      , NEW.iid
                      , NEW.date
                      , NEW.act;
      RETURN NULL;
    END IF;

    -- Female's can't have mounts, consortships, or ejactulations
    -- until they reach menarche.
    -- (The PARTS trigger will check for the existance of a female's
    -- MATUREDATES row, so here we can fail on existance.)
    PERFORM 1
      FROM parts, maturedates, biograph
      WHERE parts.iid = NEW.iid
            AND biograph.sname = parts.sname
            AND biograph.sex = 'bb_female'
            AND maturedates.sname = parts.sname
            AND NEW.date < maturedates.matured
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'INTERACT_DATA %: %: %: Female''s can''t have mounts, consortships, or ejactulations until they reach menarche'
                      , NEW.iid
                      , NEW.date
                      , NEW.act;
      RETURN NULL;
    END IF;

  END IF;

  -- Interactions with Sids that are from point sample collection
  -- must have dates that match the point sample date.
  PERFORM 1
    FROM samples
    WHERE sid = NEW.sid
          AND date <> NEW.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'INTERACT_DATA %: %: SAMPLES: %: Date does not match corresponding SAMPLES.Date'
                    , NEW.iid
                    , NEW.date
                    , NEW.sid;
    RETURN NULL;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'interact_data_commit_func' AS function;
CREATE FUNCTION interact_data_commit_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  DECLARE
    c INT;

  BEGIN
  -- Function run on transaction commit for interact_data insert and update triggers
  --
  -- GPL_notice(`  --', `2012', `The Meme Factory, Inc., http://www.meme.com/')

  -- Get the latest values of the row
  SELECT * INTO NEW
    FROM interact_data
    WHERE interact_data.iid = NEW.iid;
  IF NOT FOUND THEN
    -- Whatever row was inserted was subsequently deleted.
    -- Nothing to do.
    RETURN NULL;
  END IF;

  -- Each INTERACT_DATA row should have exactly two corresponding Parts rows
  -- unless the focal individual of an ad-lib
  -- focal point sample is the lone participant.

  -- How many parts are related to the interact_data row?
  SELECT count(*) INTO c
    FROM parts
    WHERE parts.iid = NEW.iid;

  IF NEW.sid IS NULL THEN
    -- Must have 2 Parts rows when not collected with the psion.
    IF c <> 2 THEN
      RAISE EXCEPTION 'Error on INTERACT_DATA commit: Iid %: Act %: Date %: INTERACT_DATA rows must have exactly 2 corresponding rows on PARTS unless they are collected via ad-lib focal point sampling: There are % related PARTS rows'
          , NEW.iid
          , NEW.act
          , NEW.date
          , c;
      RETURN NULL;
    END IF;

  ELSE -- NEW.sid IS NOT NULL
    -- Validate when collected with psion.  Must have 2 parts rows unless
    -- there's only 1 for the focal.
    PERFORM 1
      FROM samples
      WHERE samples.sid = NEW.sid
            AND ((c <> 2 AND c  <> 1)
                 OR (c = 1
                     AND NOT EXISTS
                            (SELECT parts.sname
                               FROM parts
                               WHERE parts.iid =  NEW.iid
                                     AND parts.sname = samples.sname)));
    IF FOUND THEN
      RAISE EXCEPTION 'Error on INTERACT_DATA commit: Iid %: Act %: Date %: INTERACT_DATA rows must have exactly 2 corresponding rows on PARTS unless they are collected via ad-lib focal point sampling: There are % related PARTS rows'
          , NEW.iid
          , NEW.act
          , NEW.date
          , c;
      RETURN NULL;
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'interact_data_before_trigger' AS trigger;
CREATE TRIGGER interact_data_before_trigger
  BEFORE INSERT OR UPDATE
  ON interact_data FOR EACH ROW
  EXECUTE PROCEDURE interact_data_before_func();

SELECT 'interact_data_trigger' AS trigger;
CREATE TRIGGER interact_data_trigger
  AFTER INSERT OR UPDATE
  ON interact_data FOR EACH ROW
  EXECUTE PROCEDURE interact_data_func();

SELECT 'interact_data_commit_trigger' AS trigger;
CREATE CONSTRAINT TRIGGER interact_data_commit_trigger
  AFTER INSERT OR UPDATE
  ON interact_data
  DEFERRABLE INITIALLY DEFERRED
  FOR EACH ROW
  EXECUTE PROCEDURE interact_data_commit_func();
