dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004-2006, 2008, 2010, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

dnl plpgsql fragment for checking that an individual is not
dnl a dad on DAD_DATA.
dnl
dnl Syntax: _not_a_dad(column)
dnl
dnl column  A dad column on DAD_DATA.
dnl
dnl Restrictions: Must be used in a FOR EACH ROW trigger function
dnl when deleting or updating on maturedates.
dnl
dnl Bugs: Does extra work testing for MATUREDATES existance when
dnl used in a delete trigger function.
changequote({,})
define({_not_a_dad},{

  -- Individuals listed as ($1) dads must be mature as of the
  -- zdate when the mother is known.
  PERFORM 1
    FROM dad_data
         JOIN biograph ON (biograph.sname = dad_data.kid)
         JOIN pregs ON (pregs.pid = biograph.pid)
         JOIN cycpoints ON (cycpoints.cpid = pregs.conceive)
    WHERE dad_data.$1 = OLD.sname
          AND NOT EXISTS (SELECT 1
                            FROM maturedates
                            WHERE maturedates.sname = OLD.sname
                            LIMIT 1)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: Cannot delete/change the Sname in a MATUREDATES row when the individual is a DAD_DATA.$1 value and the mother is known'
                  , OLD.sname;
    RETURN NULL;
  END IF;
})
changequote(`,')

dnl plpgsql fragment for checking that an individual is mature
dnl on the zdate when he's listed as a dad.
dnl
dnl Syntax: _mature_dad(column)
dnl
dnl column  A dad column on DAD_DATA.
dnl
dnl Restrictions: Must be used in a FOR EACH ROW delete trigger function
dnl on maturedates.
changequote({,})
define({_mature_dad},{

    -- Check zdates for DAD_DATA.$1.
    PERFORM 1
      FROM dad_data
           JOIN biograph ON (biograph.sname = dad_data.kid)
           JOIN pregs ON (pregs.pid = biograph.pid)
           JOIN cycpoints ON (cycpoints.cpid = pregs.conceive)
      WHERE dad_data.$1 = NEW.sname
            AND cycpoints.date < NEW.matured
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Matured %: Individual must be mature on the zdate when listed in DAD_DATA.$1 when the zdate is known'
                  , NEW.sname
                  , NEW.matured;
      RETURN NULL;
    END IF;
})
changequote(`,')

--
-- maturedates
--
SELECT 'maturedates' AS table;

SELECT 'maturedates_func' AS function;
CREATE FUNCTION maturedates_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  -- Function for maturedates insert and updates triggers
  --
  -- GPL_notice(`  --', `2004-2006, 2008, 2010', `Karl O. Pinc <kop@meme.com>')

  BEGIN

  -- BIOGRAPH

  -- Maturedate must be at least 2.7ish years of age.
  check_age_ge_days(`MATUREDATES', `Matured', bb_min_maturity_age)

  -- When we observe maturity it cannot occur after
  -- 7.5ish years of age.
  -- (Endpoint based on ALE maturity.)
  IF NEW.Mstatus = 'bb_on' THEN
    check_age_lt_days(`MATUREDATES', `Matured', bb_max_maturity_age)
  END IF;

  -- matured must be on or before statdate
  check_statdate(`MATUREDATES', `Matured')


  -- RANKDATES

  -- ranked must be after the sexual maturity date.
  -- (The constraints enforce having a sexual maturity date.)
  IF NEW.Mstatus = 'bb_on' THEN
    rev_check_gt(`RANKDATES', `Ranked', `MATUREDATES', `Matured')
  ELSE
    rev_check_ge(`RANKDATES', `Ranked', `MATUREDATES', `Matured')
  END IF;

  -- CONSORTDATES

  -- consorted must be after an existing sexual maturity date.
  -- (The constraints enforce having a sexual maturity date.)
  rev_check_gt(`CONSORTDATES', `Consorted', `MATUREDATES', `Matured')


  -- CYCPOINTS 

  -- When "ON" date maturity date must be the earliest Tdate.
  -- If there are any cycles at all.
  IF NEW.mstatus = 'bb_on' AND
     EXISTS (SELECT 1 FROM cycles WHERE cycles.sname = NEW.sname) THEN
    PERFORM 1
      FROM (SELECT cycpoints.date
              FROM cycles, cycpoints
              WHERE cycles.sname = NEW.sname
                    AND cycpoints.cid = cycles.cid
                    AND cycpoints.code = 'bb_turgesence'
              ORDER BY cycpoints.date
              LIMIT 1) AS mintdate
      WHERE mintdate.date = NEW.matured
            AND mintdate.date IS NOT NULL
      LIMIT 1;
    IF NOT FOUND THEN
      RAISE EXCEPTION '%: %: When Mstatus = bb_on then Matured must be the earliest Tdate'
          , NEW.sname
          , NEW.matured;
      RETURN NULL;
    END IF;
  END IF;

  -- No Mdates, Tdates, or Ddates before the maturity date.
  PERFORM 1
    FROM cycles, cycpoints
    WHERE cycles.sname = NEW.sname
          AND cycpoints.cid = cycles.cid
          AND cycpoints.date < NEW.matured
          AND cycpoints.code <> 'bb_bleeding'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Mdates, Tdates, or Ddates must not be before MATUREDATES.Matured'
                  , NEW.sname
                  , NEW.matured;
    RETURN NULL;
  END IF;

  -- Tdates and Mdates can be on the maturity date, when it's an "ON" date,
  -- (Tdates and Mdates can be the same date, and presumeably that'd be the case.)
  -- but Ddates can be only when it's a "BY" date..

  PERFORM 1
    FROM cycles, cycpoints
    WHERE cycles.sname = NEW.sname
          AND cycpoints.cid = cycles.cid
          AND cycpoints.date = NEW.matured
          AND cycpoints.code = 'bb_deturgesence'
          AND NEW.Mstatus = 'bb_on'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Ddates must be after MATUREDATES.Matured when Mstatus is bb_on'
                  , NEW.sname
                  , NEW.matured;
    RETURN NULL;
  END IF;
-- Leftover from when you couldn't have _any_ cycpoints on a "by" maturity date.
--  PERFORM 1
--    FROM cycles, cycpoints
--    WHERE cycles.sname = NEW.sname
--          AND cycpoints.cid = cycles.cid
--          AND cycpoints.date = NEW.matured
--          AND cycpoints.code = 'bb_turgesence'
--          AND NEW.Mstatus <> 'bb_on'
--    LIMIT 1;
--  IF FOUND THEN
--    RAISE EXCEPTION '%: %: Tdates must be after MATUREDATES.Matured when Mstatus is not bb_on'
--                  , NEW.sname
--                  , NEW.matured;
--    RETURN NULL;
--  END IF;

  -- Can't have Bdates before the ON date.
  IF NEW.mstatus = 'bb_on' THEN
    PERFORM 1
      FROM cycles, cycpoints
      WHERE cycles.sname = NEW.sname
            AND cycpoints.cid = cycles.cid
            AND cycpoints.date < NEW.matured
            AND cycpoints.code = 'bb_bleeding'
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: %: Cannot have Bdates before MATUREDATES.Matured when Mstatus is bb_on'
                    , NEW.sname
                    , NEW.matured;
      RETURN NULL;
    END IF;
  END IF;

  -- SEXSKINS

  -- Can't have SEXSKINS connected to a cycle that has no Mdate
  -- if the cycle is not the one that occurs at menarche.
  -- However, there appears to be no good way to test this
  -- so this check is done "in batch".

  -- INTERACT/PARTS

  -- Female's mounts, consortships, and ejaculations must not come
  -- before the maturity date.  (Actually, not(male) to get unknowns too.)
  PERFORM 1
    FROM parts, interact_data, biograph
    WHERE biograph.sname = NEW.sname
          AND biograph.sex <> 'bb_male'
          AND parts.sname = NEW.sname
          AND interact_data.iid = parts.iid
          AND (interact_data.act = 'bb_mounts'
               OR interact_data.act = 'bb_consortship'
               OR interact_data.act = 'bb_ejaculation')
          AND interact_data.date < NEW.matured
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Female''s mounts, consortships, and ejaculations must not come before the maturity date'
                    , NEW.sname
                    , NEW.matured;
    RETURN NULL;
  END IF;

  -- SAMPLES

  -- Female samples must be on or after 1 year before maturity date.
  PERFORM 1
    FROM samples
    WHERE samples.sname = NEW.sname
          AND samples.stype = 'bb_female_point'
          AND samples.date < NEW.matured - CAST('1 year' AS INTERVAL)
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: MATUREDATES.Matured more than 1 year after the date of an (adult) female point sample'
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- DAD_DATA

  IF TG_OP = 'UPDATE' THEN
    -- There could be DAD_DATA rows.

    IF NEW.sname <> OLD.sname THEN
      -- Don't let an sname change result in an immature dad.

      -- When you're a dad (and the mother is known) you must be mature.
      _not_a_dad(`Dad_excl')
      _not_a_dad(`Dad_1perr')
      _not_a_dad(`Dad_5perr')
      _not_a_dad(`Dad_allmales')
      _not_a_dad(`Dad_consensus')
    END IF;

    -- When the mother is known the maturedate must be before the
    -- zdate.
    _mature_dad(`Dad_excl')
    _mature_dad(`Dad_1perr')
    _mature_dad(`Dad_5perr')
    _mature_dad(`Dad_allmales')
    _mature_dad(`Dad_consensus')
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'maturedates_delete_func' AS function;
CREATE FUNCTION maturedates_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  -- Function for maturedates delete trigger
  --
  -- GPL_notice(`  --', `2004-2006, 2008, 2010', `Karl O. Pinc <kop@meme.com>')

  DECLARE

  BEGIN
  -- RANKDATES
  -- CONSORTDATES
  -- Both taken care of by key constraints.

  -- CYCPOINTS 

  -- Only Bdates can be before sexual maturity
  -- and so can exist when there's no sexual maturity.
  PERFORM 1
    FROM cycles, cycpoints
    WHERE cycles.sname = OLD.sname
          AND cycpoints.cid = cycles.cid
          AND cycpoints.code <> 'bb_bleeding'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Cannot delete MATUREDATES because there are (non bleeding) CYCPOINTS'
        , OLD.sname
        , OLD.matured;
    RETURN NULL;
  END IF;

  -- Females must be mature to partitipate in mounts, consortships,
  -- or ejaculations.
  PERFORM 1
    FROM parts, interact_data, biograph
    WHERE biograph.sname = OLD.sname
          AND biograph.sex <> 'bb_male'
          AND parts.sname = OLD.sname
          AND interact_data.iid = parts.iid
          AND (interact_data.act = 'bb_mounts'
               OR interact_data.act = 'bb_consortship'
               OR interact_data.act = 'bb_ejaculation')
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION '%: %: Cannot delete MATUREDATES because there are mount, consortship, or ejaculation interactions for the female -- the female must be mature'
                    , OLD.sname
                    , OLD.matured;
    RETURN NULL;
  END IF;

  -- When you're a dad (and the mother is known) you must be mature.
  _not_a_dad(`Dad_excl')
  _not_a_dad(`Dad_1perr')
  _not_a_dad(`Dad_5perr')
  _not_a_dad(`Dad_allmales')
  _not_a_dad(`Dad_consensus')

  RETURN NULL;
  END;
$$;

SELECT 'maturedates_trigger' AS trigger;
CREATE TRIGGER maturedates_trigger
  AFTER INSERT OR UPDATE
  ON maturedates FOR EACH ROW
  EXECUTE PROCEDURE maturedates_func();

SELECT 'maturedates_delete_trigger' AS trigger;
CREATE TRIGGER maturedates_delete_trigger
  AFTER DELETE
  ON maturedates FOR EACH ROW
  EXECUTE PROCEDURE maturedates_delete_func();
