dnl Copyright (C) 2016 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2008, 2009, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

dnl plpgsql fragment for checking there is no gap in the seq
dnl after a row has been updated or deleted
dnl
dnl Syntax: _post_gap_check()
dnl
changequote({,})
define({_post_gap_check},{
    PERFORM 1
      FROM mpi_data AS bigger
      WHERE bigger.mpiid = OLD.mpiid
            AND bigger.seq > OLD.seq
            AND NOT EXISTS ( SELECT onebigger.mpidid
                               FROM mpi_data AS onebigger
                               WHERE onebigger.mpiid = OLD.mpiid
                                     AND onebigger.seq = OLD.seq + 1
                               LIMIT 1)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Mpiid %: Mpidid %: %:Old Seq %: MPI_DATA cannot have a gap (after the Seq) in the MPI_DATA.Seq sequence'
                    , OLD.mpiid
                    , OLD.mpidid
                    , OLD.mpiact
                    , OLD.Seq;
      RETURN NULL;
    END IF;

})dnl
changequote(`,')dnl

--  
-- mpi_data
--
SELECT 'mpi_data' AS table;

SELECT 'mpi_data_func' AS function;
CREATE FUNCTION mpi_data_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  DECLARE

  BEGIN
  -- Function for mpi_data insert and update triggers
  --
  -- Copyright (C) 2011 The Meme Factory, Inc.
  -- GPL_notice(`  --', `2008', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  -- This needs to check for a change in Seq number if we want to
  -- be sure that CONSORTS.Had is a participant in one of the
  -- initial interactions.

  -- mpidid cannot change

  IF TG_OP = 'UPDATE' THEN
    cannot_change(`MPI_DATA', `Mpidid')
    cannot_change(`MPI_DATA', `Mpiid')
  END IF;

  --
  -- Check the mpi interaction date against the entrydate and statdate.
  -- Really, this is mostly checked in the MPIS triggers but
  -- there's no special rule saying date or mpiid can't change, etc.
  -- 

  -- Entrydate
  PERFORM 1
    FROM mpis, mpi_parts, biograph
    WHERE mpis.mpiid = NEW.mpiid
          AND mpi_parts.mpidid = NEW.mpidid
          AND biograph.sname = mpi_parts.sname
          AND biograph.entrydate > mpis.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of MPI_DATA'
        , DETAIL = 'Key (Mpidid) = (' || NEW.mpidid
                     || '): Value (Mpiid) = (' || NEW.mpiid
                     || '): Value (Mpidid) = (' || NEW.mpidid
                     || '): Value (Seq) = (' || NEW.seq
                     || '): MPIS.Date cannot be before '
                     || 'BIOGRAPH.Entrydate';
    RETURN NULL;
  END IF;

  -- Statdatedate
  PERFORM 1
    FROM mpis, mpi_parts, biograph
    WHERE mpis.mpiid = NEW.mpiid
          AND mpi_parts.mpidid = NEW.mpidid
          AND biograph.sname = mpi_parts.sname
          AND biograph.statdate < mpis.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Mpiid %: Mpidid %: %: Seq %: BIOGRAPH.Statdate cannot be before MPIS.Date'
                    , NEW.mpiid
                    , NEW.mpidid
                    , NEW.mpiact
                    , NEW.seq;
    RETURN NULL;
  END IF;

  -- Seq = 1 means the Mpiact must be an agonism or help request.
  IF NEW.SEQ = 1 THEN
    PERFORM 1
      FROM mpiacts
      WHERE mpiacts.mpiact = NEW.mpiact
            AND (mpiacts.kind = 'bb_agonism_mact'
                 OR mpiacts.kind = 'bb_help_request_mact')
      LIMIT 1;
    IF NOT FOUND THEN
      RAISE EXCEPTION 'Mpiid %: Mpidid %: %: Seq %: The MPI_DATA.MPIAct must be an agonism when MPI_DATA.Seq is 1'
                      , NEW.mpiid
                      , NEW.mpidid
                      , NEW.mpiact
                      , NEW.seq;
      RETURN NULL;
    END IF;
  END IF;

  --
  -- Seq must not have gaps.
  --

  -- There must be a Seq - 1 row already.
  IF NEW.seq > 1 THEN
    PERFORM 1
      FROM mpi_data
      WHERE mpi_data.mpiid = NEW.mpiid
            AND mpi_data.seq = NEW.Seq - 1
      LIMIT 1;
    IF NOT FOUND THEN
      RAISE EXCEPTION 'Mpiid %: Mpidid %: %: Seq %: MPI_DATA cannot have a gap (before the Seq) in the MPI_DATA.Seq sequence'
                      , NEW.mpiid
                      , NEW.mpidid
                      , NEW.mpiact
                      , NEW.seq;
      RETURN NULL;
    END IF;
  END IF;

  IF TG_OP = 'UPDATE' THEN
    -- Changes to the sequence must not leave a gap.
    _post_gap_check()
  END IF;

  -- Is this a request for help?
  DECLARE

    request BOOLEAN;

  BEGIN

    SELECT INTO request mpiacts.kind = 'bb_help_request_mact'
      FROM mpiacts
      WHERE mpiacts.mpiact = NEW.mpiact;

    -- Does Helped have a value when it must and the reverse?
    IF (request AND NEW.helped IS NULL)
       OR (NOT(request) AND NEW.helped IS NOT NULL) THEN
      RAISE EXCEPTION 'Mpidid %: MPIAct %: Helped %: MPI_DATA.Helped must have a value when the row is a help request, otherwise it must be NULL.'
                    , NEW.mpidid
                    , NEW.mpiact
                    , NEW.helped;
      RETURN NULL;
    END IF;

    IF request THEN
      -- Helped must be false when help is requested from an unknown
      -- individual.
      IF NOT(NEW.Helped) THEN
        PERFORM 1
          FROM mpi_parts
          WHERE mpi_parts.mpidid = NEW.mpidid
                AND mpi_parts.role = 'bb_actee'
                AND mpi_parts.unksname IS NOT NULL
          LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION 'Mpidid %: MPIAct %: MPI_DATA.Helped must be FALSE when help is requested from an unknown individual'
                        , NEW.mpidid
                        , NEW.mpiact;
          RETURN NULL;
        END IF;
      END IF;
    END IF;

    -- Does Active have a value when it must and the reverse?
    IF (request AND NEW.active IS NULL)
       OR (NOT(request) AND NEW.active IS NOT NULL) THEN
      RAISE EXCEPTION 'MPIAct %: Active %: MPI_DATA.Active must have a value when the row is a help request, otherwise it must be NULL.'
                    , NEW.mpiact
                    , NEW.active;
      RETURN NULL;
    END IF;
  END;

  -- There must be only one mpi_data row (per MPIS) with a Seq of 1,
  -- unless MPIACTS.Multi_first is TRUE.
  IF NEW.Seq = 1 THEN
    PERFORM 1
      FROM mpi_data, mpiacts AS thisact, mpiacts AS existingact
      WHERE mpi_data.mpiid = NEW.mpiid
            AND mpi_data.seq = 1
            AND mpi_data.mpidid <> NEW.mpidid
            AND thisact.mpiact = NEW.mpiact
            AND existingact.mpiact = mpi_data.mpiact
            AND NOT(thisact.multi_first AND existingact.multi_first)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Mpiid %: Mpidid %: MPIAct %: Seq %: MPI_PARTS cannot have two rows for the same multi-party interaction with a Seq of 1 unless all Seq 1 related MPIAct.Multi_first values are TRUE'
                      , NEW.mpiid
                      , NEW.mpidid
                      , NEW.mpiact
                      , NEW.seq;
      RETURN NULL;
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'mpi_data_delete_func' AS function;
CREATE FUNCTION mpi_data_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  DECLARE

  BEGIN
  -- Function for mpi_data delete triggers
  --
  -- GPL_notice(`  --', `2008', `Karl O. Pinc <kop@meme.com>')

  -- Changes to the sequence must not leave a gap.
  _post_gap_check()

  RETURN NULL;
  END;
$$;

SELECT 'mpi_data_trigger' AS trigger;
CREATE TRIGGER mpi_data_trigger
  AFTER INSERT OR UPDATE
  ON mpi_data FOR EACH ROW
  EXECUTE PROCEDURE mpi_data_func();

SELECT 'mpi_data_delete_trigger' AS trigger;
CREATE TRIGGER mpi_data_delete_trigger
  AFTER DELETE
  ON mpi_data FOR EACH ROW
  EXECUTE PROCEDURE mpi_data_delete_func();
