dnl Copyright (C) 2014, 2016 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004, 2006, 2008, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

--  
-- neighbors
--
SELECT 'neighbors' AS table;

SELECT 'neighbors_func' AS function;
CREATE FUNCTION neighbors_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for neighbors insert and update triggers
  --
  -- GPL_notice(`  --', `2004, 2006, 2008', `Karl O. Pinc <kop@meme.com>')

  -- No point sample can have two of the same ncodes.
  PERFORM 1
    FROM neighbors
    WHERE pntid = NEW.pntid
          AND ncode = NEW.ncode
          AND nghid <> NEW.nghid
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Pntid %: %: Duplicate NEIGHBORS.Ncode found for the same Pntid'
                    , NEW.pntid
                    , NEW.ncode;
    RETURN NULL;
  END IF;

  -- No point sample can have a neighbor that's the focal individual.
  PERFORM 1
    FROM samples, points
    WHERE points.pntid = NEW.pntid
          AND samples.sid = points.sid
          AND samples.sname = NEW.sname
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Pntid %: %: NEIGHBORS.Sname cannot be the SAMPLES.Sname (individual cannot be it''s own neighbor)'
                    , NEW.pntid
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- Test for sample type.
  PERFORM 1
    FROM samples, points
    WHERE points.pntid = NEW.pntid
          AND samples.sid = points.sid
          AND samples.stype = 'bb_juvenile_point'
    LIMIT 1;
  IF FOUND THEN
    -- Juvenile sample.

    -- Juvenile samples must have ncodes that go with the protocol.
    PERFORM 1
      FROM ncodes
      WHERE ncodes.ncode = NEW.ncode
            AND NOT(juvenile)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Pntid %: %: Juvenile point samples must have NEIGHBORS.Ncodes for juveniles (NCODES.Juvenile)'
                      , NEW.pntid
                      , NEW.ncode;
      RETURN NULL;
    END IF;

  ELSE
    -- Female point sample (enforced by constraint)

    -- Female samples must have ncodes that go with the protocol.
    PERFORM 1
      FROM ncodes
      WHERE ncodes.ncode = NEW.ncode
            AND NOT(female)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Pntid %: %: Female point samples must have NEIGHBORS.Ncodes for females (NCODES.Female)'
                      , NEW.pntid
                      , NEW.ncode;
      RETURN NULL;
    END IF;

  END IF;

  -- Neighbor must have entered population on or by sample date
  PERFORM biograph.sname FROM biograph, samples, points
    WHERE points.pntid = NEW.pntid
          AND samples.sid = points.sid
          AND biograph.sname = NEW.sname
          AND biograph.entrydate > samples.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of NEIGHBORS'
        , DETAIL = 'Key (Nghid) = (' || NEW.nghid
                     || '): Value (Pntid) = (' || NEW.pntid
                     || '): BIOGRAPH.Entrydate of NEIGHBORS.Sname '
                     || 'cannot be greater than SAMPLES.Date'
        , HINT = 'Neighbor not in population yet';
    RETURN NULL;
  END IF;

  -- Neighbor must be alive on sample date
  PERFORM BIOGRAPH.sname FROM BIOGRAPH, samples, points
    WHERE points.pntid = NEW.pntid
          AND samples.sid = points.sid
          AND BIOGRAPH.sname = NEW.sname
          AND BIOGRAPH.Statdate < samples.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on NEIGHBORS ' || TG_OP || ' commit'
        , DETAIL = 'Key (Nghid) = (' || NEW.nghid
                     || '): Value (Sname) = (' || NEW.sname
                     || '): Neighbor''s BIOGRAPH.Statdate cannot be less '
                     || 'than the SAMPLES.Date';
    RETURN NULL;
  END IF;

  -- NCODES.Require
  -- Some ncodes cannot be present unless others already are.
  DECLARE
    required_ncode ncodes.requires%TYPE;
    nsex           ncodes.nsex%TYPE;
    other_sname    neighbors.sname%TYPE;

  BEGIN
    -- Get the required ncode, if any.
    SELECT INTO required_ncode,  nsex
                ncodes.requires, ncodes.nsex
      FROM ncodes
      WHERE ncodes.ncode = NEW.ncode;
    IF required_ncode IS NOT NULL THEN
      -- We've got a required ncode.

      -- Check that the required row exists.
      SELECT INTO other_sname neighbors.sname
        FROM neighbors
        WHERE neighbors.pntid = NEW.pntid
              AND neighbors.ncode = required_ncode
        LIMIT 1;
      IF NOT FOUND THEN
        RAISE EXCEPTION 'Nghid %: Pntid %: neighbor %: Ncode %: required ncode %: Requires a neighbor have an Ncode and there is no such neighbor'
                        , NEW.nghid
                        , NEW.pntid
                        , NEW.sname
                        , NEW.ncode
                        , required_ncode;
        RETURN NULL;
      END IF;

      IF nsex = 'bb_nc_opposite' THEN
        -- The ncode requires the neighbor to be of a different sex.
        -- (If the neighbor has an Unksname then this_sname will
        -- be NULL and the FOUND test will consider the neighbor to be
        -- of the opposite sex.P
        PERFORM 1
          FROM biograph AS this_biograph
               , biograph AS other_biograph
          WHERE this_biograph.sname = NEW.sname
                AND other_biograph.sname = other_sname
                AND this_biograph.sex = other_biograph.sex
          LIMIT 1;
        IF FOUND THEN
          RAISE EXCEPTION 'Nghid %: Pntid %: neighbor %: Ncode %: required ncode %: Requires the neighbor be of a different sex and there is no such neighbor'
                        , NEW.nghid
                        , NEW.pntid
                        , NEW.sname
                        , NEW.ncode
                        , required_ncode;
          RETURN NULL;
        END IF;
      END IF;
    END IF;
  END;

  -- Some ncodes constrain the sex.

  -- See if the ncode requires the neighbor be male.
  PERFORM 1
    FROM ncodes, biograph
    WHERE ncodes.ncode = NEW.ncode
          AND ncodes.nsex = 'bb_nc_male'
          AND biograph.sname = NEW.sname
          AND biograph.sex <> 'bb_male'
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Nghid %: Pntid %: neighbor %: Ncode %: According to the ncode the neighbor must be male'
                    , NEW.nghid
                    , NEW.pntid
                    , NEW.sname
                    , NEW.ncode;
    RETURN NULL;
  END IF;

  -- Does the neighbor have to be unique?
  PERFORM 1
    FROM ncodes, neighbors
    WHERE ncodes.ncode = NEW.ncode
          AND ncodes.nunique
          AND neighbors.pntid = NEW.pntid
          AND neighbors.sname = NEW.sname
          AND neighbors.nghid <> NEW.nghid
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Nghid %: Pntid %: %: Ncode %: According to the Ncode the neighbor must be unique (NCODES.Nunique)'
                    , NEW.nghid
                    , NEW.pntid
                    , NEW.sname
                    , NEW.ncode;
    RETURN NULL;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'neighbors_trigger' AS trigger;
CREATE TRIGGER neighbors_trigger
  AFTER INSERT OR UPDATE
  ON neighbors FOR EACH ROW
  EXECUTE PROCEDURE neighbors_func();
