dnl Copyright (C) 2016 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012, 2014, The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004, 2006-2008, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

dnl plpgsql fragment for updating SAMPLES.Minsis
dnl
dnl Syntax: _update_minsis(sid)
dnl
dnl sid   NEW.sid or OLD.sid.  The sid of the SAMPLES row to update.
dnl
changequote({,})
define({_update_minsis},{
      UPDATE samples
        SET minsis = (SELECT count(*)
                        FROM point_data
                        WHERE point_data.sid = $1)
        WHERE sid = $1;
})dnl
changequote(`,')dnl

--  
-- point_data
--
SELECT 'point_data' AS table;

SELECT 'point_data_func' AS function;
CREATE FUNCTION point_data_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for point_data insert and update triggers
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  -- Copyright (C) 2004, 2006-2008 Karl O. Pinc <kop@meme.com>

  PERFORM 1
    FROM samples
    WHERE sid = NEW.sid
          AND stype = 'bb_juvenile_point'
    LIMIT 1;
  IF FOUND THEN
    -- Juvenile point sample

    -- Juvenile's only allowed certain activity codes.
    PERFORM 1
      FROM activities
      WHERE activity = NEW.activity
            AND NOT(juvenile)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Pntid %: %: Juvenile POINT_DATA.Activity not a juvenile activity (ACTIVITIES.Juvenile)'
                      , NEW.pntid
                      , NEW.activity;
      RETURN NULL;
    END IF;

  ELSE
    -- Female point sample (constraints enforce this)

    -- Female's only allowed certain activity codes.
    PERFORM 1
      FROM activities
      WHERE activity = NEW.activity
            AND NOT(female)
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Pntid %: %: Female POINT_DATA.Activity not a female activity (ACTIVITIES.Female)'
                      , NEW.pntid
                      , NEW.activity;
      RETURN NULL;
    END IF;

    -- Female's only allowed certain posture codes.
    PERFORM 1
      FROM postures
      WHERE postures.posture = NEW.posture
            AND postures.juvonly
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Pntid %: %: Female POINT_DATA.Posture cannot be a juvenile-only posture (POSTURES.Juvonly).'
                      , NEW.pntid
                      , NEW.posture;
      RETURN NULL;
    END IF;

    -- Integrity checks in case SID is changed

    -- The focal individual cannot be recorded as their neighbor (in NEIGHBORS)
    PERFORM 1
      FROM samples, neighbors
      WHERE samples.sid = NEW.sid
            AND neighbors.pntid = NEW.pntid
            AND samples.sname = neighbors.sname
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on POINT_DATA ' || TG_OP || ' commit'
          , DETAIL = 'Key (Pntid) = (' || NEW.pntid
                       || '): Value (Sid) = (' || NEW.sid
                       || '): NEIGHBORS.Sname cannot be the SAMPLES.Sname '
                       || '(focal cannot have itself as a neighbor)';
      RETURN NULL;
    END IF;

    -- Neighbor must have entered population on or by sample date
    PERFORM 1
      FROM biograph, samples, neighbors
      WHERE neighbors.pntid = NEW.pntid
            AND samples.sid = NEW.sid
            AND biograph.sname = neighbors.sname
            AND samples.date < biograph.entrydate
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of POINT_DATA'
          , DETAIL = 'Key (Pntid) = (' || NEW.pntid
                       || '): Value (Sid) = (' || NEW.sid
                       || '): BIOGRAPH.Entrydate of NEIGHBORS.Sname '
                       || 'cannot be greater than SAMPLES.Date'
          , HINT = 'Neighbor not in population yet';
      RETURN NULL;
    END IF;

    --Neighbor must be alive/noncensored on sample_date
    PERFORM 1
      FROM biograph, samples, neighbors
      WHERE neighbors.pntid = NEW.pntid
            AND samples.sid = NEW.sid
            AND biograph.sname = neighbors.sname
            AND samples.date > biograph.statdate
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on POINT_DATA ' || TG_OP || ' commit'
          , DETAIL = 'Key (Pntid) = (' || NEW.pntid
                       || '): Value (Sid) = (' || NEW.sid
                       || '): BIOGRAPH.Statdate of NEIGHBORS.Sname '
                       || 'cannot be less than SAMPLES.Date (dead '
                       || 'neighbor)';
      RETURN NULL;
    END IF;

  END IF;

  --
  -- All done, update SAMPLES.Minsis.
  --
  IF TG_OP = 'INSERT' THEN
    _update_minsis(NEW.sid)
  ELSE
    IF NEW.sid <> OLD.sid THEN
      _update_minsis(NEW.sid)
      _update_minsis(OLD.sid)
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'point_data_delete_func' AS function;
CREATE FUNCTION point_data_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for point_data delete trigger
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  -- Copyright (C) 2007, 2008 Karl O. Pinc <kop@meme.com>

  DECLARE

  BEGIN

  -- Update SAMPLES.Minsis
  _update_minsis(OLD.sid)

  RETURN NULL;
  END;
$$;

SELECT 'point_data_trigger' AS trigger;
CREATE TRIGGER point_data_trigger
  AFTER INSERT OR UPDATE
  ON point_data FOR EACH ROW
  EXECUTE PROCEDURE point_data_func();

SELECT 'point_data_delete_trigger' AS trigger;
CREATE TRIGGER point_data_delete_trigger
  AFTER DELETE
  ON point_data FOR EACH ROW
  EXECUTE PROCEDURE point_data_delete_func();
