dnl Copyright (C) 2012, 2017 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2009, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl
dnl
dnl
dnl plpgsql fragment for updating RAINGAUGES RGspan and EstRGspan.
dnl
dnl Syntax: update_span(row)
dnl
dnl Input:
dnl   row  OLD or NEW, the row upon which to base the update.
dnl
dnl Variables required:
dnl   next_wrid, next_daytime, next_estdaytime
dnl   prior_rgsdaytime, prior_rgsestdaytime
dnl   prior_wrdaytime, prior_wrestdaytime
dnl
dnl Restrictions: 
dnl
dnl Remarks:
dnl  This does extra work when row is NEW because we already know
dnl the raingauge setup row we're looking for (the one prior to the
dnl raingauge reading we're updating) is the NEW row.
changequote({,})
define({update_span},{
    -- Get the raingauge measurement following the $1 rain gauge
    -- setup time.
    SELECT raingauges.wrid, wreadings.wrdaytime, wreadings.estdaytime
      INTO next_wrid,       next_daytime,        next_estdaytime
      FROM wreadings, raingauges
      WHERE wreadings.wstation = $1.wstation
            AND raingauges.wrid = wreadings.wrid
            AND wreadings.wrdaytime > $1.rgsdaytime
      ORDER BY wreadings.wrdaytime
      LIMIT 1;
    IF FOUND THEN
      -- There is a raingauge measurement following the raingauge setup
      -- timestamp, so we'll need to update it.

      -- Find the latest raingauge setup prior to the raingauge
      -- measurement.
      SELECT rgsetups.rgsdaytime, rgsetups.rgsestdaytime
        INTO prior_rgsdaytime,    prior_rgsestdaytime
        FROM rgsetups
        WHERE rgsetups.wstation = $1.wstation
              AND rgsetups.rgsdaytime < next_daytime
        ORDER BY rgsetups.rgsdaytime DESC
        LIMIT 1;
      IF NOT FOUND THEN
        -- This is never reached when called from an insert trigger
        -- because the rgsetup row found should always be the
        -- inserted row.
        RAISE EXCEPTION integrity_constraint_violation USING
              MESSAGE = 'Error on ' || TG_OP || ' of RGSETUPS'
            , DETAIL = 'Key (RGSId) = (' || OLD.rgsid
                         || '): Old Value (RGSdaytime) = (' || OLD.rgsdaytime
                         || '): New Value (RGSdaytime) = (' || NEW.rgsdaytime
                         || '): Cannot change the RGSETUPS.RGSdaytime '
                         || 'because that leaves a raingauge reading without '
                         || 'a setup';
        RETURN NULL;
      END IF;

      -- Find the latest raingauge reading prior to the reading we'll
      -- be updating.
      SELECT wreadings.wrdaytime, wreadings.estdaytime
        INTO prior_wrdaytime,     prior_estdaytime
        FROM wreadings, raingauges
        WHERE wreadings.wstation = $1.wstation
              AND raingauges.wrid = wreadings.wrid
              AND wreadings.wrdaytime < next_daytime
        ORDER BY wreadings.wrdaytime DESC
        LIMIT 1;

      IF FOUND AND prior_wrdaytime > prior_rgsdaytime THEN
        -- The span is based on the prior raingauge measurement.
        UPDATE raingauges
          SET rgspan = next_daytime - prior_wrdaytime,
              estrgspan = next_estdaytime OR prior_estdaytime
          WHERE raingauges.wrid = next_wrid;
      ELSE
        -- The span is based on the initialization.
        UPDATE raingauges
          SET rgspan = next_daytime - prior_rgsdaytime,
              estrgspan = next_estdaytime OR prior_rgsestdaytime
          WHERE raingauges.wrid = next_wrid;
      END IF;
    END IF;
})
changequote(`,')dnl

--  
-- Rgsetups
--
SELECT 'rgsetups' AS table;

SELECT 'rgsetups_func' AS function;
CREATE FUNCTION rgsetups_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER -- so raingauges can be updated
  bb_function_set_search_path
  AS $$

  DECLARE
    next_wrid raingauges.wrid%TYPE;
    next_daytime wreadings.wrdaytime%TYPE;
    next_estdaytime wreadings.estdaytime%TYPE;
    prior_rgsdaytime rgsetups.rgsdaytime%TYPE;
    prior_rgsestdaytime rgsetups.rgsestdaytime%TYPE;
    prior_wrdaytime wreadings.wrdaytime%TYPE;
    prior_estdaytime wreadings.estdaytime%TYPE;

  BEGIN
  -- Function for RGSETUPS insert and update triggers
  --
  -- GPL_notice(`  --', `2009', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  --  Does not attempt to avoid redundant work, including some
  -- redundant work that's just there to keep the size of the code down.

  -- Cannot have a raingauge setup at the same time as a raingauge
  -- reading.
  PERFORM 1
    FROM wreadings, raingauges
    WHERE wreadings.wrdaytime = NEW.rgsdaytime
          AND wreadings.wstation = NEW.wstation
          AND raingauges.wrid = wreadings.wrid
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'RGSid %: Wstation %: RGSdaytime %: Cannot have a RGSTEUPS row with the same time and location as a RAINGAUGES row'
                  , NEW.rgsid
                  , NEW.wstation
                  , NEW.rgsdaytime;
    RETURN NULL;
  END IF;

  IF TG_OP = 'INSERT' THEN
    -- Update the rgspan & estrgspan for the next raingauge reading.
    update_span(`NEW')
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'rgsetups_update_func' AS function;
CREATE FUNCTION rgsetups_update_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER -- so raingauges can be updated
  bb_function_set_search_path
  AS $$

  DECLARE
    next_wrid raingauges.wrid%TYPE;
    next_daytime wreadings.wrdaytime%TYPE;
    next_estdaytime wreadings.estdaytime%TYPE;
    prior_rgsdaytime rgsetups.rgsdaytime%TYPE;
    prior_rgsestdaytime rgsetups.rgsestdaytime%TYPE;
    prior_wrdaytime wreadings.wrdaytime%TYPE;
    prior_estdaytime wreadings.estdaytime%TYPE;

  BEGIN
  -- Function for RGSETUPS update triggers
  --
  -- GPL_notice(`  --', `2009', `Karl O. Pinc <kop@meme.com>')

  -- Cannot change the RGSid.
  cannot_change(`RGSETUPS', `RGSid')

  -- Cannot change the weather station.
  -- (It's just easier that way.)
  cannot_change(`RGSETUPS', `Wstation')

  IF NEW.rgsdaytime <> OLD.wrdaytime THEN
    --
    -- Update the rgspan and estrgspan for the rain gauge measurement
    -- following the old rain gauge setup timestamp.
    --

    update_span(`OLD')

    --
    -- Update the rgspan and estrgspan for the rain gauge measurement
    -- following the new rain gauge setup timestamp.
    --

    update_span(`NEW')

  ELSE
    IF NEW.estdaytime <> OLD.estdaytime THEN
      --
      -- Update the estrgspan for the rain gauge measurement
      -- following the rain gauge setup event.
      --

      -- Get the rain gauge measurement following the setup event.
      SELECT raingauges.wrid, wreadings.estdaytime
        INTO next_wrid,       next_estdaytime
        FROM wreadings, raingauges
        WHERE wreadings.wstation = NEW.wstation
              AND raingauges.wrid = wreadings.wrid
              AND wreadings.wrdaytime > NEW.rgsdaytime
        ORDER BY wreadings.wrdaytime
        LIMIT 1;

      IF FOUND THEN
        UPDATE raingauges
          SET estrgspan = NEW.estdaytime OR next_estdaytime
          WHERE raingauges.wrid = next_wrid;
      END IF;

    END IF;
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'rgsetups_delete_func' AS function;
CREATE FUNCTION rgsetups_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER -- so raingauges can be updated
  bb_function_set_search_path
  AS $$

  DECLARE
    prior_rgsdaytime rgsetups.rgsdaytime%TYPE;
    prior_rgsestdaytime rgsetups.rgsestdaytime%TYPE;
    prior_wrdaytime wreadings.wrdaytime%TYPE;
    prior_estdaytime wreadings.estdaytime%TYPE;
    next_daytime wreadings.wrdaytime%TYPE;
    next_estdaytime wreadings.estdaytime%TYPE;
    next_wrid raingauges.wrid%TYPE;

  BEGIN
  -- Function for delete trigger
  --
  -- GPL_notice(`  --', `2009', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  --  - Control does not flow off the bottom.  Does a RETURN rather
  --    than have a large nested IF.

  -- Get the next rainguage reading.  It may need updating.
  SELECT wreadings.wrdaytime, wreadings.estdaytime, raingauges.wrid
    INTO next_daytime,        next_estdaytime,      next_wrid
    FROM wreadings, raingauges
    WHERE wreadings.wstation = OLD.wstation
          AND raingauges.wrid = wreadings.wrid
          AND wreadings.wrdaytime > OLD.rgsdaytime
    ORDER BY wreadings.wrdaytime
    LIMIT 1;
  IF NOT FOUND THEN
    -- There's no next raingauge reading so nothing to update.
    RETURN NULL;
  END IF;

  -- Find the raingauge setup for this station that is immediately prior
  -- to the next raingauge reading.
  SELECT rgsetups.rgsdaytime, rgsetups.rgsestdaytime
    INTO prior_rgsdaytime,     prior_rgsestdaytime
    FROM rgsetups
    WHERE rgsetups.wstation = OLD.wstation
          AND rgsetups.rgsdaytime < next_daytime
          AND rgsetups.rgsid <> OLD.rgsid
    ORDER BY rgsetups.rgsdaytime desc
    LIMIT 1;
  IF NOT FOUND THEN
    -- No prior rgsetups.
    -- Can't delete because it will leave a raingauge reading
    -- without a setup time.
    RAISE EXCEPTION 'RGSid %: RGSdaytime %: RGSwstation %: RGSETUPS row cannot be deleted because related RAINGAUGES row(s) won''t have a setup time'
                  , OLD.rgsid
                  , OLD.rgsdaytime
                  , OLD.rgswstation;
    RETURN NULL;
  END IF;

  -- Get the raingauge reading that is immediately prior to the
  -- next raingauge reading.
  SELECT wreadings.wrdaytime, wreadings.estdaytime
    INTO prior_wrdaytime,     prior_estdaytime
    FROM wreadings, raingauges
    WHERE wreadings.wstation = OLD.wstation
          AND raingauges.wrid = wreadings.wrid
          AND wreadings.wrdaytime < next_daytime
    ORDER BY wreadings.wrdaytime DESC
    LIMIT 1;

  IF FOUND AND prior_wrdaytime > prior_rgsdaytime THEN
    -- Compute the span from the prior rain gauge measurement.
    UPDATE raingauges
      SET rgspan = next_daytime - prior_wrdaytime,
          estrgspan = next_estdaytime OR prior_estdaytime
      WHERE raingauges.wrid = next_wrid;
  ELSE
    -- Computer the span from when the rain gauge was setup.
    UPDATE raingauges
      SET rgspan = next_daytime - prior_rgsdaytime,
          estrgspan = next_estdaytime OR prior_rgsestdaytime
      WHERE raingauges.wrid = next_wrid;
  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'rgsetups_trigger' AS trigger;
CREATE TRIGGER rgsetups_trigger
  AFTER INSERT OR UPDATE
  ON rgsetups FOR EACH ROW
  EXECUTE PROCEDURE rgsetups_func();

SELECT 'rgsetups_update_trigger' AS trigger;
CREATE TRIGGER rgsetups_update_trigger
  AFTER UPDATE
  ON rgsetups FOR EACH ROW
  EXECUTE PROCEDURE rgsetups_update_func();

SELECT 'rgsetups_delete_trigger' AS trigger;
CREATE TRIGGER rgsetups_delete_trigger
  AFTER DELETE
  ON rgsetups FOR EACH ROW
  EXECUTE PROCEDURE rgsetups_delete_func();
