dnl Copyright (C) 2014 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2004, 2006, 2008, 2010, Karl O. Pinc <kop@meme.com>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl


--  
-- samples
-- $Id: samples.m4,v 1.2 2004/11/08 18:29:01 kop Exp $
--
SELECT 'samples' AS table;

SELECT 'samples_func' AS function;
CREATE FUNCTION samples_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for samples insert and update triggers
  --
  -- GPL_notice(`  --', `2004, 2006, 2008, 2010', `Karl O. Pinc <kop@meme.com>')

  -- Can't have a sample before you entered the
  -- study population.
  check_entrydate(`SAMPLES', `Date')

  IF NEW.stype = 'bb_juvenile_point' THEN
    -- Point samples are collected on infants and juveniles 
    -- from the age of 6 months until  testicular enlargement
    -- (for males) or menarche (for females). 
    -- However, we don't really care if they are taken earlier.

    -- Maturity test
    -- Can't have a male or unknown juvenile sample more than a year
    -- after rank date.
    PERFORM 1
      FROM rankdates, biograph
      WHERE rankdates.ranked <= NEW.date - CAST ('1 year' AS Interval)
            AND rankdates.sname = NEW.sname
            AND biograph.sname = NEW.sname
            AND biograph.sex <> 'bb_female'
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Sid %: %: Juvenile point sample taken for a male (or unknown) on or after RANKDATES.Ranked plus 1 year'
                      , NEW.sname
                      , NEW.sid
                      , NEW.date;
      RETURN NULL;
    END IF;

    -- Can't have a female juvenile sample once there's a conception.
    PERFORM 1
      FROM cycles, cycpoints, pregs
      WHERE cycles.sname = NEW.sname
            AND cycpoints.cid = cycles.cid
            AND cycpoints.cpid = pregs.conceive
            AND NEW.date >= cycpoints.date
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Sid %: %: Juvenile point sample''s date is on or after the female''s first conception'
                      , NEW.sname
                      , NEW.sid
                      , NEW.date;
      RETURN NULL;
    END IF;

    -- Juvenile point samples cannot have fpoints information
    PERFORM 1
      FROM point_data, fpoints
      WHERE point_data.sid = NEW.sid
            AND fpoints.pntid = point_data.pntid
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Sid %: %: SAMPLES.Stype of bb_juvenile_point must not have related FPOINTS'
                      , NEW.sid
                      , NEW.sname;
      RETURN NULL;
    END IF;

  ELSE
    -- Female point sample (enforced by constraint)

    -- Point samples collected for females must be collected on females.
    PERFORM 1
      FROM biograph
      WHERE sname = NEW.sname
            AND sex <> 'bb_female'
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION 'Sid %: %: SAMPLES.Stype of bb_female_point must be observations of females'
                      , NEW.sid
                      , NEW.sname;
      RETURN NULL;
    END IF;

    -- Female must be within 1 year of maturity at the time of sampling.
    PERFORM 1
      FROM maturedates
      WHERE maturedates.sname = NEW.sname
            AND maturedates.matured - CAST('1 year' AS INTERVAL) > NEW.date
      LIMIT 1;
    IF FOUND THEN
      RAISE EXCEPTION '%: Sid %: %: Adult female point sample taken more than 1 year before MATUREDATES.Matured'
                      , NEW.sname
                      , NEW.sid
                      , NEW.date;
      RETURN NULL;
    END IF;

  END IF;

  -- Individual cannot be dead at the time of sampling.
  check_statdate(`SAMPLES', `Date')

  -- Neighbor must have entered the population on or by sample date
  PERFORM biograph.sname FROM biograph, point_data, neighbors
    WHERE point_data.sid = NEW.sid
          AND neighbors.pntid = point_data.pntid
          AND biograph.sname = neighbors.sname
          AND biograph.entrydate > NEW.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on ' || TG_OP || ' of SAMPLES'
        , DETAIL = 'Key (Sid) = (' || NEW.sid
                     || '): Value (Date) = (' || NEW.date
                     || '): BIOGRAPH.Entrydate of NEIGHBORS.Sname '
                     || 'cannot be greater than SAMPLES.Date'
        , HINT = 'Neighbor not in population yet';
    RETURN NULL;
  END IF;

  -- Neighbors must be alive on sample date
  PERFORM 1 FROM biograph, points, neighbors
    WHERE points.sid = NEW.sid
          AND neighbors.pntid = points.pntid
          AND biograph.sname = neighbors.sname
          AND biograph.statdate < NEW.date
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION integrity_constraint_violation USING
          MESSAGE = 'Error on SAMPLES ' || TG_OP || ' commit'
        , DETAIL = 'Key (Sid) = (' || NEW.Sid
                     || '): Value (Date) = (' || NEW.date
                     || '): Neighbor''s BIOGRAPH.Statdate cannot be less '
                     || 'than SAMPLES.Date';
    RETURN NULL;
  END IF;

  -- Focal individual cannot be a neighbor.
  PERFORM 1
    FROM point_data, neighbors
    WHERE point_data.sid = NEW.sid
          AND neighbors.pntid = point_data.pntid
          AND neighbors.sname = NEW.sname
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'Sid %: %: SAMPLES.Sname cannot be the same as a related NEIGHBORS.Sname (individual cannot be it''s own neighbor)'
                    , NEW.sid
                    , NEW.sname;
    RETURN NULL;
  END IF;

  -- Minsis must be the number of related point_data rows.
  DECLARE
    actual_minsis INT;

  BEGIN
    SELECT INTO actual_minsis count(*)
      FROM point_data
      WHERE point_data.sid = NEW.sid;

    IF NEW.minsis <> actual_minsis THEN
      RAISE EXCEPTION 'Sid %: SAMPLES.Minsis (%) must be equal to the number of related POINT_DATA rows (%)'
                      , NEW.sid
                      , NEW.minsis
                      , actual_minsis;
      RETURN NULL;
    END IF;
  END;

  --
  -- All the validation is done.  Automatically update INTERACT_DATA
  -- when needed.
  --
  IF TG_OP = 'UPDATE' THEN
    UPDATE interact_data
      SET observer = NEW.observer
      WHERE sid = NEW.sid;
  END IF;

  RETURN NULL;
  END;
$$;


SELECT 'samples_trigger' AS trigger;
CREATE TRIGGER samples_trigger
  AFTER INSERT OR UPDATE
  ON samples FOR EACH ROW
  EXECUTE PROCEDURE samples_func();
