dnl Copyright (C) 2017 Jake Gordon, <jacob.gordon@duke.edu>
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Jake Gordon <jacob.gordon@duke.edu>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl

--  
-- WeatherHawk
--
SELECT 'weatherhawk' AS table;

  -- The calculation for the TimeStampRain column uses a row's
  -- YearlyRain and subtracts the (chronological) previous row's
  -- YearlyRain from it. This is simple, except that the final row of
  -- each year has a TimeStamp of the next year, e.g. the data from
  -- 23:00-23:59 of 31 December 2010 has a TimeStamp of Janary 1,
  -- 2011, 00:00:00. So if we rely only on the year from the
  -- TimeStamp, every January 1, 20xx 01:00:00 row will has a negative
  -- TimeStampRain, because the 00:00:00 row had all of the rain from
  -- the previous year.
  --
  -- When it helps to keep the code clear, we declare a variable to
  -- indicate the TimeStamp's "effective" time (i.e. one hour
  -- previous), and use this "effective" time in calculation instead
  -- of the TimeStamp.

SELECT 'weatherhawk_func' AS function;
CREATE FUNCTION weatherhawk_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  bb_function_set_search_path
  AS $$
  BEGIN
  -- Function for WeatherHawk insert and update triggers
  --
  -- GPL_notice(`  --', `2017', `Jake Gordon, <jacob.gordon@duke.edu>')

  -- Wid, TimeStamp, and Wstation cannot change
  IF TG_OP = 'UPDATE' THEN
    cannot_change(`WEATHERHAWK', `Wid')
  END IF;

  -- If a new row was inserted, or if TimeStamp or YearlyRain were
  -- updated, are there any rows whose "effective" time is after NEW
  -- in the same year?  If so, the TimeStampRain value in the next row
  -- will need to be updated.

  IF NOT (TG_OP = 'UPDATE'
            AND (NEW.timestamprain <> OLD.timestamprain
                 OR
                 NEW.yearlyrain <> OLD.yearlyrain))
    THEN
    DECLARE
      new_eff_timestamp   weatherhawk.timestamp%TYPE
                          := (NEW.timestamp - '1 hour'::interval);

      next_wid weatherhawk.wid%TYPE;

    BEGIN
      SELECT INTO next_wid
                  weatherhawk.wid
        FROM weatherhawk
        WHERE weatherhawk.wstation = NEW.wstation
          AND (weatherhawk.timestamp - '1 hour'::interval)
              > new_eff_timestamp
          AND date_trunc('YEAR', (weatherhawk.timestamp - '1 hour'::interval))
              = date_trunc('YEAR', new_eff_timestamp)
        ORDER BY timestamp
        LIMIT 1;
      IF FOUND THEN
        -- Update the next row's TimeStampRain by invoking the BEFORE
        -- trigger.
        UPDATE weatherhawk
          SET wid = weatherhawk.wid
          WHERE wid = next_wid;
      END IF;

    END;

  END IF;

  RETURN NULL;
  END;
$$;

SELECT 'weatherhawk_before_func' AS function;
CREATE FUNCTION weatherhawk_before_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for WeatherHawk before trigger
  --
  -- GPL_notice(`  --', `2017', `Jake Gordon, <jacob.gordon@duke.edu>')
  --
  -- Calculate the TimeStampRain column.
  --

  DECLARE
    new_eff_timestamp   weatherhawk.timestamp%TYPE
                        := (NEW.timestamp - '1 hour'::interval);

    prev_yearlyrain     weatherhawk.yearlyrain%TYPE;

  BEGIN
    SELECT INTO prev_yearlyrain
                weatherhawk.yearlyrain
      FROM weatherhawk
      WHERE weatherhawk.wstation = NEW.wstation
        AND (weatherhawk.timestamp - '1 hour'::interval)
            < new_eff_timestamp
        AND DATE_TRUNC('YEAR', (weatherhawk.timestamp - '1 hour'::interval))
            = DATE_TRUNC('YEAR', new_eff_timestamp)
      ORDER BY timestamp DESC
      LIMIT 1;

    NEW.timestamprain := (NEW.yearlyrain -
                            COALESCE(prev_yearlyrain, 0));

  RETURN NEW;
  END;
$$;

SELECT 'weatherhawk_delete_func' AS function;
CREATE FUNCTION weatherhawk_delete_func ()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- Function for WeatherHawk delete trigger
  --
  -- GPL_notice(`  --', `2017', `Jake Gordon, <jacob.gordon@duke.edu>')
  --
  -- Update the TimeStampRain column for the row after the deleted
  -- one.

  DECLARE
    old_eff_timestamp   weatherhawk.timestamp%TYPE
                        := (OLD.timestamp - '1 hour'::interval);

    next_wid weatherhawk.wid%TYPE;

  BEGIN
    SELECT INTO next_wid
                weatherhawk.wid
      FROM weatherhawk
      WHERE weatherhawk.wstation = OLD.wstation
        AND (weatherhawk.timestamp - '1 hour'::interval)
            > old_eff_timestamp
        AND date_trunc('YEAR', (weatherhawk.timestamp - '1 hour'::interval))
            = date_trunc('YEAR', old_eff_timestamp)
      ORDER BY timestamp
      LIMIT 1;
    IF FOUND THEN
      -- Update the next row's TimeStampRain by invoking the BEFORE
      -- trigger.
      UPDATE weatherhawk
        SET wid = weatherhawk.wid
        WHERE wid = next_wid;
    END IF;

  RETURN NULL;
  END;
$$;

SELECT 'weatherhawk_trigger' AS trigger;
CREATE TRIGGER weatherhawk_trigger
  AFTER INSERT OR UPDATE
  ON weatherhawk FOR EACH ROW
  EXECUTE PROCEDURE weatherhawk_func();

SELECT 'weatherhawk_before_trigger' AS trigger;
CREATE TRIGGER weatherhawk_before_trigger
  BEFORE INSERT OR UPDATE
  ON weatherhawk FOR EACH ROW
  EXECUTE PROCEDURE weatherhawk_before_func();

SELECT 'weatherhawk_delete_trigger' AS trigger;
CREATE TRIGGER weatherhawk_delete_trigger
  AFTER DELETE
  ON weatherhawk FOR EACH ROW
  EXECUTE PROCEDURE weatherhawk_delete_func();
