dnl Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
dnl Copyright (C) 2009, Karl O. Pinc
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl

dnl m4 includes
include(`copyright.m4')dnl
include(`constants.m4')dnl
include(`macros.m4')dnl
dnl
dnl plpgsql fragment to find the prior raingauge measurement and the
dnl prior raingauge setup.
dnl
dnl Syntax: get_priors(this_wstation,
dnl                    this_daytime,
dnl                    this_wrid,
dnl                    error_string)
dnl
dnl Input:
dnl   this_wstation   The Wstation of the wreadings/raingauge to update.
dnl   this_daytime    The daytime of the wreadings/raingauge to update.
dnl   this_wrid       The WRid of the raingauge to update.
dnl   error_string    String used when raising exception if no prior
dnl                   RGSETUPS row is found.  Takes this_wrid as it's
dnl                   first substituion, this_daytime as it's second.
dnl
dnl Variables required:
dnl   init_daytime     Set to RGSETUPS.RGSdaytime of relevant setup.
dnl   init_estdaytime  Set to RGSETUPS.RGSestdaytime of relevant setup.
dnl   prior_daytime    Set to WREADINGS.WRdaytime of prior raingauge reading.
dnl   prior_estdaytime Set to WREADINGS.Estdaytime of prior raingauge reading.
dnl
dnl Side Effects:
dnl   Leaves FOUND set depending on whether a prior WREADINGS was found.
dnl
dnl Restrictions: 
changequote({,})
define({get_priors},{
      -- Find the prior RGSETUP.
      SELECT rgsetups.rgsdaytime, rgsetups.rgsestdaytime
        INTO init_daytime,        init_estdaytime
        FROM rgsetups
        WHERE rgsetups.wstation = this_wstation
              AND rgsetups.rgsdaytime < this_daytime
        ORDER BY rgsetups.rgsdaytime DESC
        LIMIT 1;

      IF NOT FOUND THEN
        RAISE EXCEPTION '$4'
                      , this_wrid
                      , this_daytime;
        RETURN NULL;
      END IF;

      -- Find the prior rain gauge measurement
      SELECT wreadings.wrdaytime, wreadings.estdaytime
        INTO prior_daytime,       prior_estdaytime
        FROM wreadings, raingauges
        WHERE wreadings.wstation = this_wstation
              AND wreadings.wrdaytime < this_daytime
              AND raingauges.wrid = wreadings.wrid
        ORDER BY wreadings.wrdaytime DESC
        LIMIT 1;
})
changequote(`,')dnl
dnl
dnl
dnl plpgsql fragment for updating RAINGAUGES RGspan and EstRGspan.
dnl
dnl Syntax: update_span(this_wstation,
dnl                     this_daytime,
dnl                     this_wrid,
dnl                     this_estdaytime)
dnl
dnl Input:
dnl   this_wstation   The Wstation of the wreadings/raingauge to update.
dnl   this_daytime    The daytime of the wreadings/raingauge to update.
dnl   this_wrid       The WRid of the raingauge to update.
dnl   this_estdaytime The Estdaytime of the wreadings/raingauge to udpate.
dnl
dnl Variables required:
dnl Restrictions: 
changequote({,})
define({update_span},{
    DECLARE
      init_daytime rgsetups.rgsdaytime%TYPE;
      init_estdaytime rgsetups.rgsestdaytime%TYPE;
      prior_daytime wreadings.wrdaytime%TYPE;
      prior_estdaytime wreadings.estdaytime%TYPE;
      -- Make this whole thing eaiser to read by having the macro
      -- use names instead of $ variables.
      -- (With luck it will optimize away.  :-P  )
      this_wstation CONSTANT wreadings.wstation%TYPE := $1;
      this_wrdaytime CONSTANT wreadings.wrdaytime%TYPE := $2;
      this_wrid CONSTANT wreadings.wrid%TYPE := $3;
      this_estdaytime CONSTANT wreadings.estdaytime%TYPE := $4;

    BEGIN
      get_priors(`this_wstation'
               , `this_daytime'
               , `this_wrid'
               , `WRid %: Cannot change the WRdaytime to % because there must be a prior RGSETUPS row for the Wstation')

      -- Do the update
      IF prior_daytime IS NOT NULL AND prior_daytime > init_daytime THEN
        -- Compute the span from the prior rain gauge measurement.
        UPDATE raingauges
          SET rgspan = this_daytime - prior_daytime,
              estrgspan = NEW.estdaytime OR prior_estdaytime
          WHERE raingauges.wrid = this_wrid;
      ELSE
        -- Computer the span from when the rain gauge was setup.
        UPDATE raingauges
          SET rgspan = this_daytime - init_daytime,
              estrgspan = NEW.estdaytime OR init_estdaytime
          WHERE raingauges.wrid = this_wrid;
      END IF;
    END;
})
changequote(`,')dnl

--  
-- WReadings
--
SELECT 'wreadings' AS table;

SELECT 'wreadings_update_func' AS function;
CREATE FUNCTION wreadings_update_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER -- so raingauges can be updated
  bb_function_set_search_path
  AS $$

  BEGIN
  -- Function for update triggers
  --
  -- GPL_notice(`  --', `2009', `Karl O. Pinc <kop@meme.com>')
  --
  -- Remarks:
  --  - Some potential duplicate, and one actual duplicate, work is done
  --    just to keep the code volume down.
  --  - Control does not flow off the bottom.  Does a RETURN rather
  --    than have a large nested IF.

  -- Cannot change the WRID.
  cannot_change(`WREADINGS', `WRid')


  -- Can't have a raingauge reading at the same time as a
  -- raingauge setup event.
  -- This only matters for update because at the time of insert
  -- there can't be a raingauges row.
  PERFORM 1
    FROM raingauges, rgsetups
    WHERE raingauges.wrid = NEW.wrid
          AND rgsetups.rgsdaytime = NEW.wrdaytime
          AND rgsetups.wstation = NEW.wstation
    LIMIT 1;
  IF FOUND THEN
    RAISE EXCEPTION 'WRid %: WRdaytime %: Wstation %: WREADINGS change rejected: Cannot have a raingauge reading at the same time as the raingauge is setup'
                  , NEW.wrid
                  , NEW.wrdaytime
                  , NEW.wstation;
    RETURN NULL;
  END IF;

  --
  -- The rest of the trigger only matters if there's a related
  -- RAINGAUGES row.
  --

  IF NOT EXISTS (SELECT 1
                   FROM raingauges
                   WHERE raingauges.wrid = NEW.wrid) THEN
    -- All done.
    RETURN NULL;
  END IF;

  -- Cannot change the weather station when there is a related RAINGAUGES
  -- row.
  IF NEW.wstation <> OLD.wstation THEN
    RAISE EXCEPTION 'WRid %: Wstation %: WREADINGS.Wstation cannot be changed when there is a related RAINGAUGES row'
                  , OLD.wrid
                  , NEW.wstation;
    RETURN NULL;
  END IF;


  IF NEW.wrdaytime <> OLD.wrdaytime THEN
    --
    -- Update the rgspan and estrgspan for the changed rain
    -- gauge measurement.
    --
    update_span(`NEW.wstation', `NEW.wrdaytime', `NEW.wrid', `NEW.estdaytime')

    --
    -- Update the rgspan and estrgspan for the rain gauge measurement
    -- following the old rain gauge measurement timestamp.
    --
    DECLARE
      old_wstation wreadings.wstation%TYPE;
      old_daytime wreadings.wrdaytime%TYPE;
      old_wrid wreadings.wrid%TYPE;
      old_estdaytime wreadings.estdaytime%TYPE;

    BEGIN
      SELECT wreadings.wstation
           , wreadings.wrdaytime
           , wreadings.wrid
           , wreadings.estdaytime
        INTO old_wstation
           , old_daytime
           , old_wrid
           , old_estdaytime
        FROM wreadings, raingauges
        WHERE wreadings.wstation = NEW.wstation
              AND wreadings.wrdaytime > OLD.wrdaytime
              AND raingauges.wrid = wreadings.wrid
        ORDER BY wreadings.wrdaytime
        LIMIT 1;

      IF FOUND THEN
        update_span(`old_wstation'
                  , `old_daytime'
                  , `old_wrid'
                  , `old_estdaytime')
      END IF;
    END;

    --
    -- Update the rgspan and estrgspan for the rain gauge measurement
    -- following the new rain gauge measurement timestamp.
    --
    --
    DECLARE
      afternew_wstation wreadings.wstation%TYPE;
      afternew_daytime wreadings.wrdaytime%TYPE;
      afternew_wrid wreadings.wrid%TYPE;
      afternew_estdaytime wreadings.estdaytime%TYPE;

    BEGIN
      SELECT wreadings.wstation
           , wreadings.wrdaytime
           , wreadings.wrid
           , wreadings.estdaytime
        INTO afternew_wstation
           , afternew_daytime
           , afternew_wrid
           , afternew_estdaytime
        FROM wreadings, raingauges
        WHERE wreadings.wstation = NEW.wstation
              AND wreadings.wrdaytime > NEW.wrdaytime
              AND raingauges.wrid = wreadings.wrid
        ORDER BY wreadings.wrdaytime
        LIMIT 1;

      IF FOUND THEN
        update_span(`afternew_wstation'
                  , `afternew_daytime'
                  , `afternew_wrid'
                  , `afternew_estdaytime')
      END IF;
    END;
  ELSE
    IF NEW.estdaytime <> OLD.estdaytime THEN
      DECLARE
        init_daytime rgsetups.rgsdaytime%TYPE;
        init_estdaytime rgsetups.rgsestdaytime%TYPE;
        prior_daytime wreadings.wrdaytime%TYPE;
        prior_estdaytime wreadings.estdaytime%TYPE;

        after_wstation wreadings.wstation%TYPE;
        after_daytime wreadings.wrdaytime%TYPE;
        after_wrid wreadings.wrid%TYPE;
        after_estdaytime wreadings.estdaytime%TYPE;

      BEGIN
        --
        -- Update the estrgspan for the changed rain gauge measurement.
        --
        get_priors(`NEW.wstation'
                 , `NEW.wrdaytime'
                 , `NEW.wrid'
                 , `WRid %: WRdaytime %: Found invalid WRdaytime, there must be a prior RGSETUPS row for the Wstation')

         IF prior_daytime IS NOT NULL AND prior_daytime < init_daytime THEN
           -- Calculate the est flag based on the raingauge's initialization.
           UPDATE raingauges
             SET estrgspan = NEW.estdaytime OR init_estdaytime
             WHERE raingauges.wrid = NEW.wrid;
         ELSE
           -- Calculate the est flag based on the prior raingauge measurement.
           UPDATE raingauges
             SET estrgspan = NEW.estdaytime OR prior_estdaytime
             WHERE raingagues.wrid = NEW.wrid;
         END IF;

        --
        -- Update the estrgspan for the rain gauge measurement
        -- following the changed rain gauge measurement.
        --

        -- Find the raingauge measurement following the one changed.
        SELECT wreadings.wstation
             , wreadings.wrdaytime
             , wreadings.wrid
             , wreadings.estdaytime
          INTO after_wstation
             , after_daytime
             , after_wrid
             , after_estdaytime
          FROM wreadings, raingauges
          WHERE wreadings.wstation = NEW.wstation
              AND wreadings.wrdaytime > NEW.wrdaytime
                AND raingauges.wrid = wreadings.wrid
          ORDER BY wreadings.wrdaytime
          LIMIT 1;

        IF FOUND THEN
          -- Duplicate work here, we already know that the prior row
          -- is the NEW row.
          get_priors(`NEW.wstation'
                   , `NEW.wrdaytime'
                   , `NEW.wrid'
                   , `WRid %: WRdaytime %: Found invalid WRdaytime, there must be a prior RGSETUPS row for the Wstation')

          IF prior_daytime IS NOT NULL AND prior_daytime < init_daytime THEN
            -- Calculate the est flag based on the raingauge's initialization.
            UPDATE raingauges
              SET estrgspan = NEW.estdaytime OR init_estdaytime
              WHERE raingauges.wrid = after_wrid;
          ELSE
            -- Calculate the est flag based on the prior
            -- raingauge measurement.
            UPDATE raingauges
              SET estrgspan = NEW.estdaytime OR after_estdaytime
              WHERE raingagues.wrid = after_wrid;
          END IF;
        END IF;
      END;
    END IF;
  END IF;

  RETURN NULL;
  END;
$$;

-- We only need an update trigger.

SELECT 'wreadings_update_trigger' AS trigger;
CREATE TRIGGER wreadings_update_trigger
  AFTER UPDATE
  ON wreadings FOR EACH ROW
  EXECUTE PROCEDURE wreadings_update_func();
