dnl Copyright (C) 2016 Jake Gordon, <jacob.gordon@duke.edu>
dnl Copyright (C) 2014 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Functions and triggers to allow insert, update, and delete into
dnl the actor_actees view.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

dnl plpgsql fragment to validate actor or actee grp column
dnl
dnl Syntax: _validate_grp(role)
dnl
dnl role     `Actor' or `Actee'
dnl
dnl Must be used in insert or update trigger.
dnl
changequote({,})
define({_validate_grp},{

  IF NEW.$1_grp IS NOT NULL AND NEW.$1 IS NOT NULL THEN
    -- (Written so that MEMBERS row need not exist.)
    PERFORM 1
      FROM members
      WHERE members.date = NEW.date
            AND members.sname = NEW.$1
            AND members.grp <> NEW.$1_grp
      LIMIT 1;

    IF FOUND THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of ACTOR_ACTEES'
          , DETAIL = 'Value (Actor) = ('
                     || textualize(`NEW.actor')
                     || '): Value (Actee) = ('
                     || textualize(`NEW.actee')
                     || '): Value (Date) = (' || textualize(`NEW.date')
                     || '): Value (Act) = (' || textualize(`NEW.act')
                     || '): Value ($1_Grp) = ('
                     || textualize(`NEW.$1_grp')
                     || '): $1_Grp value is not NULL and does not match MEMBERS.Grp'
         , HINT = 'Leave the $1_Grp value NULL';
      RETURN NULL;
    END IF;
  END IF;
})dnl
changequote(`,')dnl


SELECT 'actor_actees_insert_func()';
CREATE OR REPLACE FUNCTION actor_actees_insert_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of insert into actor_actees.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    -- Generate default values.

    IF NEW.actorid IS NULL AND NEW.actor IS NOT NULL THEN
      NEW.actorid := nextval('parts_partid_seq');
    END IF;

    IF NEW.acteeid IS NULL AND NEW.actee IS NOT NULL THEN
      NEW.acteeid := nextval('parts_partid_seq');
    END IF;

    IF NEW.iid IS NULL THEN
      NEW.iid := nextval('interact_data_iid_seq');
    END IF;

    -- Validate data.

    _validate_grp(`Actor')
    _validate_grp(`Actee')

    INSERT INTO interact_data (iid
                             , sid
                             , act
                             , date
                             , start
                             , stop
                             , observer
                             , handwritten)
      VALUES ( NEW.iid
             , NEW.sid
             , NEW.act
             , NEW.date
             , NEW.start
             , NEW.stop
             , NEW.observer
             , NEW.handwritten);

    IF NEW.actorid IS NOT NULL THEN
      INSERT INTO parts (partid, iid, role, sname)
        VALUES (NEW.actorid, NEW.iid, 'bb_actor', NEW.actor);
    END IF;

    IF NEW.acteeid IS NOT NULL THEN
      INSERT INTO parts (partid, iid, role, sname)
        VALUES (NEW.acteeid, NEW.iid, 'bb_actee', NEW.actee);
    END IF;

  RETURN NULL;
  END;
$$;

SELECT 'actor_actees_update_func()';
CREATE OR REPLACE FUNCTION actor_actees_update_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of update to actor_actees.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    -- Create defaults
    -- Allow creation of new PARTS rows.

    IF OLD.actorid IS NULL
       AND NEW.actorid IS NULL
       AND NEW.actor IS NOT NULL THEN
      NEW.actorid := nextval('parts_partid_seq');
    END IF;

    IF OLD.acteeid IS NULL
       AND NEW.acteeid IS NULL
       AND NEW.actee IS NOT NULL THEN
      NEW.acteeid := nextval('parts_partid_seq');
    END IF;

    -- Validate data.

    _validate_grp(`Actor')
    _validate_grp(`Actee')

    -- Update the db.

    UPDATE interact_data
      SET iid = NEW.iid
        , sid = NEW.sid
        , act = NEW.act
        , date = NEW.date
        , start = NEW.start
        , stop = NEW.stop
        , observer = NEW.observer
        , handwritten = NEW.handwritten
      WHERE iid = OLD.iid;

    IF OLD.actorid IS NOT NULL THEN
      -- There's an old PARTS row for the actor.
      IF NEW.actor IS NULL
         AND NEW.actorid IS NOT NULL
         AND NEW.actorid <> OLD.actorid THEN
        RAISE EXCEPTION data_exception USING
              MESSAGE = 'Error on UPDATE of ACTOR_ACTEES'
            , DETAIL = 'Value (Actor) = ('
                       || textualize(`NEW.actor')
                       || '): Value (Actee) = ('
                       || textualize(`NEW.actee')
                       || '): Value (Date) = (' || textualize(`NEW.date')
                       || '): Value (Act) = (' || textualize(`NEW.act')
                       || '): Value (Actorid) = ('
                       || textualize(`NEW.actorid')
                       || '): Value (Acteeid) = ('
                       || textualize(`NEW.acteeid')
                       || '): The Actorid can be NULL but otherwise cannot '
                       || 'change when deleting a PARTS row by setting the '
                       || 'Actor to NULL';
        RETURN NULL;
      END IF;

      IF NEW.actor IS NULL THEN
        -- Delete the old PARTS row for the actor.
        DELETE FROM parts
          WHERE partid = OLD.actorid;
      ELSE
        UPDATE parts
          SET partid = NEW.actorid
            , iid = NEW.iid
            , sname = NEW.actor
          WHERE partid = OLD.actorid;
      END IF;

    ELSIF NEW.actorid IS NOT NULL THEN
      -- There's a new PARTS row for the actor.
      INSERT INTO parts (partid, iid, role, sname)
        VALUES (NEW.actorid, NEW.iid, 'bb_actor', NEW.actor);
    END IF;

    IF OLD.acteeid IS NOT NULL THEN
      -- There's an old PARTS row for the actee.
      IF NEW.actee IS NULL
         AND NEW.acteeid IS NOT NULL
         AND NEW.acteeid <> OLD.acteeid THEN
        RAISE EXCEPTION data_exception USING
              MESSAGE = 'Error on UPDATE of ACTOR_ACTEES'
            , DETAIL = 'Value (Actor) = ('
                       || textualize(`NEW.actor')
                       || '): Value (Actee) = ('
                       || textualize(`NEW.actee')
                       || '): Value (Date) = (' || textualize(`NEW.date')
                       || '): Value (Act) = (' || textualize(`NEW.act')
                       || '): Value (Actorid) = ('
                       || textualize(`NEW.actorid')
                       || '): Value (Acteeid) = ('
                       || textualize(`NEW.acteeid')
                       || '): The Acteeid can be NULL but otherwise cannot '
                       || 'change when deleting a PARTS row by setting the '
                       || 'Actee to NULL';
        RETURN NULL;
      END IF;

      IF NEW.acteeid IS NULL THEN
        -- Delete the old PARTS row for the actee.
        DELETE FROM parts
          WHERE partid = OLD.acteeid;
      ELSE
        UPDATE parts
          SET partid = NEW.acteeid
            , iid = NEW.iid
            , sname = NEW.actee
          WHERE partid = OLD.acteeid;
      END IF;

    ELSIF NEW.acteeid IS NOT NULL THEN
      -- There's a new PARTS row for the actee.
      INSERT INTO parts (partid, iid, role, sname)
        VALUES (NEW.acteeid, NEW.iid, 'bb_actee', NEW.actee);
    END IF;

  RETURN NULL;
  END;
$$;


SELECT 'actor_actees_delete_func()';
CREATE OR REPLACE FUNCTION actor_actees_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of delete from actor_actees.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    DELETE FROM parts
      WHERE iid = OLD.iid;

    DELETE FROM interact_data
      WHERE iid = OLD.iid;

  RETURN NULL;
  END;
$$;


-- actor_actees
SELECT 'actor_actees_insert_trigger' AS trigger;
CREATE TRIGGER actor_actees_insert_trigger
  INSTEAD OF INSERT
  ON actor_actees FOR EACH ROW
  EXECUTE PROCEDURE actor_actees_insert_func();

SELECT 'actor_actees_update_trigger' AS trigger;
CREATE TRIGGER actor_actees_update_trigger
  INSTEAD OF UPDATE
  ON actor_actees FOR EACH ROW
  EXECUTE PROCEDURE actor_actees_update_func();

SELECT 'actor_actees_delete_trigger' AS trigger;
CREATE TRIGGER actor_actees_delete_trigger
  INSTEAD OF DELETE
  ON actor_actees FOR EACH ROW
  EXECUTE PROCEDURE actor_actees_delete_func();
