dnl Copyright (C) 2014 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Functions and triggers to allow insert, update, and delete into
dnl the points view.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

SELECT 'points_insert_func()';
CREATE OR REPLACE FUNCTION points_insert_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of insert into the points view.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: points_insert_func(viewname)
  --
  -- Input:
  --   viewname   Name of view, for error messages.
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    IF NEW.ptimespm IS NOT NULL
       AND NEW.ptimespm <> spm(NEW.ptime) THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of ' || TG_ARGV[0]
          , DETAIL = 'Value (Pntid) = ('
                     || textualize(`NEW.pntid')
                     || '): Value (Sid) = ('
                     || textualize(`NEW.sid')
                     || '): Value (Activity) = (' || textualize(`NEW.activity')
                     || '): Value (Posture) = (' || textualize(`NEW.posture')
                     || '): Value (Foodcode) = ('
                     || textualize(`NEW.foodcode')
                     || '): Value (PTime) = (' || textualize(`NEW.ptime')
                     || '): Ptimespm is not NULL and Ptimespm is not spm(PTime)'
         , HINT = 'Leave the Ptimespm value NULL';
      RETURN NULL;
    END IF;

    IF NEW.pntid IS NULL THEN
      NEW.pntid := nextval('point_data_pntid_seq');
    END IF;

    INSERT INTO point_data (pntid, sid, activity, posture, foodcode, ptime)
      values (NEW.pntid
            , NEW.sid
            , NEW.activity
            , NEW.posture
            , NEW.foodcode
            , NEW.ptime);

  RETURN NULL;
  END;
$$;

SELECT 'points_update_func()';
CREATE OR REPLACE FUNCTION points_update_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of update of the points view.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Syntax: points_update_func(viewname)
  --
  -- Input:
  --   viewname   Name of view, for error messages.
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    IF NEW.ptimespm IS NOT NULL
       AND NEW.ptimespm <> OLD.ptimespm
       AND NEW.ptimespm <> spm(NEW.ptime) THEN
      RAISE EXCEPTION integrity_constraint_violation USING
            MESSAGE = 'Error on ' || TG_OP || ' of ' || TG_ARGV[0]
          , DETAIL = 'Value (Pntid) = ('
                     || textualize(`NEW.pntid')
                     || '): Value (Sid) = ('
                     || textualize(`NEW.sid')
                     || '): Value (Activity) = (' || textualize(`NEW.activity')
                     || '): Value (Posture) = (' || textualize(`NEW.posture')
                     || '): Value (Foodcode) = ('
                     || textualize(`NEW.foodcode')
                     || '): Value (PTime) = (' || textualize(`NEW.ptime')
                     || '): Ptimespm is not NULL and Ptimespm is not spm(PTime)'
         , HINT = 'Leave the Ptimespm value NULL';
      RETURN NULL;
    END IF;

    UPDATE point_data
      SET pntid = NEW.pntid
        , sid = NEW.sid
        , activity = NEW.activity
        , posture = NEW.posture
        , foodcode = NEW.foodcode
        , ptime = NEW.ptime
      WHERE point_data.pntid = OLD.pntid;

  RETURN NULL;
  END;
$$;

SELECT 'points_delete_func()';
CREATE OR REPLACE FUNCTION points_delete_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of delete of the points view.
  --
  -- GPL_notice(`  --', `2014', `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Remarks:
  --
  -- Bugs:

  DECLARE

  BEGIN

    DELETE FROM point_data
      WHERE point_data.pntid = OLD.pntid;

  RETURN NULL;
  END;
$$;


SELECT 'points_insert_trigger' AS trigger;
CREATE TRIGGER points_insert_trigger
  INSTEAD OF INSERT
  ON points FOR EACH ROW
  EXECUTE PROCEDURE points_insert_func('points');

SELECT 'points_update_trigger' AS trigger;
CREATE TRIGGER points_update_trigger
  INSTEAD OF UPDATE
  ON points FOR EACH ROW
  EXECUTE PROCEDURE points_update_func('points');

SELECT 'points_delete_trigger' AS trigger;
CREATE TRIGGER points_delete_trigger
  INSTEAD OF DELETE
  ON points FOR EACH ROW
  EXECUTE PROCEDURE points_delete_func();


SELECT 'points_sorted_insert_trigger' AS trigger;
CREATE TRIGGER points_sorted_insert_trigger
  INSTEAD OF INSERT
  ON points_sorted FOR EACH ROW
  EXECUTE PROCEDURE points_insert_func('points_sorted');

SELECT 'points_sorted_update_trigger' AS trigger;
CREATE TRIGGER points_sorted_update_trigger
  INSTEAD OF UPDATE
  ON points_sorted FOR EACH ROW
  EXECUTE PROCEDURE points_update_func('points_sorted');

SELECT 'points_sorted_delete_trigger' AS trigger;
CREATE TRIGGER points_sorted_delete_trigger
  INSTEAD OF DELETE
  ON points_sorted FOR EACH ROW
  EXECUTE PROCEDURE points_delete_func();
