dnl Copyright (C) 2017 The Meme Factory, Inc.,  http://www.meme.com/
dnl Copyright (C) 2012-2014 The Meme Factory, Inc., http://www.meme.com/
dnl
dnl    This file is part of Babase.
dnl
dnl    Babase is free software; you can redistribute it and/or modify
dnl    it under the terms of the GNU General Public License as published by
dnl    the Free Software Foundation; either version 3 of the License, or
dnl    (at your option) any later version.
dnl
dnl    This program is distributed in the hope that it will be useful,
dnl    but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl    GNU General Public License for more details.
dnl
dnl    You should have received a copy of the GNU General Public License
dnl    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
dnl
dnl Functions and triggers to allow insert, update, and delete into
dnl the swerb_upload view.
dnl
dnl Karl O. Pinc <kop@meme.com>
dnl
dnl
dnl m4 includes
include(`copyright.m4')
include(`constants.m4')
include(`macros.m4')
dnl

SELECT 'swerb_upload_insert_func()';
CREATE OR REPLACE FUNCTION swerb_upload_insert_func()
  RETURNS trigger
  LANGUAGE plpgsql
  SECURITY DEFINER
  bb_function_set_search_path
  AS $$
  -- For triggers instead of insert into the swerb_upload view.
  --
  -- Copyright (C) 2017 The Meme Factory, Inc.,  http://www.meme.com/
  -- GPL_notice(`  --', `2012, 2013',
                `The Meme Factory, Inc.  http://www.meme.com/')
  --
  -- Remarks:
  --  Columns are:
  -- Header        Which button was pushed.
  --               Always "Waypoint". (bb_swerb_waypoint)  Raises an
  --               exception if not.
  -- Name          12 chars with user entered data, the meat of the data.
  --               See below and the Monitoring Guide.
  -- Description   The timestamp.  DD-mmm-yy space HH:MM
  --               mmm is 3 char month abbreviation in upper case.
  -- Type          More about what button was pushed.
  -- Position      Encoding position, always "37 M" space "X coordinate"
  --               space "Y coordinate".  XY units are meters.  Always
  --               positive.
  -- Altitude      In meters, except when it's in feet.
  --               If ends in space "ft" then converted to meters.
  -- Depth         Ignored.
  -- Proximity     Ignored.
  -- Display Mode  Ignored.
  -- Color         Ignored. (color you want on map at that point)
  -- Symbol        Ignored. (to put a symbol on the map at that point)
  -- Facility      Ignored.
  -- City          Ignored.
  -- State         Ignored.
  -- Country       Ignored.
-- Pdop
-- Accuracy
-- Quad
-- Date  used when manual quad record instead of Description , only one or the other is alloed, not both.
-- Timeest (ditto) swerb_bes.btimeest/swerb_bes.etimeest
-- Source swerb_bes.bsource/swerb_bes.esource
-- Is_Effort Into swerb_bes.is_effort
-- Notes  Into swerb_bes.notes
-- Lone_Animal  When non-NULL, the row is a lone animal row.
--              The value must match the animal's sex.
-- Secondary_AD  When true the begin/end rows represent a secondary ascent/
--               descent which goes into it's own bout of observation.
  --
  -- Remarks:
  --  One of the goals here is to validate all the data that's discarded.
  --
  -- Bugs:
  --   Too many duplicated is_effort validation checks.

  dnl Constants
  dnl Key value to use in our state table.
  define(`BB_SU_OUR_SKEY', `1')dnl
  define(`BB_SU_SUBGROUPC', `S')dnl -- Char that follows group letter to  dnl
  dnl                               -- indicate subgroup.                 dnl
  define(`BB_SU_BEGIN_EVENT', `B')dnl
  define(`BB_SU_END_EVENT', `E')dnl
  define(`BB_SU_OTHERG_EVENT', `G')dnl
  define(`BB_SU_OTHERG_GROUP', `O')dnl -- Group letter always O when      dnl
  dnl                                  -- other group.                    dnl
  define(`BB_SU_LONEANIMAL_EVENT_DATA', `M')dnl
  define(`BB_SU_LONEANIMAL_EVENTC', `LM')dnl
  define(`BB_SU_LONEANIMAL_GROUP', `L')dnl -- Group letter always L when  dnl
  dnl                                      -- lone animal.                dnl
  define(`BB_SU_WATER_EVENT', `W')dnl
  define(`BB_SU_TICK_EVENT', `H')dnl -- regular "interval" swerb waypoint dnl
  dnl                                -- Note, this is not in the uploaded
  dnl                                -- data.  It's strictly used internally.

  define(`BB_SU_G_LEN', `3')dnl  -- Length of group code                  dnl
  define(`BB_SU_LA_LEN', `3')dnl -- Length of the lone animal sname       dnl
  define(`BB_SU_TIME_PATTERN', `[[:digit:]]{bb_swerb_time_len}')dnl
  dnl                                               -- times are 4 digits dnl

  DECLARE
    this_lineno INT;

    time_as_str TEXT;
    this_x INT;
    this_y INT;

    this_time swerb_data.time%TYPE;
    this_date swerb_departs_data.date%TYPE;
    -- All the altitude, accuracy, and pdop values
    -- share the same datatype throughout the db.  Use SWERB_DATA
    -- as a universal template.
    this_altitude swerb_data.altitude%TYPE;
    this_accuracy swerb_data.accuracy%TYPE;
    this_pdop swerb_data.pdop%TYPE;
    -- Garmincode is TEXT on SWERB_DEPARTS_GPS, although limited
    -- to bb_garmincode_len where it appears in other tables.
    this_garmincode TEXT;
    this_quad swerb_data.quad%TYPE;

    this_timeest BOOLEAN;
    timeest_supplied BOOLEAN;

    this_is_effort BOOLEAN;
    is_effort_supplied BOOLEAN;

    this_new_lone_animal TEXT;

    this_secondary_ad BOOLEAN;
    this_in_secondary_ad BOOLEAN;

    this_source swerb_bes.bsource%TYPE;  -- also must be swerb_bes.esource%TYPE!
    source_supplied BOOLEAN;

    this_notes swerb_bes.notes%TYPE;
    notes_supplied BOOLEAN;

    this_gps_label gps_units.label%TYPE;
    this_gps_gps gps_units.gps%TYPE;

    firstc CHAR(1);
    rows_found INT; -- scratch var for sql result count

    this_beid swerb_bes.beid%TYPE;
  BEGIN

    -- Sanity checks.

    IF NEW.header <> 'bb_waypoint_button' THEN
      RAISE EXCEPTION 'Header %: must be ''bb_waypoint_button'''
                    , NEW.header;
      RETURN NULL;
    END IF;


    -- Create a temporary table for keeping track of state. *sigh*
    CREATE TEMP TABLE IF NOT EXISTS our_state
      (key INT           -- arbitrary key, there's only one row
     , lineno INT        -- current line number in uploaded file
     , label CHAR(1)     -- current gps label
     , begincnt INT      -- of N contiguous begin rows, the last fully processed
     , endcnt INT        -- of N contiguous end rows, the last fully processed
     , in_bout BOOLEAN   -- Whether we're in a bout of observation
     , secondary_begincnt INT -- of N contiguous secondary begin rows, the last
                              -- processed
     , secondary_endcnt INT   -- of N contiguous secondary end rows, the last
                              -- processed
     , in_secondary_ad BOOLEAN  -- Whether we're in a secondary AD bout
                                -- of observation.
     , last_subgroup BOOLEAN  -- Whether the last (non-other group) observation
                              -- was of a subgroup.
     , just_drank BOOLEAN  -- Whether there's been any drinking events
                           -- since the last subgroup.
     , last_beid INT)     -- The beid most recently created
                         -- (secondary A/D excepted)
      ON COMMIT DROP;

    -- Create a temporary table for remembering which drinking events
    -- have happened when the immediately preceeding (non-drinking,
    -- non-other-group) row was a subgroup.
    CREATE TEMP TABLE IF NOT EXISTS drinks
      (swid INT)          -- Same datatype as SWERB_DATA.Swid
      ON COMMIT DROP;

    -- Create a temporary table to map data entry event codes to
    -- database SWERB_DATA.Event codes.
    CREATE TEMP TABLE IF NOT EXISTS event_map
      (su_event VARCHAR(2), db_event TEXT)
      ON COMMIT DROP;
    PERFORM 1
      FROM event_map
      WHERE su_event = 'BB_SU_BEGIN_EVENT'
      LIMIT 1;
    IF NOT FOUND THEN
      -- Populate our lookup table of uploaded event codes (2nd char
      -- on the line) to the SWERB_DATA.Event codes.
      -- (We also do some pattern matching below in addition to this table.)
      INSERT INTO event_map (su_event, db_event)
        VALUES ('BB_SU_BEGIN_EVENT', 'bb_swerb_begin');
      INSERT INTO event_map (su_event, db_event)
        VALUES('BB_SU_END_EVENT', 'bb_swerb_end');
      INSERT INTO event_map (su_event, db_event)
        VALUES('BB_SU_OTHERG_EVENT', 'bb_swerb_ogrp');
      INSERT INTO event_map (su_event, db_event)
        VALUES('BB_SU_LONEANIMAL_EVENTC', 'bb_swerb_ogrp');
      INSERT INTO event_map (su_event, db_event)
        VALUES('BB_SU_WATER_EVENT', 'bb_swerb_water');
    END IF;

    -- Get the current line number.
    SELECT our_state.lineno
      INTO this_lineno
      FROM our_state
      WHERE our_state.key = BB_SU_OUR_SKEY;

    IF FOUND THEN
      -- Bump the line number.
      this_lineno := this_lineno + 1;
      UPDATE our_state
        SET lineno = this_lineno
        WHERE our_state.key = BB_SU_OUR_SKEY;
    ELSE
      -- Haven't seen any lines.
      this_lineno := 1;
      INSERT INTO our_state (key, lineno, in_bout)
             VALUES (BB_SU_OUR_SKEY, this_lineno, FALSE);
    END IF;

    --
    -- Start by getting all the GPS unit-ish stuff.
    -- We're going to need it for all the rows anyway.
    --

    --
    -- Date and time.
    -- We need the date to find the group using the one-letter-code.
    --

    IF NEW.date IS NOT NULL AND NEW.date <> '' THEN
      IF NEW.description <> '' AND NEW.description IS NOT NULL THEN
        RAISE EXCEPTION 'Name %: Date %: Description %: Both a date and a description cannot be supplied'
                      , NEW.name
                      , NEW.date
                      , NEW.description;
        RETURN NULL;
      END IF;

      -- Use the supplied date column
      this_date := NEW.date;
      this_time := NULL;
    ELSE
      IF NEW.description <> '' THEN
        -- Get date and time from the gps unit.
        IF NEW.description
             !~ '^[[:digit:]]{4}/[01]?[[:digit:]]/[0123]?[[:digit:]] [012]?[[:digit:]]:[012345]?[[:digit:]](:[012345]?[[:digit:]])?$'
           THEN
          -- We don't want to put restrictions on parsing the time stamp,
          -- but we also don't want to have extra stuff in there that's
          -- sliently ignored.
          RAISE EXCEPTION 'Description %: Invalid timestamp'
                        , NEW.description;
          RETURN NULL;
        END IF;

  -- bug  date may be omitted as well when the rows represent data taken manually.
        this_date := NEW.description;
        time_as_str := split_part(NEW.description, ' ', 2);
        IF time_as_str = '' THEN
          -- No time was supplied.
          this_time := NULL;
        ELSE
          this_time := time_as_str::TIME;
        END IF;
      ELSE
        -- Nothing from the gps unit.
        -- (Hopefully it's a median ascent/descent time that the
        -- data manager had to enter.)
        this_date := NULL;
        time_as_str := '';
        this_time := NULL;
      END IF;
    END IF;

    IF NEW.position <> '' THEN
      -- Get the X and Y coordinates from the gps unit.
      IF NEW.position
         !~ '^37 M [[:digit:]]+([.][[:digit:]]+)? [[:digit:]]+([.][[:digit:]]+)?$'
         THEN
        RAISE EXCEPTION 'Position %: Invalid coordinates'
                      , NEW.position;
        RETURN NULL;
      END IF;

      this_x := split_part(NEW.position, ' ', 3);
      this_y := split_part(NEW.position, ' ', 4);
    ELSE
      this_x := NULL;
      this_y := NULL;
    END IF;

    -- Quad
    IF NEW.quad = '' THEN
      this_quad := NULL;
    ELSE
      this_quad := NEW.quad;
    END IF;

    IF NEW.altitude <> '' THEN
      -- Get the altitude.
      IF NEW.altitude !~ ' ft$' THEN
        -- The altitude is in meters.
        IF UPPER(SUBSTRING(NEW.altitude FROM CHAR_LENGTH(NEW.altitude) - 1))
           = ' M' THEN
          -- Get rid of the trailing ' m', meaning meters.
          this_altitude
            := SUBSTRING(NEW.altitude FROM 1 FOR CHAR_LENGTH(NEW.altitude) - 2);
        ELSE
          this_altitude := NEW.altitude;
        END IF;
      ELSE
        -- The altitude is in feet.
        this_altitude := split_part(NEW.altitude, ' ', 1)::NUMERIC
                         * bb_feet_to_meters;
      END IF;
    ELSE
      this_altitude := NULL;
    END IF;

    -- Accuracy
    IF NEW.accuracy = '' THEN
      this_accuracy := NULL;
    ELSE
      this_accuracy := NEW.accuracy;
    END IF;

    -- Pdop
    IF NEW.pdop = '' THEN
      this_pdop := NULL;
    ELSE
      this_pdop := NEW.pdop;
    END IF;

    -- Garmincode
    IF NEW.name = '' THEN
      this_garmincode := NULL;
    ELSE
      -- Make sure we don't truncate the garmincode by accident.
      -- (But make an exception for the garmincode making up the first
      -- departure line because it can be of infinite length.)
      IF CHAR_LENGTH(NEW.name) > bb_garmincode_len
         AND this_lineno > 2  THEN -- past the departure lines
        RAISE EXCEPTION 'Maxlen %: Name %: is too long to fit in a Garmincode column'
                      , bb_garmincode_len
                      , NEW.name;
        RETURN NULL;
      END IF;

      IF NEW.description = '' OR NEW.description IS NULL THEN
        -- The line was manually entered by the data manager.
        -- Store only "official" garmincodes entered in the field.
        -- (There is a rule in the db about this, which we need to honor.)
        this_garmincode := NULL;
      ELSE
        this_garmincode := NEW.name;
      END IF;
    END IF;

    -- Timeest
    IF NEW.timeest = '' OR NEW.timeest IS NULL THEN
      this_timeest := FALSE;
      timeest_supplied := FALSE;
    ELSE
      this_timeest := NEW.timeest;
      timeest_supplied := TRUE;
    END IF;

    -- Is_effort
    IF NEW.is_effort = '' OR NEW.is_effort IS NULL THEN
      this_is_effort := FALSE;
      is_effort_supplied := FALSE;
    ELSE
      this_is_effort := NEW.is_effort;
      is_effort_supplied := TRUE;
    END IF;

    -- Source
    IF NEW.source = '' OR NEW.source IS NULL THEN
      this_source := 'bb_gps_time_source';
      source_supplied := FALSE;
    ELSE
      this_source := NEW.source;
      source_supplied := TRUE;
    END IF;

    -- Notes
    IF NEW.notes = '' OR NEW.notes IS NULL THEN
      this_notes := NULL;
      notes_supplied := FALSE;
    ELSE
      this_notes := NEW.notes;
      notes_supplied := TRUE;
    END IF;

    -- Lone animal
    -- Convert nulls to '' for easy testing.
    IF NEW.lone_animal IS NULL THEN
      this_new_lone_animal := '';
    ELSE
      this_new_lone_animal := NEW.lone_animal;
      IF this_new_lone_animal <> ''
         AND this_new_lone_animal <> 'bb_male'
         AND this_new_lone_animal <> 'bb_female'
         AND this_new_lone_animal <> 'bb_unk_sex' THEN
        RAISE EXCEPTION 'Lone_Animal %: Must be one of empty, NULL, bb_male, bb_female, or bb_unk_sex'
                      , this_new_lone_animal;
        RETURN NULL;
      END IF;
    END IF;

    -- Secondary_AD
    -- Convert nulls to false for easy testing.
    IF NEW.secondary_ad IS NULL THEN
      this_secondary_ad := FALSE;
    ELSE
      this_secondary_ad := NEW.secondary_ad;
    END IF;


    --
    -- Parse the Name, what the operator enters for the waypoint.
    --

    -- The first character is either the event, or is the group
    -- code (groups.one_letter_code) used in swerb records.
    firstc := SUBSTRING(NEW.name FROM 1 FOR 1);

    IF this_lineno = 1 OR this_lineno = 2 THEN
      -- First line has to be a departure, listing obsevers.

      IF firstc <> 'bb_swerb_departure_char' THEN
        -- First line, but not a departure row.
        RAISE EXCEPTION 'First char %: Name %: The first 2 data lines must have a bb_swerb_departure_char as the first character of the Name value'
                      , firstc
                      , NEW.name;
        RETURN NULL;
      END IF;

      IF this_new_lone_animal <> '' THEN
        RAISE EXCEPTION 'Name %: Lone_Animal %: The first 2 data lines must record departures and cannot represent lone animal observations'
                      , NEW.name
                      , NEW.lone_animal;
      END IF;

      IF this_secondary_ad THEN
        RAISE EXCEPTION 'Name %: Secondary_AD %: The first 2 data lines must record departures and cannot represent secondary ascent/descents'
                      , NEW.name
                      , NEW.secondary_ad;
      END IF;

      IF this_quad IS NOT NULL THEN
        RAISE EXCEPTION 'Name %: Quad %: Departure lines cannot have quad data'
                      , NEW.name
                      , this_quad;
        RETURN NULL;
      END IF;

      IF timeest_supplied THEN
        RAISE EXCEPTION 'Name %: Timeest %: Departure lines cannot have an estimated time'
                      , NEW.name
                      , NEW.timeest
          USING HINT = 'Supply NULL or the empty string in the Timeest column';
        RETURN NULL;
      END IF;

      IF is_effort_supplied THEN
        RAISE EXCEPTION 'Name %: Is_effort %: Departure lines cannot have an Is_effort value'
                      , NEW.name
                      , NEW.is_effort
          USING HINT = 'Supply NULL or the empty string in the Is_effort column';
        RETURN NULL;
      END IF;

      IF source_supplied THEN
        RAISE EXCEPTION 'Name %: Source %: Departure lines cannot have an Source value'
                      , NEW.name
                      , NEW.source
          USING HINT = 'Supply NULL or the empty string in the Source column';
        RETURN NULL;
      END IF;

      IF notes_supplied THEN
        RAISE EXCEPTION 'Name %: Notes %: Departure lines cannot have an Notes value'
                      , NEW.name
                      , NEW.notes
          USING HINT = 'Supply NULL or the empty string in the Notes column';
        RETURN NULL;
      END IF;

      -- D  Departure from camp

      -- Old swerb data has one D row, with single characters
      -- The first N letters are the observers and the last
      -- letter the driver.
      -- We can't process such old swerb datasets.
      DECLARE

        startc INT; -- generic next char pointer in NEW.name
                    -- when doing observers it's the first char of observer code
        lastc INT;  -- last char in NEW.name of observer code
        maxc INT;   -- last char in NEW.name
        this_role swerb_observers.role%TYPE;
        this_observer TEXT;
        other_observer TEXT;

      BEGIN

        -- There can be 2 D rows to start.
        -- DD means driver -- 2nd line, D means observer - first line.

        IF this_lineno = 1 THEN
          -- This is the line for the observers.
          startc := 2;  

          -- Got a departure date and time.  Insert it.
          INSERT INTO swerb_departs_data (date, time)
            VALUES(this_date, this_time);
        ELSE
          -- Second line.
          -- Second line must be a driver row.
          startc := 3;  -- skip 2nd D

          IF SUBSTRING(NEW.name FROM 1 FOR 2)
             <> ('bb_swerb_departure_char' || 'bb_swerb_departure_char') THEN
            RAISE EXCEPTION 'Name %: The second departure line must have bb_swerb_departure_charbb_swerb_departure_char as the first 2 characters of the Name value'
                          , NEW.name;
            RETURN NULL;
          END IF;
        END IF;

        -- The next char is the GPS unit.
        this_gps_label := SUBSTRING(NEW.name FROM startc FOR 1);
        IF this_time IS NOT NULL AND this_gps_label = '' THEN
          RAISE EXCEPTION 'Name %: The first line must have the gps unit label as the second character when there is a time in the Description column'
                        , NEW.name;
          RETURN NULL;
        END IF;

        startc := startc + 1;

        IF this_lineno = 1 THEN
          -- Remember the gps unit label.
          UPDATE our_state
            SET label = this_gps_label
            WHERE key = BB_SU_OUR_SKEY;

          IF NEW.Description <> '' AND NEW.Description IS NOT NULL THEN
            -- There is gps information.

            -- Find the gps unit to go with the one letter code.
            SELECT gps_units.gps
              INTO this_gps_gps
              FROM gps_units
              WHERE gps_units.label = this_gps_label
                    AND gps_units.start <= this_date
                    AND (this_date <= gps_units.finish
                         OR gps_units.finish IS NULL);

            IF NOT FOUND THEN
              RAISE EXCEPTION 'Date %: GPS label %: Name %: No GPS unit found in use on this day with this label'
                            , this_date
                            , this_gps_label
                            , NEW.name
                USING HINT = 'Have the GPS_UNITS table been updated with new GPS units?';
              RETURN NULL;
            END IF;

            -- Got the GPS info, insert the swerb_departs_gps row.
            INSERT INTO swerb_departs_gps 
              (did, xyloc, altitude, accuracy, pdop, gps, garmincode)
              VALUES(currval('swerb_departs_data_did_seq')
                   , bb_makepoint(this_x, this_y)
                   , this_altitude
                   , this_accuracy
                   , this_pdop
                   , this_gps_gps
                   , this_garmincode);
          END IF;
        ELSE
          -- this_lineno = 2
          -- Compare the GPS label to that in the first line.
          PERFORM 1
            FROM our_state
            WHERE our_state.key = BB_SU_OUR_SKEY
                  AND our_state.label = this_gps_label
            LIMIT 1;
          IF NOT FOUND THEN
            RAISE EXCEPTION 'GPS Label %: Name % : GPS label in the 2nd departure line does not match that of the first departure line'
                          , this_gps_label
                          , NEW.name;
            RETURN NULL;
          END IF;

          -- Compare the date against the date of the swerb_departs row.
          PERFORM 1
            FROM swerb_departs_data
            WHERE swerb_departs_data.did = currval('swerb_departs_data_did_seq')
                  AND swerb_departs_data.date = this_date
            LIMIT 1;
          IF NOT FOUND THEN
            RAISE EXCEPTION 'Name %: Description %: The date (%) in the second departure line does not match the date of the first departure line stored in SWERB_DEPARTS_DATA'
                          , NEW.name
                          , NEW.description
                          , this_date;
            RETURN NULL;
          END IF;
        END IF;

        -- After that are the observer codes.  (These are 2, 3 and 4 chars.
        -- make multiple guesses and if more than one works out then fail?)
        -- Fail if using one of the overlapping existing
        -- observer codes.
        -- Take each letter until we detect a valid code, then take
        -- the next letter and make sure the longer code is not also valid.
        maxc := CHAR_LENGTH(NEW.name);
        IF startc <= maxc THEN
          -- There are observers.
          lastc := startc;
          LOOP
            WHILE NOT EXISTS( SELECT 1
                                FROM observers
                                WHERE observers.initials
                                        = SUBSTRING(NEW.name
                                                    FROM startc
                                                    FOR lastc - startc + 1)
                                LIMIT 1) LOOP
              lastc := lastc + 1;
              IF lastc > maxc THEN
                this_observer
                    := SUBSTRING(NEW.name FROM startc FOR maxc - startc + 1);
                RAISE EXCEPTION 'Name %: Initials ''%'': Initials found at the end of the departure line do not exist in OBSERVERS.Initials'
                              , NEW.name
                              , this_observer;
                RETURN NULL;
              END IF;
            END LOOP;
            -- Got a valid observer.
            this_observer := SUBSTRING(NEW.name
                                       FROM startc
                                       FOR lastc - startc + 1);

            IF lastc < maxc THEN
              -- There's still more observers on the line.
              other_observer
                := SUBSTRING(NEW.name FROM startc FOR lastc - startc + 2);
              PERFORM 1
                FROM observers
                WHERE observers.initials = other_observer
                LIMIT 1;
              IF FOUND THEN
                -- Ambigiuous observer
                RAISE EXCEPTION 'Name %: Initials %: Other Initials %: Ambiguous observer initials found in departure line'
                              , NEW.name
                              , this_observer
                              , other_observer;
                RETURN NULL;
              END IF;
            END IF;

            -- What role does the observer play?
            SELECT CASE
                     WHEN this_lineno = 1 THEN
                       -- First line, observer role
                       observers.swerb_observer_role
                     ELSE
                       -- Second line, driver role
                       observers.swerb_driver_role
                    END
              INTO STRICT this_role
              FROM observers
              WHERE observers.initials = this_observer;

            -- Got an observer, insert it.
            INSERT INTO swerb_observers (did, observer, role)
              VALUES(currval('swerb_departs_data_did_seq')
                   , this_observer
                   , this_role);

            IF lastc >= maxc THEN
              -- Done with the departure line.
              EXIT;
            END IF;

            -- Get the next observer on the line. 
            startc := lastc + 1;
            lastc := startc;
          END LOOP; 
        END IF;
      END;
    ELSE
      -- This is the regular swerb data that follows line 2.
      -- (We expect line 3 to be a begin row, but let's not worry about that.)
      -- The firstc is the one letter group code.

      -- gB  Beginning of a bout of observation
      -- gE  End of a bout of observation
      --    The first/last of the day B and E lines are followed by:
      -- gMDTnnnn (median decent time) or gMATnnnn (median ascent time)
      --     nnnn is either 4 digits of time hhmm, or BA (before arrival)
      --     good only in the case of B rows, or AD (after departure
      --     good only in the case of E rows.
      --    There can be more than 1 (pair of) begin/end observations
      --    for the day when there is more than one sleeping grove.
      --    (We detect whether to create a begin or end row by
      --     remembering with a temporary table whether a B or E was
      --     last encountered.)
      -- gW  One or more baboons drinks
      -- OGgggdddd  Other group observation
      -- LMsssdddd  Lone animal observation
      -- gnnnn (time)  Hourly or half-hourly waypoint.  Always 4 digits.

      DECLARE

        pos INT := 2;  -- Position in the waypoint string
        groupc CHAR(1) := firstc;
        eventc VARCHAR(2);

        this_subgroup swerb_data.subgroup%TYPE := FALSE;
        this_seen_grp swerb_data.seen_grp%TYPE;
        this_event swerb_data.event%TYPE;

        -- state
        this_begincnt INT;
        this_endcnt INT;
        this_secondary_begincnt INT;
        this_secondary_endcnt INT;
        this_in_bout BOOLEAN;
        this_last_subgroup BOOLEAN;
        this_just_drank BOOLEAN;

      BEGIN

        eventc := SUBSTRING(NEW.name FROM pos for 1);

        IF eventc = 'BB_SU_SUBGROUPC' THEN
          -- This is a record of a subgroup.
          this_subgroup := TRUE;
          -- When the group letter is followed by the subgroup letter
          -- the event is the 3rd letter.
          pos := pos + 1;
          eventc := SUBSTRING(NEW.name FROM pos FOR 1);
        END IF;

        -- Get some state to see where we are.
        SELECT        our_state.begincnt, our_state.endcnt, our_state.in_bout
                    , our_state.last_subgroup, our_state.just_drank
                    , our_state.last_beid, our_state.in_secondary_ad
                    , our_state.secondary_begincnt, our_state.secondary_endcnt
          INTO STRICT this_begincnt     , this_endcnt     , this_in_bout
                    , this_last_subgroup, this_just_drank
                    , this_beid,           this_in_secondary_ad
                    , this_secondary_begincnt,      this_secondary_endcnt
                  FROM our_state        
                  WHERE our_state.key = BB_SU_OUR_SKEY;

        -- Handle those cases where the event is not entered by
        -- the observer but instead we must infer the event
        -- and not consume an event char from the entered waypoint.
        CASE
          WHEN NEW.name ~ '^[[:upper:]]S?BB_SU_TIME_PATTERN$'
               AND this_new_lone_animal = '' THEN
            -- Matches the pattern for a regular swerb waypoint.
            -- Make up an event code.
            eventc = 'BB_SU_TICK_EVENT';
            this_event := 'bb_swerb_obs';

          WHEN ((this_begincnt = 1 AND NOT(this_secondary_ad))
                OR (this_secondary_begincnt = 1 AND this_secondary_ad))
               AND NEW.name ~ '^[[:upper:]]S?bb_swerb_mdt.+'
               AND this_new_lone_animal = '' THEN
            -- Second line of contiguous begin lines.
            -- Make up an event code.
            eventc := 'BB_SU_BEGIN_EVENT';
            this_event := 'bb_swerb_begin';

          WHEN ((this_endcnt = 1 AND NOT(this_secondary_ad))
                OR (this_secondary_endcnt = 1 AND this_secondary_ad))
               AND NEW.name ~ '^[[:upper:]]S?bb_swerb_mat.+'
               AND this_new_lone_animal = '' THEN
            -- Second line of contiguous end lines.
            -- Make up an event code.
            eventc := 'BB_SU_END_EVENT';
            this_event := 'bb_swerb_end';

          ELSE
            -- Some other sort of event.
            pos := pos + 1;

            IF eventc = 'BB_SU_LONEANIMAL_EVENT_DATA'
               AND this_new_lone_animal <> '' THEN
              -- Frob the eventc to be "special" and indicate lone animal.
              eventc := 'BB_SU_LONEANIMAL_EVENTC';
            END IF;

            -- Look up the db event code.
            -- (Map operator entered events to codes used in SWERB_DATA.Event.)
            SELECT event_map.db_event
              INTO this_event
              FROM event_map
              WHERE event_map.su_event = eventc;
        END CASE;

        IF NOT(this_in_bout) AND this_secondary_ad THEN
          RAISE EXCEPTION 'Name %: Secondary ascents and descents must occur within bouts of observation'
                        , NEW.name;
          RETURN NULL;
        END IF;

        IF eventc <> 'BB_SU_OTHERG_EVENT'
           AND eventc <> 'BB_SU_LONEANIMAL_EVENTC'
           AND eventc <> 'BB_SU_WATER_EVENT'
           AND NOT(eventc = 'BB_SU_BEGIN_EVENT'
                   AND this_begincnt IS NOT NULL) -- 2nd begin row
           AND NOT(eventc = 'BB_SU_END_EVENT'
                   AND this_endcnt IS NOT NULL) -- 2nd end row
           THEN
          -- Guess whether the previous drinking event was of a subgroup.
          -- (Secondary ascents and descents trigger this guess too.)
          IF this_just_drank THEN
            IF this_last_subgroup AND this_subgroup THEN
              UPDATE swerb_data
                SET subgroup = TRUE
                FROM drinks
                WHERE swerb_data.swid = drinks.swid;
            END IF;
            -- Forget all the drinking events now that we've
            -- seen a non-drinking event.
            TRUNCATE TABLE drinks;
          END IF;
          -- Prepare for the future.
          UPDATE our_state
            SET just_drank = FALSE  -- This row is not a drinking event
              , last_subgroup = this_subgroup  -- Is this a subgroup?
            WHERE key = BB_SU_OUR_SKEY;
        END IF;        

        IF eventc <> 'BB_SU_OTHERG_EVENT'
           AND eventc <> 'BB_SU_LONEANIMAL_EVENTC' THEN
          -- Lookup the gid, given the one letter group code.
          SELECT groups.gid
            INTO this_seen_grp
            FROM groups
            WHERE groups.one_letter_code = groupc
                  AND groups.start <= this_date
                  AND (this_date <= groups.cease_to_exist
                       OR groups.cease_to_exist IS NULL);

          IF NOT FOUND THEN
            RAISE EXCEPTION 'Name %: Group code %: observaton date %: One letter group code not found active as of the observation date'
                          , NEW.name
                          , groupc
                          , this_date;
            RETURN NULL;
          END IF;
        END IF;

        IF eventc <> 'BB_SU_BEGIN_EVENT'
           AND eventc <> 'BB_SU_END_EVENT' THEN
          -- Not a begin or end row.

          IF this_secondary_ad THEN
            RAISE EXCEPTION integrity_constraint_violation USING
                  MESSAGE = 'Error on ' || TG_OP || ' of SWERB_UPLOAD'
                , DETAIL = 'Value (Name) = (' || NEW.name
                             || '): Only begin and end rows may be marked as '
                             || 'secondary ascents/descents with a TRUE '
                             || 'Secondary_AD column value';
            RETURN NULL;
          END IF;

          IF this_in_secondary_ad THEN
            RAISE EXCEPTION 'Secondary ascent/descent is not ended'
                  USING DETAIL = 'Name '
                                 || NEW.name
                                 || ': Row type '
                                 || eventc
                                 || ' not allowed here';
            RETURN NULL;
          END IF;

          -- Reset state so that begins or ends that are out of
          -- order are detected.
          UPDATE our_state
            SET begincnt = NULL -- no begin row yet seen
              , endcnt = NULL  -- no end row yet seen
              , secondary_begincnt = NULL -- no secondary begin row
              , secondary_endcnt = NULL -- no secondary end row
            WHERE key = BB_SU_OUR_SKEY;

          IF timeest_supplied THEN
            RAISE EXCEPTION 'Name %: Timeest %: Only begin and end lines can have an estimated time'
                          , NEW.name
                          , NEW.timeest
              USING HINT = 'Supply NULL or the empty string in the Timeest column';
            RETURN NULL;
          END IF;

          IF is_effort_supplied THEN
            RAISE EXCEPTION 'Name %: Is_effort %: Only begin lines can have an Is_effort value'
                          , NEW.name
                          , NEW.is_effort
              USING HINT = 'Supply NULL or the empty string in the Is_effort column';
            RETURN NULL;
          END IF;

          IF source_supplied THEN
            RAISE EXCEPTION 'Name %: Source %: Only begin and end lines can have a source'
                          , NEW.name
                          , NEW.source
              USING HINT = 'Supply NULL or the empty string in the Source column';
            RETURN NULL;
          END IF;

          IF notes_supplied THEN
            RAISE EXCEPTION 'Name %: Notes %: Only begin lines can have an Notes value'
                          , NEW.name
                          , NEW.notes
              USING HINT = 'Supply NULL or the empty string in the Notes column';
            RETURN NULL;
          END IF;

          IF eventc = 'BB_SU_WATER_EVENT' OR eventc = 'BB_SU_TICK_EVENT' THEN
            -- The time is what is entered by the observer.
            time_as_str := SUBSTRING(NEW.name FROM pos FOR bb_swerb_time_len);
            IF time_as_str !~ 'BB_SU_TIME_PATTERN' THEN
              RAISE EXCEPTION 'Name %: Time %: Invalid time in waypoint'
                            , NEW.name
                            , time_as_str;
              RETURN NULL;
            END IF;

            pos := pos + bb_swerb_time_len;

            -- It's a real time.  Format it properly.
            time_as_str := SUBSTRING(time_as_str FROM 1 FOR 2)
                           || ':'
                           || SUBSTRING(time_as_str FROM 3);
          END IF;
        END IF;

        CASE eventc
          WHEN 'BB_SU_OTHERG_EVENT'
             , 'BB_SU_LONEANIMAL_EVENTC' THEN
            -- It is an observation of a group other than the swerb focal
            -- or of a lone animal.

            CASE eventc
              WHEN 'BB_SU_OTHERG_EVENT' THEN
                IF groupc <> 'BB_SU_OTHERG_GROUP' THEN
                  -- Group letter not right for other group observation.
                   RAISE EXCEPTION 'Name %: Group letter must be ''BB_SU_OTHERG_GROUP'' when the observation is of a group other than the focal'
                                 , NEW.name;
                    RETURN NULL;
                END IF;            
              WHEN 'BB_SU_LONEANIMAL_EVENTC' THEN
                IF groupc <> 'BB_SU_LONEANIMAL_GROUP' THEN
                  -- Group letter not right for lone animal observation.
                   RAISE EXCEPTION 'Name %: Group letter must be ''BB_SU_LONEAMINALE_GROUP'' when the observation is of a lone animal'
                                 , NEW.name;
                    RETURN NULL;
                END IF;            
            END CASE;
            -- OGxxxdd0  Observation of a non-focal group (dd0 is optional)
            --        'xxx' is the 3 letter group code
            --        'dd0' is the distance in meters (last digit always 0)
            -- LMsssdd0  Observation is of a lone animal (dd0 is optional)
            --        'sss' is the Sname of the lone animal or a
            --        UNKSNAMES.Unksname value where UNKSNAMES.Lonemale
            --        is TRUE.
            --        'dd0' is the distance in meters (last digit always 0)
            DECLARE

              this_og TEXT;  -- other group code
              this_lone_animal TEXT;  -- lone animal sname
              this_distance TEXT; -- We need to check the string
              this_sex CHAR(1);   -- A biograph.sex value

            BEGIN
              CASE eventc
                WHEN 'BB_SU_OTHERG_EVENT' THEN
                  this_og := SUBSTRING(NEW.name FROM pos FOR BB_SU_G_LEN);
                  pos := pos + BB_SU_G_LEN;

                  -- Lookup the gid.
                  SELECT groups.gid
                    INTO this_seen_grp
                    FROM groups
                    WHERE groups.three_letter_code = this_og;

                  IF NOT FOUND THEN
                    RAISE EXCEPTION 'Name %: three letter code %: Three letter group code not found'
                                  , NEW.name
                                  , this_og;
                    RETURN NULL;
                  END IF;
                WHEN 'BB_SU_LONEANIMAL_EVENTC' THEN
                  this_lone_animal
                    := SUBSTRING(NEW.name FROM pos FOR BB_SU_LA_LEN);
                  pos := pos + BB_SU_LA_LEN;

                  -- See if it's the unknown lone male.
                  PERFORM 1
                    FROM unksnames
                    WHERE unksnames.unksname = this_lone_animal
                          AND unksnames.lonemale;

                  IF FOUND THEN
                    -- It is an unknown lone male.
                    IF this_new_lone_animal <> 'bb_male' THEN
                      RAISE EXCEPTION 'Name %: Code for lone unknown male %: Lone_Animal %: The Lone_Animal value must be bb_male when UNKSNAMES indicates an unknown lone male'
                                    , NEW.name
                                    , this_lone_animal
                                    , this_new_lone_animal;
                      RETURN NULL;
                    END IF;

                    this_lone_animal := NULL;
                  ELSE
                    -- It is not an unknown lone male.

                    -- Be sure sex is right.
                    SELECT biograph.sex
                      INTO this_sex
                      FROM biograph
                      WHERE biograph.sname = this_lone_animal;

                    IF NOT FOUND THEN
                      -- Can't find individual.
                      RAISE EXCEPTION 'Name %: Sname of lone animal %: Lone_Animal %: The indicated SNAME could not be found on BIOGRAPH when attemping to validate sex'
                                    , NEW.name
                                    , this_lone_animal
                                    , this_new_lone_animal;
                      RETURN NULL;
                    END IF;

                    IF this_new_lone_animal <> this_sex THEN
                      -- Sex mis-match.
                      RAISE EXCEPTION 'Name %: Sname of lone animal %: Lone_Animal %: BIOGRAPH.Sex %: The sex of the lone animal does not match the supplied Lone_Animal value'
                                    , NEW.name
                                    , this_lone_animal
                                    , this_new_lone_animal
                                    , this_sex;
                      RETURN NULL;
                    END IF;

                  END IF;
              END CASE;

              this_distance := NULL;
              IF CHAR_LENGTH(NEW.name) >= pos THEN
                -- Remaining chars should be distance to
                -- lone animal/other group
                this_distance := SUBSTRING(NEW.name FROM pos);

                IF TRIM(BOTH FROM this_distance) !~ '^[[:digit:]]+$' THEN
                  -- Just check for digits and let the db worry about
                  -- validating acutal values.
                  RAISE EXCEPTION 'Name %: distance %: The distance to the other group/lone animal is not made up of digits'
                                , NEW.name
                                , this_distance;
                  RETURN NULL;
                END IF;
              END IF;

              IF NOT(this_in_bout) THEN
                -- Not in a bout of observation.
                -- Make a bout of observation of the
                -- unknown group that starts and ends
                -- at the same moment the other group/lone animal
                -- observation is made.
                -- Note that we also say it does not count toward
                -- observer effort.

                -- Make the swerb_bes row.
                INSERT INTO swerb_bes (did
                                     , focal_grp
                                     , start
                                     , btimeest
                                     , bsource
                                     , stop
                                     , etimeest
                                     , esource
                                     , is_effort
                                     , notes)
                  VALUES (currval('swerb_departs_data_did_seq')
                        , 'bb_unk_grp'
                        , DATE_TRUNC('minute', this_time)
                        , this_timeest
                        , this_source
                        , DATE_TRUNC('minute', this_time)
                        , this_timeest
                        , this_source
                        , FALSE
                        , NULL);

                -- Keep the current beid (for this row only).
                this_beid := currval('swerb_bes_beid_seq');

                -- Make the swerb_data row for the begin.
                INSERT INTO swerb_data (beid
                                      , seen_grp
                                      , event
                                      , time
                                      , quad
                                      , xyloc
                                      , altitude
                                      , accuracy
                                      , subgroup
                                      , ogdistance
                                      , gps_datetime
                                      , garmincode)
                  VALUES (this_beid
                        , 'bb_unk_grp'
                        , 'bb_swerb_begin'
                        , DATE_TRUNC('minute', this_time)
                        , this_quad
                        , bb_makepoint(this_x, this_y)
                        , this_altitude
                        , this_accuracy
                        , FALSE
                        , NULL
                        , NEW.description::TIMESTAMP
                        , this_garmincode);

                -- Make the swerb_data row for the end.
                INSERT INTO swerb_data (beid
                                      , seen_grp
                                      , event
                                      , time
                                      , quad
                                      , xyloc
                                      , altitude
                                      , accuracy
                                      , subgroup
                                      , ogdistance
                                      , gps_datetime
                                      , garmincode)
                  VALUES (this_beid
                        , 'bb_unk_grp'
                        , 'bb_swerb_end'
                        , DATE_TRUNC('minute', this_time)
                        , this_quad
                        , bb_makepoint(this_x, this_y)
                        , this_altitude
                        , this_accuracy
                        , FALSE
                        , NULL
                        , NEW.description::TIMESTAMP
                        , this_garmincode);
              END IF;

              CASE eventc
                WHEN 'BB_SU_OTHERG_EVENT' THEN
                  -- Insert the other group row.
                  INSERT INTO swerb_data (beid
                                        , seen_grp
                                        , event
                                        , time
                                        , quad
                                        , xyloc
                                        , altitude
                                        , accuracy
                                        , subgroup
                                        , ogdistance
                                        , gps_datetime
                                        , garmincode)
                   VALUES (this_beid
                         , this_seen_grp
                         , this_event
                         , DATE_TRUNC('minute', this_time)
                         , this_quad
                         , bb_makepoint(this_x, this_y)
                         , this_altitude
                         , this_accuracy
                         , this_subgroup
                         , this_distance::NUMERIC(bb_swerb_ogdistance_len, 0)
                         , NEW.description::TIMESTAMP
                         , this_garmincode);

                WHEN 'BB_SU_LONEANIMAL_EVENTC' THEN
                  -- Insert the lone animal row.
                  INSERT INTO swerb_data (beid
                                        , seen_grp
                                        , lone_animal
                                        , event
                                        , time
                                        , quad
                                        , xyloc
                                        , altitude
                                        , accuracy
                                        , subgroup
                                        , ogdistance
                                        , gps_datetime
                                        , garmincode)
                   VALUES (this_beid
                         , bb_alone_grp
                         , this_lone_animal
                         , this_event
                         , DATE_TRUNC('minute', this_time)
                         , this_quad
                         , bb_makepoint(this_x, this_y)
                         , this_altitude
                         , this_accuracy
                         , this_subgroup
                         , this_distance::NUMERIC(bb_swerb_ogdistance_len, 0)
                         , NEW.description::TIMESTAMP
                         , this_garmincode);
              END CASE;
            END;

          WHEN 'BB_SU_BEGIN_EVENT', 'BB_SU_END_EVENT' THEN
            -- A beginning or end of observation.

            -- Some other-than-begin-or-end-of-group/day begins and
            -- ends have extra stuff following the begin or end
            -- code, generally a sequence number.  Since the default
            -- is to use a ADCode that requires a time and these
            -- begins and ends won't have a second line then these
            -- will generate an error unless the data manager
            -- manually removes the extra stuff.
            DECLARE

              this_cnt INT;
              this_adcode swerb_loc_data.adcode%TYPE;

            BEGIN
              IF eventc = 'BB_SU_BEGIN_EVENT' THEN
                -- Begin row.

                IF (this_begincnt IS NULL
                    AND NOT(this_secondary_ad))
                   OR (this_secondary_begincnt IS NULL
                       AND this_secondary_ad) THEN
                  -- This is the first begin row.
                  this_cnt := 1;
                  this_adcode := 'bb_swerb_descent_code';

                  -- Create the swerb_bes row.
                  INSERT INTO swerb_bes (did
                                       , focal_grp
                                       , start
                                       , btimeest
                                       , bsource
                                       , stop
                                       , etimeest
                                       , esource
                                       , is_effort
                                       , notes)
                    VALUES (currval('swerb_departs_data_did_seq')
                          , this_seen_grp
                          , DATE_TRUNC('minute', this_time)
                          , this_timeest
                          , this_source
                          , NULL
                          , FALSE
                          , 'bb_gps_time_source'
                          , this_is_effort
                          , this_notes);

                  IF this_secondary_ad THEN
                    UPDATE our_state
                      SET secondary_begincnt = 1
                        , secondary_endcnt = NULL -- no 2ndary end row seen
                        , in_secondary_ad = TRUE
                      WHERE key = BB_SU_OUR_SKEY;
                  ELSE
                    -- Regular begin row.
                    -- Update state so we know where we are.
                    this_beid := currval('swerb_bes_beid_seq');

                    UPDATE our_state
                      SET begincnt = 1
                        , endcnt = NULL  -- no end row yet seen
                        , in_bout = TRUE
                        , last_beid = this_beid
                      WHERE key = BB_SU_OUR_SKEY;
                  END IF;

                ELSIF (this_begincnt = 1
                       AND NOT(this_secondary_ad))
                      OR (this_secondary_begincnt = 1
                          AND this_secondary_ad) THEN
                  -- This is the second begin row.
                  this_cnt := 2;

                  IF is_effort_supplied THEN
                    RAISE EXCEPTION 'Name %: Is_effort %: Only the first begin line can have an Is_effort value'
                                  , NEW.name
                                  , NEW.is_effort
                      USING HINT = 'Supply NULL or the empty string in the Is_effort column';
                    RETURN NULL;
                  END IF;

                  IF source_supplied THEN
                    RAISE EXCEPTION 'Name %: Source %: Only the first begin line can have an Source value'
                                  , NEW.name
                                  , NEW.source
                      USING HINT = 'Supply NULL or the empty string in the Source column';
                    RETURN NULL;
                  END IF;

                  IF notes_supplied THEN
                    RAISE EXCEPTION 'Name %: Notes %: Only the first begin line can have an Notes value'
                                  , NEW.name
                                  , NEW.notes
                      USING HINT = 'Supply NULL or the empty string in the Notes column';
                    RETURN NULL;
                  END IF;

                  -- Update state so we know where we are.
                  IF this_secondary_ad THEN
                    UPDATE our_state
                      SET secondary_begincnt = 2
                        , secondary_endcnt = NULL -- no end row yet seen
                      WHERE key = BB_SU_OUR_SKEY;
                  ELSE
                    -- Regular 2nd end row.
                    UPDATE our_state
                      SET begincnt = 2
                        , endcnt = NULL  -- no end row yet seen
                      WHERE key = BB_SU_OUR_SKEY;
                  END IF;

                  -- Make sure we've the median descent time.
                  IF SUBSTRING(NEW.name FROM pos FOR 3) <> 'bb_swerb_mdt' THEN
                    RAISE EXCEPTION 'Name %: Expecting the characters ''bb_swerb_mdt'' in second begin line to identify the median descent time: Not found'
                                  , NEW.name;
                    RETURN NULL;
                  END IF;
                  pos := pos + 3;
                ELSE
                  -- Too many begin rows.
                  this_begincnt := this_begincnt + 1;
                  IF this_secondary_ad THEN
                    RAISE EXCEPTION 'Name %: Too many secondary begin lines, this counts as number %'
                                  , NEW.name
                                  , this_secondary_begincnt;
                  ELSE
                    RAISE EXCEPTION 'Name %: Too many begin lines, this counts as number %'
                                  , NEW.name
                                  , this_begincnt;
                  END IF;
                  RETURN NULL;
                END IF;
              ELSE
                -- 'BB_SU_END_EVENT'
                -- End row.

                IF (this_endcnt IS NULL
                    AND NOT(this_secondary_ad))
                   OR (this_secondary_endcnt IS NULL
                       AND this_secondary_ad) THEN
                  -- This is the first end row.
                  this_cnt := 1;
                  this_adcode := 'bb_swerb_ascent_code';

                  -- Update state so we know where we are.
                  IF this_secondary_ad THEN
                    UPDATE our_state
                      SET secondary_begincnt = NULL -- reset
                      , secondary_endcnt = 1
                      , in_secondary_ad = FALSE
                    WHERE key = BB_SU_OUR_SKEY;
                  ELSE -- Regular end row
                    UPDATE our_state
                      SET begincnt = NULL -- reset so no begin rows seen
                        , endcnt = 1
                        , in_bout = FALSE
                        , last_beid = NULL
                      WHERE key = BB_SU_OUR_SKEY;
                  END IF;

                  -- Update the swerb_bes row.
                  UPDATE swerb_bes
                    SET stop = DATE_TRUNC('minute', this_time)
                      , etimeest = this_timeest
                      , esource = this_source
                    WHERE beid = CASE
                                   WHEN this_secondary_ad
                                     THEN CURRVAL('swerb_bes_beid_seq'::REGCLASS)
                                   ELSE this_beid
                                 END;

                ELSIF (this_endcnt = 1 AND NOT(this_secondary_ad))
                      OR (this_secondary_endcnt = 1 AND this_secondary_ad) THEN
                  -- This is the second end row.
                  this_cnt := 2;

                  IF is_effort_supplied THEN
                    RAISE EXCEPTION 'Name %: Is_effort %: Only the first end line can have an Is_effort value'
                                  , NEW.name
                                  , NEW.is_effort
                      USING HINT = 'Supply NULL or the empty string in the Is_effort column';
                    RETURN NULL;
                  END IF;

                  IF source_supplied THEN
                    RAISE EXCEPTION 'Name %: Source %: Only the first end line can have an Source value'
                                  , NEW.name
                                  , NEW.source
                      USING HINT = 'Supply NULL or the empty string in the Source column';
                    RETURN NULL;
                  END IF;

                  IF notes_supplied THEN
                    RAISE EXCEPTION 'Name %: Notes %: Only the first end line can have an Notes value'
                                  , NEW.name
                                  , NEW.notes
                      USING HINT = 'Supply NULL or the empty string in the Notes column';
                    RETURN NULL;
                  END IF;

                  -- Update state so we know where we are.
                  IF this_secondary_ad THEN
                    UPDATE our_state
                      SET secondary_begincnt = NULL -- reset
                      , secondary_endcnt = 2
                    WHERE key = BB_SU_OUR_SKEY;
                  ELSE
                    UPDATE our_state
                      SET begincnt = NULL -- reset so no begin rows seen
                        , endcnt = 2
                      WHERE key = BB_SU_OUR_SKEY;
                  END IF;

                  -- Make sure we've the median ascent time.
                  IF SUBSTRING(NEW.name FROM pos FOR 3) <> 'bb_swerb_mat' THEN
                    RAISE EXCEPTION 'Name %: Expecting the characters ''bb_swerb_mat'' in second end line to identify the median ascent time: Not found'
                                  , NEW.name;
                    RETURN NULL;
                  END IF;
                  pos := pos + 3;
                ELSE
                  -- Too many end lines.
                  this_endcnt := this_endcnt + 1;
                  IF this_secondary_ad THEN
                    RAISE EXCEPTION 'Name %: Too many secondary end lines, this counts as number %'
                                  , NEW.name
                                  , this_secondary_endcnt;
                  ELSE
                    RAISE EXCEPTION 'Name %: Too many end lines, this counts as number %'
                                  , NEW.name
                                  , this_endcnt;
                  END IF;
                  RETURN NULL;
                END IF;
              END IF;

              -- Done validating and doing the stuff unique to a begin or end.
              -- Do the stuff common to the begin/end row.
              -- (The swerb_data, swerb_loc_data, and swerb_loc_gps rows.)
              CASE this_cnt
                WHEN 1 THEN
                  -- First of 2 lines.

                  -- Create the swerb_data begin/end row.
                  INSERT INTO swerb_data (beid
                                        , seen_grp
                                        , event
                                        , time
                                        , quad
                                        , xyloc
                                        , altitude
                                        , accuracy
                                        , subgroup
                                        , ogdistance
                                        , gps_datetime
                                        , garmincode)
                   VALUES (CASE WHEN this_secondary_ad
                                  THEN CURRVAL('swerb_bes_beid_seq'::REGCLASS)
                                ELSE this_beid
                           END
                         , this_seen_grp
                         , this_event
                         , DATE_TRUNC('minute', this_time)
                         , this_quad
                         , bb_makepoint(this_x, this_y)
                         , this_altitude
                         , this_accuracy
                         , this_subgroup
                         , NULL
                         , NEW.description::TIMESTAMP
                         , this_garmincode);

                  -- If there's a grove create the swerb_loc_data row.
                  DECLARE
                    this_grove swerb_loc_data.loc%TYPE;
                    this_loc_status swerb_loc_data.loc_status%TYPE;

                  BEGIN
                    -- Lower case 'p' preceeding grove means probable.
                    -- (But it probably won't be lower case so just
                    -- upper case it to test.)
                    IF UPPER(SUBSTRING(NEW.name FROM pos FOR 1))
                       = 'bb_swerb_probable_char' THEN
                      -- Well, P, probably means probable,
                      -- but if the grove does not exist when the P
                      -- is removed then the P is really part of the
                      -- grove code.
                      this_loc_status := 'bb_swerb_probable';
                      pos := pos + 1;

                      PERFORM 1
                        FROM swerb_gws
                        WHERE swerb_gws.loc = SUBSTRING(NEW.name FROM pos)
                              AND swerb_gws.type = 'bb_swerb_gws_grove'
                        LIMIT 1;
                      IF NOT FOUND THEN
                        -- The grove does not exist.  Undo our "probable".
                        -- The P is really part of the grove.
                        pos := pos - 1;
                        this_loc_status := 'bb_swerb_certain';
                      END IF;
                    ELSE
                      this_loc_status := 'bb_swerb_certain';
                    END IF;

                    -- Get the grove.
                    this_grove := SUBSTRING(NEW.name FROM pos);

                    IF this_grove <> '' THEN
                      -- Save the grove.
                      INSERT INTO swerb_loc_data (swid
                                                , loc
                                                , adcode
                                                , loc_status
                                                , adtime)
                        VALUES (currval('swerb_data_swid_seq')
                              , this_grove
                              , this_adcode
                              , this_loc_status
                              , NULL);
                    END IF;
                  END;
                WHEN 2 THEN
                  -- Second of 2 lines.

                  IF this_quad IS NOT NULL THEN
                    RAISE EXCEPTION 'Name %: Quad %: The second begin or end line must have a NULL quad value'
                                  , NEW.name
                                  , this_quad;
                    RETURN NULL;
                  END IF;

                  IF timeest_supplied THEN
                    RAISE EXCEPTION 'Name %: Timeest %: Only the 1st begin or the 1st end line can have an estimated time'
                                  , NEW.name
                                  , NEW.timeest
                      USING HINT = 'Supply NULL or the empty string in the Timeest column';
                    RETURN NULL;
                  END IF;

                  -- Get the median descent/ascent time
                  time_as_str := SUBSTRING(NEW.name FROM pos);
                  IF time_as_str ~ 'BB_SU_TIME_PATTERN' THEN
                    -- It's a real time.  Format it properly.
                    time_as_str := SUBSTRING(time_as_str FROM 1 FOR 2)
                                   || ':'
                                   || SUBSTRING(time_as_str FROM 3);

                    -- Save the time.
                    UPDATE swerb_loc_data
                      SET adtime = time_as_str::TIME
                      WHERE swid = currval('swerb_data_swid_seq');
                  ELSE
                    -- We don't have a real time.  Use the value as
                    -- the SWERB_LOC_DATA.ADCode value.
                    UPDATE swerb_loc_data
                      SET adcode = time_as_str
                      WHERE swid = currval('swerb_data_swid_seq');
                  END IF;

                  IF TRIM(BOTH FROM NEW.description) = ''
                     OR NEW.description IS NULL THEN
                    -- This did not come from a GPS unit.
                    -- We do not want a swerb_loc_gps row.
                    IF this_x IS NOT NULL
                       OR this_y IS NOT NULL
                       OR this_altitude IS NOT NULL
                       OR this_pdop IS NOT NULL
                       OR NEW.accuracy <> '' THEN
                      RAISE EXCEPTION 'Description %: X %: Y %: Altitude %: PDOP %: Accuracy %: Garmincode %: When there is no description there must also be no X, Y, Altitude, PDOP, or Accuracy'
                                    , NEW.description
                                    , this_x
                                    , this_y
                                    , this_altitude
                                    , this_pdop
                                    , NEW.accuracy
                                    , this_garmincode;
                      RETURN NULL;
                    END IF;

                  ELSE
                    -- Create the swerb_loc_gps row.
                    INSERT INTO swerb_loc_gps (swid
                                             , xyloc
                                             , altitude
                                             , pdop
                                             , accuracy
                                             , gps_datetime
                                             , garmincode)
                      VALUES (currval('swerb_data_swid_seq')
                            , bb_makepoint(this_x, this_y)
                            , this_altitude
                            , this_pdop
                            , this_accuracy
                            , NEW.description::TIMESTAMP
                            , this_garmincode);
                  END IF;
              END CASE;

            END;

          WHEN 'BB_SU_WATER_EVENT' THEN
            -- Drinking at a waterhole.

            -- W records look like:  gWnnnnh*, where g=group
            -- h is the waterhole and n the time.

            -- Save the swerb_data row.
            INSERT INTO swerb_data (beid
                                  , seen_grp
                                  , event
                                  , time
                                  , quad
                                  , xyloc
                                  , altitude
                                  , accuracy
                                  , subgroup
                                  , ogdistance
                                  , gps_datetime
                                  , garmincode)
             VALUES (this_beid
                   , this_seen_grp
                   , this_event
                   , time_as_str::TIME
                   , this_quad
                   , bb_makepoint(this_x, this_y)
                   , this_altitude
                   , this_accuracy
                   , this_subgroup
                   , NULL
                   , NEW.description::TIMESTAMP
                   , this_garmincode);

            -- Get the waterhole.
            DECLARE
              this_waterhole swerb_loc_data.loc%TYPE;

            BEGIN
              this_waterhole := SUBSTRING(NEW.name FROM pos);

              -- Save the swerb_loc_data row.
              INSERT INTO swerb_loc_data (swid
                                        , loc
                                        , adcode
                                        , loc_status
                                        , adtime)
                VALUES (currval('swerb_data_swid_seq')
                      , this_waterhole
                      , 'bb_swerb_notad_code'
                      , 'bb_swerb_certain'
                      , NULL);

            END;

            IF this_last_subgroup AND NOT(this_subgroup) THEN
              -- This might be a subgroup that the observation
              -- team did not write into the garmincode.
              -- Save the id in case we later want to mark
              -- it as a subgroup.
              INSERT INTO drinks (swid)
                VALUES(currval('swerb_data_swid_seq'));
              UPDATE our_state
                SET just_drank = TRUE
                WHERE key = BB_SU_OUR_SKEY;
            END IF;

          WHEN 'BB_SU_TICK_EVENT' THEN
            -- A regular interval waypoint.

            -- These look like gnnnn (or gSnnnn) where nnnn is the
            -- observer entered time.

            -- Save the swerb_data row.
            INSERT INTO swerb_data (beid
                                  , seen_grp
                                  , event
                                  , time
                                  , quad
                                  , xyloc
                                  , altitude
                                  , accuracy
                                  , subgroup
                                  , ogdistance
                                  , gps_datetime
                                  , garmincode)
             VALUES (this_beid
                   , this_seen_grp
                   , this_event
                   , time_as_str::TIME
                   , this_quad
                   , bb_makepoint(this_x, this_y)
                   , this_altitude
                   , this_accuracy
                   , this_subgroup
                   , NULL
                   , NEW.description::TIMESTAMP
                   , this_garmincode);

          ELSE
            -- An unrecognized event code.
/*
            RAISE EXCEPTION 'Name %: pos %: eventc %: this_begincnt %: this_endcnt %: Unable to parse: Unrecognized event'
                          , NEW.name
                          , pos 
                          , eventc
                          , this_begincnt
                          , this_endcnt;
*/
            RAISE EXCEPTION 'Name %: Unable to parse: Unrecognized event'
                          , NEW.name;
            RETURN NULL;
        END CASE;
      END;
    END IF;

    -- Do we need a way to manually set the SWERB_BES.Start/Stop times?

  RETURN NULL;
  END;
$$;


SELECT 'swerb_upload_insert_trigger' AS trigger;
CREATE TRIGGER swerb_upload_insert_trigger
  INSTEAD OF INSERT
  ON swerb_upload FOR EACH ROW
  EXECUTE PROCEDURE swerb_upload_insert_func();

SELECT 'swerb_upload_update_trigger' AS trigger;
CREATE TRIGGER swerb_upload_update_trigger
  INSTEAD OF UPDATE
  ON swerb_upload FOR EACH ROW
  EXECUTE PROCEDURE _error_immutable_view();

SELECT 'swerb_upload_delete_trigger' AS trigger;
CREATE TRIGGER swerb_upload_delete_trigger
  INSTEAD OF DELETE
  ON swerb_upload FOR EACH ROW
  EXECUTE PROCEDURE _error_immutable_view();
