#!/bin/sh
#
# Coypright (C) 2013 The Meme Factory, Inc., http://www.meme.com/
# Copyright (C) 2011 Karl O. Pinc <kop@meme.com>
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Load the babase schemas into a empty(ish) babase db.
#
# Usage: babase-mobile-load < foo.tar
#
# Side effects:
#  Recreates babase db from babase-mobile-dump.
#
# Remarks:
#  Must be run as root.
#
#  On Debian the following can be used to setup everything:
#
#    pg_ctlcluster 8.4 main stop   ;# only need ever run this once
#                                  ;# it might be useful to delete main
#    pg_createcluster --locale=C 8.4 babase --start
#
#  On Debian the cluster can be destroyed with:
#
#    pg_dropcluster 8.4 babase --stop
#
# Bugs:
#  Separately restoring schemas is unecessary once papio goes
# to postgresql 8.4 (or maybe earlier).  See the notes
# in babase-mobile-dump.
#
# The babase db name is hardcoded.
#

# Globals
export ME=$(basename $0)
export TMPPATH=/tmp/$ME.$$
export DUMPDIR=babase-mobile-dump  ;# top level dir in tarball
export DUMPPATH=$TMPPATH/$DUMPDIR
export SCHEMAPATH=$DUMPPATH/schemas
export PG_ADMIN=postgres
export PG84BASE=/usr/share/postgresql/8.4/contrib/   ;# Debian specific

# Helper function to get access to the db as admin.
accessdb()
{
    # if pg_hba gives root access you just need
    # $1
    # otherwise this will work with the default pg_hba.conf
    # when the user "postgres" is $PG_ADMIN.
    su -c "$1" $PG_ADMIN
}

# Make sure we don't leave cruft behind.
cleanup () {
  rm -rf $TMPPATH
}

trap cleanup 0

# Remember when this all happened.
today="$(date)"

# Untar the tarball in a unique place.
mkdir -p $TMPPATH
tar -C $TMPPATH -xzf -

# Change ownership of the import files so postgres can access them.
# (rnh, 2011-05-05)
chown -R postgres.postgres $TMPPATH

# Load the roles
accessdb psql < $DUMPPATH/roles.sql

# Re-create the database.
accessdb psql < $DUMPPATH/create_babase.sql
# Install the plpgsql language.
accessdb 'createlang plpgsql -d babase'

#
# Install postgis.
#

# Re-create the postgis schema
accessdb 'psql -d babase' < $DUMPPATH/create_postgis_schema.sql

# Install the functions.
{ printf 'set search_path=postgis;\n' ; \
  cat $PG84BASE/postgis-1.5/postgis.sql ; \
  } \
  | accessdb 'psql -d babase'
# Install the coordinate system db.
{ printf 'set search_path=postgis;\n' ; \
  cat $PG84BASE/postgis-1.5/spatial_ref_sys.sql ; \
  } \
  | accessdb 'psql -d babase'
# Install the comments.
{ printf 'set search_path=postgis;\n' ; \
  cat $PG84BASE/postgis_comments.sql ; \
  } \
  | accessdb 'psql -d babase'

#
# Restore all the schemas
#

# Instead of this:
# pg_restore --schema-only --dbname babase $DUMPPATH/schemas.custom
#   do this:
# Begin pre 8.4 hack.

# Restore the devel schema  first, then babase, babase_pending and then the 
# sandbox schema.  It's remotely possible that other
# schemas may depend on their objects.
echo debug: devel schema
accessdb 'psql -d babase' < $SCHEMAPATH/devel.sql
rm $SCHEMAPATH/devel.sql
echo debug: babase schema
accessdb 'psql -d babase' < $SCHEMAPATH/babase.sql
rm $SCHEMAPATH/babase.sql
echo debug: babase_pending schema
accessdb 'psql -d babase' < $SCHEMAPATH/babase_pending.sql
rm $SCHEMAPATH/babase_pending.sql
echo debug: sandbos schema
accessdb 'psql -d babase' < $SCHEMAPATH/sandbox.sql
rm $SCHEMAPATH/sandbox.sql

# Then restore the rest of the schemas.
for s in $(find $SCHEMAPATH -type f) ; do
  echo debug: $(basename $s .sql)
  accessdb 'psql -d babase' < $s
done

# End pre 8.4 hack

# Restore the data.
accessdb "pg_restore --data-only --disable-triggers --dbname babase $DUMPPATH/babase.custom"

# Collect stats for the query planner.
accessdb 'psql -c analyze -d babase'

#
# Go toally crazy!
# Leave a comment describing when load started.
#
# Comment may be retrieved with:
#   SELECT shobj_description((SELECT oid
#                               FROM pg_database WHERE datname = 'babase')
#                          , 'pg_database');
#
printf "COMMENT ON DATABASE babase Is 'babase-mobile database load began: %s';\n" "$today" \
  | accessdb psql
