<?php
/* Copyright (C) 2014, The Meme Factory, Inc., http://www.meme.com/
 * Copyright (C) 2007-2011, Karl O. Pinc <kop@meme.com>
 *
 *    This file is part of Babase.
 *
 *    Babase is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Upload psion data into babase.
 *
 * Karl O. Pinc <kop@meme.com>
 *
 * Remarks:
 *
 * Bugs:
 *   Slightly inefficent in that it queries the same SETUPDS row twice to get
 * the value of 2 different columns, once for each column.
 */

// Includes
require('constants.php.inc');
require('lib.php.inc');
require('lib_upload.php.inc');

//
// Constants
//

// Delimiter separating data elements in data file that's input.
define('DELIM', ',');
// Delimter separating elements of an ISO external date.
define('DATE_DELIM', '-');

// Where the row marker lives.
define('ROW_TYPE', 0);

// Markers that distinguish the different types of lines.
define('HDR_TYPE', '<HDR>');
define('PNT_TYPE', '<PNT>');
define('ADL_TYPE', '<ADL>');

//
// HDR row defines
//

// The layout of the HDR rows. (zero relative)
define('HDR_SAMPID', 1);
define('HDR_PROGID', 2);
define('HDR_SETUPID', 3);
define('HDR_PALMTOP', 4);
define('HDR_DATE', 5);
define('HDR_GROUP', 6);
define('HDR_OBSERVER', 7);
define('HDR_SNAME', 8);
define('HDR_SAMPTYPE', 9);

// Psion data structure (determined by HDR_SETUPID db lookup) that we accept.
$known_data_structures = array(BB_KNOWN_DATA_STRUCTURE);

// Psion sample type codes
define('SAMPTYPE_JUV', 'JUV');
define('SAMPTYPE_FEM', 'FEM');

// Psion header date, the offset (zero based) that references the year.
define('HDR_DATE_YEAR_START', 7);
define('HDR_DATE_YEAR_LEN', 4);
// Psion header sample id.  Zero based offsets and lengths to get data.
define('HDR_SAMPID_MONTH_START', 2);
define('HDR_SAMPID_MONTH_LEN', 2);
define('HDR_SAMPID_DAY_START', 4);
define('HDR_SAMPID_DAY_LEN', 2);
define('HDR_SAMPID_HOUR_START', 6);
define('HDR_SAMPID_HOUR_LEN', 2);
define('HDR_SAMPID_MINUTE_START', 8);
define('HDR_SAMPID_MINUTE_LEN', 2);

//
// PNT row defines.
//

// The layout of the PNT rows. (zero relative)
define('PNT_SAMPID', 1);
define('PNT_TIME', 2);
define('PNT_BLOB', 3);
define('PNT_NEIGHBOR1', 4);
define('PNT_NEIGHBOR2', 5);
define('PNT_NEIGHBOR3', 6);
define('PNT_FOODCODE', 7);

// Psion point data blob offsets, lengths, and special values.
// Special uparsed value for the whole field.
define('PNT_BLOB_OUTOFSIGHT', 'OS');
// The different pieces of the field, zero based offsets.
define('PNT_BLOB_ACTIVITY', 0);
define('PNT_BLOB_ACTIVITY_LEN', 1);
define('PNT_BLOB_POSTURE', 1);
define('PNT_BLOB_POSTURE_LEN', 1);
define('PNT_BLOB_CONTACT', 2);
define('PNT_BLOB_CONTACT_LEN', 1);
define('PNT_BLOB_SUCKLE', 3);
define('PNT_BLOB_SUCKLE_LEN', 1);

// Psion point data neighbor sname special value.
define('PNT_NOBODY_SNAME', 'XXX');

//
// ADL row defines.
//

// The layout of the ADL rows, zero relative.
define('ADL_SAMPID', 1);
define('ADL_TIME', 2);
define('ADL_KIND', 3);
define('ADL_PARTS_ACTOR', 4);
define('ADL_PARTS_ACT', 5);
define('ADL_PARTS_ACTEE', 6);

// The first field in the lump of everything that goes into allmiscs.txt.
define('ADL_TXT_START', ADL_KIND);

// ADL types that go into INTERACT/PARTS.
define('ADL_KIND_GROOMING', 'G');
define('ADL_KIND_AGONISM', 'A');
define('ADL_KIND_REQUEST', 'R');
define('ADL_KIND_APPROACH', 'P');

$adl_interact_types = array(ADL_KIND_GROOMING
                            , ADL_KIND_AGONISM
                            , ADL_KIND_REQUEST
                            , ADL_KIND_APPROACH);

//
// Functions.
// 

function require_sql(&$psion_form, &$dbconn, $line, $sql, $faulting) {
  /*
   * Execute SQL or fail.
   *
   * Input:
   *   $pison_form An instance of the Upload_form class.
   *   $dbconn  An instance of the Upload_connection class.
   *   $line  Line number in input file (zero based)
   *          May be the empty string if query is not associated with
   *          a line in the data file.
   *   $sql SQL to execute.
   *   $faulting  Boolean.  True if there were prior errors.
   *
   * Returns:
   *   TRUE if successfully exeucted, false otherwise.
   *
   * Side Effects:
   *   Yes, lots.
   *
   * Remarks:
   *   Reports an error on the form if SQL can't be executed.
   */
  if (!$dbconn->sql($sql)) {
    $psion_form->error($line + 1
                       , htmlentities($dbconn->last_error())
                       , $faulting);
    return FALSE;
  }

  return TRUE;
}

function require_sql_result(&$psion_form
                            , &$dbconn
                            , $line
                            , $sql
                            , $faulting
                            , $error
                            , $desc) {
  /*
   * Execute SQL returning a result or fail.
   *
   * Input:
   *   $pison_form An instance of the Upload_form class.
   *   $dbconn  An instance of the Upload_connection class.
   *   $line  Line number in input file (zero based)
   *          May be the empty string if query is not associated with
   *          a line in the data file.
   *   $sql SQL to execute.
   *   $faulting  Boolean.  True if there were prior errors.
   *   $error   Text to show as an error.  (Don't end in .)
   *   $desc    Further description of error, if any.
   *
   * Returns:
   *   First column of first row of result, false on error.
   *
   * Side Effects:
   *   Yes, lots.
   *
   * Remarks:
   *   Reports an error on the form if SQL can't be executed.
   */

  $r = $dbconn->sql_result($sql);

  if ($r === FALSE or $r === 0) {
    $psion_form->error($line + 1
                       , $error . ': ' .htmlentities($dbconn->last_error())
                       , $faulting
                       , $desc);
    return FALSE;
  }

  return $r;
}

function upload_psion(&$psion_form, &$dbconn) {
  /*
   * Upload the psion data and show the result.
   *
   * Input:
   *  $psion_form  An instance of the Upload_form class.
   *  $dbconn  An instance of the Upload_connection class.
   *
   * Returns: n/a
   *
   * Side Effects:
   *   Outputs html.
   *
   * Remarks:
   *  One of the main pages of the program.
   */

  global $adl_interact_types;
  global $known_data_structures;

  // Get the uploaded psion data.
  $upload_object =& new Uploaded_data($psion_form);
  if (!$upload_object->upload()) {
    return;
  }
  
  $psion = $upload_object->data;
  $lines = count($psion) - 1;

  //
  // Connect to the db.
  //
  if (!$dbconn->connect()) {
    return;
  }

  // Begin a transaction.
  if (!$dbconn->begin(TRUE)) {
    return;
  }

  $faulting = FALSE;
             
  //
  // Lookup the data structure we're expecting.
  //
  foreach ($known_data_structures as $data_structure) {
    $data_structure_code = require_sql_result($psion_form
           , $dbconn
           , ''
           , 'SELECT data_structures.data_structure'
                . ' FROM data_structures'
                . " WHERE data_structures.data_structure = '"
                              . $data_structure . "'"
                . ' LIMIT 1;'
           , FALSE
           , htmlentities($data_structure)
           , 'psionload understands this data structure; but there is no such DATA_STRUCTURES.Data_Structure value, or database failure while looking up data structure');

    if ($data_structure_code === FALSE) {
      $faulting = TRUE;
    }
  }
 
  if ($faulting) {
    // Don't have the right data structure(s).
    $dbconn->rollback('');
    return;
  }

  //
  // Go through all the data and load it.
  //
  // We load the rows when we come to them in the file.
  // The strategy used to set SAMPLES.Mins is to set it to
  // 0 initially and then increment it before we save each
  // additional point.
  //

  $header_fault = TRUE;

  for ($line = 0; $line <= $lines; $line++) {
    // Break each line into words.
    $row = explode(DELIM, rtrim($psion[$line]));
  
    // Get the row type.
    $type = $row[ROW_TYPE];
    switch ($type) {
    case HDR_TYPE:
      //
      // A header row.
      //

      // Translate the data supplied by the psion into the data expected
      // by the database.

      //
      // Get the code that goes with the setup id.
      //
      $setupid= $row[HDR_SETUPID];             
      // Lookup the setupid we're expecting.
      $setupid_code = require_sql_result($psion_form
             , $dbconn
             , $line
             , 'SELECT setupids.setupid'
                  . ' FROM setupids'
                  . " WHERE setupids.sid_string = '" . $setupid . "'"
                  . ' LIMIT 1;'
             , $faulting
             , 'Invalid setup id: ' . $setupid
             , 'No such setupid or database failure while looking up setupid');

      if ($setupid_code === FALSE) {
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      //
      // Check the data structure is one we understand.
      //
      $data_structure_code = require_sql_result($psion_form
             , $dbconn
             , $line
             , 'SELECT setupids.data_structure'
                  . ' FROM setupids'
                  . " WHERE setupids.setupid = '" . $setupid_code . "'"
                  . ' LIMIT 1;'
             , $faulting
             , 'No such setup id: ' . $setupid
             , 'The setupid does not exist or database failure while looking up setupid');
      if ($data_structure_code === FALSE) {
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      if (array_search($data_structure_code, $known_data_structures) === FALSE) {
        $psion_form->error($line + 1
                           , 'Invalid setup id: ' . $setupid
                           , $faulting
                           , 'psionload only works with files having certain data structures.  The setupid does not have a DATA_STRUCTURES.Data_Structure value recognized by psionload or database failure while looking up setupid');
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      //
      // Get the code that goes with the program id.
      //
      $programid = $row[HDR_PROGID];
      // Lookup the programid we're expecting.
      $programid_code = require_sql_result($psion_form
             , $dbconn
             , $line
             , 'SELECT programids.programid'
                  . ' FROM programids'
                  . " WHERE programids.pid_string = '" . $programid . "'"
                  . ' LIMIT 1;'
             , $faulting
             , 'Invalid program id: ' . htmlentities($programid)
             , 'No such programid or database failure while looking up programid');

      if ($programid_code === FALSE) {
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      // Check that the sample type is one we know.
      switch ($row[HDR_SAMPTYPE]) {
      case SAMPTYPE_JUV:
        $stype = BB_JUVENILE_POINT;
        break;

      case SAMPTYPE_FEM:
        $stype = BB_FEMALE_POINT;
        break;

      default:
        // A sample type we don't know about.
        $psion_form->error($line + 1
                           , 'Invalid sample type: ' . $row[HDR_SAMPTYPE]
                           , $faulting
                           , 'Sample types must either ' . SAMPTYPE_JUV
                              . ' or ' . SAMPTYPE_FEM);
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      // Get the date the sample was taken.
      $sampid = $row[HDR_SAMPID];
      // Build date in iso format.
      $date = substr($row[HDR_DATE], HDR_DATE_YEAR_START, HDR_DATE_YEAR_LEN)
              . DATE_DELIM
              . substr($sampid, HDR_SAMPID_MONTH_START, HDR_SAMPID_MONTH_LEN)
              . DATE_DELIM
              . substr($sampid, HDR_SAMPID_DAY_START, HDR_SAMPID_DAY_LEN);

      // Get the time in HH:MM (24 hour) format.
      $stime = substr($sampid, HDR_SAMPID_HOUR_START, HDR_SAMPID_HOUR_LEN)
        . ':'
        . substr($sampid, HDR_SAMPID_MINUTE_START, HDR_SAMPID_MINUTE_LEN);

      //
      // Lookup the gid based on groups.three_letter_code.
      //
      $gid = require_sql_result($psion_form
             , $dbconn
             , $line
             , 'SELECT groups.gid'
                . ' FROM groups'
                . ' WHERE three_letter_code = ' . "'" . $row[HDR_GROUP] . "';"
             , $faulting
             , htmlentities($row[HDR_GROUP])
             , 'No such group or database failure while looking up group');

      if ($gid === FALSE) {
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with next line of input file
      }

      // Save the header row.
      // (We have to look up the gid based on groups.three_letter_code.)
      if (!require_sql($psion_form
          , $dbconn
          , $line
          , 'INSERT INTO samples'
            . ' (date, stime, grp, sname, stype, mins'
            . '  , observer, palmtop, programid, setupid)'
            . ' VALUES('
                ."'$date'"                         // date
                . ", '$stime'"                     // stime
                . ", $gid"                         // grp
                . ", '" . $row[HDR_SNAME] . "'"    // sname
                . ", '$stype'"                     // stype
                . ', 0'                            // mins
                . ", '" . $row[HDR_OBSERVER] . "'" // observer
                . ", '" . $row[HDR_PALMTOP] . "'"  // palmtop
                . ", $programid_code"              // programid
                . ", $setupid_code"                // setupid
                . ');'
          , $faulting)) {
        // Can't insert the header row.
        $faulting = TRUE;
        $header_fault = TRUE;
        continue; // with the next input line
      }

      // We have successfully processed the header row.
      $header_fault = FALSE;
      // Initialize the points.min value.
      $min = 0;

      break;
        
    case PNT_TYPE:
      //
      // A point row.
      //

      if ($header_fault) {
        $psion_form->error($line + 1
                           , 'Point sample row belongs with an invalid header row'
                           , $faulting);
        $faulting = TRUE;
        continue; // with the next input line
      }

      if ($row[PNT_SAMPID] != $sampid) {
        // A bit of paranoia.  Proably a bad idea.
        $dummy = $row[PNT_SAMPID];
        $psion_form->error($line + 1
                           , "Point row sample id ($dummy) does not match"
                             . " header row sample id ($sampid)"
                           , $faulting);
        $faulting = TRUE;
        continue; // with the next input line
      }

      // Start by bumping samples.mins, it counts PNT type rows.
      // (We could pull the old value out of the db, but why?)
      $min += 1;
      if (!require_sql($psion_form
                       , $dbconn
                       , $line
                       , 'UPDATE samples'
                         . " SET mins = $min"
                         . " WHERE sid = currval('samples_sid_seq');"
                       , $faulting)) {
        $faulting = TRUE;
        continue; // with the next input line
      }

      $blob = $row[PNT_BLOB];
  
      if ($blob != PNT_BLOB_OUTOFSIGHT) {
        //
        // There is actual data in the row.
        //

        // Get the data that needs parsing.
        $activity = substr($blob, PNT_BLOB_ACTIVITY, PNT_BLOB_ACTIVITY_LEN);
        $posture = substr($blob, PNT_BLOB_POSTURE, PNT_BLOB_POSTURE_LEN);

        // Setup for neighbors.
        $neighbors = array();
        $neighbors[1] = $row[PNT_NEIGHBOR1];
        $neighbors[2] = $row[PNT_NEIGHBOR2];
        $neighbors[3] = $row[PNT_NEIGHBOR3];
        
        $ncodes = array();
        $ncodes[1] = PSION_NEIGHBOR1;
  
        // Semantics change for different sample types.
        switch($stype) {
        case BB_FEMALE_POINT:
          // Parse more out of the blob.
          $kidcontact = substr($blob, PNT_BLOB_CONTACT, PNT_BLOB_CONTACT_LEN);
          $kidsuckle = substr($blob, PNT_BLOB_SUCKLE, PNT_BLOB_SUCKLE_LEN);
  
          // Use the right neighbors.ncode values.
          $ncodes[2] = PSION_NEIGHBOR2_FEM;
          $ncodes[3] = PSION_NEIGHBOR3_FEM;

          break;
        case BB_JUVENILE_POINT:
          // Use the right neighbors.ncode values.
          $ncodes[2] = PSION_NEIGHBOR2_JUV;
          $ncodes[3] = PSION_NEIGHBOR3_JUV;

          break;
        }

        // Foodcode.
        $foodcode = $row[PNT_FOODCODE];
        if ($foodcode == '') {
          $foodcode = SQL_NULL;  

        } else {
          // Because it can be null, put the quotes SQL requires in with
          // the data.
          $foodcode = "'$foodcode'";
        }

        // Write the points row.
        if (!require_sql($psion_form
                         , $dbconn
                         , $line
                         , 'INSERT INTO point_data'
                            . ' (sid, min, ptime, activity, posture, foodcode)'
                            . ' values('
                            . "currval('samples_sid_seq')"
                            . ", '$min'"
                            . ", '" . $row[PNT_TIME] . "'"
                            . ", '$activity'"
                            . ", '$posture'"
                            . ", $foodcode);"
                         , $faulting)) {
          $faulting = TRUE;
          continue; // with the next input line
        }

        if ($stype == BB_FEMALE_POINT) {
          // Write the fpoints row.
          if (!require_sql($psion_form
                           , $dbconn
                           , $line
                           , 'INSERT INTO fpoints'
                             . ' (pntid, kidcontact, kidsuckle)'
                             . ' values('
                             . "currval('point_data_pntid_seq')"
                             . ", '$kidcontact'"
                             . ", '$kidsuckle');"
                           , $faulting)) {
            $faulting = TRUE;
            continue; // with the next input line
          }
        }

        // Write out the neighbors.
        for ($n = 1; $n <= 3; $n++) {
          $neighbor = $neighbors[$n];
          if ($neighbor != PNT_NOBODY_SNAME) {
            // It's a real neighbor.
            $ncode = $ncodes[$n];

            if ($dbconn->sql_result('SELECT 1'
                                       . ' FROM biograph'
                                       . " WHERE biograph.sname = '$neighbor'"
                                       . ' LIMIT 1')) {
              // The neighbor is on biograph.
              $neighbor_sname = "'$neighbor'";
              $neighbor_unksname = 'NULL';
            } elseif ($dbconn->sql_result('SELECT 1'
                                             . ' FROM unksnames'
                                             . ' WHERE unksnames.unksname ='
                                             . " '$neighbor'"
                                             . ' LIMIT 1')) {
              // The neighbor is one of the unknown snames.
              $neighbor_sname = 'NULL';
              $neighbor_unksname = "'$neighbor'";
            } else {
              $psion_form->error($line + 1
                                 , $neighbor 
                                   . ': is not found in BIOGRAPH or on UNKSNAMES'
                                 , $faulting);
              $faulting = TRUE;
              continue; // with the next input line
            }

            // Insert the neighbors row.
            if (!require_sql($psion_form
                             , $dbconn
                             , $line
                             , 'INSERT INTO neighbors'
                               . ' (pntid, sname, ncode, unksname)'
                               . ' values('
                               . "currval('point_data_pntid_seq')"
                               . ", $neighbor_sname"
                               . ", '$ncode'"
                               . ", $neighbor_unksname)"
                             , $faulting)) {
              $faulting = TRUE;
              continue; // with the next input line
            }
          }
        }
      }
  
      break;

    case ADL_TYPE:
      //
      // A ad-lib row.
      //

      if ($header_fault) {
        $psion_form->error($line + 1
                           , 'Ad lib row belongs with an invalid header row'
                           , $faulting);
        $faulting = TRUE;
        continue; // with the next input line
      }

      // A bit of validation.
      if ($row[ADL_SAMPID] != $sampid) {
        // A bit of paranoia.  Proably a bad idea.
        $dummy = $row[ADL_SAMPID];
        $psion_form->error($line + 1
                    , "Ad lib row sample id ($dummy) does not match"
                      . " header row sample id ($sampid)"
                    , $faulting);
        $faulting = TRUE;
        continue; // with the next input line
      }
     
      // Parse out the fields.
      $adlkind = $row[ADL_KIND];
      $time = $row[ADL_TIME];

      if (in_array($adlkind, $adl_interact_types)) {
        //
        // Put the row into the interact/parts table
        //

        // Parse the rest of the fields.
        $act = $row[ADL_PARTS_ACT];
        $actor = $row[ADL_PARTS_ACTOR];
        $actee = $row[ADL_PARTS_ACTEE];

        if ($actor == UNIDENTIFIED_SNAME and $actee == UNIDENTIFIED_SNAME) {
          $psion_form->error($line + 1
                             , 'Actor and Actee both 998'
                             , $faulting);
          $faulting = TRUE;
          continue; // with the next input line
        }

        //
        // Save the data.
        //

        // Don't go through actor_actees so that we can test for
        // unknown actor/actees and have no row in PARTS.

        // Save INTERACT_DATA.
        // (We could get the date out of the db's samples row, but why?)
        if (!require_sql($psion_form
                         , $dbconn
                         , $line
                         , 'INSERT INTO interact_data'
                           . ' (iid, sid, act, date, start, stop)'
                           . ' VALUES ('
                           . "nextval('interact_data_iid_seq')"
                           . ", currval('samples_sid_seq')"
                           . ", '$act'"
                           . ", '$date'"
                           . ", '$time'"
                           . ", '$time');"
                         , $faulting)) {
          $faulting = TRUE;
          continue; // with the next input line
        }

        // Save PARTS.
        if ($actor != UNIDENTIFIED_SNAME) {
          if (!require_sql($psion_form
                           , $dbconn
                           , $line
                           , 'INSERT INTO parts'
                             . ' (iid, role, sname)'
                             . ' values('
                             . "currval('interact_data_iid_seq')"
                             . ", '" . BB_ACTOR . "'"
                             . ", '$actor');"
                         , $faulting)) {
            $faulting = TRUE;
            continue; // with the next input line
          }
        }

        if ($actee != UNIDENTIFIED_SNAME) {
          if (!require_sql($psion_form
                           , $dbconn
                           , $line
                           , 'INSERT INTO parts'
                             . ' (iid, role, sname)'
                             . ' values('
                             . "currval('interact_data_iid_seq')"
                             . ", '" . BB_ACTEE . "'"
                             . ", '$actee');"
                         , $faulting)) {
            $faulting = TRUE;
            continue; // with the next input line
          }
        }
      } else {
        //
        // Put the row into allmiscs.
        //

        // Get the end of the line for allmiscs.txt.
        $dummy = explode(DELIM, $psion[$line], ADL_TXT_START + 1);
        $txt = $dummy[ADL_TXT_START];

        if (!require_sql($psion_form
                         , $dbconn
                         , $line
                         , 'INSERT INTO allmiscs'
                           . ' (sid, atime, txt)'
                           . ' values('
                           . "currval('samples_sid_seq')"
                           . ", '$time'"
                           . ", '$txt');"
                         , $faulting)) {
          $faulting = TRUE;
          continue; // with the next input line
        }
      }

      break;

    default:
      // Unknown type on the line.
      $psion_form->error($line + 1
                         , 'Unknown type (' . htmlentities($type) . ')'
                         , $faulting);
      $faulting = TRUE;
      continue; // with the next input line
    }
  }

  //
  // Commit the transaction (or allow user to try again.
  //

  // DEBUG
  //$dbconn->rollback('<p>Transaction deliberately rolled back.</p>');
  //return;

  if ($faulting) {
    $dbconn->rollback('');
    return;
  }

  if (!$dbconn->complete()) {
    return;
  }

  //
  // Everything seems to have worked.
  // Allow the user to do it again.
  //
  $upload_object->success();  // Remember uploaded file.
  $psion_form->show('<p class="success">PSION successfully updated!</p>');

  return;
}

//
// Body of code
//

maintain_login_session();

show_html_start('psionload -- Upload PSION data into Babase');
?>

<h1>PSION Upload</h1>

<?php

//
// Get the form data.
//

// $action
// What page we're in.
if (array_key_exists('action', $_REQUEST)) {
  $action = $_REQUEST['action'];
} else {
  $action = ACTION_NONE;
}

// $db
$db = get_session_var('db', TRUE);

// $user
$user = get_session_var('user', TRUE);

// $password
$password = get_session_var('password', FALSE);

// Instantiate the upload form.
$psion_form =& new Upload_form($db, $user, $password, 'psion');

//
// Process input values
//

//
// Dispatch on action
//
switch ($action) {
  case ACTION_NONE:
    $psion_form->show();
    break;

  case ACTION_UPLOAD:
    // Make a database connection.
    $dbconn =& new Upload_connection($db
                                           , $user
                                           , $password
                                           , $psion_form);
    // Do the deed.
    upload_psion($psion_form, $dbconn);
    break;
}

// Finish the stock page.

show_html_finish();

return;  // End of program.
