<?php
/* Copyright (C) 2006-2009, Karl O. Pinc
 *
 *    This file is part of Babase.
 *
 *    Babase is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Upload data into the census table.
 *
 * Karl O. Pinc <kop@meme.com>
 *
 * Remarks:
 *  Does not do anything to protect against SQL injection attacks,
 * but this does not matter as anybody with a login has restricted
 * permissions anyhow.
 */

// Includes
require('constants.php.inc');
require('lib.php.inc');
require('lib_upload.php.inc');

//
// Constants
//

// Delimiter separating data elements in data file that's input.
define('DELIM', "\t");

//
// Functions.
// 


function upload_census(&$census_form, &$dbconn) {
  /*
   * Upload the census data and show the result.
   *
   * Input:
   *  $census_form  An instance of the Upload_form class.
   *  $dbconn  An instance of the Upload_connection class.
   *
   * Returns: n/a
   *
   * Side Effects:
   *   Outputs html.
   *
   * Remarks:
   *  One of the main pages of the program.
   */

  // Get the uploaded census data.
  $upload_object =& new Uploaded_data($census_form);
  if (!$upload_object->upload()) {
    return;
  }
  
  $census = $upload_object->data;

  //
  // Get the group and the dates.
  //
  $row = explode(DELIM, rtrim($census[0]));
  
  $group = trim($row[0]);

  $data_rows = count($row) - 1;
  if ($data_rows < 1) {
    // Let the user try again.
    $census_form->show('<p class="error">No dates supplied.</p>');
    return;
  }
    

  for ($i = 1; $i <= $data_rows; $i++) {
    $dates[$i] = trim($row[$i]);
  }

  //
  // Connect to the db.
  //
  if (!$dbconn->connect()) {
    return;
  }

  // Begin a transaction.
  if (!$dbconn->begin(TRUE)) {
    return;
  }

  //
  // Ignore trailing blank lines.
  //
  $lines = count($census) - 1;
  for ($l = $lines; $l > 0; $l--) {
    if (trim($census[$l]) == '') {
      $lines--;
    } else {
      break;
    }
  }

  //
  // Go through all the data and load it.
  //

  $faulting = FALSE;
  for ($line = 1; $line <= $lines; $line++) {
    // Break each line into words.
    $row = explode(DELIM, rtrim($census[$line]));
  
    // Get the sname.
    $sname = trim($row[0]);
    if ($sname == '') {
      // Skip this row.
      // Let the user try again.
      $census_form->error($line + 1
                          , 'No SNAME (Empty line?)'
                          , $faulting);

      $faulting = TRUE;
      continue;  // to the next row of the input file
    }

    // Validate the amount of data on the line.
    $this_data_rows = count($row) - 1;
    if ($this_data_rows > $data_rows) {
      // There's more data than we've got dates.
      // Tell the user of the problem.
      // Skip this row.
      // Let the user try again.
      $census_form->error($line + 1
                          , 'Too many columns'
                          , $faulting
                          , "More columns ($this_data_rows)"
                            . " than dates ($data_rows)");
      $faulting = TRUE;
      continue; // to the next row of the input file
    }

    // Go through each datum on the line.
    // (Go through backwards so we don't constantantly update the statdate,
    // which causes a _lot_ of checking to be done.)
    //    for ($i = 1; $i <= $data_rows; $i++) {
    for ($i = $data_rows; $i >= 1; $i--) {
      if ($i > $this_data_rows) {
        $code = '';
      } else {
        $code = trim($row[$i]);  // Clean up extra spaces
      }
      switch ($code) {
      case '0':
        // Insert an absence.
        $status = 'A';
        break;
        
      case '':
        // Insert a census.
        $status = 'C';
        break;

      case 'N':
        // No data, do nothing.
        $status = '';
        break;

      default:
        // Bad data.
        $r = $i + 1;
        // Let the user try again.
        $census_form->error($line + 1
                            , 'Bad data'
                            , $faulting
                            , "($code) in column $r.");
        $faulting = TRUE;
        continue; // to the next datum on the line
      }

      if ($status != '') {
        // Go ahead and insert the census data.
        $date = $dates[$i];
        if (!$dbconn->sql('INSERT INTO'
                                . ' census (sname, date, grp, status, cen)'
                                . " values('$sname'"
                                      . ", '$date'"
                                      . ", '$group'"
                                      . ", '$status'"
                                      . ", TRUE);")) {
          // Give up.
          // Let the user try again.
          $census_form->error($line + 1
                              , htmlentities($dbconn->last_error())
                              , $faulting);
          $faulting = TRUE;
          continue; // to the next datum on the line
        }
      }
    }
  }

  //
  // Commit the transaction (or allow user to try again.
  //

  //$dbconn->rollback('<p>Transaction deliberately rolled back.</p>');
  //return;

  if ($faulting) {
    $dbconn->rollback('');
    return;
  }

  if (!$dbconn->complete()) {
    return;
  }

  //
  // Everything seems to have worked.
  // Allow the user to do it again.
  //
  $upload_object->success();  // Remember uploaded file.
  $census_form->show('<p class="success">CENSUS successfully updated!</p>');

  return;
}

//
// Body of code
//

maintain_login_session();

show_html_start('upcen -- Update the CENSUS table');

  ?>

<h1>Update CENSUS</h1>

<?php

//
// Get the form data.
//

// $action
// What page we're in.
if (array_key_exists('action', $_REQUEST)) {
  $action = $_REQUEST['action'];
} else {
  $action = ACTION_NONE;
}

// $db
$db = get_session_var('db', TRUE);

// $user
$user = get_session_var('user', TRUE);

// $password
$password = get_session_var('password', FALSE);

//error_log('p: ' . crypt($password, substr('12345678901234',0, CRYPT_SALT_LENGTH)) . ': ' .  ($password === ''));

// Instantiate the upload form.
$census_form =& new Upload_form($db, $user, $password, 'census');

//
// Process input values
//

//
// Dispatch on action
//
switch ($action) {
  case ACTION_NONE:
    $census_form->show();
    break;

  case ACTION_UPLOAD:
    // Make a database connection.
    $dbconn =& new Upload_connection($db
                                           , $user
                                           , $password
                                           , $census_form);
    // Do the deed.
    upload_census($census_form, $dbconn);
    break;
}

// Finish the stock page.

show_html_finish();

return;  // End of program.
