<?php
/* Copyright (C) 2012 The Meme Factory, Inc., http://www.meme.com/
 * Copyright (C) 2006-2009, Karl O. Pinc
 *
 *    This file is part of Babase.
 *
 *    Babase is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Upload tab delimited data into any table or view.
 *
 * Karl O. Pinc <kop@meme.com>
 *
 * Remarks:
 *  Does not do anything to protect against SQL injection attacks,
 * but this does not matter as anybody with a login has restricted
 * permissions anyhow.
 *
 * The first row has the column names.
 *
 * Note that this is not very efficent.
 *
 * Note also that this program puts single quotes around all the data,
 * which might result in problems with type coercion.
 */

// Includes
require('constants.php.inc');
require('lib.php.inc');
require('lib_upload.php.inc');

//
// Constants
//

// Delimiter separating data elements in data file that's input.
define('DELIM', "\t");

//
// Functions.
// 


function upload_any(&$upload_form, &$dbconn, $tablename, $upload_null, $null_rep) {
  /*
   * Upload the data and show the result.
   *
   * Input:
   *  $upload_form  An instance of the Upload_form class.
   *  $dbconn  An instance of the Upload_connection class.
   *  $tablename  The name of the table or view to load.
   *  $upload_null  Empty string "" when NULL should not be put in DB
   *  $null_rep     When $upload_null is not "", value in input data
   *                to be made NULL when put into the db.
   *
   * Returns: n/a
   *
   * Side Effects:
   *   Outputs html.
   *
   * Remarks:
   *  One of the main pages of the program.
   */

  // Get the uploaded data.
  $upload_object = new Uploaded_data($upload_form);
  if (!$upload_object->upload()) {
    return;
  }
  
  $some_data = $upload_object->data;

  // Validate table name
  if (trim($tablename) == '') {
    $upload_form->show('<p class="error">No table name supplied.</p>');
    return;
  }

  //
  // Make an INSERT statement we'll use again and again.
  //
  $insert = "INSERT INTO $tablename (";

  //
  // Get the colum names.
  //
  $columns = explode(DELIM, rtrim($some_data[0]));

  $data_columns = count($columns);
  if ($data_columns < 1) {
    // Let the user try again.
    $upload_form->show('<p class="error">No column names supplied.</p>');
    return;
  }
    

  // Add the columns into the insert statement.
  // (Quote column names in case they've got a space or something wierd.)
  $sep = '';
  for ($i = 0; $i < $data_columns; $i++) {
    $insert .= $sep . '"' . trim($columns[$i]) . '"';
    $sep = ', ';
  }

  // Finish the INPUT statement, except for the VALUES.
  $insert .= ') ';

  //
  // Connect to the db.
  //
  if (!$dbconn->connect()) {
    return;
  }

  // Begin a (sloppy) transaction.
  if (!$dbconn->begin(TRUE)) {
    return;
  }

  //
  // Ignore trailing blank lines.
  //
  $lines = count($some_data) - 1;
  for ($l = $lines; $l > 0; $l--) {
    if (trim($some_data[$l]) == '') {
      $lines--;
    } else {
      break;
    }
  }

  //
  // Go through all the data and load it.
  //
  $faulting = FALSE;

  for ($line = 1; $line <= $lines; $line++) {
    // Break each line into columns.
    $row = explode(DELIM, rtrim($some_data[$line]));

    // Validate the amount of data on the line.
    $this_data_columns = count($row);
    if ($this_data_columns > $data_columns) {
      // There's more data than we've got dates.
      // Tell the user of the problem.
      // Give up.
      // Let the user try again.
      $upload_form->error($line + 1
                          , "Too many columns,"
                            . " more data values ($this_data_columns)"
                            . " than column names ($data_columns)"
                          , $faulting);
      $faulting = TRUE;
      continue;  // to iterate over input file rows
    }

    // Go through each datum on the line and add to INSERT statement.
    $this_insert = $insert;

    $sep = 'VALUES(';
    for ($i = 0; $i < $data_columns; $i++) {
      if ($i >= $this_data_columns) {
        $datum = '';
      } else {
        $datum = $row[$i];
      }

      if ($upload_null != '' && $datum == $null_rep) {
	// Need to replace the input data with NULL.
	$datum = 'NULL';
      } else {
	// Quote.  Double all single quotes in the process.
	// (Don't clean up extra spaces.)
	$datum = "'" . str_replace("'", "''", $datum) . "'";
      }
     
      $this_insert .= $sep . $datum;
      $sep = ', ';
    }

    // Finish the insert statement.
    $this_insert .= ')';

    // Insert the row.

    if (!$dbconn->sql($this_insert)) {
      // Skip this row.
      // Let the user try again.
      $upload_form->error($line + 1
                          , 'PostgreSQL will not accept your data, it reports the error: ' 
                            . htmlentities($dbconn->last_error())
                          , $faulting
                          , 'The sql statement is: ' 
                            . htmlentities($this_insert));
      $faulting = TRUE;
      continue;  // to the next row of the input file
    }
  }

  //
  // Commit the transaction (or allow user to try again.
  //

  //$dbconn->rollback('<p>Transaction deliberately rolled back.</p>');
  //return;

  if ($faulting) {
    $dbconn->rollback('');
    return;
  }

  if (!$dbconn->complete()) {
    return;
  }

  //
  // Everything seems to have worked.
  // Allow the user to do it again.
  //
  $upload_object->success();  // Remember uploaded file.
  $upload_form->show('<p><em class="success">'
		     . $tablename
		     . ' successfully updated</em> with '
		     . $lines
		     . ' rows of data<em class="success">!</em></p>');

  return;
}

function make_tablename($value) {
  /* Make some html that inputs the name of the table
   *
   * Input:
   *  $value  The (entity escaped) name of the table or view.
   *
   * Returns:
   *   A html <input> element.
   *
   * Side Effects: none
   */

  return '<input id="tablename_id" name="tablename" tabindex="4" type="text" size="30" value="' . $value . '" />';
}

//
// Body of code
//

maintain_login_session();

show_html_start('upload -- Upload to any table or view');

  ?>

<h1>Upload to any table or view</h1>

<?php

//
// Get the form data.
//

// $action
// What page we're in.
if (array_key_exists('action', $_REQUEST)) {
  $action = $_REQUEST['action'];
} else {
  $action = ACTION_NONE;
}

// $db
$db = get_session_var('db', TRUE);

// $user
$user = get_session_var('user', TRUE);

// $password
$password = get_session_var('password', FALSE);

// $tablename
$tablename = get_session_var('tablename', TRUE);

// $upload_null_id
$upload_null = get_session_var('upload_null', TRUE);

// $null_rep
$null_rep = get_session_var('null_rep', FALSE);

//error_log('p: ' . crypt($password, substr('12345678901234',0, CRYPT_SALT_LENGTH)) . ': ' .  ($password === ''));

// Instantiate the upload form.
$upload_form =& new Upload_form($db, $user, $password, 'tab delimited data');

$upload_form->extra_input
   = array('4'
	   => array('label'
		    => '<label for="tablename_id">Table or view:</label>'
		    , 'data'
		    => make_tablename($tablename))
	   , '5'
	   => array('label'
		    => '<label for="upload_null_id">Upload NULL Values: </label>'
		    , 'data'
		    => '<input name="upload_null" id="upload_null_id" tabindex="5" type="checkbox" ' . ($upload_null == '' ? '' : 'checked="checked" ') . '/>')
	   , '6'
	   => array('label'
		    => '<label for="null_rep_id">NULL Representation: </label>'
		    , 'data'
		    => '<input name="null_rep" id="null_rep_id" tabindex="6" type="text" size="10" value="' . htmlentities($null_rep, ENT_QUOTES) . '" />'));

//
// Process input values
//

//
// Dispatch on action
//
switch ($action) {
  case ACTION_NONE:
    $upload_form->show();
    break;

  case ACTION_UPLOAD:
    // Make a database connection.
    $dbconn =& new Upload_connection($db
                                           , $user
                                           , $password
                                           , $upload_form);
    // Do the deed.
    upload_any($upload_form, $dbconn, $tablename, $upload_null, $null_rep);
    break;
}

// Finish the stock page.

show_html_finish();

return;  // End of program.
