<?php
/* Copyright (C) 2006-2011, Karl O. Pinc
 *
 *    This file is part of Babase.
 *
 *    Babase is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Upload multi-party interaction data.
 *
 * Karl O. Pinc <kop@meme.com>
 *
 * Remarks:
 *  Does not do anything to protect against SQL injection attacks,
 * but this does not matter as anybody with a login has restricted
 * permissions anyhow.
 *
 * Bugs:
 *  Hardcoded mapping of some uploaded acts to MPI_DATA.MPIAct values.
 */

// Includes
require('constants.php.inc');
require('lib.php.inc');
require('lib_upload.php.inc');

//
// Constants
//

// Delimiter separating data elements in data file that's input.
define('DELIM', "\t");

// SQL stuff

// How a null value is encoded.
define('NULL', '');

// Use dollar quoting to avoid escape and single quote issues.
define('DQ', '$$');

// Value of context column when there's a consort context (checked).
define('CONSORT_VALUE', 'CONSORT');
define('CONSORT_CONTEXT_TYPE', 'C');

// Act value that allows for multiple actors
define('VS_ACT', 'VS');
// Delimiter in the actor or actee columns when there's multiple
// participants.
define('PART_DELIM', ',');

// Translation table between uploaded act values and
// those in MPIACTS.  Values not found in the translation
// table are used as-is.
$act_map = array('+' => 'AH'
                 , '?' => 'RE'
                 , 'P' => 'PH');

// Index values into the matched subexpressions of a parsed
// consort value. (0 is the entire matched expression.)
define('CONSORT_MALE1', 1);
define('CONSORT_FEMALE', 2);
define('CONSORT_MALE2', 3);
define('CONSORT_KIND', 4);

// Columns in the input file.
define('COLUID_COL', 0); // examined, but not saved anywhere
define('COALID_COL', 1); // examined, but not saved anywhere
define('GROUP_COL', 2); // checked against "header" rows
define('DATE_COL', 3);  // checked against "header" rows
define('ACTOR_COL', 4);
define('ACT_COL', 5);
define('ACTEE_COL', 6);
define('OUTCOME_COL', 7); // MPI_DATA.Helped and MPI_DATA.Passive
define('PASSIVE_COL', 8); // More actions that occur when PH is given
                          //   The action are help by the requestee to the
                          //   requestor.  Ignore this for now because
                          //   it's unclear who's giving the help
                          //   and because multiple instances of these
                          //   actions count as a single instance of help.
// (We could put this in with the same seq number as the request for help,
// and list the requestee as the helper, and add a new MPI_DATA.Detail
// column that's a boolen set to true for this data.  But it may be
// confusing.)
define('CONTEXT_COL', 9); // context text, must be null in non-header rows
define('CONSORT_COL', 10); // consort information
define('CONTEXT_TYPE_COL', 11); // context type code
define('LAST_COL', 11);  // THIS MUST BE THE LAST INDEX

//
// Functions.
// 

function unk_consort(&$dbconn, $sname) {
  /*
   * Transform a sname, or one of the unknown codes, into an SQL literal.
   *
   * Input:
   *  $dbconn   An instance of the Upload_connection class.
   *  $sname    The sname or unknown code.
   *
   * Returns:
   *  The (quoted) string to be used in the SQL statement,
   *  or the SQL NULL representation if the sname is one of the
   *  codes for unknown individuals.
   *
   * Side Effects: none
   *
   * Bugs:
   *  Returns the input sname as data to be inserted if there's a
   *  problem querying the database.  This does not matter because
   *  if the sname is not valid the database will reject it.
   */

  $unk = $dbconn->sql_result('SELECT partunks.unksname'
                             . ' FROM partunks'
                             . ' WHERE partunks.unksname = '
                             . DQ . $sname . DQ
                             . ';');
  if ($unk === 0 or $unk === FALSE) {
    $sql_literal = DQ . $sname . DQ;
  } else {
    $sql_literal = 'NULL';
  }

  return $sql_literal;
}

function upload_mpi(&$mpi_form, &$dbconn) {
  /*
   * Upload the census data and show the result.
   *
   * Input:
   *  $mpi_form  An instance of the Upload_form class.
   *  $dbconn  An instance of the Upload_connection class.
   *
   * Returns: n/a
   *
   * Side Effects:
   *   Outputs html.
   *
   * Remarks:
   *  One of the main pages of the program.
   */

  global $act_map;

  // Get the uploaded census data.
  $upload_object =& new Uploaded_data($mpi_form);
  if (!$upload_object->upload()) {
    return;
  }
  
  $mpi = $upload_object->data;

  //
  // Ignore trailing blank lines.
  //
  $lines = count($mpi) - 1;
  for ($l = $lines; $l > 0; $l--) {
    if (trim($mpi[$l]) == '') {
      $lines--;
    } else {
      break;
    }
  }

  //
  // Ignore leading blank lines.
  //
  $firstline = -1;
  for ($l = 0; $l <= $lines; $l++) {
    if (trim($mpi[$l]) != '') {
      $firstline = $l;
      break;
    }
  }

  if ($firstline == -1) {
    $mpi_form->show('<p class="error">No data in uploaded file</p>');
    return;
  }
  
  //
  // Check the column headings.
  //
  $headers = array('mid'
                   , 'coal_id'
                   , 'grp'
                   , 'date'
                   , 'actor'
                   , 'agg_act'
                   , 'recip'
                   , 'outcome'
                   , 'form_passive_aid'
                   , 'context'
                   , 'consort'
                   , 'context_type');
  $row = explode(DELIM, rtrim($mpi[$firstline]));

  if (count($row) != count($headers)) {
    $mpi_form->show('<p class="error">Bad column headings: incorrect number of columns</p>');
    return;
  }

  foreach ($row as $k => $v) {
    if ($v != $headers[$k]) {
      $mpi_form->show('<p class="error">Bad column heading: "'
                      . htmlentities($headers[$k], ENT_QUOTES)
                      . '" should be "'
                      . $v
                      . '"</p>');
      return;
    }
  }

  $firstline += 1;

  //
  // Connect to the db.
  //
  if (!$dbconn->connect()) {
    return;
  }

  // Begin a transaction.
  if (!$dbconn->begin(TRUE)) {
    return;
  }
  $faulting = FALSE;

  //
  // Go through all the data and load it.
  // Iterate over the lines of the uploaded file.
  //
  $last_coluid = '';
  $last_coalid = '';
  $used_coalids = array();

  for ($line = $firstline; $line <= $lines; $line++) {

    //
    // Parse the line.
    //

    // Ensure that we can dollar quote everything in the row.
    if (strpos($mpi[$line], DQ) != 0) {
      $mpi_form->error($line + 1
                       , DQ . ' character sequence found on the line'
                       , $faulting
                       , 'These characters cannot be used in input to this program');
      $faulting = TRUE;
      continue; // to the next line of the input file
    }

    // Break each line into words.
    $row = explode(DELIM, rtrim($mpi[$line]));

    for ($i = count($row); $i <= LAST_COL; $i++) {
      // Make sure that there are as many columns as we might reference.
      $row[$i] = '';
    }
  
    //
    // Basic validation.
    //

    // What sort of line is it?
    // Validate the column and coalition ids (only to help
    // detect data entry errors).

    // coalid basic syntax check
    $coalid = trim($row[COALID_COL]);
    if ($coalid == '') {
      $mpi_form->error($line + 1
                       , 'The coal_id column must contain a value'
                       , $faulting);
      $faulting = TRUE;
      continue; // to the next line of the input file.
    }

    // coluid basic syntax check
    $coluid = trim($row[COLUID_COL]);
    if ($coluid == '') {
      $mpi_form->error($line + 1
                       , 'The mid column must contain a value'
                       , $faulting);
      $faulting = TRUE;
      continue; // to the next line of the input file.
    }

    //
    // Get the columns we need for both a new and continuing MPI.
    //
    $actor = trim($row[ACTOR_COL]);

    $act = trim($row[ACT_COL]);
    if (array_key_exists($act, $act_map) === TRUE) {
      $act = $act_map[$act];
    }

    $context = trim($row[CONTEXT_COL]);
    $context_type = trim($row[CONTEXT_TYPE_COL]);

    $consort = trim($row[CONSORT_COL]);

    //
    // Separate processing paths for the first line of the MPI
    // and the rest.
    //

    if ($last_coalid != $coalid) {
      //
      // This is a new MPI.
      //
      $seq = 1;

      //
      // More data validation working toward saving a MPI row.
      //

      // Validate coluid.
      if ($last_coluid != '' and !($last_coluid < $coluid)) {
        $mpi_form->error($line + 1
                         , 'The mid column must increase with each row'
                         , $faulting
                         , 'mid is '
                           . htmlentities($coluid, ENT_QUOTES)
                           . ' it should be greater than '
                           . htmlentities($last_coluid, ENT_QUOTES));
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Validate coalid
      if (array_key_exists($coalid, $used_coalids)) {
        $mpi_form->error($line + 1
                         , 'Invalid coal_id: '
                           . htmlentities($coalid, ENT_QUOTES)
                         , $faulting
                         , 'coal_id value already used');
        $faulting = TRUE;
        continue; // to the next line of the input file.
      }

      // Keep state on new MPI.
      $last_coalid = $coalid;
      $used_coalids[] = $coalid;
      $group = trim($row[GROUP_COL]);
      $date = trim($row[DATE_COL]);

      // Group basic syntax check.
      if ($group == '') {
        $mpi_form->error($line + 1
                         , 'No group supplied'
                         , $faulting);
        $faulting = TRUE;
        continue; // with next line of the input file.
      }

      // Validate and transform the context value.

      if ($context_type == CONSORT_CONTEXT_TYPE) {
        if ($context == CONSORT_VALUE) {
          // Transform the context from something good for data entry to something
          // good for the database.
          $context = NULL;
        } else {
          $mpi_form->error($line + 1
                           , 'Illegal context: '
                             . htmlentities($context, ENT_QUOTES)
                           , $faulting
                           , 'context must be '
                             . CONSORT_VALUE
                             . ' when the context type is '
                             . CONSORT_CONTEXT_TYPE);
          $faulting = TRUE;
          continue; // to the next line of the input file.
        }
      }

      // Represent the context as an SQL value.
      if ($context == NULL) {
        $context_value = 'NULL';
      } else {
        $context_value = DQ . $context . DQ;
      }
      
      //
      // Save the MPI row.
      //
      if (!$dbconn->sql('INSERT INTO MPIS (date, context_type, context) '
                        . 'values('
                        . DQ . $date . DQ
                        . ', '
                        . DQ . $context_type . DQ
                        . ', '
                        . $context_value
                        . ');')) {
        $mpi_form->error($line + 1
                         ,  htmlentities($dbconn->last_error())
                         , $faulting);
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Keep the mpiid.
      $mpiid = $dbconn->sql_result("SELECT currval('mpis_mpiid_seq');");
      if ($mpiid === FALSE or $mpiid === 0) {
        $mpi_form->error($line + 1
                         , htmlentities($dbconn->last_error())
                         , $faulting
                         , 'Internal error, likely a bug in upmpi');
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      //
      // The CONSORTS row.
      //

      if ($consort != '') {
        //
        // Parse the consorts value.
        //

        // Use the insane preg match with references to parse,
        // because matching is an easy way to test syntax
        // and because Posix regexps are depreciated.
        $parsed[0] = array();  // Silly php (4.3.9) wants initialization.
        if (preg_match_all('/^([^ ]+) +WITH *([^ ]*);([^ ]+) +(GET|KEEP) *\2$/S'
                           , $consort
                           , $parsed
                           , PREG_SET_ORDER) != 1) {
          $mpi_form->error($line + 1
                           , 'Invalid consort: '
                             . htmlentities($consort, ENT_QUOTES)
                           , $faulting
                           , 'consort values must match one of the 2 patterns'
                             . ' described in the documentation');
          $faulting = TRUE;
          continue; // with the next row of the input file
        }
        
        $match_results = $parsed[0]; // Only one match is possible.
        $male1 = $match_results[CONSORT_MALE1];
        $female = $match_results[CONSORT_FEMALE];
        $male2 = $match_results[CONSORT_MALE2];
        $match_kind = $match_results[CONSORT_KIND];

        if ($match_kind == 'KEEP' and $male1 != $male2) {
          // The pattern does not test that the male who kept the female is
          // the male that started with the female, and that's what happened.
          $mpi_form->error($line + 1
                           , 'Invalid consort results: '
                             . htmlentities($consort, ENT_QUOTES)
                           , $faulting
                           , 'The male who kept the female is not the male'
                             . ' that started with the female');
          $faulting = TRUE;
          continue; // to the next line of the input file
        }

        // Transform the consort snames if they are unknown.
        $female_value = unk_consort($dbconn, $female);
        $male1_value = unk_consort($dbconn, $male1);
        $male2_value = unk_consort($dbconn, $male2);

        //
        // Save the consorts row,
        //
        if (!$dbconn->sql('INSERT INTO consorts (mpiid, female, had, got) '
                          . "values("
                          . $mpiid
                          . ', '
                          . $female_value
                          . ', '
                          . $male1_value
                          . ', '
                          . $male2_value
                          . ');')) {
          $mpi_form->error($line + 1
                           ,  htmlentities($dbconn->last_error())
                           , $faulting);
          $faulting = TRUE;
          continue; // to the next line of the input file
        }
      }
    } else {
      //
      // Regular row within an existing MPI.
      //
      $seq += 1;

      //
      // Validation done for interactions other than the first.
      //

      // Validate coluid
      if ($coluid < $last_coluid + 1) {
        $mpi_form->error($line + 1
                         , 'mid column out of sequence'
                         , $faulting
                         , 'is '
                         . htmlentities($coluid, ENT_QUOTES)
                         . ' : should be at least: '
                         . htmlentities($last_coluid + 1));
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Validate group
      $this_group = trim($row[GROUP_COL]);
      if ($this_group != $group) {
        $mpi_form->error($line + 1
                         , 'Group cannot change within a multi-party interaction'
                         , $faulting
                         , 'the group is '
                         . htmlentities($this_group, ENT_QUOTES)
                         . ' : it should be: '
                         . htmlentities($group, ENT_QUOTES));
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Validate date
      $this_date = trim($row[DATE_COL]);
      if ($this_date != $date) {
        $mpi_form->error($line + 1
                         , 'Date cannot change within a multi-party interaction'
                         , $faulting
                         , 'the date is '
                         . htmlentities($this_date, ENT_QUOTES)
                         . ' : it should be: '
                         . htmlentities($date, ENT_QUOTES));
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Validate context
      if ($context != NULL) {
        $mpi_form->error($line + 1
                         , 'Bad context: ' . htmlentities($context, ENT_QUOTES)
                         , $faulting
                         , 'Context must be empty for lines other then the first in the multi-party transaction');
        $faulting = TRUE;
        continue; // to the next row of the input file
      }

      // Validate context type
      if ($context_type != NULL) {
        $mpi_form->error($line + 1
                         , 'Bad context type: '
                           . htmlentities($context_type, ENT_QUOTES)
                         , $faulting
                         , 'context_type must be empty for lines other then the first in the multi-party transaction');
        $faulting = TRUE;
        continue; // to the next line of the input file
      }

      // Validate consort
      if ($consort != NULL) {
        $mpi_form->error($line + 1
                         , 'Bad consort: '
                           . htmlentities($consort, ENT_QUOTES)
                         , $faulting
                         , 'consort must be empty for lines other then the first in the multi-party transaction');
        $faulting = TRUE;
        continue; // to the next line of the input file
      }
    }

    //
    // The act has been checked, get the rest of the MPI_DATA row.
    //

    // Interpret the outcome
    $outcome = trim($row[OUTCOME_COL]);
    switch ($outcome) {
    case NULL:
      $helped = 'NULL';
      $active = 'NULL';
      break;

    case 'SUCC':
      $helped = 'TRUE';
      $active = 'TRUE';
      break;

    case 'FAIL':
      $helped = 'FALSE';
      $active = 'FALSE';
      break;
      
    case 'PASS':
      $helped = 'TRUE';
      $active = 'FALSE';
      break;

    default:
      $mpi_form->error($line + 1
                       , 'Bad outcome: '
                         . htmlentities($outcome, ENT_QUOTES)
                       , $faulting
                       , 'Must be SUCC, PASS, or FAIL');
      $faulting = TRUE;
      continue; // to the next line of the input file
    }

    //
    // Save the interactions.
    // There can be multiple actors per uploaded line.
    //

    $actors = explode(PART_DELIM, $actor);
    $actees = explode(PART_DELIM, trim($row[ACTEE_COL]));

    foreach ($actees as $this_actee) {

      // See if the actee is unknown.
      $this_actee = trim($this_actee);
      $actee_value = DQ . $this_actee . DQ;
      $unk = $dbconn->sql_result('SELECT partunks.unksname'
                                 . ' FROM partunks'
                                 . " WHERE partunks.unksname = $actee_value;");
      if ($unk === 0 or $unk === FALSE) {
        // Pretend it's a real sname
        // (Or this could mean an error in running the query, but we'd 
        // get another one soon enough because we'd try to insert a bad
        // database value so don't worry about it.)
        $actee_sname = $actee_value;
        $actee_unksname = 'NULL';
      } else {
        // An unknown individual code.
        $actee_sname = 'NULL';
        $actee_unksname = $actee_value;
      }
   
      foreach ($actors as $this_actor) {

        //
        // Save the MPI_DATA row
        //
        if (!$dbconn->sql('INSERT INTO mpi_data (mpiid, seq, mpiact, helped, active) '
                          . "values("
                          . $mpiid
                          . ', '
                          . $seq
                          . ', '
                          . DQ . $act . DQ
                          . ', '
                          . $helped
                          . ', '
                          . $active
                          . ');')) {
          $mpi_form->error($line + 1
                           ,  htmlentities($dbconn->last_error())
                           , $faulting);
          $faulting = TRUE;
          continue; // to the next line of the input file
        }
        
        // Keep the mpidid.
        $mpidid = $dbconn->sql_result("SELECT currval('mpi_data_mpidid_seq');");
        if ($mpidid === FALSE or $mpidid === 0) {
          $mpi_form->error($line + 1
                           , htmlentities($dbconn->last_error())
                           , $faulting
                           , 'Internal error, likely a bug in upmpi');
          $faulting = TRUE;
          continue; // to the next line of the input file
        }

        //
        // First save the actor
        //

        $this_actor = trim($this_actor);
        $actor_value = DQ . $this_actor . DQ;

        // See if the actor is unknown.
        // We could cache the result and not have to query each iteration
        // but the case happens so infreqently that the complication is
        // not worth the trouble.
        $unk = $dbconn->sql_result('SELECT partunks.unksname'
                                   . ' FROM partunks'
                                   . " WHERE partunks.unksname = $actor_value;");
        if ($unk === 0 or $unk === FALSE) {
          // Pretend it's a real sname
          // (Or this could mean an error in running the query, but we'd 
          // get another one soon enough because we'd try to insert a bad
          // database value so don't worry about it.)
          $actor_sname = $actor_value;
          $actor_unksname = 'NULL';
        } else {
          // An unknown individual code.
          $actor_sname = 'NULL';
          $actor_unksname = $actor_value;
        }

        // Save the actor
        if (!$dbconn->sql('INSERT INTO mpi_parts (mpidid, sname, unksname, role) '
                          . "values("
                          . $mpidid
                          . ', '
                          . $actor_sname
                          . ', '
                          . $actor_unksname
                          . ", '" . BB_ACTOR . "');")) {
          $mpi_form->error($line + 1
                           ,  htmlentities($dbconn->last_error())
                           , $faulting);
          $faulting = TRUE;
          continue; // with the next actor on the line
        }

        //
        // Then save the actee.
        // PHP is lame enough that I don't feel it's worth creating
        // an abstraction to update MPI_PARTS when there's only an
        // actor and an actee. (Plus we can be marginally more efficent
        // by removing loop invarents.)
        //

        // Save the actee
        if (!$dbconn->sql('INSERT INTO mpi_parts (mpidid, sname, unksname, role) '
                          . "values("
                          . $mpidid
                          . ', '
                          . $actee_sname
                          . ', '
                          . $actee_unksname
                          . ", '" . BB_ACTEE . "');")) {
          $mpi_form->error($line + 1
                           ,  htmlentities($dbconn->last_error())
                           , $faulting);
          $faulting = TRUE;
          continue; // with the next actor on the line
        }
      }
    }

    // Prep for next iteration.
    $last_coluid = $coluid;
  }

  //
  // Commit the transaction (or allow user to try again.
  //

  //$dbconn->rollback('<p>Transaction deliberately rolled back.</p>');
  //return;

  if ($faulting) {
    $dbconn->rollback('');
    return;
  }

  if (!$dbconn->complete()) {
    return;
  }

  //
  // Everything seems to have worked.
  // Allow the user to do it again.
  //
  $upload_object->success();  // Remember uploaded file.
  $mpi_form->show('<p class="success">MPI data successfully updated!</p>');

  return;
}

//
// Body of code
//

maintain_login_session();

show_html_start('upmpi -- Update the MPI data');

  ?>

<h1>Update MPI</h1>

<?php

//
// Get the form data.
//

// $action
// What page we're in.
if (array_key_exists('action', $_REQUEST)) {
  $action = $_REQUEST['action'];
} else {
  $action = ACTION_NONE;
}

// $db
$db = get_session_var('db', TRUE);

// $user
$user = get_session_var('user', TRUE);

// $password
$password = get_session_var('password', FALSE);

//error_log('p: ' . crypt($password, substr('12345678901234',0, CRYPT_SALT_LENGTH)) . ': ' .  ($password === ''));

// Instantiate the upload form.
$mpi_form = new Upload_form($db, $user, $password, 'MPI');

//
// Process input values
//

//
// Dispatch on action
//
switch ($action) {
  case ACTION_NONE:
    $mpi_form->show();
    break;

  case ACTION_UPLOAD:
    // Make a database connection.
    $dbconn =& new Upload_connection($db
                                           , $user
                                           , $password
                                           , $mpi_form);
    // Do the deed.
    upload_mpi($mpi_form, $dbconn);
    break;
}

// Finish the stock page.

show_html_finish();

return;  // End of program.
