<?php
/* Copyright (C) 2006-2009, 2011, Karl O. Pinc
 *
 *    This file is part of Babase.
 *
 *    Babase is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Upload data into the ticks table.
 *
 * Karl O. Pinc <kop@meme.com>
 *
 * Remarks:
 *  Does not do anything to protect against SQL injection attacks,
 * but this does not matter as anybody with a login has restricted
 * permissions anyhow.
 */

// Includes
require('constants.php.inc');
require('lib.php.inc');
require('lib_upload.php.inc');
require('lib_darting.php.inc');

//
// Constants
//

// Delimiter separating data elements in data file that's input.
define('DELIM', "\t");

// Columns in the input file.
define('NAME_COL', 0);      // validated but not saved anywhere
define('SNAME_COL', 1);     // used to find the darting
define('SEX_COL', 2);       // validated but not saved anywhere
define('DARTDATE_COL', 3);  // used to find the darting
define('PCOUNT_COL', 4);
define('BODYPART_COL', 5);
define('PKIND_COL', 6);
define('PSTATUS_COL', 7);
define('PNOTES_COL', 8);
define('OTHER_NOTES_MEASURES_COL', 9);

define('LAST_COL', 9);  // THIS MUST BE THE LAST INDEX

// Use dollar quoting to avoid escape and single quote issues.
define('DQ', '$$');

//
// Functions.
// 


function upload_tick(&$tick_form, &$dbconn) {
  /*
   * Upload the tick data and show the result.
   *
   * Input:
   *  $tick_form  An instance of the Upload_form class.
   *  $dbconn  An instance of the Upload_connection class.
   *
   * Returns: n/a
   *
   * Side Effects:
   *   Outputs html.
   *
   * Remarks:
   *  One of the main pages of the program.
   */

  // Get the uploaded tick data.
  $upload_object =& new Uploaded_data($tick_form);
  if (!$upload_object->upload()) {
    return;
  }
  
  $ticks = $upload_object->data;
  $lines = count($ticks) - 1;
  
  //
  // Ignore leading blank lines.
  //
  $firstline = -1;
  for ($l = 0; $l <= $lines; $l++) {
    if (trim($ticks[$l]) != '') {
      $firstline = $l;
      break;
    }
  }

  if ($firstline == -1) {
    $tick_form->show('<p class="error">No data in uploaded file</p>');
    return;
  }
  
  //
  // Check the column headings.
  //
  $headers = array('name'
                   , 'sname'
                   , 'sex'
                   , 'dartdate'
                   , 'pcount'
                   , 'bodypart'
                   , 'pkind'
                   , 'pstatus'
                   , 'pnotes'
                   , 'other_notes_measures');
  $row = explode(DELIM, rtrim($ticks[$firstline]));

  if (count($row) != count($headers)) {
    $tick_form->show('<p class="error">Bad column headings: incorrect number of columns</p>');
    return;
  }

  foreach ($row as $k => $v) {
    if ($v != $headers[$k]) {
      $tick_form->show('<p class="error">Bad column heading: "'
                      . htmlentities($headers[$k], ENT_QUOTES)
                      . '" should be "'
                      . $v
                      . '"</p>');
      return;
    }
  }

  $firstline += 1;

  //
  // Ignore trailing blank lines.
  //
  for ($l = $lines; $l > 0; $l--) {
    if (trim($ticks[$l]) == '') {
      $lines--;
    } else {
      break;
    }
  }

  //
  // Connect to the db.
  //
  if (!$dbconn->connect()) {
    return;
  }

  // Begin a transaction.
  if (!$dbconn->begin(TRUE)) {
    return;
  }

  //
  // Go through all the data and load it.
  //

  $faulting = FALSE;
  for ($line = $firstline; $line <= $lines; $line++) {

    // Ensure that we can dollar quote everything in the row.
    if (strpos($ticks[$line], DQ) != 0) {
      $mpi_form->error($line + 1
                       , DQ . ' character sequence found on the line'
                       , $faulting
                       , 'These characters cannot be used in input to this program');
      $faulting = TRUE;
      continue; // to the next line of the input file
    }

    // Break each line into words.
    $row = explode(DELIM, rtrim($ticks[$line]));

    for ($i = count($row); $i <= LAST_COL; $i++) {
      // Make sure that there are as many columns as we might reference.
      $row[$i] = '';
    }

    // Get the name
    $name = trim($row[NAME_COL]);
    if ($name == '') {
      // Skip this row.
      // Let the user try again.
      $tick_form->error($line + 1
                          , 'No name (Empty line?)'
                          , $faulting);

      $faulting = TRUE;
      continue;  // to the next row of the input file
    }

    // Get the sname.
    $sname = trim($row[SNAME_COL]);
    if ($sname == '') {
      // Skip this row.
      // Let the user try again.
      $tick_form->error($line + 1
                          , 'No sname'
                          , $faulting);

      $faulting = TRUE;
      continue;  // to the next row of the input file
    }

    // Get the sex.
    $sex = trim($row[SEX_COL]);
    if ($sex == '') {
      // Skip this row.
      // Let the user try again.
      $tick_form->error($line + 1
                          , 'No sex'
                          , $faulting);

      $faulting = TRUE;
      continue;  // to the next row of the input file
    }

    // Get the dartdate.
    $dartdate = trim($row[DARTDATE_COL]);
    if ($dartdate == '') {
      // Skip this row.
      // Let the user try again.
      $tick_form->error($line + 1
                          , 'No dartdate?)'
                          , $faulting);

      $faulting = TRUE;
      continue;  // to the next row of the input file
    }

    //
    // Basic validation
    // This obtains the $dartid.
    //
    if (!validate_darting($dbconn
                          , $tick_form
                          , $line
                          , $dartdate
                          , $sname
                          , $name
                          , $sex
                          , $dartid
                          , TRUE
                          , $faulting)) {
      $faulting = TRUE;
      continue; // to the next line of the input file.
    }

    // Prepare data for upload.
    
    // Interpret the count.  It can be null.
    $pcount = trim($row[PCOUNT_COL]);
    if ($pcount == '') {
      $pcount = 'NULL';  // Empty data means NULL
    }      

    // Interpret the tickbpnotes.
    $pnotes = trim($row[PNOTES_COL]);
    if ($pnotes == '') {
      // Empty data means NULL
      $pnotes = 'NULL';
    } else {
      // Dollar quote the tickbpnotes.
      $pnotes = DQ . $pnotes . DQ;
    }

    //
    // Save the data.
    // 

    // Save TICKS.
    if (!$dbconn->sql('INSERT INTO'
                      . ' ticks (dartid'
                      .          ', tickcount'
                      .          ', bodypart'
                      .          ', tickkind'
                      .          ', tickstatus'
                      .          ', tickbpnotes)'
                      . " values($dartid"
                      .          ', ' . $pcount
                      .          ', ' . DQ . trim($row[BODYPART_COL]) . DQ
                      .          ', ' . DQ . trim($row[PKIND_COL]) . DQ
                      .          ', ' . DQ . trim($row[PSTATUS_COL]) . DQ
                      .          ', ' . $pnotes . ');')) {
          // Give up.
          // Let the user try again.
          $tick_form->error($line + 1
                              , htmlentities($dbconn->last_error())
                              , $faulting);
          $faulting = TRUE;
          continue; // to the next datum on the line
    }

    $ticknotes = trim($row[OTHER_NOTES_MEASURES_COL]);
    if ($ticknotes !== '') {
      // Update the DARTINGS.Ticknotes.
      if (!$dbconn->sql('UPDATE dartings SET ticknotes = '
                        . DQ . $ticknotes . DQ
                        . 'WHERE dartid = ' . $dartid . ';')) {
        $tick_form->error($line + 1
                          , htmlentites($dbconn->last_error())
                          , $faulting
                          , 'Likely upload of duplicate data');
        $faulting = TRUE;
        continue; // to the next datum on the line
      }
    }  

  }

  //
  // Commit the transaction (or allow user to try again.
  //

  //$dbconn->rollback('<p>Transaction deliberately rolled back.</p>');
  //return;

  if ($faulting) {
    $dbconn->rollback('');
    return;
  }

  if (!$dbconn->complete()) {
    return;
  }

  //
  // Everything seems to have worked.
  // Allow the user to do it again.
  //
  $upload_object->success();  // Remember uploaded file.
  $tick_form->show('<p class="success">TICK successfully updated!</p>');

  return;
}

//
// Body of code
//

maintain_login_session();

show_html_start('uptick -- Update the TICK table');

  ?>

<h1>Update TICKS</h1>

<?php

//
// Get the form data.
//

// $action
// What page we're in.
if (array_key_exists('action', $_REQUEST)) {
  $action = $_REQUEST['action'];
} else {
  $action = ACTION_NONE;
}

// $db
$db = get_session_var('db', TRUE);

// $user
$user = get_session_var('user', TRUE);

// $password
$password = get_session_var('password', FALSE);

//error_log('p: ' . crypt($password, substr('12345678901234',0, CRYPT_SALT_LENGTH)) . ': ' .  ($password === ''));

// Instantiate the upload form.
$tick_form =& new Upload_form($db, $user, $password, 'tick');

//
// Process input values
//

//
// Dispatch on action
//
switch ($action) {
  case ACTION_NONE:
    $tick_form->show();
    break;

  case ACTION_UPLOAD:
    // Make a database connection.
    $dbconn =& new Upload_connection($db
                                           , $user
                                           , $password
                                           , $tick_form);
    // Do the deed.
    upload_tick($tick_form, $dbconn);
    break;
}

// Finish the stock page.

show_html_finish();

return;  // End of program.
