# Copyright (C) 2015 Domen Kozar https://www.domenkozar.com
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Domen Kozar <domen@dev.si>

import zipfile
from io import StringIO

import mock
from webob.multidict import MultiDict
from pyramid.testing import DummyRequest

from csvdump.views import dropdowns, download, root


def test_root_view():
    r = DummyRequest()
    response = root(r)
    assert u'ABRP Genomics Data' in response.unicode_body


@mock.patch('csvdump.views.query_db')
def test_dropdowns_no_value(mock_query_db):
    mock_query_db.return_value = []
    r = DummyRequest()
    r.engine = None
    response = dropdowns(r)
    assert response == {'data_types': {}, 'studies': {}}


@mock.patch('csvdump.views.query_db')
def test_dropdowns(mock_query_db):
    mock_query_db.return_value = [
        (1, "foo_study", "foo_type"),
        (2, "foo_study", "foo_type"),
        (3, "foo_study2", "foo_type"),
        (4, "foo_study", "foo_type2"),
    ]
    r = DummyRequest()
    r.engine = None
    response = dropdowns(r)
    assert response == {
        'data_types': {
            'foo_type': [1, 2, 3],
            'foo_type2': [4],
        },
        'studies': {'foo_study': [1,2,4], 'foo_study2': [3]},
    }


@mock.patch('csvdump.views.query2')
@mock.patch('csvdump.views.DBSession')
def test_no_used_ids(mock_dbsession, mock_query2):
    r = DummyRequest()
    r.id = ''
    r.POST = MultiDict()
    r.POST['addlinfo'] = 'No'
    r.engine = None
    output = download(r)
    assert output.location == '/functional-genomic-download/?error=No+stock+ids+found%2C+please+refine+your+search.'


@mock.patch('csvdump.views.query2')
@mock.patch('csvdump.views.Dump')
@mock.patch('csvdump.views.resultproxy_to_csv')
@mock.patch('csvdump.views.DBSession')
def test_no_additional_info(mock_dbsession, mock_resultproxy_to_csv,
                            mock_dump, mock_query2):
    mock_query2.return_value = [[1], [2], [3]]
    mock_resultproxy_to_csv.return_value = StringIO()
    mock_dump().web_dump.return_value = ("foobar/", [6, 7])
    r = DummyRequest()
    r.id = ''
    r.POST = MultiDict()
    r.POST.add('studies', '1,3')
    r.POST.add('data_types', '1,2,3')
    r.POST['addlinfo'] = 'No'
    r.engine = mock.Mock()

    response = download(r)

    # assert additional information is in the zipfile
    zf = zipfile.ZipFile(response.app_iter, 'r')
    assert zf.namelist() == ['foobar/additional_information.csv']

    # assert only default columns are given
    mock_resultproxy_to_csv.assert_called_once_with(
        mock_dbsession(),
        ['stock_id', 'babid', 'ageclass', 'sex'],
        stock_ids=(6, 7),
    )


@mock.patch('csvdump.views.query2')
@mock.patch('csvdump.views.Dump')
@mock.patch('csvdump.views.resultproxy_to_csv')
@mock.patch('csvdump.views.DBSession')
def test_with_additional_information(mock_dbsession, mock_resultproxy_to_csv,
                                     mock_dump, mock_query2):
    mock_query2.return_value = [[1], [2], [3]]
    mock_resultproxy_to_csv.return_value = StringIO()
    mock_dump().web_dump.return_value = ("foobar/", [6, 7])
    r = DummyRequest()
    r.id = ''
    r.POST = MultiDict()
    r.POST.add('studies', '1,2')
    r.POST.add('data_types', '1,2,3')
    r.POST['addlinfo'] = 'Yes'
    r.engine = mock.Mock()

    response = download(r)

    # additional information is there
    assert 'age_sample_collection' in mock_resultproxy_to_csv.mock_calls[0][1][1]

    # assert additional information is in the zipfile
    zf = zipfile.ZipFile(response.app_iter, 'r')
    assert zf.namelist() == ['foobar/additional_information.csv']


@mock.patch('csvdump.views.query2')
@mock.patch('csvdump.views.Dump')
@mock.patch('csvdump.views.resultproxy_to_csv')
@mock.patch('csvdump.views.DBSession')
def test_intersect_analysis_ids(mock_dbsession, mock_resultproxy_to_csv,
                                mock_dump, mock_query2):
    mock_query2.return_value = [[1], [2], [3]]
    mock_resultproxy_to_csv.return_value = StringIO()
    mock_dump().web_dump.return_value = ("foobar/", [6, 7])
    r = DummyRequest()
    r.id = ''
    r.POST = MultiDict()
    r.POST.add('studies', '2,4')
    r.POST.add('data_types', '1,2,3')
    r.POST['addlinfo'] = 'No'
    r.engine = mock.Mock()

    download(r)

    mock_dump().web_dump.assert_called_with(
        session=mock_dbsession(),
        zfile=mock.ANY,
        stock_ids=(1, 2, 3),
        analysis_ids=[2],  # we only get intersection
        anonymize=True,
    )
