# Copyright (C) 2015 Domen Kozar https://www.domenkozar.com
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Domen Kozar <domen@dev.si>

import os
import operator
import zipfile
import urllib
import datetime
import tempfile
import functools
import logging

from pyramid.view import view_config
from pyramid.response import FileResponse, Response
from pyramid.httpexceptions import HTTPFound

from babase.chado.chado_funcgen_dump import Dump
from csvdump.sql import (query_db, QUERY1, query2,
                         DBSession, resultproxy_to_csv)


logger = logging.getLogger(__name__)
HERE = os.path.dirname(__file__)


@view_config(route_name='root',
             http_cache=3600)
def root(request):
    """Returns main index.html with rendered by react.js"""
    return FileResponse(os.path.join(HERE, 'static', 'index.html'))


@view_config(route_name='download',
             request_method='POST',
             renderer='json')
def download(request):
    """
    """
    # TODO: validate inputs
    sex = request.POST.getall('sex')
    ageclass = request.POST.getall('ageclass')
    include_add_info = request.POST['addlinfo']

    # generate analysis_ids
    def extract(raw):
        _temp = map(lambda x: x.split(','), raw)
        return functools.reduce(operator.add, _temp, [])

    studies = extract(request.POST.getall('studies'))
    data_types = extract(request.POST.getall('data_types'))
    analysis_ids = list(map(int, list(set(studies) & set(data_types))))

    logger.info(
        '[requestid={}] Download started with parameters studies={}, data_types={} '
        'sex={} ageclass={} additional_info={}'.format(
            request.id, studies, data_types, sex, ageclass, include_add_info
        )
    )

    dbsession = DBSession()

    # get stock ids
    stock_ids = query2(dbsession, sex=sex, ageclass=ageclass)
    stock_ids = tuple(map(lambda x: x[0], stock_ids))
    f = tempfile.NamedTemporaryFile(
        prefix='ABRP_Genomics_Data_{}'.format(datetime.datetime.now()),
        suffix='.zip',
        delete=True)
    with zipfile.ZipFile(
        f,
        'w',
        compression=zipfile.ZIP_DEFLATED,  # due to windows support
    ) as myzip:
        dir_name, used_ids = Dump().web_dump(
            session=dbsession,
            zfile=myzip,
            stock_ids=stock_ids or [],
            analysis_ids=analysis_ids or [],
            anonymize=True,
        )
        if not used_ids:
            logger.info('[requestid={}] No used_ids returned'.format(request.id))
            msg = urllib.parse.quote_plus('No stock ids found, please refine your search.')
            return HTTPFound(
                location='/functional-genomic-download/?error={}'.format(msg)
            )
        # additional information
        columns = ['stock_id', 'babid', 'ageclass', 'sex']
        if include_add_info == 'Yes':
            columns.extend(['birth', 'bstatus', 'collection_date',
                            'age_sample_collection', 'entrydate',
                            'entrytype', 'depart_date', 'depart_type',
                            'mature_date', 'mature_date_status', 'mom_id'])
        # get csv by querying db
        csvio = resultproxy_to_csv(
            dbsession,
            columns,
            stock_ids=tuple(used_ids))
        # write csv to the zip file
        myzip.writestr(os.path.join(dir_name, 'additional_information.csv'), csvio.getvalue())

    # log what was downloaded into a table
    # prepare response for the zip file
    response = Response(content_type='application/zip, application/octet-stream')
    f.seek(0)
    response.app_iter = f
    response.headers['Content-Disposition'] = ("attachment; filename=ABRP_Genomics_Data.zip")
    logger.info('[requestid={}] Download request ended'.format(request.id))
    return response


@view_config(route_name='dropdowns',
             renderer='json')
def dropdowns(request):
    dbsession = DBSession()
    values = query_db(dbsession, QUERY1)
    d = {
        'studies': {},
        'data_types': {},
    }

    for analysis_id, study, data_type in values:
        d['studies'].setdefault(study, []).append(analysis_id)
        d['data_types'].setdefault(data_type, []).append(analysis_id)

    return d
