# Copyright (C) 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#    This file is part of Babase.
#
#    Babase is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
import os, bz2
import rfc6266

from pyramid.response import Response
from pyramid.view import view_config
from sqlalchemy import orm, alias, text

from sqlalchemy.sql import select

from sqlalchemy.exc import DBAPIError

from babase.lib.exceptions import Error
from babase.lib.constants import *

from .models import (
    DBSession,
    Analysis,
    CVTerm,
    CV,
    Analysisprop
    )

from .lib.constants import *
from .lib.functions import build_filename

import logging
log = logging.getLogger(__name__)

class NoSuchAnalysis(Error):
    pass

class BadDBContent(Error):
    pass

class NoFileError(BadDBContent):
    pass

class BadDBXref(BadDBContent):
    pass

class BadConfiguration(BadDBContent):
    pass

class NoAnonymizedId(BadDBContent):
    pass

class MultipleAnonymizedIds(BadDBContent):
    pass

class BadFileContent(Error):
    pass

def humanize(bytes):
    '''Return string containing human readable representation of a size.
    '''
    fstr = '{:,d}{}B'
    for unit in ('','Ki','Mi','Gi','Ti','Pi','Ei','Zi'):
        if bytes < 1024.0:
            return fstr.format(bytes, unit)
        bytes /= 1024.0
        fstr = '{:0,.1f}{}B'
    return '{:0,.1f}YiB'.format(bytes)

def get_size(settings, path):
    '''Return size of the basename of path, as a human readable string.
    '''
    fname = os.path.basename(path)
    fspath = os.path.join(settings['vcfdump_directory'], fname)
    try:
        size = os.path.getsize(fspath)
    except OSError as err:
        raise NoFileError(fspath)
    return humanize(size)

def select_vcf_analysis(settings):
    '''Get a SQLAlchemy select object (fragment) that selects analysis rows
    which represent uploaded vcf files.
    
    The returned select object selects no columns.
    '''
    ap_cvterm = CVTerm.__table__.alias('ap_cvterm')
    ap_cv = CV.__table__.alias('ap_cv')
    return (select([])
            .select_from(
                Analysisprop.__table__
                .join(Analysis.__table__
                      , Analysis.analysis_id == Analysisprop.analysis_id)
                .join(ap_cvterm, ap_cvterm.c.cvterm_id == Analysisprop.type_id)
                .join(ap_cv, ap_cv.c.cv_id == ap_cvterm.c.cv_id)
                .join(CVTerm.__table__, CVTerm.cvterm_id == Analysis.type_id)
                .join(CV.__table__, CVTerm.cv_id == CV.cv_id))
            .where(ap_cvterm.c.name == settings['file_term'])
            .where(ap_cv.c.name     == settings['file_cv'])
            .where(CV.name          == settings['analysis_cv'])
            .where(CVTerm.name      == settings['analysis_term']))

def select_vcf_analysis_info(settings):
    '''Get a SQLAlchemy select object that selects basic analysis info
    from analysis which are vcf files.

    Returns a select fragment which returns the columns:
    path         The path to the vcf file
    name         The analysis.name
    '''
    return (select_vcf_analysis(settings)
            .column(Analysisprop.value)
            .column(Analysis.name))

def get_file_list(settings):
    '''Get a sorted list of vcf file names from the db.

    Returns list of dicts, each of which has the keys: name, description, size

    size is in human readable terms
    '''
    values = (DBSession.execute(
                  select_vcf_analysis_info(settings)
                  .column(Analysis.analysis_id)
                  .column(Analysis.description)
                  .order_by(Analysis.name)))
    more_values = []
    for path, name, analysis_id, description in values:
        try:
            size = get_size(settings, path)
        except NoFileError as err:
            log.error('({0}): No vcf file found for analysis_id ({1})'
                      ' when generating file size for listing of vcf files'
                      .format(path, analysis_id))
        else:
            more_values.append({'id': analysis_id
                                , 'name': name
                                , 'description': description
                                , 'size': size})
    return more_values

def get_colmap(settings, analysis_id):
    '''Get a list of dbxrefs describing a vcf file's genomic identifiers.
    The list is in the order of the columns in the vcf file and
    contains db:accession identifiers.
    '''
    map_values = DBSession.execute(
                     select([Analysisprop.value])
                     .select_from(
                         Analysisprop.__table__
                         .join(CVTerm.__table__
                               , CVTerm.cvterm_id == Analysisprop.type_id)
                         .join(CV.__table__
                               , CV.cv_id == CVTerm.cv_id))
                     .where(Analysisprop.analysis_id == analysis_id)
                     .where(CV.name == settings['sample_id_trans_cv'])
                     .where(CVTerm.name == settings['sample_id_trans_term'])
                     .order_by(Analysisprop.rank))
    colmap = [x[0] for x in map_values]
    return colmap

def get_vcf_info(settings, analysis_id):
    '''Get data on a vcf file name from the db.

    Returns tuple containing:
    path    Path to vcf file in fs
    name    Name of analysis
    colmap  List of the genomic information vcf column's dbxref identifiers.
            These are in column position order and so can be used to
            map the columns to the genome (individuals) analyzed.
    '''
    values = DBSession.execute(
                 select_vcf_analysis_info(settings)
                 .where(Analysis.analysis_id == analysis_id)).fetchone()
    if not values:
        raise NoSuchAnalysis(str(analysis_id))
    path, name = values
    colmap = get_colmap(settings, analysis_id)
    return (path, name, colmap)

def get_anonymized_ids(settings, colmap, analysis_name, analysis_id):
    '''Return a list of anonymous ids from the colmap list.'''
    # Note that we must execute SQL, since everything may not be
    # in the orm and we're not mapping to the orm anyway.
    new_ids = []
    for db_accession in colmap:
        try:
            db, accession = db_accession.split(':')
        except ValueError:
            raise BadDBXref(
                ('The DBXref ({0}) for analysis (analysis_id: {1},'
                 ' name: {2}), found in the related Analysisprop.value'
                 ' having a ({3}) CV.name'
                 ' and a ({4}) CVTerm.name, cannot be parsed')
                .format(db_accession, analysis_id, analysis_name
                        , settings['sample_id_trans_cv']
                        , settings['sample_id_trans_term']))
        chadodbmap = settings['chadodbmap']
        if db not in chadodbmap:
            raise NoSuchDB(
                ('The DB ({0}) in the DBXref ({1}) for analysis'
                 ' (analysis_id: {2},'
                 ' name: {3}), found in the related Analysisprop.value'
                 ' having a ({4}) CV.name'
                 ' and a ({5}) CVTerm.name, does not have a [columap:*]'
                 ' section in the configuration file')
                .format(db, db_accession, analysis_id, analysis_name
                        , settings['sample_id_trans_cv']
                        , settings['sample_id_trans_term']))
        table = chadodbmap[db]['table']
        key   = chadodbmap[db]['key']
        value = chadodbmap[db]['value']
        query = text("select {value} from {table} where {key} = :keyval"
                     .format(table=table, key=key, value=value))
        try:
            values = DBSession.execute(query, {'keyval': accession}).fetchall()
        except DBAPIError:
            raise BadConfiguration(
                ('The configuration for colmap:{0} produces a query'
                 ' which does not execute')
                .format(db))
        if len(values) == 0:
            raise NoAnonymizedId(
                ('The DBXref ({0}) for analysis (analysis_id: {1},'
                 ' name: {2}), found in the related Analysisprop.value'
                 ' having a ({3}) CV.name'
                 ' and a ({4}) CVTerm.name, has no corresponding anonymized'
                 ' column heading')
                .format(db_accession, analysis_id, analysis_name
                        , settings['sample_id_trans_cv']
                        , settings['sample_id_trans_term']))
        if len(values) > 1:
            raise MultipleAnonymizedIds(
                ('The DBXref ({0}) for analysis (analysis_id: {1},'
                 ' name: {2}), found in the related Analysisprop.value'
                 ' having a ({3}) CV.name'
                 ' and a ({4}) CVTerm.name, has multiple corresponding'
                 ' anonymized column headings')
                .format(db_accession, analysis_id, analysis_name
                        , settings['sample_id_trans_cv']
                        , settings['sample_id_trans_term']))
        new_ids.append(values[0][0])
    return new_ids

def fill_response(response, name, vcf_file, new_ids):
    '''Anonymize vcf file and fill the response object with it.'''
    bz2c = bz2.BZ2Compressor()
    try:
        unaltered = True
        line_no = 1
        for line in vcf_file:
            if unaltered:
                fields = line.split(BB_VCF_DELIM)
                if fields[0] == '#CHROM':
                    unaltered = False
                    replacement_fields = fields[:BB_VCF_LAST_FIXED_COL]
                    replacement_fields.extend(new_ids)
                    new_line = BB_VCF_DELIM.join(replacement_fields)
                    if line[-1] == '\n':
                        new_line += '\n'
                    line = new_line
            data = bz2c.compress(bytes(line, 'ascii'))
            response.body_file.write(data)
            line_no += 1
    except UnicodeEncodeError as err:
        raise BadFileContent(
                 'Line {0} has a byte that cannot be converted to ascii: {1}'
                 .format(line_no, err)) 
                
    data = bz2c.flush()
    response.body_file.write(data)
    response.content_type = CONTENT_TYPE
    response.content_disposition = rfc6266.build_header(build_filename(name))
    response.accept_ranges = 'none'

def connection_problem(err):
    '''Return a response when we can't connect to the db.  Log problem.

    Parameters:
    `err`  Exception notifying us of a problem
    '''
    log.critical('{0}'.format(err))
    conn_err_msg = """\
Apologies.  There is a problem connecting to the database.

Please try again later.
"""
    return Response(conn_err_msg, content_type='text/plain', status_int=500)

def missing_analysis_problem():
        return Response('Requested analysis not available.'
                        , content_type='text/plain', status_int=500)

@view_config(route_name='vcfdl')
def deliver_vcf(request):
    settings = request.registry.settings
    analysis_id = request.params.getone('id')

    try:
        path, name, colmap = get_vcf_info(settings, analysis_id)
    except DBAPIError as err:
        return connection_problem(err)
    except NoSuchAnalysis:
        log.warning(
            '({0}): Client requested a non-existant analysis.analysis_id'
            .format(str(analysis_id)))
        return missing_analysis_problem()
    try:
        new_ids = get_anonymized_ids(settings, colmap, name, analysis_id)
    except BadDBContent as err:
        log.error(str(err))
        return missing_analysis_problem()

    response = Response()
    try:
        with open(path, 'r') as vcf_file:
            try:
                fill_response(response, name, vcf_file, new_ids)
            except BadFileContent as err:
                log.error('Bad data in ({0}): {1}'.format(path, err))
                return missing_analysis_problem()
    except OSError as err:
        log.error(('{0}: analysis_id {1}: analysis.name {2}: Cannot open vcf'
                   'file for download: {3}')
                  .format(err, analysis_id, name, path))
        return missing_analysis_problem()

    log.info('Downloaded: ({0}): ({1})'.format(path, name))
    return response

@view_config(route_name='home', renderer='templates/vcfdump.mako')
def list_vcf_files(request):
    settings = request.registry.settings
    dir = settings['vcfdump_directory']
    try:
        vcf_files = get_file_list(settings)
    except DBAPIError as err:
        return connection_problem(err)
    return {'vcf_files': vcf_files
            , 'meta_description': 'View and download data sets on genetic'
                                  ' variation in the Amboseli baboon'
                                  ' population.'
           }
