#!/bin/sh
# Copyright (C) 2015 The Meme Factory, Inc.  http://www.meme.com/
#
#   This file is part of Babase.
#
#   Babase is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Babase.  If not, see <http://www.gnu.org/licenses/>.
#
# Karl O. Pinc <kop@meme.com>
#
# Syntax:
#   copy-to-abrp-genomics [--login user] --host hostname
#                         --identity keyfile dbname
#
# Input:
#   login    Name of user to login with (optional)
#            Default is "bb_receive_abrp_genomics_schema".
#   hostname Name of host to which to copy data
#   keyfile  Name of key file to use with ssh.
#   dbname   Name of db from which data is copied
#
# Remarks:
#   See docbooks docs for more info.
#
# Bugs:
#  The pg db owner is hardcoded.
#
divert(-1)dnl
dnl m4 macro definitions

include(`constants.m4')
include(`macros.m4')

dnl Copy a table from the chado schema into the abrp_genomics schema
dnl
dnl Syntax:  copy_table(tablename)
dnl
dnl Input:
dnl
dnl  tablename  The name of the table to copy
dnl
changequote([,])
define([copy_table],[dnl
create table abrp_genomics_new.$1 
  (like chado.$1 including indexes);
insert into abrp_genomics_new.$1
  select * from chado.$1;
])dnl
changequote(`,')dnl

divert`'dnl
editwarning()

# Parse command line.
export hostname=''
export keyfile=''
while [ $# -ge 2 ] ; do
    case $1 in
        --login)
             user="$2"
             shift 2
             ;;
        --host)
             hostname="$2"
             shift 2
             ;;
        --identity)
             keyfile="$2"
             shift 2
             ;;
        *)
             echo "$1: unknown argument" >&2
             exit 1
             ;;
    esac
done

if [ -z "$user" ]; then
  user=bb_receive_abrp_genomics_schema
fi

if [ -z "$hostname" ]; then
  echo 'A --host option is required' >&2
  exit 1
fi

if [ -z "$keyfile" ]; then
  echo 'A --identity option is required' >&2
  exit 1
fi

export db="$1"

export psql_args="-q -U babase_admin $db"

if [ "$db" = 'bb_live_db' ] ; then
  export role=abrp_public
else
  export role=abrp_test
fi

logger -t "$(basename $0)" -p bb_syslog_facility.INFO \
       "Starting replacment of abrp_genomics schema content on host $hostname from db $db"

psql $psql_args << EOF
\set ON_ERROR_STOP
set client_min_messages to warning;

begin;

create schema abrp_genomics_new;
grant usage on schema abrp_genomics_new to $role;

set search_path to abrp_genomics_new;

-- unique_baboons maps the usual baboon ids (uniquebaboon)
-- to anonymized ids (babid).
CREATE TABLE abrp_genomics_new.unique_baboons AS
  SELECT babid, uniquebaboon
    FROM genetic_inventory.unique_baboons;

-- Add indexes.
-- (We will only ever be mapping from the canonical name to
-- the anonymized name.)
create unique index unique_baboons_uniquebaboon
  on unique_baboons (uniquebaboon);

-- The stock_phenotype table maps stock to anonymized ids and phenotype
-- information.

CREATE TABLE abrp_genomics_new.stock_phenotypes AS
-- Query courtesy of Jake Gordon.
$(cat /etc/babase/stock_phenotypes.sql)
;

-- Add indexes
-- The stock id is really the primary key.
create unique index stock_phenotypes_stock_id on stock_phenotypes (stock_id);
create index stock_phenotypes_babid on stock_phenotypes (babid);
-- Not worth creating these.
--create index stock_phenotypes_ageclass on stock_phenotypes (ageclass);
--create index stock_phenotypes_sex on stock_phenotypes (sex);

-- Copy other tables.
-- Don't worry about order since we don't copy constraints.

-- analysis
copy_table(`analysis')

-- cvterm
copy_table(`cvterm')

-- cv
copy_table(`cv')

-- dbxref
copy_table(`dbxref')

update abrp_genomics_new.dbxref
  set accession = unique_baboons.babid
  from genetic_inventory.unique_baboons
     , chado.db
  where dbxref.db_id = db.db_id
        and dbxref.accession = unique_baboons.uniquebaboon;

-- db
copy_table(`db')

-- analysisprop
copy_table(`analysisprop')

-- analysis_nd_experiment
copy_table(`analysis_nd_experiment')

-- nd_experiment
copy_table(`nd_experiment')

-- nd_experimentprop
copy_table(`nd_experimentprop')

-- nd_experiment_stock
copy_table(`nd_experiment_stock')


GRANT SELECT ON ALL TABLES IN SCHEMA abrp_genomics_new TO $role;
REVOKE ALL ON ALL TABLES IN SCHEMA abrp_genomics_new FROM public;

commit;
EOF

export result=$?
if [ $result -ne 0 ] ; then
   logger -t "$(basename $0)" -p bb_syslog_facility.ERR -s \
          'Cannot copy the data to host $hostname'
   exit $result
fi

# Dump the new schema with it's data.
pg_dump -n abrp_genomics_new -Fc -U babase_admin $db \
  | ssh -i "$keyfile" $user@$hostname $db


# cleanup
psql $psql_args << EOF
set client_min_messages to warning;
drop schema abrp_genomics_new cascade;
EOF

logger -t "$(basename $0)" -p bb_syslog_facility.INFO \
       "Replaced abrp_genomics schema content on host $hostname from db $db"
